@@ -18,7 +18,7 @@
 
 #include "estruct.h"
 
-#if OPT_PERL
+#if OPT_PERL || OPT_PLUGIN
 
 #include "edef.h"
 #include "api.h"
@@ -186,27 +186,29 @@ lreplace(char *s, int len)
 void
 api_setup_fake_win(VileBuf * vbp, int do_delete)
 {
-    if (curwp_visible == 0)
-	curwp_visible = curwp;
+    if (vbp != 0) {
+	if (curwp_visible == 0)
+	    curwp_visible = curwp;
 
-    if (vbp->fwp) {
-	curwp = vbp->fwp;
-	curbp = curwp->w_bufp;
-    } else {
-	(void) push_fake_win(vbp->bp);
-	vbp->fwp = curwp;
-	vbp->changed = 0;
-    }
+	if (vbp->fwp) {
+	    curwp = vbp->fwp;
+	    curbp = curwp->w_bufp;
+	} else {
+	    (void) push_fake_win(vbp->bp);
+	    vbp->fwp = curwp;
+	    vbp->changed = 0;
+	}
 
-    if (vbp->ndel > 0 && do_delete) {
-	int status;
-	/* Do lazy delete; FALSE means don't put text in kill buffer */
-	status = ldelete(vbp->ndel, FALSE);
-	vbp->ndel = 0;
-	if (status == FALSE
-	    || (lforw(DOT.l) == buf_head(curbp) && DOT.o >= llength(DOT.l))) {
-	    make_local_b_val(curbp, MDNEWLINE);
-	    set_b_val(curbp, MDNEWLINE, FALSE);
+	if (vbp->ndel != 0 && do_delete) {
+	    int status;
+	    /* Do lazy delete; FALSE means don't put text in kill buffer */
+	    status = ldelete(vbp->ndel, FALSE);
+	    vbp->ndel = 0;
+	    if (status == FALSE
+		|| (lforw(DOT.l) == buf_head(curbp) && DOT.o >= llength(DOT.l))) {
+		make_local_b_val(curbp, MDNEWLINE);
+		set_b_val(curbp, MDNEWLINE, FALSE);
+	    }
 	}
     }
 }
@@ -403,7 +405,7 @@ api_gline(VileBuf * vbp, int lno, char **linep, int *lenp)
 
 #if 0				/* Not used. */
 int
-api_dotgline(VileBuf * vbp, char **linep, int *lenp, int *neednewline)
+api_dotgline(VileBuf * vbp, char **linep, B_COUNT * lenp, int *neednewline)
 {
 
     api_setup_fake_win(vbp, TRUE);
@@ -415,23 +417,35 @@ api_dotgline(VileBuf * vbp, char **linep, int *lenp, int *neednewline)
     /* FIXME: Handle rectangular regions. */
 
     if (is_header_line(DOT, curbp)
-	|| (DOT.l == vbp->region.r_end.l
-	    && (vbp->regionshape == FULLLINE
-		|| (vbp->regionshape == EXACT
-		    && DOT.o >= vbp->region.r_end.o)))) {
+	|| !lisreal(DOT.l)
+	|| DOT.o < 0) {
+	return FALSE;
+    }
+
+    if ((DOT.l == vbp->region.r_end.l
+	 && (vbp->regionshape == FULLLINE
+	     || (vbp->regionshape == EXACT
+		 && DOT.o >= vbp->region.r_end.o)))) {
 	return FALSE;
     }
 
     *linep = DOT.l->l_text + DOT.o;
-    *lenp = llength(DOT.l) - DOT.o;
 
-    if (vbp->regionshape == EXACT && DOT.l == vbp->region.r_end.l) {
-	*lenp -= llength(DOT.l) - vbp->region.r_end.o;
+    if (llength(DOT.l) >= DOT.o) {
+	*lenp = llength(DOT.l) - DOT.o;
+
+	if (vbp->regionshape == EXACT && DOT.l == vbp->region.r_end.l) {
+	    B_COUNT next = *lenp + vbp->region.r_end.o;
+	    if (next >= (B_COUNT) llength(DOT.l)) {
+		*lenp = next - llength(DOT.l);
+	    } else {
+		*lenp = 0;
+	    }
+	}
+    } else {
+	*lenp = 0;
     }
 
-    if (*lenp < 0)
-	*lenp = 0;		/* Make sure return length is non-negative */
-
     if (*lenp == 0) {
 	*linep = "";		/* Make sure we pass back a zero length,
 				   null terminated value when the length
@@ -783,7 +797,9 @@ api_command_cleanup(void)
 
     MK = DOT;			/* make sure MK is in same buffer as DOT */
 
+#if OPT_PERL
     perl_free_deferred();
+#endif
 }
 
 void
