@@ -5,7 +5,7 @@
  * functions that adjust the top line in the window and invalidate the
  * framing, are hard.
  *
- * $Header: /usr/build/vile/vile/RCS/basic.c,v 1.119 2003/07/27 19:01:43 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/basic.c,v 1.122 2005/05/27 21:48:44 tom Exp $
  *
  */
 
@@ -317,7 +317,7 @@ gotobos(int f, int n)
 int
 gotomos(int f GCC_UNUSED, int n)
 {
-    LINEPTR lp, head;
+    LINE *lp, *head;
     int half = (curwp->w_ntrows + 1) / 2;
 
     head = buf_head(curbp);
@@ -778,7 +778,7 @@ gotoeosent(int f, int n)
 	} else if (empty || !is_at_end_of_line(DOT)) {
 	    s = forwchar(TRUE, RegexpLen(exp));
 	    while (s && (is_at_end_of_line(DOT) || isSpace(char_at(DOT)))) {
-		LINEPTR lp = DOT.l;
+		LINE *lp = DOT.l;
 		s = forwchar(TRUE, 1);
 		if (lp != DOT.l)
 		    break;
@@ -853,7 +853,7 @@ next_column(int c, int col)
 int
 forwpage(int f, int n)
 {
-    LINEPTR lp;
+    LINE *lp;
     int status;
 
     if ((n = full_pages(f, n)) < 0)
@@ -884,7 +884,7 @@ forwpage(int f, int n)
 int
 backpage(int f, int n)
 {
-    LINEPTR lp;
+    LINE *lp;
     int status;
 
     if ((n = full_pages(f, n)) < 0)
@@ -921,7 +921,7 @@ backpage(int f, int n)
 int
 forwhpage(int f, int n)
 {
-    LINEPTR llp, dlp;
+    LINE *llp, *dlp;
     int status;
 
     if ((n = half_pages(f, n)) < 0)
@@ -957,7 +957,7 @@ forwhpage(int f, int n)
 int
 backhpage(int f, int n)
 {
-    LINEPTR llp, dlp;
+    LINE *llp, *dlp;
     int status;
 
     if ((n = half_pages(f, n)) < 0)
@@ -1074,9 +1074,11 @@ show_mark(int count, BUFFER *bp, MARK mark, int name)
 	    if (stop == 0) {
 		stop = ((DOT.o + tab - 1) / tab) * tab;
 	    }
-	    bprintf("%*PText", stop - DOT.o, ' ');
+	    bpadc(' ', stop - DOT.o);
+	    bprintf("Text");
 	    bprintf("\n----  -------- ------");
-	    bprintf("%*P----", stop - DOT.o, ' ');
+	    bpadc(' ', stop - DOT.o);
+	    bprintf("----");
 	}
 
 	bprintf("\n%c     %8d %8d",
@@ -1084,8 +1086,8 @@ show_mark(int count, BUFFER *bp, MARK mark, int name)
 		line_no(bp, mark.l),
 		mk_to_vcol(mark, FALSE, bp, 0, 0) + 1);
 	if (llength(mark.l) > 0) {
-	    bprintf("%*P%*S",
-		    stop - DOT.o, ' ',
+	    bpadc(' ', stop - DOT.o);
+	    bprintf("%.*s",
 		    llength(mark.l),
 		    mark.l->l_text);
 	}
@@ -1428,7 +1430,7 @@ int
 setwmark(int row, int col)
 {
     MARK save;
-    LINEPTR dlp;
+    LINE *dlp;
 
     save = DOT;
     if (row == mode_row(curwp)) {
