@@ -3,7 +3,7 @@
  *
  *	written 11-feb-86 by Daniel Lawrence
  *
- * $Header: /usr/build/vile/vile/RCS/bind.c,v 1.266 2003/05/04 13:52:55 Mark.Robinson Exp $
+ * $Header: /usr/build/vile/vile/RCS/bind.c,v 1.284 2005/07/13 23:25:16 tom Exp $
  *
  */
 
@@ -108,21 +108,6 @@ static void kbd_puts(const char *s);
 /*----------------------------------------------------------------------------*/
 
 #if OPT_NAMEBST
-static BI_NODE *
-new_namebst(BI_DATA * a)
-{
-    BI_NODE *p;
-
-    beginDisplay();
-    p = typecalloc(BI_NODE);
-    p->value = *a;
-    if (!(a->n_flags & NBST_READONLY))
-	BI_KEY(p) = strmalloc(a->bi_key);
-    endofDisplay();
-
-    return p;
-}
-
 static void
 old_namebst(BI_NODE * a)
 {
@@ -135,17 +120,42 @@ old_namebst(BI_NODE * a)
 		free(TYPECAST(char, cmd->c_help));
 #endif
 #if OPT_MACRO_ARGS
-	    if (cmd->c_args)
+	    if (cmd->c_args) {
+		int n;
+		for (n = 0; cmd->c_args[n].pi_text != 0; ++n) {
+		    free(cmd->c_args[n].pi_text);
+		}
 		free(cmd->c_args);
+	    }
 #endif
 	    free(cmd);
 	}
 	free(TYPECAST(char, BI_KEY(a)));
     }
-    free(TYPECAST(char, a));
+    free(a);
     endofDisplay();
 }
 
+static BI_NODE *
+new_namebst(BI_DATA * a)
+{
+    BI_NODE *p;
+
+    beginDisplay();
+    if ((p = typecalloc(BI_NODE)) != 0) {
+	p->value = *a;
+	if (!(a->n_flags & NBST_READONLY)) {
+	    if ((BI_KEY(p) = strmalloc(a->bi_key)) == 0) {
+		old_namebst(p);
+		p = 0;
+	    }
+	}
+    }
+    endofDisplay();
+
+    return p;
+}
+
 static void
 dpy_namebst(BI_NODE * a GCC_UNUSED, int level GCC_UNUSED)
 {
@@ -168,7 +178,7 @@ static BI_TREE redefns =
 int
 no_such_function(const char *fnp)
 {
-    mlforce("[No such function \"%s\"]", fnp != 0 ? fnp : "");
+    mlforce("[No such function \"%s\"]", NONNULL(fnp));
     return FALSE;
 }
 
@@ -197,7 +207,7 @@ vl_help(int f GCC_UNUSED, int n GCC_UNUSED)
 	}
 	alreadypopped = (bp->b_nwnd != 0);
 	/* and read the stuff in */
-	if (readin(hname, 0, bp, TRUE) == FALSE ||
+	if (readin(hname, 0, bp, TRUE) != TRUE ||
 	    popupbuff(bp) == FALSE) {
 	    (void) zotbuf(bp);
 	    returnCode(FALSE);
@@ -292,7 +302,10 @@ makechrslist(int dum1 GCC_UNUSED, void *ptr GCC_UNUSED)
     int i;
     char temp[NLINE];
 
-    bprintf("--- Terminal Character Settings %*P\n", term.cols - 1, '-');
+    bprintf("--- Terminal Character Settings ");
+    bpadc('-', term.cols - DOT.o);
+    bputc('\n');
+
     for (i = 0; TermChrs[i].name != 0; i++) {
 	bprintf("\n%s = %s",
 		TermChrs[i].name,
@@ -307,9 +320,11 @@ static int
 chr_lookup(const char *name)
 {
     int j;
-    for (j = 0; TermChrs[j].name != 0; j++)
-	if (!strcmp(name, TermChrs[j].name))
-	    return j;
+    if (name != 0) {
+	for (j = 0; TermChrs[j].name != 0; j++)
+	    if (!strcmp(name, TermChrs[j].name))
+		return j;
+    }
     return -1;
 }
 
@@ -428,7 +443,7 @@ key_to_bind(const CMDFUNC * kcmd)
 	outseq[0] = EOS;
 	if (vl_msgs)
 	    kbd_puts(kcod2prc(c, outseq));
-	hst_append_s(outseq, FALSE);
+	hst_append_s(outseq, FALSE, TRUE);
     } else {
 	mlforce("[Not a proper key-sequence]");
     }
@@ -441,7 +456,7 @@ free_KBIND(KBIND ** oldp, KBIND * newp)
     *oldp = newp->k_link;
 
     beginDisplay();
-    free(TYPECAST(char, newp));
+    free(newp);
     endofDisplay();
 
     return TRUE;
@@ -913,7 +928,7 @@ prompt_describe_key(BINDINGS * bs)
     }
 
     (void) kcod2prc(c, outseq);
-    hst_append_s(outseq, FALSE);	/* cannot replay this, but can see it */
+    hst_append_s(outseq, FALSE, TRUE);	/* cannot replay this, but can see it */
 
     /* find the function bound to the key */
     if (!fnc2ntab(&temp, kcod2fnc(bs, c))) {
@@ -1332,13 +1347,12 @@ ourstrstr(const char *haystack, const char *needle, int anchor)
 #endif /* OPT_REBIND || OPT_EVAL */
 
 static char *
-locate_fname(char *dirname, char *fname, UINT mode)
+locate_fname(char *dir_name, char *fname, UINT mode)
 {
     static char fullpath[NFILEN];	/* expanded path */
 
-    if (dirname
-	&& dirname[0] != EOS
-	&& ffaccess(pathcat(fullpath, dirname, fname), mode))
+    if (!isEmpty(dir_name)
+	&& ffaccess(pathcat(fullpath, dir_name, fname), mode))
 	return (fullpath);
 
     return 0;
@@ -1349,12 +1363,13 @@ locate_file_in_list(char *list, char *fname, UINT mode)
 {
     const char *cp;
     char *sp;
-    char dirname[NFILEN];
+    char dir_name[NFILEN];
 
-    cp = list;
-    while ((cp = parse_pathlist(cp, dirname)) != 0) {
-	if ((sp = locate_fname(dirname, fname, mode)) != 0)
-	    return sp;
+    if ((cp = list) != 0) {
+	while ((cp = parse_pathlist(cp, dir_name)) != 0) {
+	    if ((sp = locate_fname(dir_name, fname, mode)) != 0)
+		return sp;
+	}
     }
     return 0;
 }
@@ -1374,7 +1389,7 @@ PATH_value(void)
     if (!tb_length(myfiles)) {
 	char mypath[NFILEN];
 
-	(void) strcpy(mypath, prog_arg);
+	(void) strcpy(mypath, NONNULL(prog_arg));
 	if ((tmp = vms_pathleaf(mypath)) == mypath)
 	    (void) strcpy(mypath,
 			  current_directory(FALSE));
@@ -1469,13 +1484,12 @@ list_one_fname(char *fname, UINT mode)
 }
 
 static void
-list_which_fname(char *dirname, char *fname, UINT mode)
+list_which_fname(char *dir_name, char *fname, UINT mode)
 {
     char fullpath[NFILEN];	/* expanded path */
 
-    if (dirname
-	&& dirname[0] != EOS) {
-	list_one_fname(SL_TO_BSL(pathcat(fullpath, dirname, fname)), mode);
+    if (!isEmpty(dir_name)) {
+	list_one_fname(SL_TO_BSL(pathcat(fullpath, dir_name, fname)), mode);
     }
 }
 
@@ -1483,11 +1497,12 @@ static void
 list_which_file_in_list(char *list, char *fname, UINT mode)
 {
     const char *cp;
-    char dirname[NFILEN];
+    char dir_name[NFILEN];
 
-    cp = list;
-    while ((cp = parse_pathlist(cp, dirname)) != 0) {
-	list_which_fname(dirname, fname, mode);
+    if ((cp = list) != 0) {
+	while ((cp = parse_pathlist(cp, dir_name)) != 0) {
+	    list_which_fname(dir_name, fname, mode);
+	}
     }
 }
 
@@ -1581,9 +1596,9 @@ which_exec(int f, int n)
 
 /* translate a keycode to its binding-string */
 char *
-kcod2pstr(int c, char *seq)
+kcod2pstr(int c, char *seq, int limit)
 {
-    seq[0] = (char) kcod2escape_seq(c, &seq[1]);
+    seq[0] = (char) kcod2escape_seq(c, &seq[1], limit - 1);
     return seq;
 }
 
@@ -1605,48 +1620,54 @@ static const struct {
 };
 #endif
 
+#define ADD_KCODE(src) \
+	if ((ptr - base) + (len = strlen(src)) + 3 < limit) { \
+	    strcpy(ptr, src); \
+	    ptr += len; \
+	}
+
 /* Translate a 16-bit keycode to a string that will replay into the same
  * code.
  */
 int
-kcod2escape_seq(int c, char *ptr)
+kcod2escape_seq(int c, char *ptr, int limit)
 {
     char *base = ptr;
 
-    if (c & CTLA)
-	*ptr++ = (char) cntl_a;
-    else if (c & CTLX)
-	*ptr++ = (char) cntl_x;
+    if (base != 0 && limit > 5) {
+	if (c & CTLA)
+	    *ptr++ = (char) cntl_a;
+	else if (c & CTLX)
+	    *ptr++ = (char) cntl_x;
 
 #if OPT_KEY_MODIFY
-    if (c & mod_KEY) {
-	unsigned n;
-	for (n = 0; n < TABLESIZE(key_modifiers); ++n) {
-	    if (c & key_modifiers[n].code) {
-		strcpy(ptr, key_modifiers[n].name);
-		ptr += strlen(ptr);
+	if (c & mod_KEY) {
+	    int len;
+	    unsigned n;
+	    for (n = 0; n < TABLESIZE(key_modifiers); ++n) {
+		if (c & key_modifiers[n].code) {
+		    ADD_KCODE(key_modifiers[n].name);
+		}
 	    }
-	}
 #if SYS_WINNT
 #define W32INSERT "Insert"
 #define W32DELETE "Delete"
-	c &= mod_NOMOD;
-	if (c == KEY_Insert) {
-	    strcpy(ptr, W32INSERT);
-	    ptr += sizeof(W32INSERT) - 1;
-	    c = EOS;
-	} else if (c == KEY_Delete) {
-	    strcpy(ptr, W32DELETE);
-	    ptr += sizeof(W32DELETE) - 1;
-	    c = EOS;
+	    c &= mod_NOMOD;
+	    if (c == KEY_Insert) {
+		ADD_KCODE(W32INSERT);
+		c = EOS;
+	    } else if (c == KEY_Delete) {
+		ADD_KCODE(W32DELETE);
+		c = EOS;
+	    }
+#endif
 	}
 #endif
+	if (c & SPEC)
+	    *ptr++ = (char) poundc;
+	*ptr++ = (char) c;
+	*ptr = EOS;
     }
-#endif
-    if (c & SPEC)
-	*ptr++ = (char) poundc;
-    *ptr++ = (char) c;
-    *ptr = EOS;
     return (int) (ptr - base);
 }
 
@@ -1700,7 +1721,7 @@ kcod2prc(int c, char *seq)
 {
     char temp[NSTRING];
 
-    (void) kcod2pstr(c, temp);
+    (void) kcod2pstr(c, temp, sizeof(temp));
 #if OPT_KEY_MODIFY
     if (c & mod_KEY) {
 	/* Translation is complete, by defn. */
@@ -1771,14 +1792,14 @@ char *
 fnc2pstr(const CMDFUNC * f)
 {
     int c;
-    static char seq[10];
+    static char seq[40];
 
     c = fnc2kcod(f);
 
     if (c == -1)
 	return NULL;
 
-    return kcod2pstr(c, seq);
+    return kcod2pstr(c, seq, sizeof(seq));
 }
 #endif
 
@@ -2213,7 +2234,8 @@ kbd_erase_to_end(int column)
     MK = DOT;
     if (llength(DOT.l) > 0) {
 	DOT.o = column;
-	ldelete(llength(DOT.l) - DOT.o, FALSE);
+	if (llength(DOT.l) > DOT.o)
+	    ldelete(llength(DOT.l) - DOT.o, FALSE);
 	TRACE(("NULL:%2d:%s\n", llength(DOT.l), lp_visible(DOT.l)));
     }
     curbp = savebp;
@@ -2251,16 +2273,7 @@ cs_strncmp(
 #define	NEXT_DATA(p)	((p)+size_entry)
 #define	PREV_DATA(p)	((p)-size_entry)
 
-#ifdef	lint
-static				/*ARGSUSED */
-const char *
-THIS_NAME(const char *p)
-{
-    return 0;
-}
-#else
 #define	THIS_NAME(p)	(*TYPECAST(const char *const,p))
-#endif
 #define	NEXT_NAME(p)	THIS_NAME(NEXT_DATA(p))
 
 /*
@@ -2468,7 +2481,7 @@ scroll_completions(
     if ((bp = find_b_name(COMPLETIONS_BufName)) == NULL)
 	show_completions(case_insensitive, buf, len, table, size_entry);
     else {
-	LINEPTR lp;
+	LINE *lp;
 	swbuffer(bp);
 	(void) gotoeos(FALSE, 1);
 	lp = DOT.l;
@@ -2643,6 +2656,9 @@ kbd_complete(DONE_ARGS, const char *table, size_t size_entry)
     buf[cpos] = EOS;		/* terminate it for us */
     nbp = table;		/* scan for matches */
 
+    if (nbp == 0)
+	return FALSE;
+
     while (THIS_NAME(nbp) != NULL) {
 	if (StrNcmp(buf, THIS_NAME(nbp), strlen(buf)) == 0) {
 	    testcol = kbd_length();
@@ -2747,12 +2763,14 @@ kbd_complete(DONE_ARGS, const char *table, size_t size_entry)
 static int
 is_shift_cmd(const char *buffer, unsigned cpos)
 {
-    int c = *buffer;
-    if (isRepeatable(c)) {
-	while (--cpos != 0)
-	    if (*(++buffer) != c)
-		return FALSE;
-	return TRUE;
+    if (buffer != 0) {
+	int c = *buffer;
+	if (isRepeatable(c)) {
+	    while (--cpos != 0)
+		if (*(++buffer) != c)
+		    return FALSE;
+	    return TRUE;
+	}
     }
     return FALSE;
 }
@@ -2816,30 +2834,34 @@ static int
 cmd_complete(DONE_ARGS)
 {
     int status;
+    if (buf != 0 && pos != 0) {
 #if OPT_HISTORY
-    /*
-     * If the user scrolled back in 'edithistory()', the text may be a
-     * repeated-shift command, which won't match the command-table (e.g.,
-     * ">>>").
-     */
-    if ((*pos > 1) && is_shift_cmd(buf, *pos)) {
-	unsigned len = 1;
-	char tmp[NLINE];
-	tmp[0] = *buf;
-	tmp[1] = EOS;
-	status = cmd_complete(0, c, tmp, &len);
-    } else
+	/*
+	 * If the user scrolled back in 'edithistory()', the text may be a
+	 * repeated-shift command, which won't match the command-table (e.g.,
+	 * ">>>").
+	 */
+	if ((*pos > 1) && is_shift_cmd(buf, *pos)) {
+	    unsigned len = 1;
+	    char tmp[NLINE];
+	    tmp[0] = *buf;
+	    tmp[1] = EOS;
+	    status = cmd_complete(0, c, tmp, &len);
+	} else
 #endif
-    if ((*pos != 0) && isShellOrPipe(buf)) {
+	if ((*pos != 0) && isShellOrPipe(buf)) {
 #if COMPLETE_FILES
-	status = shell_complete(PASS_DONE_ARGS);
+	    status = shell_complete(PASS_DONE_ARGS);
 #else
-	status = isreturn(c);
-	if (c != NAMEC)
-	    unkeystroke(c);
+	    status = isreturn(c);
+	    if (c != NAMEC)
+		unkeystroke(c);
 #endif
+	} else {
+	    status = kbd_complete_bst(PASS_DONE_ARGS);
+	}
     } else {
-	status = kbd_complete_bst(PASS_DONE_ARGS);
+	status = ABORT;
     }
     return status;
 }
@@ -2874,25 +2896,32 @@ kbd_engl_stat(const char *prompt, char *buffer, int stated)
 int
 insert_namebst(const char *name, const CMDFUNC * cmd, int ro)
 {
-    BI_DATA temp, *p;
+    int result;
 
-    if ((p = btree_search(&namebst, name)) != 0) {
-	if ((p->n_flags & NBST_READONLY)) {
-	    if (btree_insert(&redefns, p) == 0
-		|| !btree_delete(&namebst, name)) {
+    if (name != 0) {
+	BI_DATA temp, *p;
+
+	if ((p = btree_search(&namebst, name)) != 0) {
+	    if ((p->n_flags & NBST_READONLY)) {
+		if (btree_insert(&redefns, p) == 0
+		    || !btree_delete(&namebst, name)) {
+		    return FALSE;
+		}
+		mlwrite("[Redefining builtin '%s']", name);
+	    } else if (!delete_namebst(name, TRUE)) {
 		return FALSE;
 	    }
-	    mlwrite("[Redefining builtin '%s']", name);
-	} else if (!delete_namebst(name, TRUE)) {
-	    return FALSE;
 	}
+
+	temp.bi_key = name;
+	temp.n_cmd = cmd;
+	temp.n_flags = (UCHAR) (ro ? NBST_READONLY : 0);
+
+	result = (btree_insert(&namebst, &temp) != 0);
+    } else {
+	result = FALSE;
     }
-
-    temp.bi_key = name;
-    temp.n_cmd = cmd;
-    temp.n_flags = (UCHAR) (ro ? NBST_READONLY : 0);
-
-    return (btree_insert(&namebst, &temp) != 0);
+    return result;
 }
 
 static void
