@@ -5,7 +5,7 @@
  * keys. Like everyone else, they set hints
  * for the display system.
  *
- * $Header: /usr/build/vile/vile/RCS/buffer.c,v 1.263 2003/06/18 22:14:57 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/buffer.c,v 1.294 2005/07/17 14:56:19 tom Exp $
  *
  */
 
@@ -311,7 +311,7 @@ FreeBuffer(BUFFER *bp)
 
 #if !WINMARK
     if (is_header_line(MK, bp)) {
-	MK.l = null_ptr;
+	MK.l = 0;
 	MK.o = 0;
     }
 #endif
@@ -323,7 +323,7 @@ FreeBuffer(BUFFER *bp)
 #if OPT_HILITEMATCH
 	clobber_save_curbp(bp);
 #endif
-#if OPT_PERL || OPT_TCL
+#if OPT_PERL || OPT_TCL || OPT_PLUGIN
 	api_free_private(bp->b_api_private);
 #endif
 	free((char *) bp);	/* Release buffer block */
@@ -358,23 +358,6 @@ hist_lookup(int c)
     return valid_buffer(bp) ? bp->b_bname : 0;
 }
 
-/* returns the buffer corresponding to the given number in the history */
-static int
-lookup_hist(BUFFER *bp1)
-{
-    BUFFER *bp;
-    int count = -1;
-
-    for_each_buffer(bp) {
-	if (!b_is_temporary(bp)) {
-	    count++;
-	    if (bp == bp1)
-		return count;
-	}
-    }
-    return -1;			/* no match */
-}
-
 /*
  * Run the $buffer-hook procedure, if it is defined.  Note that we must not do
  * this if curwp->w_bufp != curbp, since that would break the use of DOT and MK
@@ -395,37 +378,47 @@ run_buffer_hook(void)
 #define run_buffer_hook()	/*EMPTY */
 #endif
 
+#define HIST_SHOW 9
+
 static int
-hist_show(void)
+hist_show(int lo_limit, int hi_limit, int cycle)
 {
-    BUFFER *bp;
-    int i = 0;
-    char line[NLINE];
+    int i;
+    int first = -1;
+    char line[(HIST_SHOW + 1) * (NBUFN + 10)];
     BUFFER *abp = (BUFFER *) 0;
 
+    TRACE((T_CALLED "hist_show(%d..%d) %d\n", lo_limit, hi_limit, cycle));
+
     if (!global_g_val(GMDABUFF))
 	abp = find_alt();
 
     (void) strcpy(line, "");
-    for_each_buffer(bp) {
-	if (!b_is_temporary(bp)) {
-	    if (bp != curbp) {	/* don't bother with current */
-		(void) lsprintf(line + strlen(line), "  %d%s%s %s",
-				i,
-				b_is_changed(bp) ? "*" : "",
-				(abp && abp == bp) ? "#" : "",
-				bp->b_bname);
-	    }
-	    if (++i > 9)	/* limit to single-digit */
-		break;
+    for (i = lo_limit; i <= hi_limit; ++i) {
+	int j = ((i - lo_limit) + cycle) % (hi_limit + 1 - lo_limit) + lo_limit;
+	BUFFER *bp = find_b_hist(j);
+
+	if (bp != 0 && bp != curbp) {
+	    if (first < 0)
+		first = j;
+	    (void) lsprintf(line + strlen(line), "  %d%s%s %s",
+			    j,
+			    b_is_changed(bp) ? "*" : "",
+			    (abp && abp == bp) ? "#" : "",
+			    bp->b_bname);
 	}
     }
-    if (strcmp(line, "")) {
+    if (first >= 0) {
 	mlforce("%s", line);
-	return TRUE;
-    } else {
-	return FALSE;
     }
+    returnCode(first);
+}
+
+static int
+valid_history_index(int hist)
+{
+    BUFFER *bp;
+    return ((bp = find_b_hist(hist)) != 0 && bp != curbp);
 }
 
 /*
@@ -450,23 +443,47 @@ histbuff0(int f, int n, int this_window)
 {
     int thiskey, c;
     BUFFER *bp = 0;
+    int cycle = 0;
+    int first;
     char *bufn;
 
     if (f == FALSE) {
-	if (!hist_show())
-	    return FALSE;
-	thiskey = lastkey;
-	c = keystroke8();
-	mlerase();
-	if (c == thiskey) {
-	    c = lookup_hist(bp = find_alt());
-	} else if (isDigit(c)) {
-	    c = c - '0';
-	} else {
-	    if (!isreturn(c))
-		unkeystroke(c);
-	    return FALSE;
+	int hi_limit = HIST_SHOW;
+	int lo_limit = 1;
+
+	/* verify that we have a list which we can display */
+	while (!valid_history_index(hi_limit) && hi_limit > 1) {
+	    --hi_limit;
 	}
+	if (valid_history_index(0))
+	    lo_limit = 0;
+	if (lo_limit > hi_limit)
+	    return FALSE;
+
+	thiskey = lastkey;
+	do {
+	    if ((first = hist_show(lo_limit, hi_limit, cycle)) < 0)
+		return FALSE;
+	    c = keystroke();
+	    mlerase();
+	    if (c == KEY_Tab) {
+		if (++cycle >= hi_limit)
+		    cycle = 0;
+		c = -1;
+	    } else if (isBackTab(c)) {
+		if (--cycle < 0)
+		    cycle = hi_limit - 1;
+		c = -1;
+	    } else if (c == thiskey) {
+		c = first;
+	    } else if (isDigit(c)) {
+		c = c - '0';
+	    } else {
+		if (!isreturn(c))
+		    unkeystroke(c);
+		return FALSE;
+	    }
+	} while (c < 0);
     } else {
 	c = n;
     }
@@ -502,7 +519,6 @@ histbuff(int f, int n)
     if (!clexec)
 	vile_is_busy = FALSE;
     return (rc);
-
 }
 
 /*
@@ -577,18 +593,21 @@ imply_alt(char *fname, int copy, int lockfl)
     char nfname[NFILEN];
     char *stripped;
 
-    if (interrupted() || fname == 0)	/* didn't really have a filename */
-	return;
-    if (nested)
-	return;
+    TRACE((T_CALLED "imply_alt(fname=%s, copy=%d, lockfl=%d)\n",
+	   NONNULL(fname), copy, lockfl));
+
+    if (interrupted()
+	|| isEmpty(fname)
+	|| nested)
+	returnVoid();
 
     if ((stripped = is_appendname(fname)) != 0)
 	fname = stripped;
 
     /* if fname is a pipe cmd, it can be arbitrarily long */
     vl_strncpy(nfname, fname, sizeof(nfname));
-    nfname[sizeof(nfname) - 1] = '\0';
     (void) lengthen_path(nfname);
+
     if (global_g_val(GMDIMPLYBUFF)
 	&& valid_buffer(curbp)
 	&& curbp->b_fname != 0
@@ -596,64 +615,80 @@ imply_alt(char *fname, int copy, int lockfl)
 	&& !isInternalName(fname)) {
 	savebp = curbp;
 	if ((bp = find_b_file(nfname)) == 0) {
-	    L_NUM top, now;
+	    L_NUM top = -1;
+	    L_NUM now = -1;
 
 	    if ((bp = make_bp(fname, 0)) == 0) {
 		mlforce("[Cannot create buffer]");
-		return;
-	    }
+	    } else {
+		/* fill the buffer */
+		b_clr_flags(bp, BFINVS | BFCHG);
+		b_set_flags(bp, BFIMPLY);
+		bp->b_active = TRUE;
+		ch_fname(bp, nfname);
+		make_local_b_val(bp, MDNEWLINE);
+		if (valid_window(curwp) && curwp->w_bufp == curbp) {
+		    top = line_no(curbp, curwp->w_line.l);
+		    now = line_no(curbp, DOT.l);
+		}
 
-	    /* fill the buffer */
-	    b_clr_flags(bp, BFINVS | BFCHG);
-	    b_set_flags(bp, BFIMPLY);
-	    bp->b_active = TRUE;
-	    ch_fname(bp, nfname);
-	    make_local_b_val(bp, MDNEWLINE);
-	    if (valid_window(curwp) && curwp->w_bufp == curbp) {
-		top = line_no(curbp, curwp->w_line.l);
-		now = line_no(curbp, DOT.l);
-	    } else
-		top = now = -1;
-
-	    if (copy) {
-		for_each_line(lp, savebp) {
-		    if (addline(bp, lp->l_text, lp->l_used) != TRUE) {
-			mlforce("[Copy-buffer failed]");
-			return;
+		if (copy) {
+		    for_each_line(lp, savebp) {
+			if (addline(bp, lp->l_text, lp->l_used) != TRUE) {
+			    mlforce("[Copy-buffer failed]");
+			    zotbuf(bp);
+			    bp = 0;
+			    break;
+			}
+		    }
+		    if (bp != 0) {
+			set_b_val(bp, MDNEWLINE, b_val(savebp, MDNEWLINE));
+		    }
+		} else {
+		    if (readin(fname, lockfl, bp, FALSE) != TRUE) {
+			zotbuf(bp);
+			bp = 0;
 		    }
 		}
-		set_b_val(bp, MDNEWLINE, b_val(savebp, MDNEWLINE));
-	    } else {
-		readin(fname, lockfl, bp, FALSE);
 	    }
 
 	    /* setup so that buffer-toggle works as in vi (i.e.,
 	     * the top/current lines of the screen are the same).
 	     */
-	    if (now >= 0) {
-		for_each_line(lp, bp) {
-		    if (--now == 0) {
-			bp->b_dot.l = lp;
-			bp->b_dot.o = curbp->b_dot.o;
-			break;
-		    }
-		    if (--top == 0) {
-			bp->b_wline.l = lp;
+	    if (bp != 0) {
+		if (now >= 0) {
+		    for_each_line(lp, bp) {
+			if (--now == 0) {
+			    bp->b_dot.l = lp;
+			    bp->b_dot.o = curbp->b_dot.o;
+			    break;
+			}
+			if (--top == 0) {
+			    bp->b_wline.l = lp;
+			}
 		    }
 		}
 	    }
 	} else if (copy && !b_is_changed(bp)) {
 	    nested++;
-	    if (bp->b_active == TRUE)
-		bp2readin(bp, TRUE);
+	    if (bp->b_active == TRUE) {
+		if (bp2readin(bp, TRUE) != TRUE) {
+		    zotbuf(bp);
+		    bp = 0;
+		}
+	    }
 	    nested--;
 	}
+
 	DisableHook(&bufhook);
-	make_current(bp);
-	infer_majormode(bp);
+	if (bp != 0) {
+	    make_current(bp);
+	    infer_majormode(bp);
+	}
 	make_current(savebp);
 	EnableHook(&bufhook);
     }
+    returnVoid();
 }
 
 /* switch back to the most recent buffer */
@@ -679,6 +714,16 @@ qs_bname_cmp(const void *a, const void *b)
     return strcmp(*(const char *const *) a, (*(const char *const *) b));
 }
 
+static void
+free_bname_cmpl(char **list)
+{
+    int n;
+    for (n = 0; list[n] != 0; ++n) {
+	free(list[n]);
+    }
+    free((char *) list);
+}
+
 static char **
 init_bname_cmpl(void)
 {
@@ -691,8 +736,16 @@ init_bname_cmpl(void)
 	beginDisplay();
 	if ((list = typeallocn(char *, count + 1)) != 0) {
 	    for_each_buffer(bp) {
-		list[used++] = strmalloc(add_backslashes(bp->b_bname));
+		list[used] = strmalloc(add_backslashes(bp->b_bname));
+		if (list[used] == 0) {
+		    free_bname_cmpl(list);
+		    list = 0;
+		    break;
+		}
+		++used;
 	    }
+	}
+	if (list != 0) {
 	    list[used] = 0;
 	    qsort(list, used, sizeof(char *), qs_bname_cmp);
 	}
@@ -701,16 +754,6 @@ init_bname_cmpl(void)
     return list;
 }
 
-static void
-free_bname_cmpl(char **list)
-{
-    int n;
-    for (n = 0; list[n] != 0; ++n) {
-	free(list[n]);
-    }
-    free((char *) list);
-}
-
 static int
 bname_complete(DONE_ARGS)
 {
@@ -754,6 +797,30 @@ ask_for_bname(char *prompt, char *bufn, size_t len)
     return status;
 }
 
+/*
+ * Prompt for a buffer name, ignoring if we cannot find it (creating it then).
+ */
+#define ASK_FOR_NEW_BNAME(prompt, status, bufn, bp) \
+    bufn[0] = EOS; \
+    if ((status = ask_for_bname(prompt, bufn, sizeof(bufn))) != TRUE \
+	&& (status != SORTOFTRUE)) \
+	returnCode(status); \
+    if ((bp = find_buffer(bufn)) == 0) { \
+	bp = bfind(bufn, 0); \
+    }
+
+/*
+ * Prompt for a existing-buffer name, warning if we cannot find it.
+ */
+#define ASK_FOR_OLD_BNAME(prompt, status, bufn, bp) \
+    bufn[0] = EOS; \
+    if ((status = ask_for_bname(prompt, bufn, sizeof(bufn))) != TRUE \
+	&& (status != SORTOFTRUE)) \
+	returnCode(status); \
+    if ((bp = find_any_buffer(bufn)) == 0) { \
+	returnCode(FALSE); \
+    }
+
 /*
  * Attach a buffer to a window. The
  * values of dot and mark come from the buffer
@@ -769,17 +836,13 @@ usebuffer(int f GCC_UNUSED, int n GCC_UNUSED)
     char bufn[NBUFN];
 
     TRACE((T_CALLED "usebuffer()\n"));
-    bufn[0] = EOS;
-    if ((s = ask_for_bname("Use buffer: ", bufn, sizeof(bufn))) != TRUE
-	&& (s != SORTOFTRUE))
-	returnCode(s);
-    if ((bp = find_any_buffer(bufn)) == 0)	/* Try buffer */
-	returnCode(FALSE);
+    ASK_FOR_OLD_BNAME("Use buffer: ", s, bufn, bp);
     returnCode(swbuffer(bp));
 }
 
 /*
- * Just like "select-buffer", but we will create the buffer if it does not exist.
+ * Just like "select-buffer", but we will create the buffer if it does not
+ * exist.
  */
 int
 edit_buffer(int f GCC_UNUSED, int n GCC_UNUSED)
@@ -789,18 +852,14 @@ edit_buffer(int f GCC_UNUSED, int n GCC_UNUSED)
     char bufn[NBUFN];
 
     TRACE((T_CALLED "edit_buffer()\n"));
-    bufn[0] = EOS;
-    if ((s = ask_for_bname("Use buffer: ", bufn, sizeof(bufn))) != TRUE
-	&& (s != SORTOFTRUE))
-	returnCode(s);
-    if ((bp = find_buffer(bufn)) == 0) {
-	bp = bfind(bufn, 0);
-    }
+    ASK_FOR_NEW_BNAME("Edit buffer: ", s, bufn, bp);
     returnCode(swbuffer(bp));
 }
 
 #if !SMALLER
-/* ARGSUSED */
+/*
+ * Set the current window to display the given buffer.
+ */
 int
 set_window(int f GCC_UNUSED, int n GCC_UNUSED)
 {
@@ -808,15 +867,43 @@ set_window(int f GCC_UNUSED, int n GCC_UNUSED)
     int s;
     char bufn[NBUFN];
 
-    bufn[0] = EOS;
-    if ((s = ask_for_bname("Set window to buffer: ", bufn, sizeof(bufn))) != TRUE
-	&& (s != SORTOFTRUE))
-	return s;
-    if ((bp = find_any_buffer(bufn)) == 0)	/* Try buffer */
-	return FALSE;
-    return swbuffer_lfl(bp, TRUE, TRUE);
+    TRACE((T_CALLED "set_window()\n"));
+    ASK_FOR_OLD_BNAME("Set window to buffer: ", s, bufn, bp);
+    returnCode(swbuffer_lfl(bp, TRUE, TRUE));
 }
-#endif
+
+/*
+ * Open a window for the given buffer.
+ */
+int
+popup_buffer(int f GCC_UNUSED, int n GCC_UNUSED)
+{
+    BUFFER *bp;
+    int s;
+    char bufn[NBUFN];
+
+    TRACE((T_CALLED "popup_buffer()\n"));
+    ASK_FOR_OLD_BNAME("Popup buffer: ", s, bufn, bp);
+    if (bp->b_nwnd == 0)
+	returnCode(popupbuff(bp));
+    returnCode(TRUE);
+}
+
+/*
+ * Close all windows which display the given buffer.
+ */
+int
+popdown_buffer(int f GCC_UNUSED, int n GCC_UNUSED)
+{
+    BUFFER *bp;
+    int s;
+    char bufn[NBUFN];
+
+    TRACE((T_CALLED "popdown_buffer()\n"));
+    ASK_FOR_OLD_BNAME("Popdown buffer: ", s, bufn, bp);
+    returnCode(zotwp(bp));
+}
+#endif /* !SMALLER */
 
 /* switch back to the first buffer (i.e., ":rewind") */
 /* ARGSUSED */
@@ -955,7 +1042,7 @@ swbuffer(BUFFER *bp)
 static int
 suckitin(BUFFER *bp, int copy, int lockfl)
 {
-    int s = TRUE;
+    int s = FALSE;
 
     TRACE((T_CALLED "suckitin(%s, %s)\n", bp->b_bname, copy ? "copy" : "new"));
     if (copy) {
@@ -974,7 +1061,7 @@ suckitin(BUFFER *bp, int copy, int lockfl)
     }
 #ifdef MDCHK_MODTIME
     else
-	(void) check_file_changed(bp, bp->b_fname);
+	s = check_file_changed(bp, bp->b_fname);
 #endif
     updatelistbuffers();
     run_buffer_hook();
@@ -1015,7 +1102,7 @@ swbuffer_lfl(BUFFER *bp, int lockfl, int this_window)
     MK = nullmark;
 #endif
 
-    if (curbp) {
+    if (valid_buffer(curbp)) {
 	/* if we'll have to take over this window, and it's the last */
 	if ((this_window || bp->b_nwnd == 0)
 	    && curbp->b_nwnd != 0 && --(curbp->b_nwnd) == 0) {
@@ -1090,7 +1177,7 @@ swbuffer_lfl(BUFFER *bp, int lockfl, int this_window)
 	else
 	    clone_window(curwp, wp);
     } else
-#endif
+#endif /* !SMALLER */
 	/* get it already on the screen if possible */
     if (bp->b_nwnd != 0) {	/* then it's on the screen somewhere */
 	if ((wp = bp2any_wp(bp)) == 0)
@@ -1231,6 +1318,147 @@ kill_that_buffer(BUFFER *bp)
     return s;
 }
 
+static void
+free_buffer_list(char **list)
+{
+    int n;
+    for (n = 0; list[n] != 0; ++n) {
+	free(list[n]);
+    }
+    free(list);
+}
+
+/*
+ * Make a list of buffer names to process.
+ */
+static char **
+make_buffer_list(char *bufn)
+{
+    BUFFER *bp;
+    int count = 0;
+    int limit = countBuffers();
+    char **result = typecallocn(char *, limit + 1);
+
+    if (result != 0) {
+#if OPT_FORBUFFERS_CHOICES
+#if OPT_GLOB_RANGE
+	/*
+	 * If it looks like a scratch buffer, try to find exactly that before
+	 * possibly getting confused with range expressions, etc.
+	 */
+	if (global_g_val(GVAL_FOR_BUFFERS) == FB_MIXED
+	    && is_scratchname(bufn)
+	    && (bp = find_b_name(bufn)) != 0) {
+	    result[count++] = strmalloc(bp->b_bname);
+	} else
+#endif
+	    if (global_g_val(GVAL_FOR_BUFFERS) == FB_MIXED
+		|| global_g_val(GVAL_FOR_BUFFERS) == FB_GLOB) {
+	    for_each_buffer(bp) {
+		if (glob_match_leaf(bp->b_bname, bufn)) {
+		    result[count++] = strmalloc(bp->b_bname);
+		}
+	    }
+	} else if (global_g_val(GVAL_FOR_BUFFERS) == FB_REGEX) {
+	    regexp *exp;
+
+	    if ((exp = regcomp(bufn, strlen(bufn), TRUE)) != 0) {
+		for_each_buffer(bp) {
+		    if (regexec(exp, bp->b_bname, (char *) 0, 0, -1)) {
+			result[count++] = strmalloc(bp->b_bname);
+		    }
+		}
+		free(exp);
+	    }
+	}
+	if (count == 0) {
+	    if ((bp = find_buffer(bufn)) != 0) {	/* Try buffer */
+		result[count++] = strmalloc(bp->b_bname);
+	    }
+	}
+#else
+	if ((bp = find_buffer(bufn)) != 0) {	/* Try buffer */
+	    result[count++] = strmalloc(bp->b_bname);
+	}
+#endif
+	if (count == 0) {
+	    free_buffer_list(result);
+	    result = 0;
+	}
+    }
+    return result;
+}
+
+#if !SMALLER
+/*
+ * Select a set of buffers with the first parameter, and prompt for a command
+ * to execute on each buffer.
+ */
+int
+for_buffers(int f, int n)
+{
+    BUFFER *bp;
+    int s;
+    char bufn[NFILEN];
+    char command[NSTRING];
+    char **list;
+    int inlist;
+    int updated = 0;
+
+    bufn[0] = EOS;
+    if ((s = ask_for_bname("Buffers: ", bufn, sizeof(bufn))) != TRUE)
+	return s;
+
+    if ((list = make_buffer_list(bufn)) != 0) {
+	*command = EOS;
+	/*
+	 * Don't do command-completion, since we want to allow range
+	 * expressions, e.g.,
+	 *      :fb *.h %trim
+	 *
+	 * FIXME: split-out logic in execute_named_command() to allow proper
+	 * parsing to linespec and command-verb.
+	 */
+	if ((s = kbd_string("Command: ",
+			    command,
+			    sizeof(command),
+			    '\n',
+			    KBD_QUOTES,
+			    no_completion)) == TRUE) {
+	    char *save_execstr = execstr;
+	    int save_clexec = clexec;
+
+	    for (inlist = 0; list[inlist] != 0; ++inlist) {
+		if ((bp = find_b_name(list[inlist])) != 0
+		    && swbuffer(bp)) {
+		    char this_command[NSTRING];
+
+		    /*
+		     * FIXME: does namedcmd() modify what execstr points to?
+		     */
+		    execstr = strcpy(this_command, command);
+		    clexec = TRUE;
+
+		    if (namedcmd(f, n))
+			++updated;
+		}
+	    }
+	    execstr = save_execstr;
+	    clexec = save_clexec;
+	}
+	free_buffer_list(list);
+    } else {
+	mlforce("[No matching buffers]");
+    }
+
+    if (updated) {
+	mlforce("[Processed %d buffer%s]", updated, PLURAL(updated));
+	s = TRUE;
+    }
+    return s;
+}
+#endif /* !SMALLER */
+
 /*
  * Dispose of a buffer, by name.  If this is a screen-command, pick the name up
  * from the screen.  Otherwise, ask for the name.  Look it up (don't get too
@@ -1251,11 +1479,15 @@ killbuffer(int f, int n)
     BUFFER *bp;
     int s;
     char bufn[NFILEN];
+    char **list;
+    int inlist;
+    int removed = 0;
 
 #if OPT_UPBUFF
     C_NUM save_COL;
     MARK save_DOT;
     MARK save_TOP;
+
     int animated = (f
 		    && valid_window(curwp)
 		    && (n > 1)
@@ -1277,34 +1509,18 @@ killbuffer(int f, int n)
     if (!f)
 	n = 1;
     for_ever {
-	BUFFER *bp_next;
 	bufn[0] = EOS;
 	if ((s = ask_for_bname("Kill buffer: ", bufn, sizeof(bufn))) != TRUE)
 	    break;
 
-#if !SMALLER			/* allow user to kill a glob'd expression. */
-	s = 0;
-	for (bp = bheadp; bp; bp = bp_next) {
-	    /* Fetch next buffer prior to possible destruction */
-	    bp_next = bp->b_bufp;
-	    /* Attempt match and kill if match okay */
-	    if (glob_match_leaf(bp->b_bname, bufn)) {
-		if (kill_that_buffer(bp))
-		    s++;
+	if ((list = make_buffer_list(bufn)) != 0) {
+	    for (inlist = 0; list[inlist] != 0; ++inlist) {
+		if ((bp = find_b_name(list[inlist])) != 0
+		    && kill_that_buffer(bp)) {
+		    ++removed;
+		}
 	    }
-	}
-	if (s != 0) {
-	    mlforce("[Removed %d buffer%s]", s, PLURAL(s));
-	    s = TRUE;
-	} else			/* perhaps filename or buffer number */
-#endif
-	{
-	    if ((bp = find_any_buffer(bufn)) == 0) {	/* Try buffer */
-		s = FALSE;
-		break;
-	    }
-	    if ((s = kill_that_buffer(bp)) != TRUE)
-		break;
+	    free_buffer_list(list);
 	}
 
 	if (--n > 0) {
@@ -1319,6 +1535,10 @@ killbuffer(int f, int n)
 	    break;
     }
 
+    if (removed) {
+	mlforce("[Removed %d buffer%s]", removed, PLURAL(removed));
+	s = TRUE;
+    }
 #if OPT_UPBUFF
     if (animated && !special) {
 	curgoal = save_COL;
@@ -1377,12 +1597,14 @@ strip_brackets(char *dst, const char *src)
 char *
 add_brackets(char *dst, const char *src)
 {
-    size_t len = strlen(src);
-    if (len > NBUFN - 3)
-	len = NBUFN - 3;
-    dst[0] = SCRTCH_LEFT[0];
-    (void) strncpy(&dst[1], src, len);
-    (void) strcpy(&dst[len + 1], SCRTCH_RIGHT);
+    if (dst != 0 && src != 0) {
+	size_t len = strlen(src);
+	if (len > NBUFN - 3)
+	    len = NBUFN - 3;
+	dst[0] = SCRTCH_LEFT[0];
+	(void) strncpy(&dst[1], src, len);
+	(void) strcpy(&dst[len + 1], SCRTCH_RIGHT);
+    }
     return dst;
 }
 
@@ -1533,7 +1755,7 @@ popupbuff(BUFFER *bp)
     WINDOW *wp;
 
     TRACE((T_CALLED "popupbuff(%s) nwnd=%d\n", bp->b_bname, bp->b_nwnd));
-    if (!curbp) {
+    if (!valid_buffer(curbp)) {
 	curbp = bp;		/* possibly at startup time */
 	curwp->w_bufp = curbp;
 	++curbp->b_nwnd;
@@ -1682,7 +1904,7 @@ footnote(int c)
 #else
 #define	MakeNote(c)	bputc(c)
 #define	ShowNotes()
-#endif
+#endif /* !SMALLER */
 
 /*
  * This routine rebuilds the text in the buffer that holds the buffer list.  It
@@ -1696,12 +1918,12 @@ static void
 makebufflist(int unused GCC_UNUSED, void *dummy GCC_UNUSED)
 {
     BUFFER *bp;
-    LINEPTR curlp;		/* entry corresponding to buffer-list */
+    LINE *curlp;		/* entry corresponding to buffer-list */
     int nbuf = 0;		/* no. of buffers */
     int this_or_that;
     char temp[NFILEN];
 
-    curlp = null_ptr;
+    curlp = 0;
 
     if (this_bp == 0)
 	this_bp = curbp;
@@ -1710,7 +1932,13 @@ makebufflist(int unused GCC_UNUSED, void *dummy GCC_UNUSED)
 
     bprintf("      %7s %*s %s\n", "Size", NBUFN - 1, "Buffer name",
 	    "Contents");
-    bprintf("      %7Q %*Q %30Q\n", '-', NBUFN - 1, '-', '-');
+    bpadc(' ', 6);
+    bpadc('-', 7);
+    bputc(' ');
+    bpadc('-', NBUFN - 1);
+    bputc(' ');
+    bpadc('-', term.cols - DOT.o);
+    bputc('\n');
 
     /* output the list of buffers */
     for_each_buffer(bp) {
@@ -1740,7 +1968,7 @@ makebufflist(int unused GCC_UNUSED, void *dummy GCC_UNUSED)
 	    bprintf(" %2d%c ", nbuf++, this_or_that);
 
 	(void) bsizes(bp);
-	bprintf("%7ld %*S ", bp->b_bytecount, NBUFN - 1, bp->b_bname);
+	bprintf("%7lu %.*s ", bp->b_bytecount, NBUFN - 1, bp->b_bname);
 	{
 	    char *p;
 
@@ -1759,9 +1987,9 @@ makebufflist(int unused GCC_UNUSED, void *dummy GCC_UNUSED)
 	    current_directory(FALSE));
 
     /* show the actual size of the buffer-list */
-    if (curlp != null_ptr) {
+    if (curlp != 0) {
 	(void) bsizes(curbp);
-	(void) lsprintf(temp, "%7ld", curbp->b_bytecount);
+	(void) lsprintf(temp, "%7lu", curbp->b_bytecount);
 	(void) memcpy(curlp->l_text + 6, temp, strlen(temp));
     }
 }
@@ -1893,17 +2121,17 @@ addline(BUFFER *bp, const char *text, int len)
  * Add a LINE filled with the given text after the specified LINE.
  */
 int
-add_line_at(BUFFER *bp, LINEPTR prevp, const char *text, int len)
+add_line_at(BUFFER *bp, LINE *prevp, const char *text, int len)
 {
-    LINEPTR newlp;
-    LINEPTR nextp;
+    LINE *newlp;
+    LINE *nextp;
     LINE *lp;
     int ntext;
 
     nextp = lforw(prevp);
-    ntext = (len < 0) ? (int) strlen(text) : len;
+    ntext = (len < 0) ? (int) strlen(NONNULL(text)) : len;
     newlp = lalloc(ntext, bp);
-    if (newlp == null_ptr)
+    if (newlp == 0)
 	return (FALSE);
 
     lp = newlp;
@@ -1964,7 +2192,7 @@ next_buffer_line(const char *bname)
 
     /* how much left on the current line? */
     blen = llength(bp->b_dot.l);
-    if (blen > bp->b_dot.o)
+    if (blen > (B_COUNT) bp->b_dot.o)
 	blen -= bp->b_dot.o;
     else
 	blen = 0;
@@ -2085,7 +2313,7 @@ BUFFER *
 bfind(const char *bname, UINT bflag)
 {
     BUFFER *bp;
-    LINEPTR lp;
+    LINE *lp;
     BUFFER *lastb = NULL;	/* buffer to insert after */
     BUFFER *bp2;
 
@@ -2105,7 +2333,7 @@ bfind(const char *bname, UINT bflag)
 	set_bname(bp, bname);
 
 	lp = lalloc(0, bp);
-	if (lp == null_ptr) {
+	if (lp == 0) {
 	    FreeAndNull(bp);
 	    (void) no_memory("BUFFER head");
 	} else {
@@ -2138,10 +2366,10 @@ bfind(const char *bname, UINT bflag)
 	    } else
 		bp->b_cryptkey[0] = EOS;
 #endif
-	    bp->b_udstks[0] = bp->b_udstks[1] = null_ptr;
-	    bp->b_ulinep = null_ptr;
-	    bp->b_udtail = null_ptr;
-	    bp->b_udlastsep = null_ptr;
+	    bp->b_udstks[0] = bp->b_udstks[1] = 0;
+	    bp->b_ulinep = 0;
+	    bp->b_udtail = 0;
+	    bp->b_udlastsep = 0;
 
 	    b_set_counted(bp);	/* buffer is empty */
 	    set_lforw(lp, lp);
@@ -2162,6 +2390,7 @@ bfind(const char *bname, UINT bflag)
 #if OPT_PERL || OPT_TCL
 	    bp->b_api_private = 0;
 #endif
+	    set_record_sep(bp, (RECORD_SEP) global_b_val(VAL_RECORD_SEP));
 	}
     }
     endofDisplay();
@@ -2213,7 +2442,7 @@ make_ro_bp(const char *bname, UINT flags)
 int
 bclear(BUFFER *bp)
 {
-    LINEPTR lp;
+    LINE *lp;
 
     if (!b_is_temporary(bp)	/* Not invisible or scratch */
 	&&b_is_changed(bp)) {	/* Something changed    */
@@ -2240,9 +2469,9 @@ bclear(BUFFER *bp)
 	lfree(lp, bp);
     }
 
-    if (bp->b_ulinep != null_ptr) {
+    if (bp->b_ulinep != 0) {
 	lfree(bp->b_ulinep, bp);
-	bp->b_ulinep = null_ptr;
+	bp->b_ulinep = 0;
     }
     FreeAndNull(bp->b_ltext);
     bp->b_ltext_end = NULL;
@@ -2281,29 +2510,32 @@ bclear(BUFFER *bp)
 int
 bsizes(BUFFER *bp)
 {
-    LINE *lp;			/* current line */
-    B_COUNT numchars = 0;	/* # of chars in file */
-    L_NUM numlines = 0;		/* # of lines in file */
-    L_NUM ending = len_record_sep(bp);
+    int code = FALSE;
 
-    if (b_is_counted(bp))
-	return FALSE;
+    if (valid_buffer(curbp)
+	&& !b_is_counted(bp)) {
+	LINE *lp;		/* current line */
+	B_COUNT numchars = 0;	/* # of chars in file */
+	L_NUM numlines = 0;	/* # of lines in file */
+	L_NUM ending = len_record_sep(bp);
 
-    /* count chars and lines */
-    for_each_line(lp, bp) {
-	++numlines;
-	numchars += llength(lp) + ending;
+	/* count chars and lines */
+	for_each_line(lp, bp) {
+	    ++numlines;
+	    numchars += llength(lp) + ending;
 #if !SMALLER			/* tradeoff between codesize & data */
-	lp->l_number = numlines;
+	    lp->l_number = numlines;
 #endif
-    }
-    if (!b_val(bp, MDNEWLINE))
-	numchars -= ending;
+	}
+	if (!b_val(bp, MDNEWLINE))
+	    numchars -= ending;
 
-    bp->b_bytecount = numchars;
-    bp->b_linecount = numlines;
-    b_set_counted(bp);
-    return TRUE;
+	bp->b_bytecount = numchars;
+	bp->b_linecount = numlines;
+	b_set_counted(bp);
+	code = TRUE;
+    }
+    return code;
 }
 
 /*
