@@ -5,7 +5,7 @@
  * functions use hints that are left in the windows by the commands.
  *
  *
- * $Header: /usr/build/vile/vile/RCS/display.c,v 1.380 2003/07/27 16:53:03 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/display.c,v 1.407 2005/06/17 23:32:03 tom Exp $
  *
  */
 
@@ -18,6 +18,7 @@
 #define vMAXNEG (-vMAXINT)	/* 0x80000001 */
 
 #define	NU_WIDTH 8
+#define NU_GUTTER 1
 
 #define reset_term_attrs() term.rev(0)
 
@@ -96,8 +97,8 @@ static char *
 right_num(char *buffer, int len, long value)
 {
     char temp[NSTRING];
-    register char *p = lsprintf(temp, "%ld", value);
-    register char *q = buffer + len;
+    char *p = lsprintf(temp, "%ld", value);
+    char *q = buffer + len;
 
     *q = EOS;
     while (q != buffer)
@@ -106,18 +107,27 @@ right_num(char *buffer, int len, long value)
 }
 
 /*
- * Do format a string.
+ * Do format a string.  Return the number of bytes by which 'width' exceeds
+ * that actually written to the output.
  */
 static int
-dfputsn(OutFunc outfunc, const char *s, int n)
+dfputsn(OutFunc outfunc, const char *s, int width, int limit)
 {
-    int c;
     int l = 0;
 
-    TRACE2(("...str=%s\n", visible_buff(s, (n < 0) ? strlen(s) : n, TRUE)));
     if (s != 0) {
-	while ((n-- != 0) && ((c = *s++) != EOS)) {
-	    (*outfunc) (c);
+	int length = strlen(s);
+
+	if (width < 0)
+	    width = length;
+	if (limit > 0 && width > limit)
+	    width = limit;
+	if (width > length)
+	    width = length;
+
+	TRACE2(("...str=%s\n", visible_buff(s, width, TRUE)));
+	while (width-- > 0) {
+	    (*outfunc) (*s++);
 	    l++;
 	}
     }
@@ -128,7 +138,7 @@ dfputsn(OutFunc outfunc, const char *s, int n)
 static int
 dfputs(OutFunc outfunc, const char *s)
 {
-    return dfputsn(outfunc, s, -1);
+    return dfputsn(outfunc, s, -1, -1);
 }
 
 /*
@@ -190,46 +200,76 @@ dfputf(OutFunc outfunc, double s)
     return n;
 }
 
+/*
+ * On entry, *fmt may point to either '*' or a digit.  If either, decode the
+ * appropriate width or limit.
+ */
+static void
+decode_length(const char **fmt, va_list *app, int *result)
+{
+    int c = **fmt;
+    int value = 0;
+    int found = FALSE;
+
+    if (c == '*') {
+	value = va_arg(*app, int);
+	found = TRUE;
+	(*fmt)++;
+    } else {
+	while (isDigit(c)) {
+	    value = (value * 10) + c - '0';
+	    found = TRUE;
+	    (*fmt)++;
+	    c = **fmt;
+	}
+    }
+    if (found)
+	*result = value;
+}
+
 /*
  * Generic string formatter.  Takes printf-like args, and calls
  * the global function (*dfoutfn)(c) for each c
  */
 static void
-dofmt(const char *fmt, va_list * app)
+dofmt(const char *fmt, va_list *app)
 {
-    register int c;		/* current char in format string */
-    register int wid;
-    register int n;
-    register int nchars = 0;
+    int c;			/* current char in format string */
+    int the_width;
+    int the_limit;
+    int n;
     int islong;
+    UINT uint_value;
+    ULONG ulong_value;
     int int_value;
     long long_value;
-    UINT radix;
     OutFunc outfunc = dfoutfn;	/* local copy, for recursion */
 
-    TRACE2(("dofmt fmt='%s'\n", fmt));
+    TRACE2(("dofmt fmt='%s'\n", visible_buff(fmt, strlen(fmt), FALSE)));
     while ((c = *fmt++) != 0) {
 	if (c != '%') {
 	    (*outfunc) (c);
-	    nchars++;
 	    continue;
 	}
-	c = *fmt++;
-	wid = 0;
+	the_width = -1;
 	islong = FALSE;
-	if (c == '*') {
-	    wid = va_arg(*app, int);
-	    c = *fmt++;
+	decode_length(&fmt, app, &the_width);
+	if (*fmt == '.') {
+	    ++fmt;
+	    decode_length(&fmt, app, &the_limit);
+	    if (the_width < 0)
+		the_width = the_limit;
 	} else {
-	    while (isDigit(c)) {
-		wid = (wid * 10) + c - '0';
-		c = *fmt++;
-	    }
+	    the_limit = the_width;
 	}
+	c = *fmt++;
 	if (c == 'l') {
 	    islong = TRUE;
 	    c = *fmt++;
 	}
+	TRACE2(("... fmt='%%%d.%d%c%s'\n",
+		the_width, the_limit,
+		c, islong ? "L" : ""));
 	switch (c) {
 	case EOS:
 	    n = 0;
@@ -249,7 +289,6 @@ dofmt(const char *fmt, va_list * app)
 		    }
 		    int_value = -int_value;
 		    (*outfunc) ('-');
-		    nchars++;
 		}
 		n = dfputi(outfunc, (UINT) int_value, 10);
 		break;
@@ -258,11 +297,20 @@ dofmt(const char *fmt, va_list * app)
 	    if (long_value < 0) {
 		long_value = -long_value;
 		(*outfunc) ('-');
-		nchars++;
 	    }
 	    n = dfputli(outfunc, (ULONG) long_value, 10);
 	    break;
 
+	case 'u':
+	    if (!islong) {
+		uint_value = va_arg(*app, UINT);
+		n = dfputi(outfunc, uint_value, 10);
+		break;
+	    }
+	    ulong_value = va_arg(*app, ULONG);
+	    n = dfputli(outfunc, ulong_value, 10);
+	    break;
+
 	case 'o':
 	    n = dfputi(outfunc, va_arg(*app, UINT), 8);
 	    break;
@@ -277,58 +325,21 @@ dofmt(const char *fmt, va_list * app)
 	    n = dfputli(outfunc, va_arg(*app, ULONG), 16);
 	    break;
 
-	case 'r':
-	case 'R':
-	    radix = va_arg(*app, UINT);
-	    if (radix < 2 || radix > 36)
-		radix = 10;
-	    if (islong || c == 'R')
-		n = dfputli(outfunc,
-			    va_arg(*app, ULONG), radix);
-	    else
-		n = dfputi(outfunc,
-			   va_arg(*app, UINT), radix);
-	    break;
-
 	case 's':
-	    if (wid <= 0) {
-		n = dfputs(outfunc, va_arg(*app, char *));
-		break;
-	    }
-	    /* FALLTHROUGH */
-
-	case 'S':		/* use wid as max width */
-	    n = dfputsn(outfunc, va_arg(*app, char *), wid);
+	    n = dfputsn(outfunc, va_arg(*app, char *), the_width, the_limit);
 	    break;
 
 	case 'f':
 	    n = dfputf(outfunc, va_arg(*app, double));
 	    break;
 
-	case 'P':		/* output padding -- pads total output to
-				   "wid" chars, using c as the pad char */
-	    wid -= nchars;
-	    /* FALLTHROUGH */
-
-	case 'Q':		/* field padding -- puts out "wid"
-				   copies of c */
-	    n = 0;
-	    c = va_arg(*app, int);
-	    while (n < wid) {
-		(*outfunc) (c);
-		n++;
-	    }
-	    break;
-
 	default:
 	    (*outfunc) (c);
 	    n = 1;
 	}
-	wid -= n;
-	nchars += n;
-	while (wid-- > 0) {
+	the_width -= n;
+	while (the_width-- > 0) {
 	    (*outfunc) (' ');
-	    nchars++;
 	}
     }
 }
@@ -366,7 +377,7 @@ set_vattrs(int row, int col, VIDEO_ATTR attr, size_t len)
 #endif
 
 static void
-freeVIDEO(register VIDEO * vp)
+freeVIDEO(VIDEO * vp)
 {
     if (vp != 0) {
 #if OPT_VIDEO_ATTRS
@@ -379,7 +390,7 @@ freeVIDEO(register VIDEO * vp)
 int
 video_alloc(VIDEO ** vpp)
 {
-    register VIDEO *vp;
+    VIDEO *vp;
     /* struct VIDEO already has 4 of the bytes */
     vp = typeallocplus(VIDEO, term.maxcols - 4);
     if (vp == 0)
@@ -401,7 +412,7 @@ video_alloc(VIDEO ** vpp)
 static int
 vtalloc(void)
 {
-    register int i, first;
+    int i, first;
     static int vcols, vrows;
 
     if (term.maxrows > vrows) {
@@ -440,7 +451,7 @@ vtalloc(void)
 static void
 vtfree(void)
 {
-    register int i;
+    int i;
 
     if (vscreen) {
 	for (i = 0; i < term.maxrows; ++i) {
@@ -521,7 +532,7 @@ vtputc(int c)
     /* since we don't allow wrapping on the message line, we only need
      * to evaluate this once.  */
     int lastcol = vtrow == term.rows - 1 ? term.cols - 1 : term.cols;
-    register VIDEO *vp;
+    VIDEO *vp;
 
 #ifdef WMDLINEWRAP
     if (vtrow < 0) {
@@ -601,8 +612,10 @@ static void
 vtputsn(const char *s, int n)
 {
     int c;
-    while (n-- > 0 && (c = *s++) != EOS)
-	vtputc(c);
+    if (s != 0) {
+	while (n-- > 0 && (c = *s++) != EOS)
+	    vtputc(c);
+    }
 }
 
 /*
@@ -610,10 +623,10 @@ vtputsn(const char *s, int n)
  * line-wrap or right-shifting.
  */
 static void
-vtset(LINEPTR lp, WINDOW *wp)
+vtset(LINE *lp, WINDOW *wp)
 {
-    register char *from;
-    register int n = llength(lp);
+    char *from;
+    int n = llength(lp);
     BUFFER *bp = wp->w_bufp;
     int skip = -vtcol;
     int list = w_val(wp, WMDLIST);
@@ -634,21 +647,22 @@ vtset(LINEPTR lp, WINDOW *wp)
     } else
 #endif
     if (w_val(wp, WMDNUMBER)) {
-	register int j, k, jk;
+	int j, k, jk;
 	L_NUM line = line_no(bp, lp);
 	int fill = ' ';
 	char temp[NU_WIDTH + 2];
 
 	vtcol = 0;		/* make sure we always see line numbers */
-	vtputsn(right_num(temp, NU_WIDTH - 2, (long) line), NU_WIDTH - 2);
-	vtputsn("  ", 2);
+	vtputsn(right_num(temp, NU_WIDTH - NU_GUTTER, (int) line),
+		NU_WIDTH - NU_GUTTER);
+	vtputsn("  ", NU_GUTTER);
 	horscroll = skip - vtcol;
 
 	/* account for leading fill; this repeats logic in vtputc so
 	 * I don't have to introduce a global variable... */
 	from = lp->l_text;
 	for (j = k = jk = 0; (j < n) && (k < skip); j++) {
-	    register int c = from[j];
+	    int c = from[j];
 	    if ((list || (c != '\t')) && !isPrint(c)) {
 		if (c & HIGHBIT) {
 		    k += 4;
@@ -723,7 +737,7 @@ vtset(LINEPTR lp, WINDOW *wp)
 	else if (!b_val(bp, MDNEWLINE) && (lforw(lp) == buf_head(bp)))
 	    /*EMPTY */ ;
 	else {
-	    char *s = get_record_sep(bp);
+	    const char *s = get_record_sep(bp);
 	    while (*s != EOS)
 		vtlistc(*s++);
 	}
@@ -767,7 +781,24 @@ upscreen(int f GCC_UNUSED, int n GCC_UNUSED)
 
 static UINT scrflags;
 
-/* line to virtual column */
+/*
+ * Convert MARK's offset to virtual column.  The virtual column counts the left
+ * margin, the sideways shift of the window, etc., up to the given MARK.
+ *
+ * The buffer may have a left-margin, e.g., [Registers].  Use the 'bp'
+ * parameter to get that value.
+ *
+ * To allow this to be used in the OPT_CACHE_VCOL logic, it accepts a 'col'
+ * parameter which in that case may be nonzero, e.g., the nominal column of the
+ * left margin of the display (the column at which the sideways-shift
+ * coincides).
+ *
+ * Finally, use 'adjust' to adjust for cases where the sideways-shift does
+ * not exactly coincide with the 'col', where nonprinting data is displayed
+ * in hex/octal/uparrow format and is split across the display's left margin.
+ * After adding the left margin, that gives the index into the line for
+ * reading bytes whose width is added til we read the mark's offset.
+ */
 int
 mk_to_vcol(MARK mark, int expanded, BUFFER *bp, int col, int adjust)
 {
@@ -775,9 +806,12 @@ mk_to_vcol(MARK mark, int expanded, BUFFER *bp, int col, int adjust)
     int c;
     int lim;
     int t = tabstop_val(bp);
-    LINEPTR lp;
+    LINE *lp;
     int extra = ((!global_g_val(GMDALTTABPOS) && !insertmode) ? 1 : 0);
 
+    TRACE2((T_CALLED "mk_to_vcol(mark.o=%d, col=%d, adjust=%d) extra %d\n",
+	    mark.o, col, adjust, extra));
+
     if (i < 0) {
 	i = 0;
 	col = 0;
@@ -801,7 +835,7 @@ mk_to_vcol(MARK mark, int expanded, BUFFER *bp, int col, int adjust)
     }
     if (extra && (col != 0) && (mark.o < llength(lp)))
 	col--;
-    return col;
+    return2Code(col);
 }
 
 /*
@@ -818,22 +852,25 @@ dot_to_vcol(WINDOW *wp)
 #if OPT_CACHE_VCOL
     W_TRAITS *wt = &(wp->w_traits);
     int need_col = FALSE;
-    int shift = w_val(wp, WVAL_SIDEWAYS)
-    ? w_val(wp, WVAL_SIDEWAYS) - 1
-    : 0;
-    int use_off = w_left_margin(wp);
+    int shift = (w_val(wp, WVAL_SIDEWAYS)
+		 ? w_val(wp, WVAL_SIDEWAYS) - 1
+		 : 0);
+    int use_off = -b_left_margin(bp);
 #if OPT_TRACE
     int check;
 #endif
 
+    TRACE2((T_CALLED "dot_to_vcol(%s)\n", bp->b_bname));
+
 #if OPT_TRACE && OPT_DEBUG
-    TRACE(("CHECK left_col %d, shift %d, length %d\n",
-	   wt->w_left_col, shift, llength(wp->w_dot.l)));
+    TRACE(("HAVE: dot.o %d, left_dot %d+%d, left_col %d\n",
+	   wp->w_dot.o,
+	   wt->w_left_dot.o, use_off,
+	   wt->w_left_col));
     TRACE(("%s\n", lp_visible(wp->w_dot.l)));
-    TRACE(("DOT:%s\n", visible_buff(
-				       wp->w_dot.l->l_text + wp->w_dot.o,
-				       llength(wp->w_dot.l) - wp->w_dot.o,
-				       FALSE)));
+    TRACE(("DOT:%s\n", visible_buff(wp->w_dot.l->l_text + wp->w_dot.o,
+				    llength(wp->w_dot.l) - wp->w_dot.o,
+				    FALSE)));
 #endif
 
     /*
@@ -879,8 +916,6 @@ dot_to_vcol(WINDOW *wp)
 	    TRACE(("...adjust:%d\n", col));
 	    wt->w_left_col -= col;
 	}
-	if (wt->w_left_dot.o)
-	    use_off = 0;
     } else
 #endif
 	if ((wt->w_left_col + wt->w_left_adj < shift)
@@ -924,8 +959,11 @@ dot_to_vcol(WINDOW *wp)
 	    if (col < wt->w_left_col) {
 		/* if there was a multi-column character at the left
 		 * side of the shifted screen, adjust */
-		TRACE(("...adjust for multi-column character %d\n", col));
 		wt->w_left_adj = wt->w_left_col - col;
+		TRACE(("...adjust for multi-column character %d..%d -> %d\n",
+		       col,
+		       wt->w_left_col,
+		       wt->w_left_adj));
 		wt->w_left_col = col;
 	    }
 	    need_col = FALSE;
@@ -943,28 +981,31 @@ dot_to_vcol(WINDOW *wp)
 	    TRACE(("...cache w_left_col %d\n", wt->w_left_col));
 	}
     }
+    if (use_off != 0) {
+	if (wt->w_left_col == 0)
+	    use_off = -wt->w_left_dot.o;
+    }
     result = mk_to_vcol(wp->w_dot,
 			w_val(wp, WMDLIST),
 			bp,
 			wt->w_left_col,
-			wt->w_left_dot.o + use_off - b_left_margin(bp));
+			wt->w_left_dot.o + use_off);
 #if OPT_TRACE
     check = mk_to_vcol(wp->w_dot, w_val(wp, WMDLIST), bp, 0, 0);
     if (check != result) {
-	TRACE(("dot_to_vcol result %d check %d (off=%d, shift=%d)\n",
-	       result, check, wp->w_dot.o, shift));
+	TRACE(("MISMATCH result %d check %d\n", result, check));
 	kbd_alarm();
-	TRACE(("-> OOPS:%s %d vs %d+%d %d\n",
-	       bp->b_bname,
+	TRACE(("-> OOPS: dot.o %d, left_dot %d+%d, left_col %d\n",
 	       wp->w_dot.o,
 	       wt->w_left_dot.o, use_off,
 	       wt->w_left_col));
+	result = check;
     }
 #endif
 #else
     result = mk_to_vcol(wp->w_dot, w_val(wp, WMDLIST), bp, 0, 0);
 #endif
-    return result;
+    return2Code(result);
 }
 
 /*
@@ -984,8 +1025,8 @@ dot_to_vcol(WINDOW *wp)
 static void
 update_line(int row, int colfrom, int colto)
 {
-    register struct VIDEO *vp1 = vscreen[row];	/* virtual screen image */
-    register int req = (vp1->v_flag & VFREQ) == VFREQ;
+    struct VIDEO *vp1 = vscreen[row];	/* virtual screen image */
+    int req = (vp1->v_flag & VFREQ) == VFREQ;
 
 #if OPT_COLOR
     CurFcolor(vp1) = ReqFcolor(vp1);
@@ -1019,8 +1060,8 @@ update_line(int row, int colfrom, int colto)
 static void
 update_line(int row, int colfrom, int colto)
 {
-    register char *vc, *pc, *evc;
-    register VIDEO_ATTR *va, *pa, xx;
+    char *vc, *pc, *evc;
+    VIDEO_ATTR *va, *pa, xx;
     int nchanges = 0;
 
     if ((vscreen[row]->v_flag & VFCHG) == 0)
@@ -1063,17 +1104,17 @@ update_line(int row, int colfrom, int colto)
 {
     struct VIDEO *vp1 = vscreen[row];	/* virtual screen image */
     struct VIDEO *vp2 = PSCREEN[row];	/* physical screen image */
-    register int xl = colfrom;
-    register int xr = colto;
-    register int xx;
+    int xl = colfrom;
+    int xr = colto;
+    int xx;
 
-    register char *cp1 = VideoText(vp1);
-    register char *cp2 = VideoText(vp2);
-    register int nbflag;	/* non-blanks to the right flag? */
+    char *cp1 = VideoText(vp1);
+    char *cp2 = VideoText(vp2);
+    int nbflag;			/* non-blanks to the right flag? */
 
 #if OPT_VIDEO_ATTRS
-    register VIDEO_ATTR *ap1 = VideoAttr(vp1);
-    register VIDEO_ATTR *ap2 = VideoAttr(vp2);
+    VIDEO_ATTR *ap1 = VideoAttr(vp1);
+    VIDEO_ATTR *ap2 = VideoAttr(vp2);
     VIDEO_ATTR Blank = 0;	/* FIXME: Color? */
 #else
     UINT rev;			/* reverse video flag */
@@ -1201,7 +1242,7 @@ update_line(int row, int colfrom, int colto)
     movecursor(row, xl - colfrom);	/* Go to start of line. */
 #if OPT_VIDEO_ATTRS
     while (xl < xx) {
-	register int j = xl;
+	int j = xl;
 	VIDEO_ATTR attr = VATTRIB(ap1[j]);
 	while ((j < xx) && (attr == VATTRIB(ap1[j])))
 	    j++;
@@ -1266,15 +1307,13 @@ kbd_openup(void)
 #if !FRAMEBUF && !OPT_PSCREEN
     if (pscreen != 0) {
 	for (i = 0; i < term.rows - 1; ++i) {
-	    (void) memcpy(
-			     pscreen[i]->v_text,
-			     pscreen[i + 1]->v_text,
-			     (size_t) (term.cols));
+	    (void) memcpy(pscreen[i]->v_text,
+			  pscreen[i + 1]->v_text,
+			  (size_t) (term.cols));
 #if OPT_VIDEO_ATTRS
-	    (void) memcpy(
-			     pscreen[i]->v_attrs,
-			     pscreen[i + 1]->v_attrs,
-			     alen);
+	    (void) memcpy(pscreen[i]->v_attrs,
+			  pscreen[i + 1]->v_attrs,
+			  alen);
 #endif
 	}
 	(void) memset(pscreen[i]->v_text, ' ', (size_t) (term.cols));
@@ -1353,12 +1392,11 @@ kbd_flush(void)
  * to know if we're in the record-separator.
  */
 static int
-offs2col0(
-	     WINDOW *wp,
-	     LINEPTR lp,
-	     C_NUM offset,
-	     C_NUM * cache_offset,
-	     int *cache_column)
+offs2col0(WINDOW *wp,
+	  LINE *lp,
+	  C_NUM offset,
+	  C_NUM * cache_offset,
+	  int *cache_column)
 {
     int column;
 
@@ -1413,7 +1451,7 @@ offs2col0(
 }
 
 int
-offs2col(WINDOW *wp, LINEPTR lp, C_NUM offset)
+offs2col(WINDOW *wp, LINE *lp, C_NUM offset)
 {
     return offs2col0(wp, lp, offset, 0, 0);
 }
@@ -1425,7 +1463,7 @@ offs2col(WINDOW *wp, LINEPTR lp, C_NUM offset)
  */
 #if OPT_MOUSE || defined(WMDLINEWRAP)
 int
-col2offs(WINDOW *wp, LINEPTR lp, C_NUM col)
+col2offs(WINDOW *wp, LINE *lp, C_NUM col)
 {
     int tabs = tabstop_val(wp->w_bufp);
     int list = w_val(wp, WMDLIST);
@@ -1436,10 +1474,10 @@ col2offs(WINDOW *wp, LINEPTR lp, C_NUM col)
     w_val(wp, WVAL_SIDEWAYS);
     int goal = col + left - nu_width(wp) - w_left_margin(wp);
 
-    register C_NUM n;
-    register C_NUM offset;
-    register C_NUM len = llength(lp);
-    register char *text = lp->l_text;
+    C_NUM n;
+    C_NUM offset;
+    C_NUM len = llength(lp);
+    char *text = lp->l_text;
 
     if (lp == win_head(wp)) {
 	offset = 0;
@@ -1447,7 +1485,7 @@ col2offs(WINDOW *wp, LINEPTR lp, C_NUM col)
 	for (offset = w_left_margin(wp), n = 0;
 	     (offset < len) && (n < goal);
 	     offset++) {
-	    register int c = text[offset];
+	    int c = text[offset];
 	    if (isPrint(c)) {
 		n++;
 	    } else if (list || (c != '\t')) {
@@ -1468,7 +1506,7 @@ col2offs(WINDOW *wp, LINEPTR lp, C_NUM col)
  */
 #ifdef WMDLINEWRAP
 int
-line_height(WINDOW *wp, LINEPTR lp)
+line_height(WINDOW *wp, LINE *lp)
 {
     int hi = 1;
     if (w_val(wp, WMDLINEWRAP)) {
@@ -1499,7 +1537,7 @@ line_height(WINDOW *wp, LINEPTR lp)
 WINDOW *
 row2window(int row)
 {
-    register WINDOW *wp;
+    WINDOW *wp;
 
     for_each_visible_window(wp) {
 	if (row >= wp->w_toprow && row <= mode_row(wp))
@@ -1528,8 +1566,8 @@ static size_t recomp_len;
 static void
 recompute_buffer(BUFFER *bp)
 {
-    register WINDOW *wp;
-    register SAVEWIN *tbl;
+    WINDOW *wp;
+    SAVEWIN *tbl;
 
     struct VAL b_vals[MAX_B_VALUES];
     size_t num = 0;
@@ -1623,7 +1661,7 @@ TypeAhead(int force)
  * Update virtual screen line, given a LINE pointer.
  */
 static void
-update_screen_line(WINDOW *wp, LINEPTR lp, int sline)
+update_screen_line(WINDOW *wp, LINE *lp, int sline)
 {
     C_NUM left;
 
@@ -1635,8 +1673,8 @@ update_screen_line(WINDOW *wp, LINEPTR lp, int sline)
 #ifdef WMDLINEWRAP
     if (w_val(wp, WMDLINEWRAP)) {
 	int top_line = sline - wp->w_line.o;
-	register int m = (top_line >= 0) ? top_line : 0;
-	register int n = top_line + line_height(wp, lp);
+	int m = (top_line >= 0) ? top_line : 0;
+	int n = top_line + line_height(wp, lp);
 	while (n > m)
 	    if (--n < mode_row(wp)) {
 		vscreen[n]->v_flag |= VFCHG;
@@ -1657,7 +1695,7 @@ update_screen_line(WINDOW *wp, LINEPTR lp, int sline)
 	vtmove(sline, -left);
 	vtset(lp, wp);
 	if (left && sline >= 0) {
-	    register int zero = nu_width(wp);
+	    int zero = nu_width(wp);
 	    vscreen[sline]->v_text[zero] = MRK_EXTEND_LEFT;
 	    if (vtcol <= zero)
 		vtcol = zero + 1;
@@ -1683,8 +1721,8 @@ update_screen_line(WINDOW *wp, LINEPTR lp, int sline)
 static void
 update_oneline(WINDOW *wp)
 {
-    register LINEPTR lp;	/* line to update */
-    register int sline;		/* physical screen line to update */
+    LINE *lp;			/* line to update */
+    int sline;			/* physical screen line to update */
 
     /* search down the line we want */
     lp = wp->w_line.l;
@@ -1706,8 +1744,8 @@ update_oneline(WINDOW *wp)
 static void
 update_all(WINDOW *wp)
 {
-    register LINEPTR lp;	/* line to update */
-    register int sline;		/* physical screen line to update */
+    LINE *lp;			/* line to update */
+    int sline;			/* physical screen line to update */
 
     /* search down the lines, updating them */
     lp = wp->w_line.l;
@@ -1758,7 +1796,7 @@ static void
 update_line_attrs(WINDOW *wp)
 {
     int row;
-    LINEPTR lp;
+    LINE *lp;
     int linewrap;
 
 #ifdef WMDLINEWRAP
@@ -1784,6 +1822,9 @@ update_line_attrs(WINDOW *wp)
 		else
 		    i = 0;
 	    }
+	    if (!linewrap && w_val(wp, WMDNUMBER) && w_val(wp, WVAL_SIDEWAYS)) {
+		i += NU_WIDTH;
+	    }
 	    /* If i > 0, make sure it indexes a valid attribute.  */
 	    if (i > 0) {
 		int ii;
@@ -1837,7 +1878,7 @@ update_window_attrs(WINDOW *wp)
     int i;
 
     L_NUM start_wlnum, end_wlnum;
-    LINEPTR lp;
+    LINE *lp;
     int rows;
 
     /*
@@ -1964,9 +2005,9 @@ static void
 update_garbaged_screen(void)
 {
 #if !FRAMEBUF && !OPT_PSCREEN
-    register int j;
+    int j;
 #endif
-    register int i;
+    int i;
 
     for (i = 0; i < term.rows; ++i) {
 	vscreen[i]->v_flag |= VFCHG;
@@ -2010,8 +2051,8 @@ update_garbaged_screen(void)
 static int
 update_extended_line(int col, int excess, int use_excess)
 {
-    register int rcursor;
-    register int zero = nu_width(curwp);
+    int rcursor;
+    int zero = nu_width(curwp);
     int scrollsiz;
 
     /* calculate what column the real cursor will end up in */
@@ -2049,12 +2090,12 @@ update_extended_line(int col, int excess, int use_excess)
 static int
 update_cursor_position(int *screenrowp, int *screencolp)
 {
-    register LINEPTR lp;
+    LINE *lp;
 #ifdef WMDLINEWRAP
-    register int i;
+    int i;
 #endif
-    register int col, excess;
-    register int collimit;
+    int col, excess;
+    int collimit;
     int moved = FALSE;
     int nuadj = is_empty_buf(curwp->w_bufp) ? 0 : nu_width(curwp);
     int liadj = (w_val(curwp, WMDLIST)) ? 1 : 0;
@@ -2377,7 +2418,7 @@ simple_scroll(int inserts)
 static void
 update_physical_screen(int force GCC_UNUSED)
 {
-    register int i;
+    int i;
 
 #if CAN_SCROLL
     if (scrflags & WFKILLS)
@@ -2409,8 +2450,8 @@ update_physical_screen(int force GCC_UNUSED)
 static void
 mlfs_prefix(const char **fsp, char **msp, int lchar)
 {
-    register const char *fs = *fsp;
-    register char *ms = *msp;
+    const char *fs = *fsp;
+    char *ms = *msp;
     if (*fs == ':') {
 	fs++;
 	while (*fs && *fs != ':') {
@@ -2454,7 +2495,7 @@ mlfs_suffix(const char **fsp, char **msp, int lchar)
 static void
 mlfs_skipfix(const char **fsp)
 {
-    register const char *fs = *fsp;
+    const char *fs = *fsp;
     if (*fs == ':') {
 	for (fs++; *fs && *fs != ':'; fs++) ;
 	if (*fs == ':')
@@ -2480,7 +2521,7 @@ mlfs_skipfix(const char **fsp)
 	if (b_val(bp, mode)) PutModename("%c%s", name)
 
 #if OPT_MAJORMODE
-#define PutMajormode(bp) if (bp->majr != 0) PutModename("%c%smode", bp->majr->name)
+#define PutMajormode(bp) if (bp->majr != 0) PutModename("%c%s", bp->majr->longname)
 #else
 #define PutMajormode(bp)	/*nothing */
 #endif
@@ -2488,8 +2529,8 @@ mlfs_skipfix(const char **fsp)
 static int
 modeline_modes(BUFFER *bp, char **msptr)
 {
-    register char *ms = msptr ? *msptr : 0;
-    register size_t mcnt = 0;
+    char *ms = msptr ? *msptr : 0;
+    size_t mcnt = 0;
 
     PutMajormode(bp);
 #if COMPLETE_FILES || COMPLETE_DIRS
@@ -2559,8 +2600,8 @@ modeline_modes(BUFFER *bp, char **msptr)
 static int
 modeline_show(WINDOW *wp, int lchar)
 {
-    register int ic = lchar;
-    register BUFFER *bp = wp->w_bufp;
+    int ic = lchar;
+    BUFFER *bp = wp->w_bufp;
 
     if (b_val(bp, MDSHOWMODE)) {
 #ifdef insertmode		/* insert mode is a trait for each window */
@@ -2632,7 +2673,7 @@ percentage(WINDOW *wp)
 #endif
 	val = line_no(bp, wp->w_dot.l);
 
-    return (denom == 0) ? 100 : ((val * 100) / denom);
+    return PERCENT(val, denom);
 }
 
 /*
@@ -2640,7 +2681,7 @@ percentage(WINDOW *wp)
  * a number of special variables that we would like to output quickly.
  */
 void
-special_formatter(TBUFF ** result, const char *fs, WINDOW *wp)
+special_formatter(TBUFF **result, const char *fs, WINDOW *wp)
 {
     BUFFER *bp;
     char *ms;
@@ -2655,6 +2696,9 @@ special_formatter(TBUFF ** result, const char *fs, WINDOW *wp)
     int right_len;
     int n;
 
+    if (fs == 0)
+	return;
+
     tb_init(result, EOS);
 
     left_ms[0] = right_ms[0] = EOS;
@@ -2768,6 +2812,20 @@ special_formatter(TBUFF ** result, const char *fs, WINDOW *wp)
 		    mlfs_skipfix(&fs);
 		break;
 
+#endif
+#ifdef WMDSHOWCHAR
+	    case 'C':
+		if (w_val(wp, WMDSHOWCHAR) && !is_empty_buf(wp->w_bufp)) {
+		    int curchar = (is_at_end_of_line(wp->w_dot)
+				   ? '\n'
+				   : char_at(wp->w_dot));
+		    sprintf(temp, "%02X", curchar);
+		    mlfs_prefix(&fs, &ms, lchar);
+		    ms = lsprintf(ms, "%s", temp);
+		    mlfs_suffix(&fs, &ms, lchar);
+		} else
+		    mlfs_skipfix(&fs);
+		break;
 #endif
 	    case 'P':
 		ms = lsprintf(ms, "%d", percentage(wp));
@@ -2800,7 +2858,11 @@ special_formatter(TBUFF ** result, const char *fs, WINDOW *wp)
 		    execstr = temp;
 
 		    strncpy0(temp, save_fs, fs + 1 - save_fs);
-		    execstr = get_token(execstr, &tok, EOS, (int *) 0);
+		    execstr = get_token(execstr,
+					&tok,
+					eol_null,
+					EOS,
+					(int *) 0);
 		    vl_strncpy(temp, tokval(temp), sizeof(temp));
 
 		    tb_free(&tok);
@@ -2850,9 +2912,11 @@ special_formatter(TBUFF ** result, const char *fs, WINDOW *wp)
 
     tb_bappend(result, left_ms, strlen(left_ms));
 
-    if (((int) tb_length(*result) < term.cols)
+    if (tb_values(*result) != 0 && ((int) tb_length(*result) < term.cols)
 	&& (right_len = strlen(right_ms)) != 0) {
-	for (n = term.cols - (int) tb_length(*result) - right_len; n > 0; n--)
+	for (n = term.cols - (int) tb_length(*result) - right_len;
+	     n > 0;
+	     n--)
 	    tb_append(result, lchar);
 	if ((n = term.cols - right_len) < 0) {
 	    col = right_len + n;
@@ -2866,8 +2930,10 @@ special_formatter(TBUFF ** result, const char *fs, WINDOW *wp)
 	    tb_bappend(result, right_ms + n, term.cols - col);
     }
 
-    if (need_eighty_column_indicator) {		/* mark column 80 */
+    /* mark column 80 */
+    if (tb_values(*result) != 0 && need_eighty_column_indicator) {
 	int left = -nu_width(wp);
+	char *ss;
 #ifdef WMDLINEWRAP
 	if (!w_val(wp, WMDLINEWRAP))
 #endif
@@ -2878,8 +2944,10 @@ special_formatter(TBUFF ** result, const char *fs, WINDOW *wp)
 	if ((n > 80) && (col >= 0)) {
 	    for (n = tb_length(*result); n < col; n++)
 		tb_append(result, lchar);
-	    if (tb_values(*result)[col] == lchar)
-		tb_values(*result)[col] = '|';
+	    if ((ss = tb_values(*result)) != 0
+		&& ss[col] == lchar) {
+		ss[col] = '|';
+	    }
 	}
     }
     tb_append(result, EOS);
@@ -2892,6 +2960,7 @@ update_modeline(WINDOW *wp)
 #if OPT_MLFORMAT
     static TBUFF *result;
 #else
+    static char lchar_pad[4];
     BUFFER *bp;
     char temp[NFILEN];
     int col;
@@ -2901,7 +2970,7 @@ update_modeline(WINDOW *wp)
     char *ms;
     char right_ms[NFILEN * 2];
 #endif
-    register int n;
+    int n;
 
     term.cursorvis(FALSE);
 
@@ -2969,13 +3038,14 @@ update_modeline(WINDOW *wp)
 	}
 	ms = lsprintf(ms, " %s ", temp);
     }
+    memset(lchar_pad, lchar, 3);
 #ifdef WMDRULER
     if (w_val(wp, WMDRULER))
-	(void) lsprintf(right_ms, " (%d,%d) %3P",
-			wp->w_ruler_line, wp->w_ruler_col, lchar);
+	(void) lsprintf(right_ms, " (%d,%d) %s",
+			wp->w_ruler_line, wp->w_ruler_col, lchar_pad);
     else
 #endif
-	(void) lsprintf(right_ms, " %s %3P", rough_position(wp), lchar);
+	(void) lsprintf(right_ms, " %s %s", rough_position(wp), lchar_pad);
 
     *ms++ = EOS;
     right_len = strlen(right_ms);
@@ -3013,10 +3083,11 @@ update_modeline(WINDOW *wp)
 void
 upmode(void)
 {
-    register WINDOW *wp;
+    WINDOW *wp;
 
-    for_each_window(wp)
+    for_each_window(wp) {
 	wp->w_flag |= WFMODE;
+    }
 }
 
 /*
@@ -3026,21 +3097,21 @@ upmode(void)
 static void
 reframe_cursor_position(WINDOW *wp)
 {
-    register LINEPTR dlp;
-    register LINEPTR lp;
-    register int i = 0;
-    register int rows;
+    LINE *dlp;
+    LINE *lp;
+    int i = 0;
+    int rows;
     int founddot = FALSE;	/* set to true iff we find dot */
     int tildecount;
 
     /* if not a requested reframe, check for a needed one */
     if ((wp->w_flag & WFFORCE) == 0) {
 	/* initial update in main.c may not set these first... */
-	if (wp->w_dot.l == (LINE *) 0) {
+	if (!valid_line_wp(wp->w_dot.l, wp)) {
 	    wp->w_dot.l = lforw(win_head(wp));
 	    wp->w_dot.o = 0;
 	}
-	if (wp->w_line.l == (LINE *) 0) {
+	if (!valid_line_wp(wp->w_line.l, wp)) {
 	    wp->w_line.l = wp->w_dot.l;
 	    wp->w_line.o = 0;
 	}
@@ -3217,29 +3288,36 @@ reframe_cursor_position(WINDOW *wp)
 static void
 de_extend_lines(void)
 {
-    register WINDOW *wp;
-    register LINEPTR lp;
-    register int i;
+    WINDOW *wp;
+    LINE *lp;
+    int i;
 
     for_each_visible_window(wp) {
-	lp = wp->w_line.l;
-	i = TopRow(wp);
+	if ((lp = wp->w_line.l) == NULL) {
+	    /* this should not happen */
+	    mlforce("BUG:  lost w_line for %s", wp->w_bufp->b_bname);
+	    lp = win_head(wp);
+	    wp->w_toprow = 0;
+	}
+	if (valid_line_wp(lp, wp)) {
+	    i = TopRow(wp);
 
-	while (i < mode_row(wp)) {
-	    if (i >= 0
-		&& (vscreen[i]->v_flag & VFEXT) != 0) {
-		if ((wp != curwp)
-		    || (lp != wp->w_dot.l)
-		    || ((i != currow)
-			&& (curcol < col_limit(wp)))) {
-		    update_screen_line(wp, lp, i);
-		    vteeol();
-		    /* this line no longer is extended */
-		    vscreen[i]->v_flag &= ~VFEXT;
+	    while (i < mode_row(wp)) {
+		if (i >= 0
+		    && (vscreen[i]->v_flag & VFEXT) != 0) {
+		    if ((wp != curwp)
+			|| (lp != wp->w_dot.l)
+			|| ((i != currow)
+			    && (curcol < col_limit(wp)))) {
+			update_screen_line(wp, lp, i);
+			vteeol();
+			/* this line no longer is extended */
+			vscreen[i]->v_flag &= ~VFEXT;
+		    }
 		}
+		i += line_height(wp, lp);
+		lp = lforw(lp);
 	    }
-	    i += line_height(wp, lp);
-	    lp = lforw(lp);
 	}
     }
 }
@@ -3254,7 +3332,7 @@ de_extend_lines(void)
 int
 update(int force /* force update past type ahead? */ )
 {
-    register WINDOW *wp;
+    WINDOW *wp;
     int origrow, origcol;
     int screenrow, screencol;
     int updated = FALSE;
@@ -3265,14 +3343,22 @@ update(int force /* force update past type ahead? */ )
     origrow = ttrow;
     origcol = ttcol;
 
-    if (clhide || !is_visible_window(curwp))
+    if (clhide || (valid_window(curwp) && !is_visible_window(curwp)))
 	return TRUE;
-    if (!curbp || !vscreen || !curwp)	/* not initialized */
+
+    /*
+     * If not initialized, just return.
+     */
+    if (!valid_buffer(curbp)
+	|| !vscreen
+	|| !valid_window(curwp))
 	return FALSE;
-    /* don't try to update if we got called via a read-hook on a window
+
+    /*
+     * Don't try to update if we got called via a read-hook on a window
      * that isn't complete.
      */
-    if (curwp->w_bufp == 0
+    if (!valid_buffer(curwp->w_bufp)
 	|| curwp->w_bufp->b_nwnd == 0
 	|| curwp->w_ntrows < 1)
 	return FALSE;
@@ -3305,7 +3391,7 @@ update(int force /* force update past type ahead? */ )
 	if (wp->w_flag & WFMODE) {
 	    if (wp->w_bufp->b_nwnd > 1) {
 		/* make sure all previous windows have this */
-		register WINDOW *owp;
+		WINDOW *owp;
 		for_each_visible_window(owp) {
 		    if (owp->w_bufp == wp->w_bufp)
 			owp->w_flag |= WFMODE;
@@ -3346,6 +3432,10 @@ update(int force /* force update past type ahead? */ )
 		wp->w_ruler_col = col;
 		wp->w_flag |= WFMODE;
 	    }
+#ifdef WMDSHOWCHAR
+	} else if (w_val(wp, WMDSHOWCHAR)) {
+	    wp->w_flag |= WFMODE;
+#endif
 	} else if (wp->w_flag & WFSTAT) {
 	    wp->w_flag |= WFMODE;
 	}
@@ -3443,7 +3533,7 @@ void
 hilite(int row, int colfrom, int colto, int on)
 {
 #if !OPT_VIDEO_ATTRS
-    register VIDEO *vp1 = vscreen[row];
+    VIDEO *vp1 = vscreen[row];
 #endif
 #ifdef WMDLINEWRAP
     WINDOW *wp = row2window(row);
@@ -3542,7 +3632,7 @@ mlsavec(int c)
  * line" flag TRUE.
  */
 static void
-mlmsg(const char *fmt, va_list * app)
+mlmsg(const char *fmt, va_list *app)
 {
     static int recur;
     int end_at;
@@ -3581,7 +3671,7 @@ mlmsg(const char *fmt, va_list * app)
 	 * so save it now */
 	tb_init(&mlsave, EOS);
 #if OPT_POPUP_MSGS
-	if (global_g_val(GMDPOPUP_MSGS) || (curwp == 0)) {
+	if (global_g_val(GMDPOPUP_MSGS) || !valid_window(curwp)) {
 	    TRACE(("mlmsg popup_msgs #1 for '%s'\n", fmt));
 	    popup_msgs();
 	    msg_putc('\n');
@@ -3646,15 +3736,17 @@ mlmsg(const char *fmt, va_list * app)
 void
 mlwrite(const char *fmt,...)
 {
-    va_list ap;
-    if (global_b_val(MDTERSE) || kbd_replaying(FALSE) || !vl_msgs) {
-	if (!clhide)
-	    bottomleft();
-	return;
+    if (fmt != 0) {
+	va_list ap;
+	if (global_b_val(MDTERSE) || kbd_replaying(FALSE) || !vl_msgs) {
+	    if (!clhide)
+		bottomleft();
+	    return;
+	}
+	va_start(ap, fmt);
+	mlmsg(fmt, &ap);
+	va_end(ap);
     }
-    va_start(ap, fmt);
-    mlmsg(fmt, &ap);
-    va_end(ap);
 }
 
 /*
@@ -3664,72 +3756,85 @@ mlwrite(const char *fmt,...)
 void
 mlforce(const char *fmt,...)
 {
-    va_list ap;
-    va_start(ap, fmt);
-    mlmsg(fmt, &ap);
-    va_end(ap);
+    if (fmt != 0) {
+	va_list ap;
+	va_start(ap, fmt);
+	mlmsg(fmt, &ap);
+	va_end(ap);
+    }
 }
 
 /* VARARGS1 */
 void
 mlprompt(const char *fmt,...)
 {
-    va_list ap;
-    int osgarbf = sgarbf;
-    if (!vl_msgs) {
-	bottomleft();
-	return;
+    if (fmt != 0) {
+	va_list ap;
+	int osgarbf = sgarbf;
+	if (!vl_msgs) {
+	    bottomleft();
+	    return;
+	}
+	sgarbf = FALSE;
+	va_start(ap, fmt);
+	mlmsg(fmt, &ap);
+	va_end(ap);
+	sgarbf = osgarbf;
     }
-    sgarbf = FALSE;
-    va_start(ap, fmt);
-    mlmsg(fmt, &ap);
-    va_end(ap);
-    sgarbf = osgarbf;
 }
 
 /* VARARGS */
 void
 dbgwrite(const char *fmt,...)
 {
-    char temp[80];
+    if (fmt != 0) {
+	char temp[80];
 
-    va_list ap;			/* ptr to current data field */
-    va_start(ap, fmt);
-    lsprintf(temp, "[press ^G to continue] %s", fmt);
-    mlmsg(temp, &ap);
-    va_end(ap);
-    beginDisplay();
-    while (term.getch() != '\007') {
-	;
+	va_list ap;		/* ptr to current data field */
+	va_start(ap, fmt);
+	lsprintf(temp, "[press ^G to continue] %s", fmt);
+	mlmsg(temp, &ap);
+	va_end(ap);
+	beginDisplay();
+	while (term.getch() != '\007') {
+	    ;
+	}
+	endofDisplay();
     }
-    endofDisplay();
 }
 
 /*
  * Do the equivalent of 'perror()' on the message line
  */
 void
-mlerror(const char *s)
+mlerror(const char *str)
 {
+    int save_err = errno;
     const char *t = 0;
 #ifdef HAVE_STRERROR
 #if SYS_VMS
-    if (errno == EVMSERR)
-	t = strerror(errno, vaxc$errno);
+    if (save_err == EVMSERR)
+	t = strerror(save_err, vaxc$errno);
     else
 #endif /* SYS_VMS */
-    if (errno > 0)
-	t = strerror(errno);
+    if (save_err > 0)
+	t = strerror(save_err);
 #else
 #ifdef HAVE_SYS_ERRLIST
-    if (errno > 0 && errno < sys_nerr)
-	t = sys_errlist[errno];
+    if (save_err > 0 && save_err < sys_nerr)
+	t = sys_errlist[save_err];
 #endif /* HAVE_SYS_ERRLIST */
 #endif /* HAVE_STRERROR */
-    if (t != 0)
-	mlwarn("[Error %s: %s]", s, t);
-    else
-	mlwarn("[Error %s: unknown system error %d]", s, errno);
+    if (t != 0) {
+	/* Borland's strerror() returns newlines on the end of the strings */
+	static TBUFF *tt;
+	if (tb_scopy(&tt, t) != 0) {
+	    t = mktrimmed(tb_values(tt));
+	}
+	mlwarn("[Error %s: %s]", str, t);
+    } else {
+	mlwarn("[Error %s: unknown system error %d]", str, save_err);
+    }
 }
 
 /*
@@ -3766,19 +3871,30 @@ lspputc(int c)
 char *
 lsprintf(char *buf, const char *fmt,...)
 {
-    va_list ap;
-    va_start(ap, fmt);
+    if ((lsp = buf) != 0) {
+	va_list ap;
+	va_start(ap, fmt);
 
-    lsp = buf;
-    dfoutfn = lspputc;
+	dfoutfn = lspputc;
 
-    dofmt(fmt, &ap);
-    va_end(ap);
+	dofmt(fmt, &ap);
+	va_end(ap);
 
-    *lsp = EOS;
+	*lsp = EOS;
+    }
     return lsp;
 }
 
+int
+format_int(char *buf, UINT number, UINT radix)
+{
+    if ((lsp = buf) != 0) {
+	dfputi(lspputc, number, radix);
+	*lsp = EOS;
+    }
+    return (lsp - buf);
+}
+
 /*
  * Buffer printf -- like regular printf, but puts characters
  *	into the BUFFER.
@@ -3792,6 +3908,13 @@ bputc(int c)
 	(void) linsert(1, c);
 }
 
+void
+bpadc(int c, int count)
+{
+    while (count-- > 0)
+	bputc(c);
+}
+
 /* printf into curbp, at DOT */
 /* VARARGS */
 void
@@ -3806,6 +3929,57 @@ bprintf(const char *fmt,...)
     va_end(ap);
 }
 
+/*
+ * Print into the given internal buffer, returning a pointer to the line
+ * written.
+ */
+LINE *
+b2vprintf(BUFFER *bp, const char *fmt, va_list ap)
+{
+    LINE *result = 0;
+    WINDOW *save_wp;
+    BUFFER *save_bp = curbp;
+    W_TRAITS save_w_traits;
+    OutFunc save_outfn;
+
+    if ((save_wp = push_fake_win(bp)) != 0) {
+
+	save_outfn = dfoutfn;
+	dfoutfn = bputc;
+
+	save_w_traits = curwp->w_traits;
+
+	(void) gotoeob(FALSE, 1);
+	(void) gotoeol(FALSE, 1);
+
+	dofmt(fmt, &ap);
+
+	result = lback(DOT.l);
+
+	b_clr_changed(bp);
+	curwp->w_traits = save_w_traits;
+
+	pop_fake_win(save_wp, save_bp);
+	dfoutfn = save_outfn;
+    }
+    return result;
+}
+
+/*
+ * Write text (not necessarily a whole line) to the given buffer.
+ */
+LINE *
+b2printf(BUFFER *bp, const char *fmt,...)
+{
+    LINE *result;
+    va_list ap;
+
+    va_start(ap, fmt);
+    result = b2vprintf(bp, fmt, ap);
+    va_end(ap);
+    return result;
+}
+
 #if OPT_EVAL || OPT_DEBUGMACROS
 /* printf into [Trace] */
 void
@@ -3813,37 +3987,20 @@ tprintf(const char *fmt,...)
 {
     static int nested;
 
-    WINDOW *save_wp;
     BUFFER *bp;
-    BUFFER *save_bp = curbp;
-    W_TRAITS save_w_traits;
-    OutFunc save_outfn;
+    LINE *line;
+    va_list ap;
 
     if (!nested
-	&& (bp = make_ro_bp(TRACE_BufName, BFINVS)) != 0
-	&& (save_wp = push_fake_win(bp)) != 0) {
-	va_list ap;
-
+	&& (bp = make_ro_bp(TRACE_BufName, BFINVS)) != 0) {
 	nested = TRUE;
 
-	save_outfn = dfoutfn;
-	dfoutfn = bputc;
-
-	save_w_traits = curwp->w_traits;
-	DOT.l = lback(buf_head(curbp));
-	DOT.o = 0;
-
 	va_start(ap, fmt);
-	dofmt(fmt, &ap);
+	line = b2vprintf(bp, fmt, ap);
 	va_end(ap);
 
-	TRACE(("tprintf {%.*s}\n", llength(lback(DOT.l)), lback(DOT.l)->l_text));
+	TRACE(("tprintf {%.*s}\n", llength(line), line->l_text));
 
-	b_clr_changed(bp);
-	curwp->w_traits = save_w_traits;
-
-	pop_fake_win(save_wp, save_bp);
-	dfoutfn = save_outfn;
 	nested = FALSE;
     }
 }
