@@ -6,7 +6,7 @@
  */
 
 /*
- * $Header: /usr/build/vile/vile/RCS/edef.h,v 1.313 2003/08/04 22:34:38 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/edef.h,v 1.320 2005/03/10 19:54:57 tom Exp $
  */
 
 #ifndef VILE_EDEF_H
@@ -16,14 +16,16 @@
 extern "C" {
 #endif
 
+#include "patchlev.h"
+
 /* I know this declaration stuff is really ugly, and I probably won't ever
  *	do it again.  promise.  but it _does_ make it easy to add/change
  *	globals. -pgf
  */
 #ifdef realdef
-# define decl_init_const(thing,value) EXTERN_CONST thing = value
-# define decl_init(thing,value) thing = value
-# define decl_uninit(thing) thing
+# define decl_init_const(thing,value) DECL_EXTERN_CONST(thing) = value
+# define decl_init(thing,value) DECL_EXTERN(thing) = value
+# define decl_uninit(thing) DECL_EXTERN(thing)
 #else
 # define decl_init_const(thing,value) extern const thing
 # define decl_init(thing,value) extern thing
@@ -46,7 +48,7 @@ decl_init( char *exec_pathname, ".");	/* replaced at runtime with path-head of a
 #endif /* PROGRAM_NAME */
 
 decl_init( char prognam[], PROGRAM_NAME);
-decl_init( char version[], "version 9.4");
+decl_init( char version[], "version " VILE_VERSION);
 
 #ifdef SYSTEM_NAME
 decl_init( char opersys[], SYSTEM_NAME);
@@ -161,7 +163,7 @@ decl_init( int	eolexist, TRUE );	/* does clear to EOL exist	*/
 decl_uninit( int revexist );		/* does reverse video exist?	*/
 
 #ifdef realdef
-	MARK	nullmark = { NULL, 0 };
+DECL_EXTERN(MARK nullmark) = { NULL, 0 };
 #else
 extern	MARK	nullmark;
 #endif
@@ -318,6 +320,14 @@ decl_uninit( size_t fflinelen );	/* fflinebuf length */
 
 /*--------------------------------------------------------------------------*/
 
+#if OPT_LOCALE
+decl_uninit( char *vl_encoding );
+decl_uninit( char *vl_locale );
+#else
+#define vl_encoding "8bit"
+#define vl_locale "built-in"
+#endif
+
 decl_init ( L_NUM help_at, -1 );	/* position in help-file */
 decl_uninit( char *helpfile );
 decl_init( char vl_pathchr, PATHCHR );	/* $pathlist-separator */
@@ -384,6 +394,7 @@ decl_init_const( char TRACE_BufName[],		"[Trace]");
 decl_init( TBUFF *filename_expr, 0 );
 decl_init( TBUFF *error_expr, 0 );
 decl_init( TBUFF *error_match, 0 );
+decl_init( int error_tabstop, 8 );
 decl_init_const( char ERRORS_BufName[],		"[Error Expressions]");
 decl_init_const( char ERR_REGEX_BufName[],	"[Error Patterns]");
 #endif
@@ -428,6 +439,9 @@ decl_init_const( char TAGSTACK_BufName[],	"[Tag Stack]");
 #if OPT_TAGS
 decl_init_const( char TAGFILE_BufName[],	"[Tags %d]");
 #endif
+#ifdef MDFILTERMSGS
+decl_init_const( char FLTMSGS_BufName[],	"[Filter Messages]");
+#endif
 
 /* defined in nebind.h and nename.h */
 extern const NTAB nametbl[];
