@@ -12,7 +12,7 @@
 */
 
 /*
- * $Header: /usr/build/vile/vile/RCS/estruct.h,v 1.536 2003/07/01 00:32:51 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/estruct.h,v 1.584 2005/07/09 15:49:55 tom Exp $
  */
 
 #ifndef _estruct_h
@@ -69,11 +69,16 @@
 #define SYS_OS2    1
 #endif
 
-#ifdef __GO32__ 	/* DJ's GCC version 1.09 */
+#ifdef __GO32__		/* DJ's GCC version 1.09 */
 #define SYS_MSDOS  1
 #define CC_DJGPP   1
 #endif
 
+#ifdef __LCC__
+#define SYS_WINNT	1
+#define CC_LCC_WIN32	1
+#endif
+
 #ifdef _MSC_VER
 #define SYS_WINNT	1
 #define CC_MSVC		1
@@ -85,6 +90,10 @@
 # define SYS_SUNOS 1		/* FIXME: need to tweak lint ifdefs */
 #endif
 
+#if defined(__CYGWIN__) || defined(__CYGWIN32__)
+# define SYS_CYGWIN 1
+#endif
+
 #ifdef __EMX__
 # define SYS_OS2_EMX 1		/* makefile.emx, or configure-script */
 #endif
@@ -106,6 +115,10 @@
 #define CC_DJGPP		0	/* DJ's GCC version 1.09 */
 #endif
 
+#ifndef CC_LCC_WIN32
+#define CC_LCC_WIN32		0	/* Lcc-win32 */
+#endif
+
 #ifndef CC_MSC
 #define CC_MSC			0	/* Microsoft C versions 3 & 4 & 5 & 6 */
 #endif
@@ -122,16 +135,20 @@
 #define CC_WATCOM		0	/* WATCOM C/386 version 9.0 or above */
 #endif
 
+#ifndef SYS_CYGWIN
+#define SYS_CYGWIN		0	/* Unix'ed Win32		*/
+#endif
+
 #ifndef SYS_MSDOS
-#define SYS_MSDOS       	0       /* MS-DOS                       */
+#define SYS_MSDOS		0	/* MS-DOS			*/
 #endif
 
 #ifndef SYS_OS2
-#define SYS_OS2         	0       /* son of DOS                   */
+#define SYS_OS2			0	/* son of DOS			*/
 #endif
 
 #ifndef SYS_OS2_EMX
-#define SYS_OS2_EMX     	0       /* Unix'ed OS2                  */
+#define SYS_OS2_EMX		0	/* Unix'ed OS2			*/
 #endif
 
 #ifndef SYS_SUNOS
@@ -139,15 +156,15 @@
 #endif
 
 #ifndef SYS_UNIX
-#define SYS_UNIX       		0       /* Unix & friends               */
+#define SYS_UNIX		0	/* Unix & friends		*/
 #endif
 
 #ifndef SYS_VMS
-#define SYS_VMS       		0       /* OpenVMS                      */
+#define SYS_VMS			0	/* OpenVMS			*/
 #endif
 
 #ifndef SYS_WINNT
-#define SYS_WINNT       	0       /* Windows/NT                   */
+#define SYS_WINNT		0	/* Windows/NT			*/
 #endif
 
 /* As of version 3.51 of vile, CC_NEWDOSCC should be correct for Turbo,
@@ -155,7 +172,7 @@
  * probably correct for MSC (Microsoft C) and ZTC (Zortech), but I'm not
  * sure of those.  (It implies a lot of ANSI and POSIX behavior.)
  */
-#if CC_TURBO || CC_WATCOM || CC_MSC || CC_DJGPP || SYS_WINNT || CC_CSETPP || CC_MSVC
+#if CC_TURBO || CC_WATCOM || CC_MSC || CC_DJGPP || SYS_WINNT || CC_CSETPP || CC_MSVC || CC_LCC_WIN32
 # define CC_NEWDOSCC 1
 #endif
 
@@ -168,7 +185,8 @@
 #ifndef HAVE_CONFIG_H		/* we did not run the configure script */
 
 #if CC_NEWDOSCC
-# define HAVE_GETCWD 1
+# define HAVE_GETCWD		1
+# define HAVE_LIMITS_H		1
 #endif
 
 #if CC_CSETPP
@@ -183,11 +201,14 @@
 #endif
 
 #if SYS_WINNT && !CC_TURBO
-# define HAVE_PUTENV		1
 # define HAVE_SYS_UTIME_H	1
 # define HAVE_UTIME		1
 #endif
 
+#if SYS_WINNT
+# define HAVE_PUTENV		1
+#endif
+
 /*
  * Porting constraints: supply the normally assumed values that we get from
  * the "configure" script, for systems on which we cannot run "configure"
@@ -240,6 +261,9 @@
 # endif
 # define SIGT void
 # define SIGRET
+# if DISP_X11 && defined(NEED_X_INCLUDES)
+#  define XTSTRINGDEFINES
+# endif
 #endif
 
 #endif /* HAVE_CONFIG_H */
@@ -249,6 +273,7 @@
 #endif
 
 #include <stdio.h>
+
 #if SYS_VMS && (! defined(__DECC_VER))
 # include <types.h>
 # include <stat.h>
@@ -420,6 +445,10 @@
 #define OPT_FILTER 0	/* normally set by configure-script */
 #endif
 
+#ifndef OPT_ICONV_FUNCS
+#define OPT_ICONV_FUNCS 0 /* normally set by configure-script */
+#endif
+
 #ifndef OPT_LOCALE
 #define OPT_LOCALE 0	/* normally set by configure-script */
 #endif
@@ -428,6 +457,10 @@
 #define OPT_PERL 0	/* normally set by configure-script */
 #endif
 
+#ifndef OPT_PLUGIN
+#define OPT_PLUGIN 0	/* normally set by configure-script */
+#endif
+
 #ifndef OPT_SHELL
 #define OPT_SHELL 1	/* we'll disable this only as an exercise */
 #endif
@@ -464,8 +497,8 @@
 #define VILE_SOMEDAY 0
 
 /* various terminal stuff */
-#define IBM_VIDEO 	(SYS_MSDOS || SYS_OS2 || SYS_WINNT)
-#define CRLF_LINES 	(SYS_MSDOS || SYS_OS2 || SYS_WINNT)
+#define IBM_VIDEO	(SYS_MSDOS || SYS_OS2 || SYS_WINNT)
+#define CRLF_LINES	(SYS_MSDOS || SYS_OS2 || SYS_WINNT)
 
 /* various color stuff */
 #define	OPT_COLOR (DISP_ANSI || IBM_VIDEO || DISP_TERMCAP || DISP_CURSES || DISP_X11)
@@ -481,7 +514,7 @@
 #define	OPT_REVSTA	1	/* Status line appears in reverse video       */
 #define	OPT_CFENCE	1	/* do fence matching in CMODE		      */
 #define OPT_LCKFILES	0	/* create lock files (file.lck style)	      */
-#define OPT_TAGS	1	/* tags support 			      */
+#define OPT_TAGS	1	/* tags support				      */
 #define	OPT_PROCEDURES	1	/* macro language procedures		      */
 #define	OPT_PATHLOOKUP	1	/* search $PATH for startup and help files    */
 #define	OPT_SCROLLCODE	1	/* code in display.c for scrolling the screen.
@@ -514,6 +547,11 @@
 # define OPT_ICURSOR    0
 #endif
 
+#if SYS_WINNT
+# define MAX_RECENT_FILES 20
+# define MAX_RECENT_FLDRS 20
+#endif
+
 /*
  * The $findpath statevar and find-cfg mode features require:
  *
@@ -543,7 +581,11 @@
 /* the "working..." message -- we must have the alarm() syscall, and
    system calls must be restartable after an interrupt by default or be
    made restartable with sigaction() */
-#define OPT_WORKING (!SMALLER && HAVE_ALARM && HAVE_RESTARTABLE_PIPEREAD)
+#if !SMALLER && defined(HAVE_ALARM) && defined(HAVE_RESTARTABLE_PIPEREAD)
+#define OPT_WORKING 1
+#else
+#define OPT_WORKING 0
+#endif
 
 #define OPT_SCROLLBARS (XTOOLKIT | DISP_NTWIN)	/* scrollbars */
 
@@ -551,9 +593,14 @@
 #define OPT_VMS_PATH    (SYS_VMS)  /* vax/vms path parsing (testing/porting)*/
 #endif
 
+#if OPT_VMS_PATH
+#undef SYSTEM_NAME
+#define SYSTEM_NAME	"vms"	/* fakevms pretends to scripts it's vms */
+#endif
+
 /* systems with MSDOS-like filename syntax */
 #define OPT_MSDOS_PATH  (SYS_MSDOS || SYS_OS2 || SYS_WINNT || SYS_OS2_EMX)
-#define OPT_CASELESS	(SYS_MSDOS || SYS_OS2 || SYS_WINNT || SYS_OS2_EMX || SYS_VMS)
+#define OPT_CASELESS	(SYS_MSDOS || SYS_OS2 || SYS_WINNT || SYS_OS2_EMX || SYS_CYGWIN || SYS_VMS)
 #define OPT_UNC_PATH	(SYS_WINNT)
 
 /* individual features that are (normally) controlled by SMALLER */
@@ -592,8 +639,8 @@
 #define OPT_SHOW_COLORS	(!SMALLER && OPT_COLOR)	/* "show-colors" */
 #define OPT_SHOW_CTYPE	!SMALLER		/* "show-printable" */
 #define OPT_SHOW_EVAL   (!SMALLER && OPT_EVAL)	/* "show-variables" */
-#define OPT_SHOW_MAPS   !SMALLER 		/* display mapping for ":map" */
-#define OPT_SHOW_MARKS  !SMALLER 		/* "show-marks" */
+#define OPT_SHOW_MAPS   !SMALLER		/* display mapping for ":map" */
+#define OPT_SHOW_MARKS  !SMALLER		/* "show-marks" */
 #define OPT_SHOW_REGS   !SMALLER		/* "show-registers" */
 #define OPT_SHOW_TAGS   (!SMALLER && OPT_TAGS)	/* ":tags" displays tag-stack */
 
@@ -620,7 +667,7 @@
 #define	OPT_XTERM	0	/* vile doesn't recognize xterm mouse */
 #endif
 
- 	/* implement window title */
+	/* implement window title */
 #define OPT_TITLE	(SYS_WINNT | DISP_X11 | OPT_XTERM)
 
 	/* combine select/yank (for mouse support) */
@@ -666,13 +713,16 @@
 #define OPT_CHARCLASS_CHOICES	 OPT_SHOW_CTYPE
 #define OPT_COLOR_CHOICES	 (OPT_ENUM_MODES && OPT_COLOR)
 #define OPT_DIRECTIVE_CHOICES    !SMALLER
+#define OPT_FORBUFFERS_CHOICES   !SMALLER
 #define OPT_HILITE_CHOICES	 (OPT_ENUM_MODES && OPT_HILITEMATCH)
 #define OPT_LOOKUP_CHOICES       !SMALLER
 #define OPT_MMQUALIFIERS_CHOICES OPT_MAJORMODE
 #define OPT_PARAMTYPES_CHOICES   OPT_MACRO_ARGS
 #define OPT_PATH_CHOICES         !SMALLER
 #define OPT_POPUP_CHOICES	 (OPT_ENUM_MODES && OPT_POPUPCHOICE)
+#define OPT_READERPOLICY_CHOICES !SMALLER
 #define OPT_RECORDFORMAT_CHOICES (OPT_ENUM_MODES && SYS_VMS)
+#define OPT_RECORDATTRS_CHOICES  (OPT_ENUM_MODES && SYS_VMS)
 #define OPT_RECORDSEP_CHOICES    !SMALLER
 #define OPT_SHOWFORMAT_CHOICES   !SMALLER
 #define OPT_VIDEOATTRS_CHOICES   (OPT_ENUM_MODES && OPT_COLOR_SCHEMES)
@@ -703,6 +753,14 @@
 #define OPT_TRACE		0  /* turn on debug/trace (link with trace.o) */
 #endif
 
+#ifndef CAN_TRACE
+#define CAN_TRACE		1  /* (link with trace.o) */
+#endif
+
+#ifndef CAN_VMS_PATH
+#define CAN_VMS_PATH		1  /* (link with fakevms.o) */
+#endif
+
 /* That's the end of the user selections -- the rest is static definition */
 /* (i.e. you shouldn't need to touch anything below here */
 /* ====================================================================== */
@@ -762,9 +820,20 @@ extern char *rindex (const char *s, int c);
 #include <sys/param.h>
 #endif
 
+#ifdef HAVE_LIMITS_H
+#include <limits.h>
+#endif
+
 /*	System dependent library redefinitions, structures and includes	*/
 
-#if CC_NEWDOSCC && ! CC_CSETPP
+#if CC_CSETPP
+#elif CC_LCC_WIN32
+#include <direct.h>		/* chdir */
+#define cwait(i,j,k) _cwait(i,j,k)
+#define utime(i,j)   _utime(i,j)
+#define SP_NOTREPORTED 0x4000	/* wingdi.h */
+#define WHEEL_DELTA 120		/* winuser.h */
+#elif CC_NEWDOSCC && ! CC_CSETPP
 #include <dos.h>
 # define HAVE_STRERROR		1
 #endif
@@ -780,7 +849,7 @@ extern char *rindex (const char *s, int c);
 
 /* on MS-DOS we have to open files in binary mode to see the ^Z characters. */
 
-#if SYS_MSDOS || SYS_OS2 || SYS_WINNT || defined(__CYGWIN32__) || defined(__CYGWIN__) || defined(__EMX__)
+#if SYS_MSDOS || SYS_OS2 || SYS_WINNT || SYS_CYGWIN || SYS_OS2_EMX
 #define FOPEN_READ	"rb"
 #define FOPEN_WRITE	"wb"
 #define FOPEN_APPEND	"ab"
@@ -795,6 +864,11 @@ extern char *rindex (const char *s, int c);
 #if SYS_VMS
 #undef  FOPEN_READ
 #define FOPEN_READ "r", "shr=get,upd"
+#define RECORD_ATTRS(ftn,cr,prn,blk) \
+			(ftn<<FAB$V_FTN) | \
+			(cr <<FAB$V_CR)  | \
+			(prn<<FAB$V_PRN) | \
+			(blk<<FAB$V_BLK)
 #endif
 
 #if OPT_MSDOS_PATH && !SYS_OS2_EMX	/* DOS path / to \ conversions */
@@ -853,12 +927,12 @@ extern int MainProgram(int argc, char *argv[]);
 
 	/* semaphore may be needed to prevent interrupt of display-code */
 #if defined(SIGWINCH) || OPT_WORKING
-# if OPT_TRACE > 1
+# if OPT_TRACE > 2
 extern void beginDisplay(void);
 extern void endofDisplay(void);
 # else
 # define beginDisplay() ++im_displaying
-# define endofDisplay() if (im_displaying) --im_displaying
+# define endofDisplay() if (im_displaying > 0) --im_displaying; else assert(im_displaying > 0)
 # endif
 #else
 # define beginDisplay() /* nothing */
@@ -991,13 +1065,18 @@ extern void endofDisplay(void);
 
 #define	char2int(c)	((int)(c & 0xff)) /* mask off sign-extension, etc. */
 
-#define	PLURAL(n)	((n!=1)?"s":"")
+#define	PLURAL(n)	((n) != 1 ? "s" : "")
+#define NONNULL(s)	((s) != 0 ? (s) : "")
+#define isEmpty(s)	((s) == 0 || *(s) == EOS)
 
 #define EOS        '\0'
 #define BQUOTE     '`'
 #define SQUOTE     '\''
 #define DQUOTE     '"'
 #define BACKSLASH  '\\'
+#define CH_TILDE   '~'
+
+#define isEscaped(s)	((s)[-1] == BACKSLASH)
 
 /* protect against losing namespaces */
 #undef	FALSE
@@ -1005,7 +1084,7 @@ extern void endofDisplay(void);
 #undef	ABORT
 #undef	SORTOFTRUE
 
-#define FALSE	0			/* False, no, bad, etc. 	*/
+#define FALSE	0			/* False, no, bad, etc.		*/
 #define TRUE	1			/* True, yes, good, etc.	*/
 #define ABORT	2			/* Death, ESC, abort, etc.	*/
 #define	SORTOFTRUE	3		/* really!	*/
@@ -1068,6 +1147,12 @@ typedef	enum {
 #endif
 } DIRECTIVE;
 
+typedef enum {
+	FB_MIXED = 0
+	, FB_GLOB
+	, FB_REGEX
+} FOR_BUFFERS;
+
 typedef enum {
 	MMQ_ANY = 0
 	, MMQ_ALL
@@ -1125,6 +1210,12 @@ typedef enum {
 	, RS_CRLF		/* dos */
 } RECORD_SEP;
 
+typedef enum {
+	RP_QUICK
+	, RP_SLOW
+	, RP_BOTH
+} READERPOLICY_CHOICES;
+
 #if CRLF_LINES
 #define RS_DEFAULT RS_CRLF
 #else
@@ -1212,15 +1303,18 @@ typedef UINT WATCHTYPE;
 #define	FORWARD	TRUE
 #define REVERSE	FALSE
 
+typedef enum {
 	/* nonfatal codes */
-#define FIOBAD  -1			/* File I/O, truncation		*/
-#define FIOSUC  0			/* File I/O, success.		*/
-#define FIOEOF  1			/* File I/O, end of file.	*/
+    FIOBAD = -1			/* File I/O, truncation		*/
+    ,FIOSUC			/* File I/O, success.		*/
+    ,FIOEOF			/* File I/O, end of file.	*/
 	/* error codes */
-#define FIOFNF  2			/* File I/O, file not found.	*/
-#define FIOERR  3			/* File I/O, error.		*/
-#define FIOMEM  4			/* File I/O, out of memory	*/
-#define FIOABRT 5			/* File I/O, aborted		*/
+    ,FIONUL			/* File I/O, empty file		*/
+    ,FIOFNF			/* File I/O, file not found.	*/
+    ,FIOERR			/* File I/O, error.		*/
+    ,FIOMEM			/* File I/O, out of memory	*/
+    ,FIOABRT			/* File I/O, aborted		*/
+} FIO_TYPES;
 
 /* three flavors of insert mode	*/
 /* it's FALSE, or one of:	*/
@@ -1391,12 +1485,21 @@ typedef USHORT CHARTYPE;
 #define ispath(c)	istype(vl_pathn, c)
 #define iswild(c)	istype(vl_wild, c)
 
+#define KEY_Space	' '
+#define KEY_Tab		'\t'
+
 /* macro for cases where return & newline are equivalent */
 #define	isreturn(c)	((c == '\r') || (c == '\n'))
 
 /* macro for whitespace (non-return) */
 #define	isBlank(c)      ((c == '\t') || (c == ' '))
 
+#if OPT_KEY_MODIFY
+#define isBackTab(c)	((c) == KEY_BackTab || (((c) & mod_SHIFT) != 0 && CharOf(c) == KEY_Tab))
+#else
+#define isBackTab(c)	((c) == KEY_BackTab)
+#endif
+
 #define	isGraph(c)	(!isSpecial(c) && !isSpace(c) && isPrint(c))
 
 /* DIFCASE represents the difference between upper
@@ -1406,8 +1509,8 @@ typedef USHORT CHARTYPE;
 #define	DIFCNTRL	0x40
 #define toUpper(c)	vl_uppercase[CharOf(c)]
 #define toLower(c)	vl_lowercase[CharOf(c)]
-#define tocntrl(c)	((c)^DIFCNTRL)
-#define toalpha(c)	((c)^DIFCNTRL)
+#define tocntrl(c)	(((unsigned)(c))^DIFCNTRL)
+#define toalpha(c)	(((unsigned)(c))^DIFCNTRL)
 
 #define nocase_eq(bc,pc) (CharOf(bc) == CharOf(pc) || (toUpper(bc) == toUpper(pc)))
 
@@ -1479,6 +1582,8 @@ typedef	struct	vl_tbuff	{
 	int	tb_errs;	/* true if we copied error_val here */
 	} TBUFF;
 
+#define isTB_ERRS(p) ((p) != 0 && (p)->tb_errs)
+
 /*
  * Definitions for 'itbuff.c' (temporary/dynamic int-buffers)
  */
@@ -1501,7 +1606,7 @@ typedef	struct {
 } L_FLAG;		/* LINE-flags */
 
 typedef	ULONG		CMDFLAGS;	/* CMDFUNC flags */
-typedef	long		B_COUNT;	/* byte-count */
+typedef	ULONG		B_COUNT;	/* byte-count */
 
 /*
  * Normally we build with ANSI C compilers, but allow for the possibility of
@@ -1517,6 +1622,19 @@ typedef	long		B_COUNT;	/* byte-count */
 #define EXTERN_CONST const
 #endif
 
+/*
+ * The Intel compiler's warnings regarding externs which have not been
+ * previously declared applies to data items as well as functions.  Use these
+ * macros to hide the extra declaration.
+ */
+#if defined(__INTEL_COMPILER)
+# define DECL_EXTERN_CONST(name) extern const name; EXTERN_CONST name
+# define DECL_EXTERN(name)       extern       name;              name
+#else
+# define DECL_EXTERN_CONST(name) EXTERN_CONST name
+# define DECL_EXTERN(name)                    name
+#endif
+
 /*
  * Control structures
  */
@@ -1541,11 +1659,9 @@ typedef	long		B_COUNT;	/* byte-count */
  *
  * Lines are additionally sometimes stacked in undo lists.
  */
-typedef	struct	LINE*	LINEPTR;
-
 typedef struct	LINE {
-	LINEPTR l_fp;			/* Link to the next line	*/
-	LINEPTR l_bp;			/* Link to the previous line	*/
+	struct LINE *l_fp;		/* Link to the next line	*/
+	struct LINE *l_bp;		/* Link to the previous line	*/
 	union {
 		size_t	l_sze;		/* Allocated size		*/
 		C_NUM	l_fo;		/* forward undo dot offs (undo only) */
@@ -1557,16 +1673,16 @@ typedef struct	LINE {
 	int	l_used;			/* Used size (may be negative)	*/
 	union {
 	    char *l_txt;		/* The data for this line	*/
-	    LINEPTR l_nxt;		/* if an undo stack separator,	*/
+	    struct LINE *l_nxt;		/* if an undo stack separator,	*/
 	} lt;				/*  a pointer to the next one	*/
 	union
 	{
-	    LINEPTR	l_stklnk;	/* Link for undo stack		*/
+	    struct LINE	*l_stklnk;	/* Link for undo stack		*/
 	    L_FLAG	l_flg;		/* flags for undo ops		*/
 	} l;
 #if OPT_LINE_ATTRS
 	UCHAR  *l_attrs;		/* indexes into the line_attr_tbl
-	                                   hash table */
+					   hash table */
 #endif
 }	LINE;
 
@@ -1601,7 +1717,6 @@ typedef struct	LINE {
 #define PURESTACKSEP	((int)(-3)) /* as above, but buffer unmodified before */
 					/* this change */
 
-#define	null_ptr	(LINE *)0
 #define set_lforw(a,b)	lforw(a) = (b)
 #define set_lback(a,b)	lback(a) = (b)
 #define lforw(lp)	(lp)->l_fp
@@ -1611,10 +1726,10 @@ typedef struct	LINE {
 	 * Macros for referencing fields in the LINE struct.
 	 */
 #define lgetc(lp, n)		char2int(lvalue(lp)[(n)])
-#define lputc(lp, n, c) 	(lvalue(lp)[(n)]=(char)(c))
+#define lputc(lp, n, c)		(lvalue(lp)[(n)]=(char)(c))
 #define lvalue(lp)		((lp)->l_text)
 #define llength(lp)		((lp)->l_used)
-#define line_length(lp)		(llength(lp)+len_rs) /* counting recordsep */
+#define line_length(lp)		((B_COUNT)(llength(lp))+len_rs) /* counting recordsep */
 
 #define liscopied(lp)		((lp)->l.l_undo_cookie == current_undo_cookie)
 #define lsetcopied(lp)		((lp)->l.l_undo_cookie = current_undo_cookie)
@@ -1640,7 +1755,7 @@ typedef struct	LINE {
 
 /* marks are a line and an offset into that line */
 typedef struct MARK {
-	LINEPTR l;
+	LINE *l;
 	C_NUM o;
 } MARK;
 
@@ -1660,11 +1775,11 @@ typedef struct MARK {
  * characters, is kept in a region structure.  Used by the region commands.
  */
 typedef struct	{
-	MARK	r_orig; 		/* Origin LINE address. 	*/
-	MARK	r_end;			/* Ending LINE address. 	*/
-	C_NUM	r_leftcol;		/* Leftmost column. 		*/
-	C_NUM	r_rightcol;		/* Rightmost column. 		*/
-	B_COUNT r_size; 		/* Length in characters.	*/
+	MARK	r_orig;			/* Origin LINE address.		*/
+	MARK	r_end;			/* Ending LINE address.		*/
+	C_NUM	r_leftcol;		/* Leftmost column.		*/
+	C_NUM	r_rightcol;		/* Rightmost column.		*/
+	B_COUNT r_size;			/* Length in characters.	*/
 #if OPT_SELECTIONS
 	USHORT	r_attr_id;		/* id of corresponding display  */
 #endif
@@ -1772,7 +1887,7 @@ typedef struct vl_aregion {
 
 /* Hash table entry for line attribute indices. */
 typedef struct vl_line_attr_entry {
-    	VIDEO_ATTR	vattr;
+	VIDEO_ATTR	vattr;
 	char		in_use;
 }	LINE_ATTR_ENTRY;
 
@@ -1800,7 +1915,9 @@ typedef	struct	{
 	struct VAL      *global;
 } VALARGS;
 
-#define CHGD_ARGS VALARGS *args, int glob_vals, int testing
+struct	BUFFER;
+
+#define CHGD_ARGS struct BUFFER *bp, VALARGS *args, int glob_vals, int testing
 
 typedef	int	(*ChgdFunc) ( CHGD_ARGS );
 
@@ -1930,13 +2047,14 @@ typedef struct MINORMODE {
 } MINORMODE;
 
 typedef struct {
-	char *name;
+	char *shortname;	/* name without "mode" suffix */
+	char *longname;		/* name with "mode" suffix */
 	M_VALUES mm;
 	Q_VALUES mq;
 	MINORMODE *sm;
 } MAJORMODE;
 
-#define is_c_mode(bp) (bp->majr != 0 && !strcmp(bp->majr->name, "c"))
+#define is_c_mode(bp) (bp->majr != 0 && !strcmp(bp->majr->shortname, "c"))
 #define fix_cmode(bp,value)	/* nothing */
 #define for_each_modegroup(bp,n,m,vals) \
 	for (vals = get_submode_vals(bp, n = m); vals != 0; vals = get_submode_valx(bp, m, &n))
@@ -1971,13 +2089,11 @@ typedef int FUID;
  * read in at "use buffer" time.
  */
 
-struct	BUFFER;
-
 typedef	int	(*UpBuffFunc) ( struct BUFFER * );
 
 typedef struct	BUFFER {
-	MARK	b_line; 	/* Link to the header LINE (offset unused) */
-	struct	BUFFER *b_bufp; 	/* Link to next BUFFER		*/
+	MARK	b_line;		/* Link to the header LINE (offset unused) */
+	struct	BUFFER *b_bufp;		/* Link to next BUFFER		*/
 	MARK	*b_nmmarks;		/* named marks a-z		*/
 #if OPT_SELECTIONS
 	AREGION	*b_attribs;		/* attributed regions		*/
@@ -1992,27 +2108,29 @@ typedef struct	BUFFER {
 	B_COUNT	b_bytecount;		/* # of chars			*/
 	L_NUM	b_linecount;		/* no. lines in buffer		*/
 	L_NUM	b_lines_on_disk;	/* no. lines as of last read/write */
-	LINEPTR b_udstks[2];		/* undo stack pointers		*/
+	LINE	*b_udstks[2];		/* undo stack pointers		*/
 	MARK	b_uddot[2];		/* Link to "." before undoable op */
 	short	b_udstkindx;		/* which of above to use	*/
-	LINEPTR b_udtail;		/* tail of undo backstack	*/
-	LINEPTR b_udlastsep;		/* last stack separator pushed	*/
+	LINE	*b_udtail;		/* tail of undo backstack	*/
+	LINE	*b_udlastsep;		/* last stack separator pushed	*/
 	int	b_udcount;		/* how many undo's we can do	*/
-	LINEPTR	b_LINEs;		/* block-malloced LINE structs	*/
-	LINEPTR b_LINEs_end;		/* end of	"	"	*/
-	LINEPTR b_freeLINEs;		/* list of free "	"	*/
+	LINE	*b_LINEs;		/* block-malloced LINE structs	*/
+	LINE	*b_LINEs_end;		/* end of	"	"	*/
+	LINE	*b_freeLINEs;		/* list of free "	"	*/
 	UCHAR	*b_ltext;		/* block-malloced text		*/
 	UCHAR	*b_ltext_end;		/* end of block-malloced text	*/
-	LINEPTR	b_ulinep;		/* pointer at 'Undo' line	*/
+	LINE	*b_ulinep;		/* pointer at 'Undo' line	*/
 	int	b_active;		/* window activated flag	*/
 	int	b_refcount;		/* counts levels of source'ing	*/
-	UINT	b_nwnd; 		/* Count of windows on buffer	*/
-	UINT	b_flag; 		/* Flags			*/
+	UINT	b_nwnd;			/* Count of windows on buffer	*/
+	UINT	b_flag;			/* Flags			*/
 	short	b_inuse;		/* nonzero if executing macro	*/
 	short	b_acount;		/* auto-save count		*/
+	const char *b_recordsep_str;	/* string for recordsep		*/
+	int	b_recordsep_len;	/* ...its length		*/
 	char	*b_fname;		/* File name			*/
 	int	b_fnlen;		/* length of filename		*/
-	char	b_bname[NBUFN]; 	/* Buffer name			*/
+	char	b_bname[NBUFN];		/* Buffer name			*/
 #if	OPT_ENCRYPT
 	char	b_cryptkey[NKEYLEN];	/* encryption key		*/
 #endif
@@ -2040,7 +2158,7 @@ typedef struct	BUFFER {
 #if OPT_HILITEMATCH
 	USHORT	b_highlight;
 #endif
-#if OPT_PERL || OPT_TCL
+#if OPT_PERL || OPT_TCL || OPT_PLUGIN
 	void *	b_api_private;		/* pointer to private perl, tcl, etc.
 					   data */
 #endif
@@ -2127,6 +2245,9 @@ typedef struct	BUFFER {
 #define BFUPBUFF   iBIT(8)	/* set if buffer should be updated */
 #define BFRCHG     iBIT(9)	/* Changed since last reset of this flag*/
 #define BFISREAD   iBIT(10)	/* set if we are reading data into buffer */
+#define BFREGD     iBIT(11)	/* set if file path written to registry
+				 * (winvile feature)
+				 */
 
 /* macros for manipulating b_flag */
 #define b_is_set(bp,flags)        (((bp)->b_flag & (flags)) != 0)
@@ -2141,6 +2262,12 @@ typedef struct	BUFFER {
 #define b_is_recentlychanged(bp)  b_is_set(bp, BFRCHG)
 #define b_is_scratch(bp)          b_is_set(bp, BFSCRTCH)
 #define b_is_temporary(bp)        b_is_set(bp, BFINVS|BFSCRTCH)
+/*
+ * don't write file paths to registry if already written _or_ if dealing
+ * with a non-file and/or transient buffer.
+ */
+#define b_is_registered(bp)        b_is_set(bp, \
+                                       BFINVS|BFSCRTCH|BFREGD|BFEXEC|BFDIRS)
 
 #define b_set_flags(bp,flags)     (bp)->b_flag |= (flags)
 #define b_set_changed(bp)         b_set_flags(bp, BFCHG)
@@ -2150,6 +2277,7 @@ typedef struct	BUFFER {
 #define b_set_reading(bp)         b_set_flags(bp, BFISREAD)
 #define b_set_recentlychanged(bp) b_set_flags(bp, BFRCHG)
 #define b_set_scratch(bp)         b_set_flags(bp, BFSCRTCH)
+#define b_set_registered(bp)      b_set_flags(bp, BFREGD)
 
 #define b_clr_flags(bp,flags)     (bp)->b_flag &= ~(flags)
 #define b_clr_changed(bp)         b_clr_flags(bp, BFCHG)
@@ -2159,6 +2287,7 @@ typedef struct	BUFFER {
 #define b_clr_reading(bp)         b_clr_flags(bp, BFISREAD)
 #define b_clr_recentlychanged(bp) b_clr_flags(bp, BFRCHG)
 #define b_clr_scratch(bp)         b_clr_flags(bp, BFSCRTCH)
+#define b_clr_registered(bp)      b_clr_flags(bp, BFREGD)
 
 #if OPT_HILITEMATCH
 #define b_match_attrs_dirty(bp)	(bp)->b_highlight |= HILITE_DIRTY
@@ -2179,9 +2308,12 @@ typedef struct	BUFFER {
 #define HILITE_DIRTY	2
 #endif
 
+#define len_record_sep(bp)	((bp)->b_recordsep_len)
+#define get_record_sep(bp)	((bp)->b_recordsep_str)
+
 /* macro for iterating over the marks associated with the current buffer */
 
-#if OPT_PERL || OPT_TCL
+#if OPT_PERL || OPT_TCL || OPT_PLUGIN
 extern MARK *api_mark_iterator(BUFFER *bp, int *iter);
 #define api_do_mark_iterate_helper(mp, statement)	\
 	{						\
@@ -2247,7 +2379,7 @@ extern MARK *api_mark_iterator(BUFFER *bp, int *iter);
 typedef struct	WINDOW {
 	W_TRAITS w_traits;		/* features of the window we should */
 					/* remember between displays	*/
-	struct	WINDOW *w_wndp; 	/* Next window			*/
+	struct	WINDOW *w_wndp;		/* Next window			*/
 	BUFFER	*w_bufp;		/* Buffer displayed in window	*/
 	int	w_toprow;		/* Origin 0 top row of window	*/
 	int	w_ntrows;		/* # of rows of text in window	*/
@@ -2323,10 +2455,10 @@ typedef struct	WINDOW {
  * one terminal type.
  */
 typedef struct	{
-	int	maxrows;		/* max rows count 		*/
-	int	rows; 			/* current row count		*/
-	int	maxcols; 		/* max column count		*/
-	int	cols; 			/* current column count		*/
+	int	maxrows;		/* max rows count		*/
+	int	rows;			/* current row count		*/
+	int	maxcols;		/* max column count		*/
+	int	cols;			/* current column count		*/
 	int	pausecount;		/* # times thru update to pause */
 	void	(*open) (void);		/* Open terminal at the start.	*/
 	void	(*close) (void);	/* Close terminal at end.	*/
@@ -2347,7 +2479,7 @@ typedef struct	{
 	void	(*setpal) (const char *p); /* set color palette	*/
 	void	(*setccol) (int c);	/* set cursor color		*/
 	void	(*scroll) (int from, int to, int n); /* scroll region	*/
-	void	(*pflush) (void);	/* really flush 		*/
+	void	(*pflush) (void);	/* really flush			*/
 	void	(*icursor) (int c);	/* set cursor shape for insertion */
 	void	(*set_title) (const char *t); /* set window title	*/
 	int	(*watchfd)(int, WATCHTYPE, long *);
@@ -2391,7 +2523,7 @@ typedef struct  VIDEO {
 #define ReqBcolor(vp) gbcolor
 #endif
 
-#define VFCHG	iBIT(0) 		/* Changed flag 		*/
+#define VFCHG	iBIT(0)			/* Changed flag			*/
 #define	VFEXT	iBIT(1)			/* extended (beyond column 80)	*/
 #define	VFREV	iBIT(2)			/* reverse video status		*/
 #define	VFREQ	iBIT(3)			/* reverse video request	*/
@@ -2505,7 +2637,7 @@ typedef struct {
 #endif
 
 typedef struct  k_bind {
-	int	k_code; 		/* Key code			*/
+	int	k_code;			/* Key code			*/
 	const CMDFUNC *k_cmd;
 #if OPT_REBIND
 	struct  k_bind *k_link;
@@ -2592,6 +2724,7 @@ typedef enum {
 	,file_is_external
 	,file_is_pipe
 	,file_is_internal
+	,file_is_new          /* winvile-specific functionality */
 } FFType;
 
 /* definitions for 'mlreply_file()' and other filename-completion */
@@ -2683,36 +2816,33 @@ extern void _exit (int code);
 /* array/table size */
 #define	TABLESIZE(v)	(sizeof(v)/sizeof(v[0]))
 
+#define PERCENT(num,den) ((den) ? (int)((100.0 * (num))/(den)) : 100)
+
 /* Quiet compiler warnings on places where we're being blamed incorrectly,
  * e.g., for casting away const, or for alignment problems.  It's always
- * legal to cast a pointer to long w/o loss of precision.
+ * legal in c89 to cast a pointer to long w/o loss of precision.
+ *
+ * FIXME: c99 may require a wider type.
  */
+#ifdef __GNUC__
 #define TYPECAST(type,ptr) (type*)((long)(ptr))
-
-/* structure-allocate, for linting */
-#if	0 /* this was useful for K&R lint, but we don't use it anymore */
-#define	castalloc(cast,nbytes)		((cast *)0)
-#define	castrealloc(cast,ptr,nbytes)	((ptr)+(nbytes))
-#define	typecalloc(cast)		((cast *)0)
-#define	typecallocn(cast,ntypes)	(((cast *)0)+(ntypes))
-#define	typealloc(cast)			((cast *)0)
-#define	typeallocn(cast,ntypes)		(((cast *)0)+(ntypes))
-#define	typereallocn(cast,ptr,ntypes)	(TYPECAST(cast,ptr)+(ntypes))
-#define	typeallocplus(cast,extra)	(((cast *)0)+(extra))
 #else
+#define TYPECAST(type,ptr) (type*)(ptr)
+#endif
+
+/* structure-allocate, for appeasing strict compilers */
 #define	castalloc(cast,nbytes)		(cast *)malloc(nbytes)
-#define	castrealloc(cast,ptr,nbytes)	(cast *)realloc((char *)(ptr),(nbytes))
+#define	castrealloc(cast,ptr,nbytes)	(cast *)realloc((ptr),(nbytes))
 #define	typecalloc(cast)		(cast *)calloc(sizeof(cast),1)
 #define	typecallocn(cast,ntypes)	(cast *)calloc(sizeof(cast),ntypes)
 #define	typealloc(cast)			(cast *)malloc(sizeof(cast))
 #define	typeallocn(cast,ntypes)		(cast *)malloc((ntypes)*sizeof(cast))
-#define	typereallocn(cast,ptr,ntypes)	(cast *)realloc((char *)(ptr),\
+#define	typereallocn(cast,ptr,ntypes)	(cast *)realloc((ptr),\
 							(ntypes)*sizeof(cast))
 #define	typeallocplus(cast,extra)	(cast *)calloc((extra)+sizeof(cast),1)
-#endif
 
-#define	FreeAndNull(p)	if ((p) != 0) { free((char *)p); p = 0; }
-#define	FreeIfNeeded(p)	if ((p) != 0) free((char *)(p))
+#define	FreeAndNull(p)	if ((p) != 0)	{ free(p); p = 0; }
+#define	FreeIfNeeded(p)	if ((p) != 0)	free(p)
 
 #if defined(VILE_ERROR_ABORT)
 extern void ExitProgram(int code);
@@ -2726,6 +2856,8 @@ extern void ExitProgram(int code);
 #endif
 #endif
 
+#define VILE_SUBKEY "Software\\VI Like Emacs"
+
 /*
  * We cannot define these in config.h, since they require parameters to be
  * passed (that's non-portable).
@@ -2750,7 +2882,7 @@ extern void ExitProgram(int code);
 #define	GCC_UNUSED /* nothing */
 #endif
 
-#if 0				/* requires a patch to gcc */
+#if 1				/* requires a patch to gcc */
 #define VILE_PRINTF(fmt,var) GCC_PRINTFLIKE(fmt,var)
 #else
 #define VILE_PRINTF(fmt,var) /*nothing*/
@@ -2813,6 +2945,9 @@ extern void ExitProgram(int code);
 #ifndef	USE_DMALLOC	/* test malloc/free/strcpy/memcpy, etc. */
 #define	USE_DMALLOC	0
 #endif
+#ifndef	USE_MPATROL	/* test malloc/free/strcpy/memcpy, etc. */
+#define	USE_MPATROL	0
+#endif
 #ifndef	NO_LEAKS	/* free permanent memory, analyze leaks */
 #define	NO_LEAKS	0
 #endif
@@ -2852,11 +2987,15 @@ extern void ExitProgram(int code);
 #    include <dmalloc.h>
 #    define show_alloc() dmalloc_log_unfreed()
 #  endif
-#  if OPT_TRACE
+#  if USE_MPATROL
+#    include <mpatrol.h>
+#    define show_alloc() void __mp_summary()
+#  endif
+#  if CAN_TRACE && OPT_TRACE
 #    include "trace.h"
 #  endif
 #else
-#  if NO_LEAKS || DOALLOC || OPT_TRACE
+#  if CAN_TRACE && (NO_LEAKS || DOALLOC || OPT_TRACE)
 #    include "trace.h"
 #  endif
 #endif	/* USE_DBMALLOC */
@@ -2867,10 +3006,16 @@ extern void ExitProgram(int code);
 
 /* extra checking if we're tracing */
 #if !OPT_TRACE
-#define valid_buffer(bp) ((bp) != NULL)
-#define valid_window(wp) ((wp) != NULL)
+#define NDEBUG			/* turn off assert's */
+#define valid_buffer(bp)        ((bp) != NULL)
+#define valid_window(wp)        ((wp) != NULL)
+#define valid_line_bp(lp,bp)    ((lp) != NULL)
+#define valid_line_wp(lp,wp)    ((lp) != NULL)
 #endif
 
+/* this must be after NDEBUG is defined */
+#include <assert.h>
+
 /* Normally defined in "trace.h" */
 #ifndef TRACE
 #define TRACE(p) /* nothing */
@@ -2881,12 +3026,14 @@ extern void ExitProgram(int code);
 
 #if OPT_TRACE > 1
 #define TRACE2(params) TRACE(params)
+#define return2Code(c)  returnCode(c)
 #else
 #define TRACE2(params) /*nothing*/
+#define return2Code(c)  return(c)
 #endif
 
 #if OPT_EVAL || OPT_DEBUGMACROS
-#define TPRINTF(p) if (tracemacros) tprintf p
+#define TPRINTF(p) { TRACE(p); if (tracemacros) tprintf p; }
 #else
 #define TPRINTF(p) /* nothing */
 #endif
@@ -2923,7 +3070,7 @@ extern void ExitProgram(int code);
 #endif
 
 /* for debugging VMS pathnames on UNIX... */
-#if SYS_UNIX && OPT_VMS_PATH
+#if CAN_VMS_PATH && (SYS_UNIX && OPT_VMS_PATH)
 #include "fakevms.h"
 #endif
 
