@@ -2,7 +2,7 @@
  *	eval.c -- function and variable evaluation
  *	original by Daniel Lawrence
  *
- * $Header: /usr/build/vile/vile/RCS/eval.c,v 1.322 2003/05/25 23:34:52 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/eval.c,v 1.348 2005/05/27 22:21:58 tom Exp $
  *
  */
 
@@ -65,21 +65,25 @@ makectypelist(int dum1 GCC_UNUSED, void *ptr GCC_UNUSED)
     CHARTYPE k;
     const char *s;
 
-    bprintf("--- Printable Characters %*P\n", term.cols - 1, '-');
+    bprintf("--- Printable Characters for locale %s (%s) ",
+	    NONNULL(vl_locale),
+	    NONNULL(vl_encoding));
+    bpadc('-', term.cols - DOT.o);
+    bputc('\n');
+
     for (i = 0; i < N_chars; i++) {
-	bprintf("\n%d\t", i);
+	bprintf("\n%d\t", (int) i);
 	if ((i == '\n') || (i == '\t'))		/* vtlistc() may not do these */
-	    bprintf("^%c", '@' | i);
+	    bprintf("^%c", (int) ('@' | i));
 #if OPT_LOCALE
 	else if (!isPrint(i) && i > 127 && i < 160)	/* C1 controls? */
-	    bprintf(
-		       global_w_val(WMDNONPRINTOCTAL)
-		       ? "\\%3o"
-		       : "\\x%2x",
-		       i);
+	    bprintf(global_w_val(WMDNONPRINTOCTAL)
+		    ? "\\%3o"
+		    : "\\x%2x",
+		    i);
 #endif
 	else
-	    bprintf("%c", i);
+	    bprintf("%c", (int) i);
 	bputc('\t');
 	for (j = 0; j != vl_UNUSED; j++) {
 	    if ((s = choice_to_name(fsm_charclass_choices, j)) != 0) {
@@ -87,7 +91,7 @@ makectypelist(int dum1 GCC_UNUSED, void *ptr GCC_UNUSED)
 		if (j != 0)
 		    bputc(' ');
 		bprintf("%*s",
-			strlen(s),
+			(int) strlen(s),
 			(vl_chartypes_[i] & k)
 			? s
 			: "-");
@@ -147,10 +151,9 @@ get_charclass_code(void)
 		       KBD_NOEVAL | KBD_LOWERC,
 		       cclass_complete);
     return (status == TRUE)
-	? choice_to_code(
-			    fsm_charclass_choices,
-			    tb_values(var),
-			    tb_length(var))
+	? choice_to_code(fsm_charclass_choices,
+			 tb_values(var),
+			 tb_length(var))
 	: -1;
 }
 
@@ -169,9 +172,9 @@ get_charclass_regexp(void)
 	tb_scopy(&var, "");
     status = kbd_reply("Character pattern: ", &var,
 		       eol_history, '=', KBD_NOEVAL, no_completion);
-    return (status == TRUE)
-	? new_regexval(tb_values(var), TRUE)
-	: 0;
+    return ((status == TRUE)
+	    ? new_regexval(tb_values(var), TRUE)
+	    : 0);
 }
 
 static int
@@ -283,7 +286,7 @@ charclass_of(const char *arg)
  * Make display of the intersection of the character classes in the argument.
  */
 static void
-show_charclass(TBUFF ** result, const char *arg)
+show_charclass(TBUFF **result, const char *arg)
 {
     CHARTYPE k = charclass_of(arg);
 #if OPT_SHOW_CTYPE
@@ -315,12 +318,15 @@ makevarslist(int dum1 GCC_UNUSED, void *ptr)
     UVAR *p;
     int j, k;
 
-    bprintf("--- State variables %*P\n", term.cols - 1, '-');
+    bprintf("--- State variables ");
+    bpadc('-', term.cols - DOT.o);
+    bputc('\n');
+
     bprintf("%s", (char *) ptr);
     if (arg_stack != 0 && ((k = arg_stack->num_args) != 0)) {
-	bprintf("--- %s parameters %*P",
-		tb_values(arg_stack->all_args[0]),
-		term.cols - 1, '-');
+	bprintf("--- %s parameters ",
+		tb_values(arg_stack->all_args[0]));
+	bpadc('-', term.cols - DOT.o);
 	for (j = 1; j <= k; j++) {
 	    bprintf("\n$%d = %s", j,
 		    tb_values(arg_stack->all_args[j]));
@@ -330,7 +336,8 @@ makevarslist(int dum1 GCC_UNUSED, void *ptr)
 	bputc('\n');
     for (p = temp_vars, j = 0; p != 0; p = p->next) {
 	if (!j++)
-	    bprintf("--- Temporary variables %*P", term.cols - 1, '-');
+	    bprintf("--- Temporary variables ");
+	bpadc('-', term.cols - DOT.o);
 	bprintf("\n%%%s = %s", p->u_name, p->u_value);
     }
 }
@@ -357,7 +364,7 @@ static const char *
 get_listvalue(const char *name, int showall)
 {
     VALARGS args;
-    register int s;
+    int s;
     int vnum;
 
     if ((s = is_mode_name(name, showall, &args)) == TRUE)
@@ -381,13 +388,13 @@ static int
 show_VariableList(BUFFER *bp GCC_UNUSED)
 {
     char *values;
-    register WINDOW *wp = curwp;
-    register int s;
-    register size_t t;
-    register char *v;
-    register const char *vv;
+    WINDOW *wp = curwp;
+    int s;
+    size_t t;
+    char *v;
+    const char *vv;
     static const char fmt[] =
-    {"$%s = %*S\n"};
+    {"$%s = %.*s\n"};
     const char *const *Names;
     int showall;
     int rc = FALSE;
@@ -395,34 +402,36 @@ show_VariableList(BUFFER *bp GCC_UNUSED)
     TRACE((T_CALLED "show_VariableList()\n"));
 
     Names = show_vars_f ? list_of_modes() : statevars;
-    showall = show_vars_f ? (show_vars_n > 1) : FALSE;
-    /* collect data for state-variables, since some depend on window */
-    for (s = t = 0; Names[s] != 0; s++) {
-	if ((vv = get_listvalue(Names[s], showall)) != 0)
-	    t += strlen(Names[s]) + strlen(fmt) + strlen(vv);
-    }
-
-    beginDisplay();
-    if ((values = typeallocn(char, t)) != 0) {
-	for (s = 0, v = values; Names[s] != 0; s++) {
-	    if ((vv = get_listvalue(Names[s], showall)) != 0) {
-		t = strlen(vv);
-		if (t == 0) {
-		    t = 1;
-		    vv = "";
-		} else if (vv[t - 1] == '\n')
-		    t--;	/* suppress trailing newline */
-		v = lsprintf(v, fmt, Names[s], t, vv);
-	    }
+    if (Names != 0) {
+	showall = show_vars_f ? (show_vars_n > 1) : FALSE;
+	/* collect data for state-variables, since some depend on window */
+	for (s = t = 0; Names[s] != 0; s++) {
+	    if ((vv = get_listvalue(Names[s], showall)) != 0)
+		t += strlen(Names[s]) + strlen(fmt) + strlen(vv);
 	}
-	rc = liststuff(VARIABLES_BufName, FALSE,
-		       makevarslist, 0, (void *) values);
-	free(values);
 
-	/* back to the buffer whose modes we just listed */
-	swbuffer(wp->w_bufp);
+	beginDisplay();
+	if ((values = typeallocn(char, t)) != 0) {
+	    for (s = 0, v = values; Names[s] != 0; s++) {
+		if ((vv = get_listvalue(Names[s], showall)) != 0) {
+		    t = strlen(vv);
+		    if (t == 0) {
+			t = 1;
+			vv = "";
+		    } else if (vv[t - 1] == '\n')
+			t--;	/* suppress trailing newline */
+		    v = lsprintf(v, fmt, Names[s], t, vv);
+		}
+	    }
+	    rc = liststuff(VARIABLES_BufName, FALSE,
+			   makevarslist, 0, (void *) values);
+	    free(values);
+
+	    /* back to the buffer whose modes we just listed */
+	    swbuffer(wp->w_bufp);
+	}
+	endofDisplay();
     }
-    endofDisplay();
     returnCode(rc);
 }
 
@@ -471,6 +480,7 @@ vl_lookup_func(const char *name)
 	for (n = 0; n < NFUNCS; n++) {
 	    if (!strncmp(downcased, vl_ufuncs[n].f_name, m)) {
 		if ((n + 1 >= NFUNCS
+		     || m == strlen(vl_ufuncs[n].f_name)
 		     || strncmp(downcased, vl_ufuncs[n + 1].f_name, m)))
 		    fnum = n;
 		break;
@@ -505,7 +515,7 @@ search_token(const char *token, const char *delims, const char *string)
  * result only if we run past the end of the number of tokens.
  */
 static void
-extract_token(TBUFF ** result, const char *count, const char *delims, const char *string)
+extract_token(TBUFF **result, const char *count, const char *delims, const char *string)
 {
     int num = scan_int(count);
 
@@ -524,16 +534,18 @@ extract_token(TBUFF ** result, const char *count, const char *delims, const char
  * Translate string 'from' 'to'.
  */
 static void
-translate_string(TBUFF ** result, const char *from, const char *to, const char *string)
+translate_string(TBUFF **result, const char *from, const char *to, const char *string)
 {
     int len_to = strlen(to);
     char *s, *t;
 
     if (tb_scopy(result, string) != 0) {
-	for (s = tb_values(*result); *s != EOS; ++s) {
-	    if ((t = strchr(from, *s)) != 0
-		&& (t - from) < len_to) {
-		*s = to[t - from];
+	if ((s = tb_values(*result)) != 0) {
+	    for (; *s != EOS; ++s) {
+		if ((t = strchr(from, *s)) != 0
+		    && (t - from) < len_to) {
+		    *s = to[t - from];
+		}
 	    }
 	}
     }
@@ -568,10 +580,10 @@ path_trim(char *path)
  * Extract the head of the given path
  */
 static char *
-path_head(TBUFF ** result, const char *path)
+path_head(TBUFF **result, const char *path)
 {
-    char *temp;
-    if ((temp = tb_values(tb_scopy(result, path))) != error_val) {
+    char *temp = tb_values(tb_scopy(result, path));
+    if (temp != error_val && temp != 0) {
 	*pathleaf(temp) = EOS;
 	path_trim(temp);
     }
@@ -583,7 +595,7 @@ path_head(TBUFF ** result, const char *path)
  * simply copy it.
  */
 static void
-path_quote(TBUFF ** result, const char *path)
+path_quote(TBUFF **result, const char *path)
 {
 #if SYS_VMS
     tb_scopy(result, path);	/* no point in quoting here */
@@ -612,7 +624,7 @@ path_quote(TBUFF ** result, const char *path)
 }
 
 static void
-render_date(TBUFF ** result, char *format, time_t stamp)
+render_date(TBUFF **result, char *format, time_t stamp)
 {
 #ifdef HAVE_STRFTIME
     struct tm *tm = localtime(&stamp);
@@ -634,7 +646,7 @@ render_date(TBUFF ** result, char *format, time_t stamp)
 }
 
 static void
-default_mode_value(TBUFF ** result, char *name)
+default_mode_value(TBUFF **result, char *name)
 {
     VALARGS args;
     VWRAP vd;
@@ -674,65 +686,162 @@ default_mode_value(TBUFF ** result, char *name)
     }
 }
 
-#define MAXARGS 3
-
-/* quote string, so toktyp() will know what it is */
-static void
-render_string(TBUFF ** rp)
+/*
+ * Based on mac_tokval(), but unlike that, strips quotes which came from
+ * evaluating via run_func().
+ */
+static char *
+dequoted_parameter(TBUFF **tok)
 {
-    char *value = tb_values(*rp);
+    char *previous;
+    const char *newvalue;
+    int strip;
 
-    if (value != error_val
-	&& *value != EOS
-	&& toktyp(value) != TOK_LITSTR
-	&& toktyp(value) != TOK_QUOTSTR) {
-	UINT j;
-	UINT have = strlen(value);	/* FIXME: assumes we used EOS */
-	UINT need = 2 + have;
+    if (mac_token(tok) != 0) {
+	previous = tb_values(*tok);
 
-	for (j = 0; j < have; ++j) {
-	    if (value[j] == SQUOTE
-		|| value[j] == BACKSLASH) {
-		++need;
+	switch (toktyp(previous)) {
+	default:
+	case TOK_NULL:
+	case TOK_QUERY:
+	case TOK_BUFLINE:
+	case TOK_TEMPVAR:
+	case TOK_DIRECTIVE:
+	case TOK_LABEL:
+	case TOK_LITSTR:
+	    newvalue = tokval(previous);
+	    strip = FALSE;
+	    break;
+	case TOK_FUNCTION:
+	    newvalue = tokval(previous);
+	    strip = (toktyp(newvalue) == TOK_QUOTSTR);
+	    break;
+	case TOK_QUOTSTR:
+	    strip = TRUE;
+	    newvalue = previous;
+	    break;
+	}
+
+	if (newvalue == (const char *) error_val) {
+	    return error_val;
+	} else if (strip) {
+	    if (((const char *) previous) != newvalue) {
+		TBUFF *fix = 0;
+		tb_scopy(&fix, newvalue);
+		tb_free(tok);
+		*tok = fix;
 	    }
+	    tb_dequote(tok);
+	} else if (((const char *) previous) != newvalue) {
+	    tb_scopy(tok, newvalue);
 	}
-	tb_alloc(rp, need + 1);
-	(*rp)->tb_used = need + 1;
-#if 0
-	/*
-	 * FIXME:  We _should_ quote the result just as if it had come from a
-	 * file.  But some inputs come from a prompt, which does not (yet)
-	 * quote its input.  Taking out the check for empty string above, the
-	 * following loop would quote the string "properly".  But returning ''
-	 * would break for various reasons.
-	 */
-	value[need] = EOS;
-	value[need - 1] = SQUOTE;
-	for (j = 0; j < have; ++j) {
-	    UINT i = have - j - 1;
-	    UINT k = need - j - 2;
-	    value[k] = value[i];
-	    if (value[k] == SQUOTE
-		|| value[k] == BACKSLASH) {
-		--need;
-		value[k - 1] = BACKSLASH;
-	    }
-	}
-#else
-	/*
-	 * What we actually do is simpler: put a quote character at the front
-	 * of the data so it will be treated as a string.
-	 */
-	value[have + 1] = EOS;
-	for (j = 0; j < have; ++j) {
-	    UINT i = have - j;
-	    value[i] = value[i - 1];
-	}
-#endif
-	value[0] = SQUOTE;
+	return (tb_values(*tok));
     }
+    tb_free(tok);
+    return (0);
 }
 
+#define UNI_CLASSNAME "universal"
+#define BUF_CLASSNAME "buffer"
+#define WIN_CLASSNAME "window"
+#define SUB_CLASSNAME "submode"
+#define MAJ_CLASSNAME "Majormode"
+
+#define COLOR_CLASSNAME "color"
+#define MODES_CLASSNAME "mode"
+
+/*
+ * Check if the given value is an instance of the given class.  That's easier
+ * to work with in scripts than trying to use a feature and (then) trying to
+ * recover if it doesn't exist.
+ *
+ * Note:  The classnames are chosen to allow single character abbreviations.
+ */
+static int
+test_isa_class(const char *classname, const char *value)
+{
+    VALARGS args;
+    int result = FALSE;
+    size_t len = strlen(classname);
+
+    if (!strncmp(classname, BUF_CLASSNAME, len)) {
+	result = (find_b_name(value) != 0);
+    } else if (!strncmp(classname, COLOR_CLASSNAME, len)) {
+#if OPT_COLOR_CHOICES
+	result = (choice_to_code(fsm_color_choices, value, len) != ENUM_ILLEGAL);
+#endif
+    } else if (!strncmp(classname, MODES_CLASSNAME, len)) {
+	if (find_mode_class(curbp, value, TRUE, &args, UNI_MODE)
+	    || find_mode_class(curbp, value, TRUE, &args, BUF_MODE)
+	    || find_mode_class(curbp, value, TRUE, &args, WIN_MODE))
+	    result = TRUE;
+#if OPT_MAJORMODE
+    } else if (!strncmp(classname, SUB_CLASSNAME, len)) {
+	result = find_mode_class(curbp, value, TRUE, &args, SUB_MODE);
+    } else if (!strncmp(classname, MAJ_CLASSNAME, len)) {
+	result = find_mode_class(curbp, value, TRUE, &args, MAJ_MODE);
+#endif
+    }
+    return result;
+}
+
+/*
+ * Reverse of the "&isa" operator.
+ */
+static void
+find_classof(TBUFF **result, const char *value)
+{
+    static const char *table[] =
+    {
+	BUF_CLASSNAME,
+	COLOR_CLASSNAME,
+	MODES_CLASSNAME,
+	SUB_CLASSNAME,
+	MAJ_CLASSNAME
+    };
+    size_t n;
+    int found = FALSE;
+
+    tb_init(result, EOS);
+    for (n = 0; n < TABLESIZE(table); ++n) {
+	if (test_isa_class(table[n], value)) {
+	    found = TRUE;
+	    if (tb_length(*result))
+		tb_sappend0(result, " ");
+	    tb_sappend0(result, table[n]);
+	    break;
+	}
+    }
+    if (!found)
+	tb_error(result);
+}
+
+/*
+ * Find the mode's class (universal, buffer, window, major).
+ */
+static void
+find_modeclass(TBUFF **result, const char *value)
+{
+    VALARGS args;
+
+    if (find_mode_class(curbp, value, TRUE, &args, UNI_MODE))
+	tb_scopy(result, UNI_CLASSNAME);
+    else if (find_mode_class(curbp, value, TRUE, &args, BUF_MODE))
+	tb_scopy(result, BUF_CLASSNAME);
+    else if (find_mode_class(curbp, value, TRUE, &args, WIN_MODE))
+	tb_scopy(result, WIN_CLASSNAME);
+#if OPT_MAJORMODE
+    else if (find_mode_class(curbp, value, TRUE, &args, SUB_MODE))
+	tb_scopy(result, SUB_CLASSNAME);
+    else if (find_mode_class(curbp, value, TRUE, &args, MAJ_MODE))
+	tb_scopy(result, MAJ_CLASSNAME);
+#endif
+    else
+	tb_error(result);
+}
+
+#define MAXARGS 3
+
 /*
  * execute a builtin function
  */
@@ -741,8 +850,10 @@ run_func(int fnum)
 {
     static TBUFF *result;	/* function result */
 
+    REGEXVAL *exp;
     BUFFER *bp;
     TBUFF *args[MAXARGS];
+    TBUFF *juggle = 0;
     char *arg[MAXARGS];		/* function arguments */
     char *cp;
     const char *sp;
@@ -753,7 +864,7 @@ run_func(int fnum)
     long value = 0;
     long nums[MAXARGS];
 
-    TRACE((T_CALLED "run_func(%d)\n", fnum));
+    TRACE((T_CALLED "run_func(%d:%s)\n", fnum, vl_ufuncs[fnum].f_name));
 
     nargs = vl_ufuncs[fnum].f_code & NARGMASK;
     args_numeric = vl_ufuncs[fnum].f_code & NUM;
@@ -762,24 +873,29 @@ run_func(int fnum)
     ret_numeric = vl_ufuncs[fnum].f_code & NRET;
     ret_boolean = vl_ufuncs[fnum].f_code & BRET;
 
-    TPRINTF(("** evaluate '%s' (0x%x), %d args\n",
+    TPRINTF(("** evaluate '%s' (0x%x), %d %s args returning %s\n",
 	     vl_ufuncs[fnum].f_name,
 	     vl_ufuncs[fnum].f_code,
-	     nargs));
+	     nargs,
+	     (args_numeric
+	      ? "numeric"
+	      : (args_boolean
+		 ? "boolean"
+		 : "string")),
+	     (ret_numeric
+	      ? "numeric"
+	      : (ret_boolean
+		 ? "boolean"
+		 : "string"))));
 
     /* fetch required arguments */
     for (i = 0; i < nargs; i++) {
 	args[i] = 0;
-	if ((arg[i] = mac_tokval(&args[i])) == 0
+	if ((arg[i] = dequoted_parameter(&args[i])) == 0
 	    || (arg[i] == error_val)
-	    || (args[i]->tb_errs)) {
+	    || isTB_ERRS(args[i])) {
 	    arg[i] = error_val;
 	    is_error = TRUE;
-	} else if (arg[i][0] == SQUOTE) {
-	    /* trim leading quote from recursive result from this function */
-	    for (cp = arg[i] + 1; (cp[-1] = cp[0]) != EOS; ++cp) {
-		;
-	    }
 	}
 	tb_free(&result);	/* in case mac_tokval() called us */
 	TPRINTF(("...arg[%d] = '%s'\n", i, arg[i]));
@@ -820,6 +936,15 @@ run_func(int fnum)
 	tb_scopy(&result, arg[0]);
 	tb_sappend0(&result, arg[1]);
 	break;
+    case UFISA:
+	value = test_isa_class(arg[0], arg[1]);
+	break;
+    case UFCLASSOF:
+	find_classof(&result, arg[0]);
+	break;
+    case UFMCLASS:
+	find_modeclass(&result, arg[0]);
+	break;
     case UFCCLASS:
 	show_charclass(&result, arg[0]);
 	break;
@@ -920,8 +1045,20 @@ run_func(int fnum)
 	tb_append(&result, EOS);
 	break;
     case UFGTSEQ:
-	(void) kcod2escape_seq(kbd_seq_nomap(), tb_values(result));
-	result->tb_used = strlen(tb_values(result));
+	(void) kcod2escape_seq(kbd_seq_nomap(), tb_values(result), result->tb_size);
+	tb_setlen(&result, -1);
+	break;
+    case UFCMATCH:
+	if ((exp = new_regexval(arg[0], TRUE)) != 0) {
+	    int save_flag = ignorecase;
+	    ignorecase = TRUE;
+	    value = regexec(exp->reg, arg[1], (char *) 0, 0, -1);
+	    ignorecase = save_flag;
+	}
+	break;
+    case UFMATCH:
+	if ((exp = new_regexval(arg[0], TRUE)) != 0)
+	    value = regexec(exp->reg, arg[1], (char *) 0, 0, -1);
 	break;
     case UFRANDOM:		/* FALLTHRU */
     case UFRND:
@@ -937,7 +1074,7 @@ run_func(int fnum)
 	break;
     case UFENV:
 	if (!is_error)
-	    tb_scopy(&result, vile_getenv(arg[0]));
+	    tb_scopy(&result, safe_getenv(arg[0]));
 	break;
     case UFBIND:
 	if (!is_error) {
@@ -1032,22 +1169,37 @@ run_func(int fnum)
 		    tb_scopy(&result, SL_TO_BSL(cp));
 		break;
 	    case PATH_FULL:
-		tb_scopy(&result, arg[1]);
-		tb_alloc(&result, NFILEN);
-		SL_TO_BSL(lengthen_path(tb_values(result)));
+		if (tb_scopy(&juggle, arg[1])
+		    && tb_alloc(&juggle, NFILEN)) {
+		    lengthen_path(tb_values(juggle));
+		    tb_setlen(&juggle, -1);
+		    tb_scopy(&result, SL_TO_BSL(tb_values(juggle)));
+		} else {
+		    is_error = TRUE;
+		}
 		break;
 	    case PATH_HEAD:
-		SL_TO_BSL(path_head(&result, arg[1]));
+		path_head(&juggle, arg[1]);
+		tb_setlen(&juggle, -1);
+		tb_scopy(&result, SL_TO_BSL(tb_values(juggle)));
 		break;
 	    case PATH_ROOT:
 		tb_scopy(&result, SL_TO_BSL(pathleaf(arg[1])));
-		if ((cp = strchr(tb_values(result), '.')) != 0)
+		if (tb_values(result) != 0
+		    && (cp = strchr(tb_values(result), '.')) != 0) {
 		    *cp = EOS;
+		    tb_setlen(&result, -1);
+		}
 		break;
 	    case PATH_SHORT:
-		tb_scopy(&result, arg[1]);
-		tb_alloc(&result, NFILEN);
-		SL_TO_BSL(shorten_path(tb_values(result), FALSE));
+		if (tb_scopy(&juggle, arg[1])
+		    && tb_alloc(&juggle, NFILEN)) {
+		    shorten_path(tb_values(juggle), FALSE);
+		    tb_setlen(&juggle, -1);
+		    tb_scopy(&result, SL_TO_BSL(tb_values(juggle)));
+		} else {
+		    is_error = TRUE;
+		}
 		break;
 	    case PATH_TAIL:
 		tb_scopy(&result, pathleaf(arg[1]));
@@ -1059,8 +1211,10 @@ run_func(int fnum)
 	}
 	break;
     case UFPATHCAT:
-	tb_alloc(&result, NFILEN);
-	SL_TO_BSL(pathcat(tb_values(result), arg[0], arg[1]));
+	tb_alloc(&juggle, NFILEN);
+	pathcat(tb_values(juggle), arg[0], arg[1]);
+	tb_setlen(&juggle, -1);
+	tb_scopy(&result, SL_TO_BSL(tb_values(juggle)));
 	break;
     case UFPATHQUOTE:
 	path_quote(&result, SL_TO_BSL(arg[0]));
@@ -1108,12 +1262,13 @@ run_func(int fnum)
     else if (ret_boolean)
 	render_boolean(&result, value);
     else
-	render_string(&result);
+	tb_enquote(&result);
 
-    TPRINTF(("-> %s'%s'\n",
+    TPRINTF(("-> %s%s\n",
 	     is_error ? "*" : "",
-	     is_error ? error_val : tb_values(result)));
+	     is_error ? error_val : NONNULL(tb_values(result))));
 
+    tb_free(&juggle);
     for (i = 0; i < nargs; i++) {
 	tb_free(&args[i]);
     }
@@ -1124,7 +1279,7 @@ run_func(int fnum)
 static UVAR *
 lookup_tempvar(const char *name)
 {
-    register UVAR *p;
+    UVAR *p;
 
     if (*name) {
 	for (p = temp_vars; p != 0; p = p->next)
@@ -1138,7 +1293,7 @@ lookup_tempvar(const char *name)
 static int
 lookup_statevar(const char *name)
 {
-    register int vnum;		/* ordinal number of var referenced */
+    int vnum;			/* ordinal number of var referenced */
     if (!*name)
 	return ILLEGAL_NUM;
 
@@ -1226,14 +1381,13 @@ FindVar(char *var, VWRAP * vd)
 	if (vl_lookup_func(var) == UFINDIRECT) {
 	    TBUFF *tok = 0;
 	    /* grab token, and eval it */
-	    execstr = get_token(execstr, &tok, EOS, (int *) 0);
+	    execstr = get_token(execstr, &tok, eol_null, EOS, (int *) 0);
 	    (void) vl_strncpy(var, tokval(tb_values(tok)), NLINE);
 	    FindVar(var, vd);	/* recursive, but probably safe */
 	    tb_free(&tok);
 	}
 	break;
     }
-
 }
 
 /*
@@ -1265,7 +1419,7 @@ vars_complete(DONE_ARGS)
 }
 
 static int
-PromptForVariableName(TBUFF ** result)
+PromptForVariableName(TBUFF **result)
 {
     int status;
     static TBUFF *var;
@@ -1319,7 +1473,9 @@ PromptAndSet(const char *name, int f, int n)
 	    }
 	    (void) lsprintf(prompt, "Value of %s: ", var);
 	    status = mlreply2(prompt, &tmp);
-	    if (status != TRUE)
+	    if (status == ABORT)
+		tb_error(&tmp);
+	    else if (status != TRUE)
 		returnCode(status);
 	}
 
@@ -1354,7 +1510,7 @@ free_UVAR_value(UVAR * p)
 int
 rmv_tempvar(const char *name)
 {
-    register UVAR *p, *q;
+    UVAR *p, *q;
 
     if (*name == '%')
 	name++;
@@ -1465,6 +1621,9 @@ SetVarValue(VWRAP * var, const char *name, const char *value)
     char *savestr;
     VALARGS args;
 
+    if (isEmpty(name) || value == 0)
+	return FALSE;
+
     status = TRUE;
     switch (var->v_type) {
     case VW_NOVAR:
@@ -1499,7 +1658,7 @@ SetVarValue(VWRAP * var, const char *name, const char *value)
 	break;
     }
 
-#if	OPT_DEBUGMACROS
+#if OPT_DEBUGMACROS
     if (tracemacros) {		/* we tracing macros? */
 	if (var->v_type == TOK_STATEVAR)
 	    mlforce("(((%s:&%s:%s)))",
@@ -1591,23 +1750,26 @@ set_ctrans(const char *thePalette)
     char *next;
 
     TRACE(("set_ctrans(%s)\n", thePalette));
-    while (*thePalette != EOS) {
-	thePalette = skip_cblanks(thePalette);
-	value = (int) strtol(thePalette, &next, 0);
-	if (next == thePalette)
-	    break;
-	thePalette = next;
-	ctrans[n] = value;
-	if (++n >= NCOLORS)
-	    break;
-    }
+
+    if (thePalette != 0) {
+	while (*thePalette != EOS) {
+	    thePalette = skip_cblanks(thePalette);
+	    value = (int) strtol(thePalette, &next, 0);
+	    if (next == thePalette)
+		break;
+	    thePalette = next;
+	    ctrans[n] = value;
+	    if (++n >= NCOLORS)
+		break;
+	}
 #if OPT_TRACE
-    TRACE(("...ctrans, where not 1-1:\n"));
-    for (n = 0; n < NCOLORS; n++) {
-	if (n != ctrans[n])
-	    TRACE(("\t[%d] = %d\n", n, ctrans[n]));
-    }
+	TRACE(("...ctrans, where not 1-1:\n"));
+	for (n = 0; n < NCOLORS; n++) {
+	    if (n != ctrans[n])
+		TRACE(("\t[%d] = %d\n", n, ctrans[n]));
+	}
 #endif
+    }
 }
 #endif
 
@@ -1615,11 +1777,12 @@ set_ctrans(const char *thePalette)
 static void
 show_attr(int color, const char *attr, const char *name)
 {
-    bprintf(" %c%d%sC", CONTROL_A, strlen(name), attr);
+    bprintf(" %c%d%sC", (char) CONTROL_A, (int) strlen(name), attr);
     if (color >= 0)
 	bprintf("%X", color);
     bprintf(":%s", name);
 }
+
 /*
  * This will show the foreground colors, which we can display with attributes.
  */
@@ -1627,12 +1790,13 @@ show_attr(int color, const char *attr, const char *name)
 static void
 makecolorlist(int dum1 GCC_UNUSED, void *ptr GCC_UNUSED)
 {
-    int j, k;
+    int j, k, next;
     REGION region;
-    char temp[11];
-    char *s;
 
-    bprintf("--- Color palette %*P\n", term.cols - 1, '-');
+    bprintf("--- Color palette ");
+    bpadc('-', term.cols - DOT.o);
+    bputc('\n');
+
     bprintf("\nColor name       Internal  $palette   Examples\n");
     for (j = -1; j < NCOLORS; j++) {
 	if (!is_color_code(j))
@@ -1644,18 +1808,20 @@ makecolorlist(int dum1 GCC_UNUSED, void *ptr GCC_UNUSED)
 	 * filters.  That's 'C' followed by a single digit, in
 	 * hexadecimal because we allow up to 16 colors.
 	 */
+	next = DOT.o + 10;
 	if (j >= 0)
-	    bprintf("C%X%*P", j, 10, ' ');
-	else
-	    bprintf("%*P", 10, ' ');
+	    bprintf("C%X", j);
+	bpadc(' ', next - DOT.o);
+
 	/*
 	 * $palette values are in decimal, but it's useful to see the
 	 * hexadecimal equivalents.
 	 */
-	s = lsprintf(temp, "%d", k);
+	next = DOT.o + 10;
+	bprintf("%d", k);
 	if (k > 9)
-	    (void) lsprintf(s, " (%X)", k);
-	bprintf("%s%*P", temp, sizeof(temp) - 1, ' ');
+	    bprintf(" (%X)", k);
+	bpadc(' ', next - DOT.o);
 
 	show_attr(j, "", "Normal");
 	show_attr(j, "B", "Bold");
@@ -1721,25 +1887,27 @@ set_palette(const char *value)
 
 /* represent integer as string */
 char *
-render_int(TBUFF ** rp, int i)
+render_int(TBUFF **rp, int i)
 {
     char *p, *q;
 
     p = tb_values(tb_alloc(rp, 32));
     q = lsprintf(p, "%d", i);
-    (*rp)->tb_used = (q - p + 1);
+    if (rp != 0 && q != 0 && p != 0)
+	(*rp)->tb_used = (q - p + 1);
     return p;
 }
 
 /* represent long integer as string */
 char *
-render_long(TBUFF ** rp, long i)
+render_long(TBUFF **rp, long i)
 {
     char *p, *q;
 
     p = tb_values(tb_alloc(rp, 32));
     q = lsprintf(p, "%ld", i);
-    (*rp)->tb_used = (q - p + 1);
+    if (rp != 0 && q != 0 && p != 0)
+	(*rp)->tb_used = (q - p + 1);
     return p;
 }
 
@@ -1747,7 +1915,7 @@ render_long(TBUFF ** rp, long i)
 
 /* represent boolean as string */
 char *
-render_boolean(TBUFF ** rp, int val)
+render_boolean(TBUFF **rp, int val)
 {
     static char *bools[] =
     {"FALSE", "TRUE"};
@@ -1757,13 +1925,14 @@ render_boolean(TBUFF ** rp, int val)
 #if (SYS_WINNT||SYS_VMS)
 /* unsigned to hex */
 char *
-render_hex(TBUFF ** rp, unsigned i)
+render_hex(TBUFF **rp, unsigned i)
 {
     char *p, *q;
 
     p = tb_values(tb_alloc(rp, 32));
     q = lsprintf(p, "%x", i);
-    (*rp)->tb_used = (q - p + 1);
+    if (rp != 0 && q != 0 && p != 0)
+	(*rp)->tb_used = (q - p + 1);
     return p;
 }
 #endif
@@ -1810,7 +1979,7 @@ scan_bool(const char *val)
 int
 toktyp(const char *tokn)
 {
-    if (tokn[0] == EOS)
+    if (isEmpty(tokn))
 	return TOK_NULL;
     if (tokn[0] == DQUOTE)
 	return TOK_QUOTSTR;
@@ -1968,7 +2137,7 @@ complete_vars(DONE_ARGS)
 }
 
 static int
-read_argument(TBUFF ** paramp, const PARAM_INFO * info)
+read_argument(TBUFF **paramp, const PARAM_INFO * info)
 {
     int status = TRUE;
     char *prompt;
@@ -2099,8 +2268,10 @@ save_arguments(BUFFER *bp)
     char temp[NBUFN];
     const CMDFUNC *cmd = engl2fnc(strip_brackets(temp, bp->b_bname));
     int ok = TRUE;
-    int status = FALSE;
+    int fail = FALSE;
+    int status;
     const char *cp;
+    TBUFF **all_args;
 
     if (cmd != 0 && cmd->c_args != 0) {
 	for (max_args = 0;
@@ -2112,37 +2283,76 @@ save_arguments(BUFFER *bp)
 
     TRACE((T_CALLED "save_arguments(%s) max_args=%d\n", bp->b_bname, max_args));
 
-    beginDisplay();
-    if ((p = typealloc(PROC_ARGS)) != 0) {
-	status = TRUE;
-	p->nxt_args = arg_stack;
-	arg_stack = p;
-	p->all_args = typecallocn(TBUFF *, max_args + 1);
-	tb_scopy(&(p->all_args[num_args++]), bp->b_bname);
+#define tb_allocn(n) typecallocn(TBUFF *, n)	/* FIXME: indent */
 
-	while (num_args < max_args + 1) {
-	    if (ok) {
-		status = read_argument(&(p->all_args[num_args]),
-				       &(cmd->c_args[num_args - 1]));
-		if (status == ABORT)
-		    break;
-		ok = (status == TRUE);
-	    }
-	    if (ok) {
-		if ((cp = tokval(tb_values(p->all_args[num_args]))) != error_val) {
-		    tb_scopy(&(p->all_args[num_args]), cp);
+    beginDisplay();
+    if ((p = typealloc(PROC_ARGS)) == 0) {
+	status = no_memory("save_arguments");
+    } else if ((all_args = tb_allocn(max_args + 1)) == 0) {
+	status = no_memory("save_arguments");
+	free(p);
+    } else {
+	status = TRUE;
+
+	p->nxt_args = arg_stack;
+	p->all_args = all_args;
+	arg_stack = p;
+
+	tb_scopy(&(p->all_args[num_args]), bp->b_bname);
+	if (p->all_args[num_args] == 0) {
+	    fail = TRUE;
+	} else {
+	    ++num_args;
+
+	    while (num_args < max_args + 1) {
+		int copied;
+
+		if (ok) {
+		    status = read_argument(&(p->all_args[num_args]),
+					   &(cmd->c_args[num_args - 1]));
+		    if (status == ABORT) {
+			break;
+		    }
+		    ok = (status == TRUE);
+		}
+
+		if (ok) {
+		    cp = tokval(tb_values(p->all_args[num_args]));
+		    if (cp != error_val) {
+			tb_scopy(&(p->all_args[num_args]), cp);
+			copied = TRUE;
+		    } else {
+			status = ABORT;
+			break;
+		    }
 		} else {
-		    status = ABORT;
+		    tb_scopy(&(p->all_args[num_args]), "");
+		    copied = TRUE;
+		}
+
+		if (copied && p->all_args[num_args] == 0) {
+		    status = no_memory("save_arguments");
+		    fail = TRUE;
 		    break;
 		}
-	    } else {
-		tb_scopy(&(p->all_args[num_args]), "");
+
+		TPRINTF(("...ARG%d:%s\n",
+			 num_args,
+			 tb_values(p->all_args[num_args])));
+		num_args++;
 	    }
-	    TPRINTF(("...ARG%d:%s\n", num_args, tb_values(p->all_args[num_args])));
-	    num_args++;
+
+	    p->num_args = num_args - 1;
 	}
 
-	p->num_args = num_args - 1;
+	if (fail) {
+	    arg_stack = p->nxt_args;
+	    while (max_args >= 0) {
+		tb_free(&(p->all_args[max_args--]));
+	    }
+	    free(p->all_args);
+	    free(p);
+	}
     }
     endofDisplay();
 
@@ -2227,7 +2437,7 @@ get_argument(const char *name)
 	    }
 	}
     }
-    return result;
+    return NONNULL(result);
 }
 
 #endif /* OPT_MACRO_ARGS */
@@ -2272,7 +2482,7 @@ tempvar_arg_eval(char *argp)
     } else {
 	result = error_val;
     }
-    return result;
+    return NONNULL(result);
 }
 
 /* state variables are expanded.  if it's
@@ -2331,6 +2541,7 @@ static const char *
 directive_arg_eval(char *argp)
 {
     static TBUFF *tkbuf;
+    char *result;
 #if !SYS_UNIX
     /*
      * major hack alert -- the directives start with '~'.  on unix, we'd also
@@ -2342,8 +2553,14 @@ directive_arg_eval(char *argp)
     if (strncmp("~/", argp, 2))
 	return error_val;
 #endif
+    if (argp == 0)
+	argp = "";
     tb_alloc(&tkbuf, NFILEN + strlen(argp));
-    return lengthen_path(strcpy(tb_values(tkbuf), argp));
+    if (tb_values(tkbuf) != 0)
+	result = lengthen_path(strcpy(tb_values(tkbuf), argp));
+    else
+	result = "";
+    return result;
 }
 
 typedef const char *(ArgEvalFunc) (char *argp);
@@ -2375,6 +2592,8 @@ tokval(char *tokn)
 	int toknum = toktyp(tokn);
 	if (toknum < 0 || toknum > MAXTOKTYPE)
 	    result = error_val;
+	else if (toknum == TOK_DIRECTIVE || toknum == TOK_LABEL)
+	    result = tokn;
 	else
 	    result = (*eval_func[toknum]) (tokn);
 	TRACE2(("tokval(%s) = %s\n", TRACE_NULL(tokn), TRACE_NULL(result)));
@@ -2418,6 +2637,27 @@ is_falsem(const char *val)
 }
 
 #if OPT_EVAL
+/*
+ * Check if the next token will be something that we can evaluate, i.e.,
+ * a function.
+ */
+static int
+can_evaluate(void)
+{
+    int result = FALSE;
+    int type;
+    char *save = execstr;
+    TBUFF *tok = 0;
+
+    if (mac_token(&tok)) {
+	type = toktyp(tb_values(tok));
+	result = (type == TOK_FUNCTION);
+    }
+    tb_free(&tok);
+    execstr = save;
+    return result;
+}
+
 int
 evaluate(int f, int n)
 {
@@ -2428,15 +2668,24 @@ evaluate(int f, int n)
     int code = FALSE;
 
     if (read_argument(&params, &info) == TRUE) {
-	if (tb_values(params) != error_val) {
+	if ((tmp = tb_values(params)) != error_val && tmp != 0) {
 	    old = execstr;
-	    execstr = tb_values(params);
+	    execstr = tmp;
 	    TRACE(("EVAL %s\n", execstr));
-	    while ((tmp = mac_tokval(&tok)) != 0) {
-		if (tb_length(cmd))
-		    tb_sappend0(&cmd, " ");
-		tb_sappend0(&cmd, tmp);
-		TRACE(("...token {%s}\n", tmp));
+	    for (;;) {
+		if (can_evaluate()) {
+		    if ((tmp = mac_tokval(&tok)) != 0) {
+			if (tb_length(cmd))
+			    tb_sappend0(&cmd, " ");
+			tb_sappend0(&cmd, tmp);
+		    }
+		} else if (mac_token(&tok)) {
+		    if (tb_length(cmd))
+			tb_sappend0(&cmd, " ");
+		    tb_sappend0(&cmd, tb_values(tok));
+		} else {
+		    break;
+		}
 	    }
 	    if ((tmp = tb_values(cmd)) != 0) {
 		execstr = tmp;
@@ -2486,28 +2735,31 @@ s2offset(const char *s, const char *n)
 LINE *
 label2lp(BUFFER *bp, const char *label)
 {
-    LINE *glp;
     LINE *result = 0;
-    size_t len = strlen(label);
-    int need = len + 1;
-    int col;
 
-    if (len > 1) {
-	for_each_line(glp, bp) {
-	    if (llength(glp) >= need) {
-		for (col = 0; col < llength(glp); ++col)
-		    if (!isSpace(glp->l_text[col]))
+    if (bp != 0 && label != 0 && *label != EOS) {
+	LINE *glp;
+	size_t len = strlen(label);
+	int need = len + 1;
+	int col;
+
+	if (len > 1) {
+	    for_each_line(glp, bp) {
+		if (llength(glp) >= need) {
+		    for (col = 0; col < llength(glp); ++col)
+			if (!isSpace(glp->l_text[col]))
+			    break;
+		    if (llength(glp) >= need + col
+			&& glp->l_text[col] == '*'
+			&& !memcmp(&glp->l_text[col + 1], label, len)) {
+			result = glp;
 			break;
-		if (llength(glp) >= need + col
-		    && glp->l_text[col] == '*'
-		    && !memcmp(&glp->l_text[col + 1], label, len)) {
-		    result = glp;
-		    break;
+		    }
 		}
 	    }
 	}
+	TRACE(("label2lp(%s) ->%d\n", label, line_no(curbp, result)));
     }
-    TRACE(("label2lp(%s) ->%d\n", label, line_no(curbp, result)));
     return result;
 }
 
@@ -2517,7 +2769,7 @@ label2lp(BUFFER *bp, const char *label)
 char *
 mkupper(char *s)
 {
-    if (s != error_val) {
+    if (s != error_val && s != 0) {
 	char *sp;
 
 	for (sp = s; *sp; sp++)
@@ -2532,7 +2784,7 @@ mkupper(char *s)
 char *
 mklower(char *s)
 {
-    if (s != error_val) {
+    if (s != error_val && s != 0) {
 	char *sp;
 
 	for (sp = s; *sp; sp++)
@@ -2549,7 +2801,7 @@ mktrimmed(char *str)
     char *base = str;
     char *dst = str;
 
-    if (str != error_val) {
+    if (str != error_val && str != 0) {
 	while (*str != EOS) {
 	    if (isSpace(*str)) {
 		if (dst != base)
@@ -2582,7 +2834,7 @@ must_quote_token(const char *values, unsigned last)
 {
     unsigned n;
 
-    if (last != 0) {
+    if (last != 0 && values != 0) {
 	for (n = 0; n < last; n++) {
 	    int ch = CharOf(values[n]);
 	    if (ch == SQUOTE
@@ -2599,7 +2851,7 @@ must_quote_token(const char *values, unsigned last)
  * Appends the buffer, with quotes
  */
 void
-append_quoted_token(TBUFF ** dst, const char *values, unsigned last)
+append_quoted_token(TBUFF **dst, const char *values, unsigned last)
 {
     unsigned n;
 
