@@ -4,7 +4,7 @@
  *	original by Daniel Lawrence, but
  *	much modified since then.  assign no blame to him.  -pgf
  *
- * $Header: /usr/build/vile/vile/RCS/exec.c,v 1.259 2003/06/21 14:59:47 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/exec.c,v 1.279 2005/07/17 18:00:32 tom Exp $
  *
  */
 
@@ -13,10 +13,10 @@
 #include "nefunc.h"
 #include "nefsms.h"
 
-#define isSPorTAB(c) ((c) == ' ' || (c) == '\t')
-
-static int rangespec(const char *specp, LINEPTR * fromlinep, LINEPTR *
-		     tolinep, CMDFLAGS * flagp);
+static int rangespec(const char *specp,
+		     LINE **fromlinep,
+		     LINE **tolinep,
+		     CMDFLAGS * flagp);
 
 /* while loops are described by a list of small structs.  these point
  * at the line on which they were found, and at the line to which one
@@ -36,8 +36,8 @@ static int rangespec(const char *specp, LINEPTR * fromlinep, LINEPTR *
  */
 
 typedef struct WHLOOP {
-    LINEPTR w_here;		/* line containing this directive */
-    LINEPTR w_there;		/* if endwhile, ptr to while, else to endwhile */
+    LINE *w_here;		/* line containing this directive */
+    LINE *w_there;		/* if endwhile, ptr to while, else to endwhile */
     DIRECTIVE w_type;		/* block type: D_WHILE, D_BREAK, D_ENDWHILE */
     struct WHLOOP *w_next;
 } WHLOOP;
@@ -106,7 +106,7 @@ eol_range(const char *buffer, unsigned cpos, int c, int eolchar GCC_UNUSED)
 static int
 end_of_cmd(void)
 {
-    register int c = end_string();
+    int c = end_string();
     return isreturn(c);
 }
 
@@ -153,12 +153,12 @@ static				/* next procedure is local */
 int
 execute_named_command(int f, int n)
 {
-    register int status;
-    register const CMDFUNC *cfp;	/* function to execute */
-    register char *fnp;		/* ptr to the name of the cmd to exec */
+    int status;
+    const CMDFUNC *cfp;		/* function to execute */
+    char *fnp;			/* ptr to the name of the cmd to exec */
 
-    LINEPTR fromline;		/* first linespec */
-    LINEPTR toline;		/* second linespec */
+    LINE *fromline;		/* first linespec */
+    LINE *toline;		/* second linespec */
     MARK save_DOT;
     MARK last_DOT;
     static TBUFF *lspec;	/* text of line spec */
@@ -198,7 +198,8 @@ execute_named_command(int f, int n)
 	    } else if (isreturn(c) && (status == FALSE)) {
 		return FALSE;
 	    } else {
-		unkeystroke(c);	/* ...so we can splice */
+		if (!clexec)	/* if there's an I/O vs a buffer */
+		    unkeystroke(c);	/* ...so we can splice */
 	    }
 	    cmode = 1;
 	    repeat_cmd = EOS;
@@ -233,8 +234,9 @@ execute_named_command(int f, int n)
 			c = end_string();
 			if (isPunct(c)
 			    && strchr(global_g_val_ptr(GVAL_EXPAND_CHARS),
-				      c) != 0)
+				      c) != 0) {
 			    unkeystroke(c);
+			}
 		    }
 		    break;
 		}
@@ -283,7 +285,7 @@ execute_named_command(int f, int n)
 
     /* parse the accumulated lspec */
     if (rangespec(tb_values(lspec), &fromline, &toline, &lflag) != TRUE) {
-	mlforce("[Improper line range]");
+	mlforce("[Improper line range '%s']", tb_values(lspec));
 	return FALSE;
     }
 
@@ -327,7 +329,7 @@ execute_named_command(int f, int n)
 	    mlforce("[Can't use address 0 with \"%s\" command]", fnp);
 	    return FALSE;
 	}
-	if (fromline == null_ptr)
+	if (fromline == 0)
 	    fromline = buf_head(curbp);		/* buffer is empty */
 
 	/*  we're positioned at fromline == curbp->b_linep, so commands
@@ -389,7 +391,7 @@ execute_named_command(int f, int n)
 	} else if (flags & DFLNONE) {
 #if OPT_SHELL
 	    if (cfp == &f_operfilter) {
-		cfp = &f_spawn;
+		cfp = &f_vl_spawn;
 		(void) setmark();	/* not that it matters */
 	    } else
 #endif
@@ -402,7 +404,7 @@ execute_named_command(int f, int n)
 		mlforce("[Configuration error: DFLNONE]");
 		return FALSE;
 	    }
-	    fromline = toline = null_ptr;
+	    fromline = toline = 0;
 	    lflag |= (FROM | TO);	/* avoid prompt for line-count */
 	} else
 	    lflag &= ~DFLALL;
@@ -457,20 +459,20 @@ execute_named_command(int f, int n)
 
     save_DOT = DOT;
     if (!(flags & NOMOVE)
-	&& ((toline != null_ptr) || (fromline != null_ptr))) {
+	&& ((toline != 0) || (fromline != 0))) {
 	/* assume it's an absolute motion */
 	/* we could probably do better */
 	curwp->w_lastdot = DOT;
     }
-    if ((toline != null_ptr) && (flags & TO)) {
+    if ((toline != 0) && (flags & TO)) {
 	DOT.l = toline;
 	(void) firstnonwhite(FALSE, 1);
 	(void) setmark();
     }
-    if ((fromline != null_ptr) && (flags & FROM)) {
+    if ((fromline != 0) && (flags & FROM)) {
 	DOT.l = fromline;
 	(void) firstnonwhite(FALSE, 1);
-	if (toline == null_ptr)
+	if (toline == 0)
 	    (void) setmark();
     }
     if (!sameline(last_DOT, DOT))
@@ -497,8 +499,9 @@ execute_named_command(int f, int n)
 	&& end_string() != ' '
 	&& !end_of_cmd()) {
 	status = FALSE;
-	if (!mapped_c_avail())
+	if (!mapped_c_avail()) {
 	    unkeystroke(end_string());
+	}
 	(void) kbd_seq();
 	mlforce("[Extra characters after \"%s\" command]", fnp);
     } else {
@@ -538,11 +541,11 @@ namedcmd(int f, int n)
  * Steve Kirkendall
  */
 static const char *
-parse_linespec(const char *s, LINEPTR * markptr)
+parse_linespec(const char *s, LINE **markptr)
 {
     int num;
     LINE *lp;			/* where the linespec takes us */
-    register const char *t;
+    const char *t;
     int status;
 
     (void) setmark();
@@ -555,7 +558,7 @@ parse_linespec(const char *s, LINEPTR * markptr)
 	    s++;
 
 	/* skip leading spaces */
-	while (isSPorTAB(*s))
+	while (isBlank(*s))
 	    s++;
 
 	/* dot means current position */
@@ -660,15 +663,18 @@ parse_linespec(const char *s, LINEPTR * markptr)
  * Steve Kirkendall
  */
 static int
-rangespec(const char *specp, LINEPTR * fromlinep, LINEPTR * tolinep,
+rangespec(const char *specp, LINE **fromlinep, LINE **tolinep,
 	  CMDFLAGS * flagp)
 {
-    register const char *scan;	/* used to scan thru specp */
-    LINEPTR fromline;		/* first linespec */
-    LINEPTR toline;		/* second linespec */
+    const char *scan;		/* used to scan thru specp */
+    LINE *fromline;		/* first linespec */
+    LINE *toline;		/* second linespec */
 
     *flagp = 0;
 
+    if (specp == 0)
+	return FALSE;
+
     /* ignore command lines that start with a double-quote */
     if (*specp == '"') {
 	*fromlinep = *tolinep = DOT.l;
@@ -676,7 +682,7 @@ rangespec(const char *specp, LINEPTR * fromlinep, LINEPTR * tolinep,
     }
 
     /* permit extra colons at the start of the line */
-    while (isSPorTAB(*specp) || *specp == ':') {
+    while (isBlank(*specp) || *specp == ':') {
 	specp++;
     }
 
@@ -695,7 +701,7 @@ rangespec(const char *specp, LINEPTR * fromlinep, LINEPTR * tolinep,
     } else {
 	scan = parse_linespec(scan, &toline);
 	*flagp |= FROM;
-	if (toline == null_ptr)
+	if (toline == 0)
 	    toline = DOT.l;
 	fromline = toline;
 	while (*scan == ',') {
@@ -703,7 +709,7 @@ rangespec(const char *specp, LINEPTR * fromlinep, LINEPTR * tolinep,
 	    scan++;
 	    scan = parse_linespec(scan, &toline);
 	    *flagp |= TO;
-	    if (toline == null_ptr) {
+	    if (toline == 0) {
 		/* faulty line spec */
 		return FALSE;
 	    }
@@ -711,13 +717,13 @@ rangespec(const char *specp, LINEPTR * fromlinep, LINEPTR * tolinep,
     }
 
     if (is_empty_buf(curbp))
-	fromline = toline = null_ptr;
+	fromline = toline = 0;
 
     if (scan == specp)
 	*flagp |= DFLALL;
 
     /* skip whitespace */
-    while (isSPorTAB(*scan))
+    while (isBlank(*scan))
 	scan++;
 
     if (*scan) {
@@ -763,7 +769,7 @@ docmd(char *cline, int execflag, int f, int n)
 	    returnCode(FALSE);
 	}
 	if (*token == ':') {	/* allow leading ':' on line */
-	    register int j;
+	    int j;
 	    for (j = 0; (token[j] = token[j + 1]) != EOS; j++) ;
 	}
     } while (!*token);
@@ -839,8 +845,8 @@ call_cmdfunc(const CMDFUNC * p, int f, int n)
 int
 execute(const CMDFUNC * execfunc, int f, int n)
 {
-    register int status;
-    register CMDFLAGS flags;
+    int status;
+    CMDFLAGS flags;
 
     if (execfunc == NULL) {
 #if OPT_REBIND
@@ -848,11 +854,20 @@ execute(const CMDFUNC * execfunc, int f, int n)
 #else
 	mlwarn("[Not a command]");	/* complain             */
 #endif
-	returnCode(FALSE);
+	return (FALSE);
     }
 
-    TRACE((T_CALLED "execute(execfunc=%p(%s), f=%d, n=%d)\n",
-	   execfunc, execfunc->c_name ? execfunc->c_name : "?", f, n));
+    TRACE((T_CALLED "execute(execfunc=%p(%s:%s), f=%d, n=%d)\n",
+	   execfunc,
+	   (execfunc->c_flags & CMD_PERL
+	    ? "perl"
+	    : (execfunc->c_flags & CMD_PROC
+	       ? "proc"
+	       : "func")),
+	   (execfunc->c_name
+	    ? execfunc->c_name
+	    : "?"),
+	   f, n));
 
     flags = execfunc->c_flags;
 
@@ -867,12 +882,14 @@ execute(const CMDFUNC * execfunc, int f, int n)
 	if (flags & UNDO) {
 	    /* undoable command can't be permitted when read-only */
 	    if (!(flags & VIEWOK)) {
-		if (b_val(curbp, MDVIEW))
+		if (b_val(curbp, MDVIEW)) {
 		    returnCode(rdonly());
+		}
 #ifdef MDCHK_MODTIME
 		if (!b_is_changed(curbp) &&
-		    !ask_shouldchange(curbp))
+		    !ask_shouldchange(curbp)) {
 		    returnCode(FALSE);
+		}
 #endif
 	    }
 	    if (!kbd_replaying(FALSE))
@@ -899,7 +916,7 @@ execute(const CMDFUNC * execfunc, int f, int n)
 	    ukb = dotcmdkreg;
     }
 
-    if (curwp->w_tentative_lastdot.l == null_ptr)
+    if (curwp->w_tentative_lastdot.l == 0)
 	curwp->w_tentative_lastdot = DOT;
 
     status = call_cmdfunc(execfunc, f, n);
@@ -926,10 +943,10 @@ execute(const CMDFUNC * execfunc, int f, int n)
  * actually found.
  */
 char *
-get_token(char *src, TBUFF ** tok, int eolchar, int *actual)
+get_token(char *src, TBUFF **tok, int (*endfunc) (EOL_ARGS), int eolchar, int *actual)
 {
     tb_init(tok, EOS);
-    return get_token2(src, tok, eolchar, actual);
+    return get_token2(src, tok, endfunc, eolchar, actual);
 }
 
 /*
@@ -937,7 +954,7 @@ get_token(char *src, TBUFF ** tok, int eolchar, int *actual)
  * re-gluing a "!" to a quoted string from kbd_reply().
  */
 char *
-get_token2(char *src, TBUFF ** tok, int eolchar, int *actual)
+get_token2(char *src, TBUFF **tok, int (*endfunc) (EOL_ARGS), int eolchar, int *actual)
 {
     int quotef = EOS;		/* nonzero iff the current string quoted */
     int c, i, d, chr;
@@ -950,7 +967,7 @@ get_token2(char *src, TBUFF ** tok, int eolchar, int *actual)
 	return src;
 
     /* first scan past any whitespace in the source string */
-    while (isSPorTAB(*src))
+    while (isBlank(*src))
 	++src;
 
     /* scan through the source string, which may be quoted */
@@ -1046,14 +1063,18 @@ get_token2(char *src, TBUFF ** tok, int eolchar, int *actual)
 		if (c == eolchar) {
 		    if (actual != 0)
 			*actual = *src;
-		    if (!isSPorTAB(c))
+		    if (!isBlank(c))
 			src++;
 		    break;
+		} else if (endfunc(tb_values(*tok), tb_length(*tok), c, eolchar)) {
+		    if (actual != 0)
+			*actual = *src;
+		    break;
 		} else if (c == DQUOTE) {
 		    quotef = c;
 		    /* note that leading quote
 		       is included */
-		} else if (isSPorTAB(c)) {
+		} else if (isBlank(c)) {
 		    if (actual != 0)
 			*actual = *src;
 		    break;
@@ -1072,7 +1093,7 @@ get_token2(char *src, TBUFF ** tok, int eolchar, int *actual)
     }
 
     /* scan past any whitespace remaining in the source string */
-    while (isSPorTAB(*src))
+    while (isBlank(*src))
 	++src;
     token_ended_line = isreturn(*src) || *src == EOS;
 
@@ -1091,30 +1112,36 @@ get_token2(char *src, TBUFF ** tok, int eolchar, int *actual)
  * commands are multiple tokens.
  */
 int
-macroize(TBUFF ** p, TBUFF * src, int skip)
+macroize(TBUFF **p, TBUFF *src, int skip)
 {
-    register int c;
-    char *ref = tb_values(src);	/* FIXME */
-    int multi = !isShellOrPipe(ref);	/* shift command? */
-    int count = 0;
+    int c;
+    char *ref;
+    char *txt;
 
-    if (tb_init(p, esc_c) != 0) {
-	size_t n, len = tb_length(src);
-	char *txt = tb_values(src);
+    if ((ref = tb_values(src)) != 0) {	/* FIXME */
+	int multi = !isShellOrPipe(ref);	/* shift command? */
+	int count = 0;
 
-	TRACE(("macroizing %s\n", tb_visible(src)));
-	(void) tb_append(p, DQUOTE);
-	for (n = skip; n < len; n++) {
-	    c = txt[n];
-	    if (multi && count++)
-		(void) tb_sappend(p, "\" \"");
-	    if (c == BACKSLASH || c == DQUOTE)
-		(void) tb_append(p, BACKSLASH);
-	    (void) tb_append(p, c);
+	if (tb_init(p, esc_c) != 0) {
+	    size_t n, len = tb_length(src);
+
+	    if ((txt = tb_values(src)) != 0) {
+
+		TRACE(("macroizing %s\n", tb_visible(src)));
+		(void) tb_append(p, DQUOTE);
+		for (n = skip; n < len; n++) {
+		    c = txt[n];
+		    if (multi && count++)
+			(void) tb_sappend(p, "\" \"");
+		    if (c == BACKSLASH || c == DQUOTE)
+			(void) tb_append(p, BACKSLASH);
+		    (void) tb_append(p, c);
+		}
+		(void) tb_append(p, DQUOTE);
+		TRACE(("macroized %s\n", tb_visible(*p)));
+		return (tb_append(p, EOS) != 0);
+	    }
 	}
-	(void) tb_append(p, DQUOTE);
-	TRACE(("macroized %s\n", tb_visible(*p)));
-	return (tb_append(p, EOS) != 0);
     }
     return FALSE;
 }
@@ -1139,7 +1166,7 @@ mac_tokens(void)
 
 /* fetch and isolate the next token from execstr */
 int
-mac_token(TBUFF ** tok)
+mac_token(TBUFF **tok)
 {
     int savcle;
     const char *oldstr = execstr;
@@ -1148,7 +1175,7 @@ mac_token(TBUFF ** tok)
     clexec = TRUE;
 
     /* get and advance past token */
-    execstr = get_token(execstr, tok, EOS, (int *) 0);
+    execstr = get_token(execstr, tok, eol_null, EOS, (int *) 0);
 
     clexec = savcle;
     return (execstr != oldstr);
@@ -1156,11 +1183,31 @@ mac_token(TBUFF ** tok)
 
 /* fetch and isolate and evaluate the next token from execstr */
 char *
-mac_tokval(TBUFF ** tok)
+mac_tokval(TBUFF **tok)
 {
     if (mac_token(tok) != 0) {
+	char *previous = tb_values(*tok);
+	const char *newvalue = tokval(previous);
 	/* evaluate the token */
-	(void) tb_scopy(tok, tokval(tb_values(*tok)));
+	if (previous != 0
+	    && newvalue != 0
+	    && (const char *) previous != newvalue) {
+	    /*
+	     * Check for the special case where we're just shifting the result
+	     * down by one since we're stripping a leading quote.  I don't want
+	     * to use memmove() in tb_buff.c just for this instance since it's
+	     * less portable -TD
+	     */
+	    if (((const char *) previous) + 1 == newvalue) {
+		TBUFF *fix = 0;
+		tb_scopy(&fix, newvalue);
+		tb_free(tok);
+		*tok = fix;
+	    } else {
+		(void) tb_scopy(tok, newvalue);
+	    }
+	    tb_dequote(tok);
+	}
 	return (tb_values(*tok));
     }
     tb_free(tok);
@@ -1171,7 +1218,7 @@ mac_tokval(TBUFF ** tok)
  * get a macro line argument
  */
 int
-mac_literalarg(TBUFF ** tok)
+mac_literalarg(TBUFF **tok)
 {				/* buffer to place argument */
     /* grab everything on this line, literally */
     (void) tb_scopy(tok, execstr);
@@ -1185,7 +1232,7 @@ mac_literalarg(TBUFF ** tok)
  * Parameter info is given as a keyword, optionally followed by a prompt string.
  */
 static int
-decode_parameter_info(TBUFF * tok, PARAM_INFO * result)
+decode_parameter_info(TBUFF *tok, PARAM_INFO * result)
 {
     char name[NSTRING];
     char text[NSTRING];
@@ -1219,7 +1266,7 @@ decode_parameter_info(TBUFF * tok, PARAM_INFO * result)
 #endif
 
 static int
-setup_macro_buffer(TBUFF * name, int flag)
+setup_macro_buffer(TBUFF *name, int flag)
 {
 #if OPT_MACRO_ARGS || OPT_ONLINEHELP
     static TBUFF *temp;
@@ -1286,6 +1333,7 @@ setup_macro_buffer(TBUFF * name, int flag)
 		    }
 		    if (decode_parameter_info(temp, &(cf->c_args[count]))) {
 			cf->c_args[++count].pi_type = PT_UNKNOWN;
+			cf->c_args[count].pi_text = 0;
 			break;
 		    }
 		    /* FALLTHRU */
@@ -1356,7 +1404,7 @@ int
 storeproc(int f, int n)
 {
     static TBUFF *name;		/* procedure name */
-    register int status;	/* return status */
+    int status;			/* return status */
 
     /* if a number was given, then they must mean macro N */
     if (f)
@@ -1488,7 +1536,10 @@ push_variable(char *name)
     }
 
     /* We've not saved it - do it now */
-    p = typealloc(LOCALS);
+    if ((p = typealloc(LOCALS)) == 0) {
+	no_memory("push_variable");
+	return FALSE;
+    }
     p->next = ifstk.locals;
     p->name = strmalloc(name);
     p->m_value = 0;
@@ -1507,7 +1558,8 @@ push_variable(char *name)
 	    TBUFF *tmp = 0;
 	    append_quoted_token(&tmp, p->p_value, strlen(p->p_value));
 	    tb_append(&tmp, EOS);
-	    p->p_value = tb_values(tmp);
+	    if (tb_values(tmp) != 0)
+		p->p_value = tb_values(tmp);
 	    free(tmp);
 	} else {
 	    p->m_value = strmalloc(p->p_value);
@@ -1618,7 +1670,7 @@ unbalanced_directive(DIRECTIVE dirnum)
 }
 
 static int
-navigate_while(LINEPTR * lpp, WHLOOP * whlist)
+navigate_while(LINE **lpp, WHLOOP * whlist)
 {
     WHLOOP *wht;
 
@@ -1646,7 +1698,7 @@ begin_directive(char **const cmdpp,
 		DIRECTIVE dirnum,
 		WHLOOP * whlist,
 		BUFFER *bp,
-		LINEPTR * lpp)
+		LINE **lpp)
 {
     int status = DDIR_COMPLETE;	/* assume directive is self-contained */
     TBUFF *argtkn = 0;
@@ -1657,12 +1709,14 @@ begin_directive(char **const cmdpp,
 
     switch (dirnum) {
     case D_LOCAL:
-	while (mac_token(&argtkn) == TRUE) {
-	    value = tb_values(argtkn);
-	    if (push_variable(value) != TRUE) {
-		mlforce("[cannot save '%s']", value);
-		status = DDIR_FAILED;
-		break;
+	if (!ifstk.disabled) {
+	    while (mac_token(&argtkn) == TRUE) {
+		value = tb_values(argtkn);
+		if (push_variable(value) != TRUE) {
+		    mlforce("[cannot save '%s']", value);
+		    status = DDIR_FAILED;
+		    break;
+		}
 	    }
 	}
 	break;
@@ -1756,10 +1810,14 @@ begin_directive(char **const cmdpp,
     case D_GOTO:
 	if (!ifstk.disabled) {
 	    TBUFF *label = 0;
-	    LINEPTR glp;	/* line to goto */
+	    LINE *glp;		/* line to goto */
 
 	    /* grab label to jump to */
-	    *cmdpp = (char *) get_token(*cmdpp, &label, EOS, (int *) 0);
+	    *cmdpp = (char *) get_token(*cmdpp,
+					&label,
+					eol_null,
+					EOS,
+					(int *) 0);
 #if VILE_MAYBE
 	    /* i think a simple re-eval would get us variant
 	     * targets, i.e. ~goto %somelabel.  */
@@ -1795,20 +1853,26 @@ begin_directive(char **const cmdpp,
 	break;
 
     case D_ELSEWITH:
-	if (!tb_length(with_prefix)) {
-	    status = unbalanced_directive(dirnum);
-	    break;
+	if (!ifstk.disabled) {
+	    if (!tb_length(with_prefix)) {
+		status = unbalanced_directive(dirnum);
+		break;
+	    }
 	}
 	/* FALLTHRU */
     case D_WITH:
-	execstr = skip_space_tab(execstr);
-	tb_scopy(&with_prefix, execstr);
-	status = DDIR_COMPLETE;
+	if (!ifstk.disabled) {
+	    execstr = skip_space_tab(execstr);
+	    tb_scopy(&with_prefix, execstr);
+	    status = DDIR_COMPLETE;
+	}
 	break;
 
     case D_ENDWITH:
-	tb_free(&with_prefix);
-	status = DDIR_COMPLETE;
+	if (!ifstk.disabled) {
+	    tb_free(&with_prefix);
+	    status = DDIR_COMPLETE;
+	}
 	break;
 
     case D_UNKNOWN:
@@ -1831,7 +1895,7 @@ begin_directive(char **const cmdpp,
 }
 
 static WHLOOP *
-alloc_WHLOOP(WHLOOP * whp, DIRECTIVE dirnum, LINEPTR lp)
+alloc_WHLOOP(WHLOOP * whp, DIRECTIVE dirnum, LINE *lp)
 {
     WHLOOP *nwhp;
 
@@ -1852,7 +1916,7 @@ static int
 setup_dobuf(BUFFER *bp, WHLOOP ** result)
 {
     int status = TRUE;
-    LINEPTR lp;
+    LINE *lp;
     char *cmdp;
     WHLOOP *twhp, *whp = 0;
     DIRECTIVE dirnum;
@@ -2056,7 +2120,7 @@ perform_dobuf(BUFFER *bp, WHLOOP * whlist)
 {
     int status = TRUE;
     int glue = 0;		/* nonzero to append lines */
-    LINEPTR lp;
+    LINE *lp;
     size_t linlen;
     DIRECTIVE dirnum;
     WINDOW *wp;
@@ -2138,7 +2202,7 @@ perform_dobuf(BUFFER *bp, WHLOOP * whlist)
 	 */
 	if (lforw(lp) != buf_head(bp)
 	    && linlen != 0
-	    && cmdp[linlen - 1] == BACKSLASH) {
+	    && isEscaped(cmdp + linlen)) {
 	    glue = linlen + (size_t) (cmdp - linebuf) - 1;
 	    continue;
 	}
@@ -2262,8 +2326,12 @@ perform_dobuf(BUFFER *bp, WHLOOP * whlist)
 	} else if (*cmdp != DIRECTIVE_CHAR) {
 	    /* prefix lines with "WITH" value, if any */
 	    if (tb_length(with_prefix)) {
-		char *temp = (char *) malloc(tb_length(with_prefix)
-					     + strlen(cmdp) + 2);
+		char *temp = typeallocn(char, tb_length(with_prefix)
+					+ strlen(cmdp) + 2);
+		if (temp == 0) {
+		    status = no_memory("performing ~with");
+		    break;
+		}
 		(void) lsprintf(temp, "%s %s", tb_values(with_prefix), cmdp);
 		free(linebuf);
 		cmdp = linebuf = temp;
@@ -2416,7 +2484,7 @@ do_source(char *fname, int n, int optional)
 int
 execfile(int f GCC_UNUSED, int n)
 {
-    register int status;
+    int status;
     char fname[NFILEN];		/* name of file to execute */
     static TBUFF *last;
 
@@ -2472,14 +2540,14 @@ set_b_lineno(BUFFER *bp, L_NUM n)
 int
 dofile(char *fname)
 {
-    register BUFFER *bp;	/* buffer to place file to execute */
-    register int status;	/* results of various calls */
+    BUFFER *bp;			/* buffer to place file to execute */
+    int status;			/* results of various calls */
     int clobber = FALSE;
     int original;
 
     TRACE((T_CALLED "dofile(%s)\n", TRACE_NULL(fname)));
 
-    if (fname == 0 || *fname == EOS)
+    if (isEmpty(fname))
 	returnCode(no_file_found());
 
     /*
@@ -2517,6 +2585,8 @@ dofile(char *fname)
 	    (void) zotbuf(bp);
 	else
 	    set_b_lineno(bp, original);
+    } else if (clobber) {
+	(void) zotbuf(bp);
     }
     returnCode(status);
 }
