@@ -5,7 +5,7 @@
  * reading and writing of the disk are
  * in "fileio.c".
  *
- * $Header: /usr/build/vile/vile/RCS/file.c,v 1.351 2003/06/18 22:25:36 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/file.c,v 1.383 2005/07/10 00:45:47 tom Exp $
  */
 
 #include "estruct.h"
@@ -36,6 +36,28 @@
 #define isFileMode(mode) (mode & S_IFMT) == S_IFREG
 #endif
 
+#define NO_BNAME "NoName"
+
+#if SYS_WINNT && DISP_NTWIN
+    /*
+     * Create a shadow copy of ffstatus for the purposes of distinguishing
+     * a new file from an existing file.  A new file does not yet exist on
+     * disk and the user may decide not to write back its contents before
+     * exiting the editor.
+     *
+     * A shadow copy is used here to prevent breaking a _lot_ of existing
+     * code that references the value of ffstatus.
+     */
+static FFType ffshadow;
+#endif
+
+#define FIO2Status(fio) \
+	  ((fio) < FIOERR) \
+	   ? TRUE \
+	   : (((fio) == FIOERR) \
+	      ? FALSE \
+	      : ABORT)
+
 /*--------------------------------------------------------------------------*/
 
 /* Returns the modification time iff the path corresponds to an existing file,
@@ -250,6 +272,7 @@ CleanAfterPipe(int Wrote)
     if (must_clean_pipe) {
 #if SYS_UNIX || SYS_MSDOS
 	ttunclean();		/* may clear the screen as a side-effect */
+	term.kopen();
 	term.flush();
 	if (Wrote)
 	    pressreturn();
@@ -321,12 +344,35 @@ explicit_version(char *ver)
 }
 #endif /* OPT_VMS_PATH */
 
-#if SYS_VMS
+#if OPT_VMS_PATH
 static char *
 resolve_filename(BUFFER *bp)
 {
     char temp[NFILEN];
+#if SYS_VMS
     ch_fname(bp, fgetname(ffp, temp));
+#else
+    /*
+     * The fgetname() should be just a shortcut; we should already have a
+     * usable filename in bp->b_fname.
+     */
+    if (bp->b_fname == 0) {
+	TRACE(("resolve_filename NULL\n"));
+    } else if (is_vms_pathname(bp->b_fname, FALSE)
+	       && strcspn(bp->b_fname, ":[];") != strlen(bp->b_fname)) {
+	TRACE(("resolve_filename NONEED: %s\n", bp->b_fname));
+    } else {
+	char temp2[NFILEN];
+	mklower(strcpy(temp2, bp->b_fname));
+	if (realpath(temp2, temp)) {
+	    (void) unix2vms_path(temp, temp);
+	    TRACE(("resolve_filename %s -> %s\n", bp->b_fname, temp));
+	    ch_fname(bp, temp);
+	} else {
+	    TRACE(("resolve_filename FAILED: %s\n", bp->b_fname));
+	}
+    }
+#endif
     markWFMODE(bp);
     return bp->b_fname;
 }
@@ -367,7 +413,7 @@ same_fname(const char *fname, BUFFER *bp, int lengthen)
 	if ((!explicit_version(s)
 	     || !explicit_version(t))
 	    && ((s - bname) == (t - fname))) {
-	    status = !strncmp(fname, bname, (size_t) (s - bname));
+	    status = !strnicmp(fname, bname, (size_t) (s - bname));
 	    TRACE(("=>%d\n", status));
 	    return (status);
 	}
@@ -480,6 +526,36 @@ cannot_reread(void)
     return FALSE;
 }
 
+/*
+ * Zap the given buffer without trying to close its window, by renaming it
+ * to the unnamed-buffer and resetting its local modes.  Keeping the window
+ * sidesteps the issue of whether it was the only buffer, while allowing us
+ * to free the associated memory.
+ *
+ * FIXME: this may also have a majormode associated with it, should remove it.
+ */
+static void
+reset_to_unnamed(BUFFER *bp)
+{
+    WINDOW *wp;
+    int n;
+
+    for (n = 0; n < MAX_B_VALUES; ++n) {
+	if (is_local_b_val(bp, n)) {
+	    make_global_b_val(bp, n);
+	}
+    }
+    bclear(bp);
+    set_bname(bp, UNNAMED_BufName);
+    ch_fname(bp, "");
+
+    for_each_window(wp) {
+	if (wp->w_bufp == bp) {
+	    wp->w_flag |= (WFHARD | WFMODE);
+	}
+    }
+}
+
 /*
  * Read a file into the current
  * buffer. This is really easy; all you do it
@@ -490,18 +566,20 @@ cannot_reread(void)
 int
 fileread(int f GCC_UNUSED, int n GCC_UNUSED)
 {
-    int s;
+    int status;
     char fname[NFILEN];
     W_VALUES *save_wvals;
 
+    TRACE((T_CALLED "fileread(%d,%d)\n", f, n));
+
     if (more_named_cmd()) {
-	if ((s = mlreply_file("Replace with file: ", (TBUFF **) 0,
-			      FILEC_REREAD, fname)) != TRUE)
-	    return s;
+	if ((status = mlreply_file("Replace with file: ", (TBUFF **) 0,
+				   FILEC_REREAD, fname)) != TRUE)
+	    returnCode(status);
     } else if (b_is_temporary(curbp)) {
-	return cannot_reread();
+	returnCode(cannot_reread());
     } else if (!(global_g_val(GMDWARNREREAD) || curbp->b_fname[0] == EOS)
-	       || ((s = mlyesno("Reread current buffer")) == TRUE)) {
+	       || ((status = mlyesno("Reread current buffer")) == TRUE)) {
 	(void) vl_strncpy(fname, curbp->b_fname, sizeof(fname));
 	/* Check if we are rereading an unnamed-buffer if it is not
 	 * associated with a file.
@@ -509,16 +587,16 @@ fileread(int f GCC_UNUSED, int n GCC_UNUSED)
 	if (is_internalname(curbp->b_fname)) {
 	    if (eql_bname(curbp, STDIN_BufName)
 		|| eql_bname(curbp, UNNAMED_BufName)) {
-		s = bclear(curbp);
-		if (s == TRUE)
+		status = bclear(curbp);
+		if (status == TRUE)
 		    mlerase();
-		curwp->w_flag |= WFMODE | WFHARD;
-		return s;
+		curwp->w_flag |= (WFMODE | WFHARD);
+		returnCode(status);
 	    }
-	    return cannot_reread();
+	    returnCode(cannot_reread());
 	}
     } else {
-	return FALSE;
+	returnCode(FALSE);
     }
 
 #if OPT_LCKFILES
@@ -533,11 +611,15 @@ fileread(int f GCC_UNUSED, int n GCC_UNUSED)
     b_clr_changed(curbp);
 
     save_wvals = save_window_modes(curbp);
-    s = readin(fname, TRUE, curbp, TRUE);
-    set_buffer_name(curbp);
+    status = readin(fname, TRUE, curbp, TRUE);
+    if (status == ABORT) {
+	reset_to_unnamed(curbp);
+    } else {
+	set_buffer_name(curbp);
+    }
     restore_window_modes(curbp, save_wvals);
 
-    return s;
+    returnCode(status);
 }
 
 /*
@@ -598,6 +680,8 @@ set_files_to_edit(const char *prompt, int appflag)
     char *actual;
     BUFFER *firstbp = 0;
 
+    TRACE((T_CALLED "set_files_to_edit(%s, %d)\n", NONNULL(prompt), appflag));
+
     if ((status = mlreply_file(prompt, &lastfileedited,
 			       FILEC_READ | FILEC_EXPAND,
 			       fname)) == TRUE) {
@@ -619,10 +703,13 @@ set_files_to_edit(const char *prompt, int appflag)
 		}
 	    }
 	}
-	if (find_bp(firstbp) != 0)
+	if (find_bp(firstbp) != 0) {
 	    status = bp2swbuffer(firstbp, FALSE, TRUE);
+	    if (status != TRUE)
+		reset_to_unnamed(firstbp);
+	}
     }
-    return status;
+    returnCode(status);
 }
 
 /* ARGSUSED */
@@ -707,7 +794,7 @@ writelinesmsg(char *fn, int nline, B_COUNT nchar)
 	    action = "Wrote";
 	}
 	sprintf(strlines, "%d", nline);
-	sprintf(strchars, "%ld", nchar);
+	sprintf(strchars, "%lu", nchar);
 	lines_len = strlen(strlines);
 	chars_len = strlen(strchars);
 	outlen = (term.cols - 1) -
@@ -786,9 +873,8 @@ ff_load_directory(char *fname)
 #endif
 
 /*
- * Insert file "fname" into the kill register
- * Called by insert file command. Return the final
- * status of the read.
+ * Insert file "fname" into the kill register.  Called by insert file command.
+ * Return the final status of the read as true/false/abort.
  */
 static int
 kifile(char *fname)
@@ -798,73 +884,86 @@ kifile(char *fname)
     int nline;
     int nbytes;
 
-    ksetup();
-    if ((s = ffropen(fname)) == FIOERR)		/* Hard file open */
-	goto out;
-    else if (s == FIOFNF)	/* File not found */
-	return no_such_file(fname);
-#if COMPLETE_FILES || COMPLETE_DIRS
-    else if ((s = ff_load_directory(fname)) == FIOERR)
-	goto out;
-#endif
+    TRACE((T_CALLED "kifile(%s)\n", NONNULL(fname)));
 
-    nline = 0;
-#if OPT_ENCRYPT
-    if ((s = vl_resetkey(curbp, fname)) == TRUE)
+    ksetup();
+    if ((s = ffropen(fname)) == FIOERR) {	/* Hard file open */
+	goto out;
+    } else if (s == FIOFNF) {	/* File not found */
+	returnCode(no_such_file(fname));
+#if COMPLETE_FILES || COMPLETE_DIRS
+    } else if ((s = ff_load_directory(fname)) == FIOERR) {
+	goto out;
 #endif
-    {
-	mlwrite("[Reading...]");
-	CleanToPipe(FALSE);
-	while ((s = ffgetline(&nbytes)) <= FIOSUC) {
-	    for (i = 0; i < nbytes; ++i)
-		if (!kinsert(fflinebuf[i]))
-		    return FIOMEM;
-	    if ((s == FIOSUC) && !kinsert('\n'))
-		return FIOMEM;
-	    ++nline;
-	    if (s < FIOSUC)
-		break;
+    } else {
+
+	nline = 0;
+#if OPT_ENCRYPT
+	if ((s = vl_resetkey(curbp, fname)) == TRUE)
+#endif
+	{
+	    mlwrite("[Reading...]");
+	    CleanToPipe(FALSE);
+	    while ((s = ffgetline(&nbytes)) <= FIOSUC) {
+		for (i = 0; i < nbytes; ++i)
+		    if (!kinsert(fflinebuf[i]))
+			returnCode(FIOMEM);
+		if ((s == FIOSUC) && !kinsert('\n')) {
+		    s = FIOMEM;
+		    goto out;
+		}
+		++nline;
+		if (s < FIOSUC)
+		    break;
+	    }
+	    CleanAfterPipe(FALSE);
 	}
-	CleanAfterPipe(FALSE);
+	kdone();
+	(void) ffclose();	/* Ignore errors.       */
+	readlinesmsg(nline, s, fname, FALSE);
     }
-    kdone();
-    (void) ffclose();		/* Ignore errors.       */
-    readlinesmsg(nline, s, fname, FALSE);
 
   out:
-    return (s != FIOERR);
+    returnCode(FIO2Status(s));
 }
 
 /*
- * Insert a file into the current
- * buffer. This is really easy; all you do it
- * find the name of the file, and call the standard
- * "insert a file into the current buffer" code.
+ * Insert a file into the current buffer.  This is really easy; all you do is
+ * find the name of the file, and call the standard "insert a file into the
+ * current buffer" code.
  */
 /* ARGSUSED */
 int
 insfile(int f GCC_UNUSED, int n GCC_UNUSED)
 {
-    int s;
+    int status;
     char fname[NFILEN];
     static TBUFF *last;
 
+    TRACE((T_CALLED "insfile(%d, %d)\n", f, n));
+
     if (!calledbefore) {
-	if ((s = mlreply_file("Insert file: ", &last,
-			      FILEC_READ | FILEC_PROMPT, fname)) != TRUE)
-	    return s;
+	if ((status = mlreply_file("Insert file: ", &last,
+				   FILEC_READ | FILEC_PROMPT, fname)) != TRUE)
+	    returnCode(status);
     }
+
     if (ukb == 0)
-	return ifile(fname, TRUE, (FILE *) 0);
+	status = ifile(fname, TRUE, (FILE *) 0);
     else
-	return kifile(fname);
+	status = kifile(fname);
+
+    if (status == ABORT) {
+	TRACE(("undo insert to recover memory\n"));
+	undo(FALSE, 1);
+    }
+    returnCode(status);
 }
 
 BUFFER *
-getfile2bp(
-	      const char *fname,	/* file name to find */
-	      int ok_to_ask,
-	      int cmdline)
+getfile2bp(const char *fname,	/* file name to find */
+	   int ok_to_ask,
+	   int cmdline)
 {
     BUFFER *bp = cmdline ? NULL : find_b_name(fname);
     int s;
@@ -958,13 +1057,12 @@ no_file_name(const char *fname)
 }
 
 int
-getfile(
-	   char *fname,		/* file name to find */
-	   int lockfl)
-{				/* check the file for locks? */
+getfile(char *fname,		/* file name to find */
+	int lockfl)		/* check the file for locks? */
+{
     BUFFER *bp;
 
-    if (fname == 0 || *fname == EOS)
+    if (isEmpty(fname))
 	return no_file_found();
 
     /* if there are no path delimiters in the name, then the user
@@ -1020,11 +1118,7 @@ apply_dosmode(BUFFER *bp, int doslines, int unixlines)
      * Make 'dos' and 'recordseparator' modes local, since this
      * transformation should apply only to the specified file.
      */
-    make_local_b_val(bp, MDDOS);
-    make_local_b_val(bp, VAL_RECORD_SEP);
-
-    set_b_val(bp, MDDOS, result);
-    set_b_val(bp, VAL_RECORD_SEP, result ? RS_CRLF : RS_LF);
+    set_record_sep(bp, result ? RS_CRLF : RS_LF);
 }
 
 static void
@@ -1095,11 +1189,8 @@ explicit_dosmode(BUFFER *bp, RECORD_SEP record_sep)
 {
     TRACE(("explicit_dosmode(%s)\n",
 	   choice_to_name(fsm_recordsep_choices, record_sep)));
-    make_local_b_val(bp, MDDOS);
-    make_local_b_val(bp, VAL_RECORD_SEP);
 
-    set_b_val(bp, MDDOS, (record_sep == RS_CRLF));
-    set_b_val(bp, VAL_RECORD_SEP, record_sep);
+    set_record_sep(bp, record_sep);
 }
 
 /*
@@ -1111,7 +1202,6 @@ modified_record_sep(RECORD_SEP record_sep)
     explicit_dosmode(curbp, record_sep);
     guess_dosmode(curbp);
     explicit_dosmode(curbp, record_sep);	/* ignore the guess - only want to strip CR's */
-    set_record_sep(curbp, (RECORD_SEP) b_val(curbp, VAL_RECORD_SEP));
     return TRUE;
 }
 
@@ -1151,16 +1241,19 @@ grab_lck_file(BUFFER *bp, char *fname)
 	!isShellOrPipe(fname) &&
 	!b_val(bp, MDVIEW)) {
 	char locker[100];
+	char *locked_by;
 
 	if (!set_lock(fname, locker, sizeof(locker))) {
 	    /* we didn't get it */
-	    make_local_b_val(bp, MDVIEW);
-	    set_b_val(bp, MDVIEW, TRUE);
-	    make_local_b_val(bp, MDLOCKED);
-	    set_b_val(bp, MDLOCKED, TRUE);
-	    make_local_b_val(bp, VAL_LOCKER);
-	    set_b_val_ptr(bp, VAL_LOCKER, strmalloc(locker));
-	    markWFMODE(bp);
+	    if ((locked_by = strmalloc(locker)) != 0) {
+		make_local_b_val(bp, MDVIEW);
+		set_b_val(bp, MDVIEW, TRUE);
+		make_local_b_val(bp, MDLOCKED);
+		set_b_val(bp, MDLOCKED, TRUE);
+		make_local_b_val(bp, VAL_LOCKER);
+		set_b_val_ptr(bp, VAL_LOCKER, locked_by);
+		markWFMODE(bp);
+	    }
 	}
     }
 }
@@ -1186,7 +1279,6 @@ init_b_traits(BUFFER *bp)
 #endif
 }
 
-#if ! SYS_MSDOS
 /*
  * Analyze the file to determine its format
  */
@@ -1291,6 +1383,7 @@ static int
 quickreadf(BUFFER *bp, int *nlinep)
 {
     B_COUNT length;
+    B_COUNT request;
     B_COUNT offset;
     LINE *lp;
     L_NUM lineno = 0;
@@ -1302,13 +1395,13 @@ quickreadf(BUFFER *bp, int *nlinep)
     TRACE((T_CALLED "quickreadf(buffer=%s, file=%s)\n", bp->b_bname, bp->b_fname));
 
     beginDisplay();
-    if ((length = ffsize()) < 0) {
+    if (ffsize(&request) < 0) {
 	mlwarn("[Can't size file]");
 	rc = FIOERR;
-    }
-    /* avoid malloc(0) problems down below; let slowreadf() do the work */
-    else if (length == 0
-	     || (buffer = castalloc(UCHAR, (size_t) length)) == NULL) {
+    } else if (request == 0) {
+	/* avoid malloc(0) problems down below; let slowreadf() do the work */
+	rc = FIONUL;
+    } else if ((buffer = castalloc(UCHAR, request)) == NULL) {
 	rc = FIOMEM;
     }
 #if OPT_ENCRYPT
@@ -1316,7 +1409,17 @@ quickreadf(BUFFER *bp, int *nlinep)
 	free(buffer);
     }
 #endif
-    else if ((length = ffread((char *) buffer, length)) < 0) {
+    else if (ffread((char *) buffer, request, &length) < 0
+#if !SYS_VMS
+	/*
+	 * For most systems, the advertised size of the file will match the
+	 * number of chars that we can read from it.  However, VMS's structured
+	 * filetypes such as VFC or VAR/CR will return fewer since we are not
+	 * using the structure information.
+	 */
+	     || (length != request)
+#endif
+	) {
 	free(buffer);
 	mlerror("reading");
 	rc = FIOERR;
@@ -1336,7 +1439,7 @@ quickreadf(BUFFER *bp, int *nlinep)
 	/*
 	 * Modify readin()'s setting for newline mode if needed:
 	 */
-	if (buffer[length - 1] != (rscode == RS_CR ? '\r' : '\n')) {
+	if (buffer[length - 1] != (UCHAR) (rscode == RS_CR ? '\r' : '\n')) {
 	    set_b_val(bp, MDNEWLINE, FALSE);
 	}
 
@@ -1370,7 +1473,7 @@ quickreadf(BUFFER *bp, int *nlinep)
 		if (lp != bp->b_LINEs)
 		    set_lback(lp, lp - 1);
 		lsetclear(lp);
-		lp->l_nxtundo = null_ptr;
+		lp->l_nxtundo = 0;
 #if OPT_LINE_ATTRS
 		lp->l_attrs = NULL;
 #endif
@@ -1400,7 +1503,6 @@ quickreadf(BUFFER *bp, int *nlinep)
     endofDisplay();
     returnCode(rc);
 }
-#endif /* ! SYS_MSDOS */
 
 /*
  * Read file "fname" into a buffer, blowing away any text found there.  Returns
@@ -1409,7 +1511,7 @@ quickreadf(BUFFER *bp, int *nlinep)
  * fname  - name of file to read
  * lockfl - check for file locks?
  * bp     - read into this buffer
- * mflg   - print messages? 
+ * mflg   - print messages?
  */
 /* ARGSUSED */
 int
@@ -1423,16 +1525,23 @@ readin(char *fname, int lockfl, BUFFER *bp, int mflg)
     && is_local_val(bp->b_values.bv, MDCRYPT);
 #endif
 
+    TRACE((T_CALLED "readin(fname=%s, lockfl=%d, bp=%p, mflg=%d)\n",
+	   NONNULL(fname), lockfl, bp, mflg));
+
+#if SYS_WINNT && DISP_NTWIN
+    ffshadow = file_is_closed;	/* an assumption */
+#endif
+
     if (bp == 0)		/* doesn't hurt to check */
-	return FALSE;
+	returnCode(FALSE);
 
     if (*fname == EOS) {
-	TRACE(("readin called with NULL fname\n"));
-	return TRUE;
+	TRACE(("...called with NULL fname\n"));
+	returnCode(FALSE);	/* we do not want to do that */
     }
 
     if ((s = bclear(bp)) != TRUE)	/* Might be old.    */
-	return s;
+	returnCode(s);
 
 #if OPT_ENCRYPT
     /* bclear() gets rid of local flags */
@@ -1462,6 +1571,9 @@ readin(char *fname, int lockfl, BUFFER *bp, int mflg)
 	   and it will appear as empty buffer */
 	/*EMPTY */ ;
     } else if (s == FIOFNF) {	/* File not found */
+#if SYS_WINNT && DISP_NTWIN
+	ffshadow = file_is_new;
+#endif
 	if (mflg)
 	    mlwrite("[New file]");
 #if COMPLETE_FILES || COMPLETE_DIRS
@@ -1482,7 +1594,7 @@ readin(char *fname, int lockfl, BUFFER *bp, int mflg)
 		    path_trunc(fname, outlen, tmp, sizeof(tmp)));
 #undef READING_FILE_FMT
 	}
-#if SYS_VMS
+#if OPT_VMS_PATH
 	if (!isInternalName(bp->b_fname))
 	    fname = resolve_filename(bp);
 #endif
@@ -1491,10 +1603,18 @@ readin(char *fname, int lockfl, BUFFER *bp, int mflg)
 #if OPT_WORKING
 	max_working = cur_working = old_working = 0;
 #endif
-#if ! SYS_MSDOS
-	if ((ffstatus == file_is_pipe) || (s = quickreadf(bp, &nline)) == FIOMEM)
-#endif
+
+	if (ffstatus == file_is_pipe
+	    || global_g_val(GVAL_READER_POLICY) == RP_SLOW) {
 	    s = slowreadf(bp, &nline);
+	} else {
+	    s = quickreadf(bp, &nline);
+	    if (s == FIONUL
+		|| (s == FIOMEM
+		    && global_g_val(GVAL_READER_POLICY) == RP_BOTH))
+		s = slowreadf(bp, &nline);
+	}
+
 #if OPT_WORKING
 	cur_working = 0;
 #endif
@@ -1526,14 +1646,16 @@ readin(char *fname, int lockfl, BUFFER *bp, int mflg)
     /*
      * Set the majormode if the file's suffix matches.
      */
-    infer_majormode(bp);
+    if (s < FIOERR)
+	infer_majormode(bp);
 
     for_each_window(wp) {
 	if (wp->w_bufp == bp) {
 	    init_window(wp, bp);
 	}
     }
-    imply_alt(fname, FALSE, lockfl);
+    if (s < FIOERR)
+	imply_alt(fname, FALSE, lockfl);
     updatelistbuffers();
 
 #if OPT_LCKFILES
@@ -1549,7 +1671,7 @@ readin(char *fname, int lockfl, BUFFER *bp, int mflg)
 	set_directory_from_file(bp);
 #endif
     b_match_attrs_dirty(bp);
-    return (s != FIOERR);
+    returnCode(FIO2Status(s));
 }
 
 /*
@@ -1558,9 +1680,25 @@ readin(char *fname, int lockfl, BUFFER *bp, int mflg)
 int
 bp2readin(BUFFER *bp, int lockfl)
 {
-    int s = readin(bp->b_fname, lockfl, bp, TRUE);
-    bp->b_active = TRUE;
-    return s;
+    int status = readin(bp->b_fname, lockfl, bp, TRUE);
+
+    if (status == TRUE) {
+	bp->b_active = TRUE;
+
+#if SYS_WINNT && DISP_NTWIN
+	if (status &&
+	    (!(isShellOrPipe(bp->b_fname) ||
+	       ffshadow == file_is_new ||
+	       b_is_registered(bp)))) {
+	    /* save file path to windows registry for later recall */
+
+	    store_recent_file_or_folder(bp->b_fname, TRUE);
+	    b_set_registered(bp);
+	}
+#endif
+    }
+
+    return status;
 }
 
 /*
@@ -1673,7 +1811,7 @@ slowreadf(BUFFER *bp, int *nlinep)
 }
 
 /*
- * Take a (null-terminated) file name, and from it fabricate a buffer name. 
+ * Take a (null-terminated) file name, and from it fabricate a buffer name.
  * This routine knows about the syntax of file names on the target system.  I
  * suppose that this information could be put in a better place than a line of
  * code.
@@ -1692,7 +1830,9 @@ makename(char *bname, const char *fname)
     fcp = skip_string(vl_strncpy(temp, fname, sizeof(temp)));
 #if OPT_VMS_PATH
     if (is_vms_pathname(temp, TRUE)) {
-	;
+	vms2unix_path(temp, temp);
+	(void) strncpy0(bname, pathleaf(temp), NBUFN);
+	strip_version(bname);
     } else if (is_vms_pathname(temp, FALSE)) {
 	for (; fcp > temp && !strchr(":]", fcp[-1]); fcp--) {
 	    ;
@@ -1735,7 +1875,7 @@ makename(char *bname, const char *fname)
 	    (void) vl_strncpy(bcp, pathleaf(fcp), NBUFN);
 
 #if SYS_UNIX
-	    /* UNIX filenames can have any characters (other than EOS!). 
+	    /* UNIX filenames can have any characters (other than EOS!).
 	     * Refuse (rightly) to deal with leading/trailing blanks, but allow
 	     * embedded blanks.  For this special case, ensure that the buffer
 	     * name has no blanks, otherwise it is difficult to reference from
@@ -1764,7 +1904,7 @@ makename(char *bname, const char *fname)
 #endif
     }
     if (*bname == EOS)
-	(void) strcpy(bname, "NoName");
+	(void) strcpy(bname, NO_BNAME);
     TRACE((" -> '%s'\n", bname));
 }
 
@@ -1782,11 +1922,12 @@ unqname(char *name)
 
     j = strlen(name);
     if (j == 0)
-	j = strlen(strcpy(name, "NoName"));
+	j = strlen(strcpy(name, NO_BNAME));
 
     /* check to see if this name is in use */
     vl_strncpy(newname, name, NBUFN);
     adjust = is_scratchname(newname);
+    strcpy(suffixbuf, "-");
     while (find_b_name(newname) != NULL) {
 	/* from "foo" create "foo-1" or "foo-a1b5" */
 	/* from "thisisamuchlongernam" create
@@ -1795,7 +1936,7 @@ unqname(char *name)
 	/* that is, put suffix at end if it fits, or else
 	   overwrite some of the name to make it fit */
 	/* the suffix is in "base 36" */
-	suffixlen = (int) (lsprintf(suffixbuf, "-%r", 36, ++i) - suffixbuf);
+	suffixlen = format_int(suffixbuf + 1, ++i, 36) + 1;
 	k = NBUFN - 1 - suffixlen;
 	if (j < k)
 	    k = j;
@@ -1871,10 +2012,10 @@ actually_write(REGION * rp, char *fn, int msgf, BUFFER *bp, int forced, int enco
     int s;
     LINE *lp;
     int nline;
-    int i;
+    B_COUNT i;
     B_COUNT nchar;
     const char *ending = get_record_sep(bp);
-    int len_rs = strlen(ending);
+    int len_rs = len_record_sep(bp);
     C_NUM offset = rp->r_orig.o;
 
     /* this is adequate as long as we cannot write parts of lines */
@@ -1914,8 +2055,9 @@ actually_write(REGION * rp, char *fn, int msgf, BUFFER *bp, int forced, int enco
      * it may be a quasi-interactive process that we don't want to modify its
      * display.
      */
-    if (isShellOrPipe(fn))
+    if (isShellOrPipe(fn)) {
 	beginDisplay();
+    }
 
     /* tell us we're writing */
     if (msgf == TRUE)
@@ -1936,9 +2078,13 @@ actually_write(REGION * rp, char *fn, int msgf, BUFFER *bp, int forced, int enco
 	 * after the line), allow 'newline' mode to suppress the
 	 * trailing newline.
 	 */
-	if ((rp->r_size -= line_length(lp)) <= 0
-	    && !b_val(bp, MDNEWLINE))
-	    ending = "";
+	if (rp->r_size <= line_length(lp)) {
+	    rp->r_size = 0;
+	    if (!b_val(bp, MDNEWLINE))
+		ending = "";
+	} else {
+	    rp->r_size -= line_length(lp);
+	}
 #if OPT_SELECTIONS
 	if (encoded) {
 	    TBUFF *temp;
@@ -1973,7 +2119,7 @@ actually_write(REGION * rp, char *fn, int msgf, BUFFER *bp, int forced, int enco
 
   out:
     if (s == FIOSUC) {		/* No write error.      */
-#if SYS_VMS
+#if OPT_VMS_PATH
 	if (same_fname(fn, bp, FALSE))
 	    fn = resolve_filename(bp);
 #endif
@@ -1991,8 +2137,9 @@ actually_write(REGION * rp, char *fn, int msgf, BUFFER *bp, int forced, int enco
 
     CleanAfterPipe(TRUE);
 
-    if (isShellOrPipe(fn))
+    if (isShellOrPipe(fn)) {
 	endofDisplay();
+    }
 
     if (s != FIOSUC)		/* Some sort of error.      */
 	return FALSE;
@@ -2047,7 +2194,8 @@ writereg(REGION * rp,
 
     if (no_file_name(given_fn)) {
 	status = FALSE;
-    } else if (!forced && b_val(bp, MDREADONLY)) {
+    } else if (!forced && b_val(bp, MDREADONLY)
+	       && (bp->b_fname == 0 || !strcmp(given_fn, bp->b_fname))) {
 	mlwarn("[Buffer mode is \"readonly\"]");
 	status = FALSE;
     } else {
@@ -2069,14 +2217,37 @@ writereg(REGION * rp,
 		else
 #endif
 		{
-		    if (forced
-			&& !ffaccess(fn, FL_WRITEABLE)
-			&& (protection = file_protection(fn)) >= 0) {
-			chmod(SL_TO_BSL(fn), protection | 0600);
+#if SYS_WINNT
+		    ULONG prev_access_mask = 0;
+		    int write_acl_added = FALSE;
+#endif
+		    if (forced) {
+#if SYS_WINNT
+			/*
+			 * If running cygwin on an NTFS partition, cygwin's
+			 * chmod utility adds ACLs to files to simulate
+			 * unix file permissions.  In the case of readonly
+			 * files (e.g., $ chmod a=r fn), the win32 posix
+			 * interfaces don't work well, if at all.  So make
+			 * the current file object a bit more Unix API
+			 * friendly, if necessary.
+			 */
+
+			write_acl_added = w32_add_write_acl(fn,
+							    &prev_access_mask);
+#endif
+			if (!ffaccess(fn, FL_WRITEABLE)
+			    && (protection = file_protection(fn)) >= 0) {
+			    chmod(SL_TO_BSL(fn), protection | 0600);
+			}
 		    }
 		    status = actually_write(rp, fn, msgf, bp, forced, encoded);
 		    if (protection > 0)
 			chmod(SL_TO_BSL(fn), protection);
+#if SYS_WINNT
+		    if (write_acl_added)
+			(void) w32_remove_write_acl(fn, prev_access_mask);
+#endif
 		}
 	    }
 	}
@@ -2092,10 +2263,22 @@ int
 writeout(const char *fn, BUFFER *bp, int forced, int msgf)
 {
     REGION region;
+    int status;
 
     setup_file_region(bp, &region);
 
-    return writereg(&region, fn, msgf, bp, forced, FALSE);
+    status = writereg(&region, fn, msgf, bp, forced, FALSE);
+
+#if SYS_WINNT && DISP_NTWIN
+    if (status && (!(isShellOrPipe(fn) || b_is_registered(bp)))) {
+	/* save file path to windows registry for later recall */
+
+	store_recent_file_or_folder(fn, TRUE);
+	b_set_registered(bp);
+    }
+#endif
+
+    return (status);
 }
 
 /*
@@ -2240,33 +2423,42 @@ vl_filename(int f GCC_UNUSED, int n GCC_UNUSED)
     } else
 #endif
 	ch_fname(curbp, fname);
+
+#if SYS_WINNT && DISP_NTWIN
+    /* remember new filename if later written out to disk */
+    b_clr_registered(curbp);
+#endif
+
     curwp->w_flag |= WFMODE;
     updatelistbuffers();
     return TRUE;
 }
 
 /*
- * Insert file "fname" into the current buffer, called by insert file command. 
+ * Insert file "fname" into the current buffer, called by insert file command.
  * Return the final status of the read.
  */
 int
 ifile(char *fname, int belowthisline, FILE *haveffp)
 {
-    LINEPTR prevp;
-    LINEPTR newlp;
-    LINEPTR nextp;
-    int s;
+    LINE *prevp;
+    LINE *newlp;
+    LINE *nextp;
+    int status;
     int nbytes;
     int nline;
 
+    TRACE((T_CALLED "ifile(fname=%s, belowthisline=%d, haveffp=%p)\n",
+	   NONNULL(fname), belowthisline, haveffp));
+
     b_clr_invisible(curbp);	/* we are not temporary */
     if (haveffp == 0) {
-	if ((s = ffropen(fname)) == FIOERR)	/* Hard file open */
+	if ((status = ffropen(fname)) == FIOERR)	/* Hard file open */
 	    goto out;
-	else if (s == FIOFNF)	/* File not found */
-	    return no_such_file(fname);
+	else if (status == FIOFNF)	/* File not found */
+	    returnCode(no_such_file(fname));
 #if COMPLETE_FILES || COMPLETE_DIRS
-	else if ((s = ff_load_directory(fname)) == FIOERR)
+	else if ((status = ff_load_directory(fname)) == FIOERR)
 	    goto out;
 	if (b_is_directory(curbp)) {
 	    /* special case: contents are already added */
@@ -2274,8 +2466,8 @@ ifile(char *fname, int belowthisline, FILE *haveffp)
 	}
 #endif
 #if OPT_ENCRYPT
-	if ((s = vl_resetkey(curbp, fname)) != TRUE)
-	    return s;
+	if ((status = vl_resetkey(curbp, fname)) != TRUE)
+	    returnCode(status);
 #endif
 	mlwrite("[Inserting...]");
 	CleanToPipe(FALSE);
@@ -2288,8 +2480,8 @@ ifile(char *fname, int belowthisline, FILE *haveffp)
     MK = DOT;
 
     nline = 0;
-    nextp = null_ptr;
-    while ((s = ffgetline(&nbytes)) <= FIOSUC) {
+    nextp = 0;
+    while ((status = ffgetline(&nbytes)) <= FIOSUC) {
 #if OPT_DOSFILES
 	if (b_val(curbp, MDDOS)
 	    && (nbytes > 0)
@@ -2303,22 +2495,34 @@ ifile(char *fname, int belowthisline, FILE *haveffp)
 
 	beginDisplay();
 	if (add_line_at(curbp, prevp, fflinebuf, nbytes) != TRUE) {
-	    s = FIOMEM;
-	    break;
+	    status = FIOMEM;
+	    newlp = 0;
+	} else {
+	    newlp = lforw(prevp);
+	    if ((tag_for_undo(newlp)) == ABORT) {
+		status = FIOMEM;
+		/*
+		 * De-link the line added for which we cannot undo.
+		 * If we don't do this, undoworker() won't find a match
+		 * against the buffer, and will be deadlocked.
+		 */
+		lremove(curbp, prevp);
+	    }
 	}
-	newlp = lforw(prevp);
-	tag_for_undo(newlp);
 	endofDisplay();
 
+	if (status == FIOMEM)
+	    break;
+
 	prevp = belowthisline ? newlp : nextp;
 	++nline;
-	if (s < FIOSUC)
+	if (status < FIOSUC)
 	    break;
     }
     if (!haveffp) {
 	CleanAfterPipe(FALSE);
 	(void) ffclose();	/* Ignore errors.       */
-	readlinesmsg(nline, s, fname, FALSE);
+	readlinesmsg(nline, status, fname, FALSE);
     }
 
     /* mark the window for changes.  could this be moved up to
@@ -2330,7 +2534,8 @@ ifile(char *fname, int belowthisline, FILE *haveffp)
     /* copy window parameters back to the buffer structure */
     copy_traits(&(curbp->b_wtraits), &(curwp->w_traits));
 
-    imply_alt(fname, FALSE, FALSE);
+    if (status < FIOERR)
+	imply_alt(fname, FALSE, FALSE);
 
 #if COMPLETE_FILES || COMPLETE_DIRS
     if (b_is_directory(curbp)) {
@@ -2340,7 +2545,7 @@ ifile(char *fname, int belowthisline, FILE *haveffp)
 	/* advance to the next line */
 	DOT.l = lforw(DOT.l);
 
-    return (s != FIOERR);
+    returnCode(FIO2Status(status));
 }
 
 /* try really hard to create a private subdirectory in some tmp
