@@ -5,7 +5,7 @@
  * Written by T.E.Dickey for vile (march 1993).
  *
  *
- * $Header: /usr/build/vile/vile/RCS/filec.c,v 1.113 2003/06/18 21:38:00 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/filec.c,v 1.121 2005/01/20 01:58:58 tom Exp $
  *
  */
 
@@ -42,6 +42,10 @@
 extern char **environ;
 #endif
 
+#if CC_LCC_WIN32
+#define environ _environ
+#endif
+
 #define	SLASH (EOS+1)		/* less than everything but EOS */
 
 #if OPT_VMS_PATH
@@ -224,7 +228,7 @@ bs_init(const char *name)
  * lpp	 - in/out line pointer, for iteration 
  */
 static int
-bs_find(char *fname, size_t len, BUFFER *bp, LINEPTR * lpp)
+bs_find(char *fname, size_t len, BUFFER *bp, LINE **lpp)
 {
     LINE *lp;
     int doit = FALSE;
@@ -309,7 +313,7 @@ already_scanned(BUFFER *bp, char *path)
     LINE *lp;
     size_t len;
     char fname[NFILEN];
-    LINEPTR slp;
+    LINE *slp;
 
     len = force_slash(strcpy(fname, path));
 
@@ -547,7 +551,7 @@ sortMyBuff(BUFFER *bp)
 #endif /* USE_QSORT */
 
 int
-fill_directory_buffer(BUFFER *bp, char *path, int dots)
+fill_directory_buffer(BUFFER *bp, char *path, int dots GCC_UNUSED)
 {
     int count = 0;
     char *s;
@@ -599,7 +603,7 @@ fill_directory_buffer(BUFFER *bp, char *path, int dots)
 #endif
 	    TRACE(("> '%s'\n", path));
 	    if_dots(s, dots) count++;
-	    (void) bs_find(path, strlen(path), bp, (LINEPTR *) 0);
+	    (void) bs_find(path, strlen(path), bp, (LINE **) 0);
 
 	} while (entries = 1,
 		 DosFindNext(hdir, &fb, sizeof(fb), &entries) == NO_ERROR
@@ -657,13 +661,15 @@ fill_directory_buffer(BUFFER *bp, char *path, int dots)
 	    if_dots(leaf, dots) count++;
 #if USE_QSORT
 #if OPT_VMS_PATH
-	    vms2hybrid(s = strcpy(temp, path));
+	    if (temp != path)
+		strcpy(temp, path);
+	    vms2hybrid(s = temp);
 #else
 	    s = path;
 #endif
 	    (void) makeString(bp, buf_head(bp), s, strlen(s));
 #else /* !USE_QSORT */
-	    (void) bs_find(path, strlen(path), bp, (LINEPTR *) 0);
+	    (void) bs_find(path, strlen(path), bp, (LINE **) 0);
 #endif /* USE_QSORT/!USE_QSORT */
 	}
 	(void) closedir(dp);
@@ -696,7 +702,7 @@ fillMyBuff(BUFFER *bp, char *name)
 
 	/**********************************************************************/
     if (is_environ(name)) {
-	LINEPTR lp;
+	LINE *lp;
 	int n;
 	size_t len = strlen(name) - 1;
 
@@ -741,7 +747,7 @@ fillMyBuff(BUFFER *bp, char *name)
 #if USE_QSORT
 	    (void) makeString(bp, buf_head(bp), path, strlen(path));
 #else /* !USE_QSORT */
-	    (void) bs_find(path, strlen(path), bp, (LINEPTR *) 0);
+	    (void) bs_find(path, strlen(path), bp, (LINE **) 0);
 #endif /* USE_QSORT/!USE_QSORT */
 	    TRACE(("> '%s'\n", path));
 	}
@@ -841,16 +847,20 @@ makeMyList(BUFFER *bp, char *name)
 	    bp->b_index_list = typereallocn(char *, bp->b_index_list, bp->b_index_size);
     }
 
-    n = 0;
-    for_each_line(lp, bp) {
-	/* exclude listings of subdirectories below
-	   current directory */
-	if (llength(lp) >= len
-	    && ((slashocc = strchr(lp->l_text + len, SLASHC)) == NULL
-		|| slashocc[1] == EOS))
-	    bp->b_index_list[n++] = lp->l_text;
+    if (bp->b_index_list != 0) {
+	n = 0;
+	for_each_line(lp, bp) {
+	    /* exclude listings of subdirectories below
+	       current directory */
+	    if (llength(lp) >= len
+		&& ((slashocc = strchr(lp->l_text + len, SLASHC)) == NULL
+		    || slashocc[1] == EOS))
+		bp->b_index_list[n++] = lp->l_text;
+	}
+	bp->b_index_list[n] = 0;
+    } else {
+	bp->b_index_size = 0;
     }
-    bp->b_index_list[n] = 0;
     endofDisplay();
 }
 
@@ -906,7 +916,7 @@ path_completion(DONE_ARGS)
 {
     int code = FALSE;
     int action = (c == NAMEC || c == TESTC);
-    int ignore = (*buf != EOS && isInternalName(buf));
+    int ignore;
 #if USE_DOTNAMES
     size_t dots = 0;
     int count;
@@ -915,6 +925,11 @@ path_completion(DONE_ARGS)
     TRACE(("path_completion(%#x '%c' %d:\"%s\")\n",
 	   flags, c, *pos, visible_buff(buf, (int) *pos, TRUE)));
     (void) flags;
+
+    if (buf == 0)
+	return FALSE;
+
+    ignore = (*buf != EOS && isInternalName(buf));
 #if OPT_VMS_PATH
     if (ignore && action) {	/* resolve scratch-name conflict */
 	if (is_vms_pathname(buf, -TRUE))
@@ -1171,7 +1186,7 @@ strip_non_graphics(char *path)
  * flag - +1 to read, -1 to write, 0 don't care
  */
 int
-mlreply_file(const char *prompt, TBUFF ** buffer, UINT flag, char *result)
+mlreply_file(const char *prompt, TBUFF **buffer, UINT flag, char *result)
 {
     int status;
     static TBUFF *last;
@@ -1297,7 +1312,7 @@ mlreply_file(const char *prompt, TBUFF ** buffer, UINT flag, char *result)
  * Prompt for a directory name, allowing completion via tab and '?'
  */
 int
-mlreply_dir(const char *prompt, TBUFF ** buffer, char *result)
+mlreply_dir(const char *prompt, TBUFF **buffer, char *result)
 {
     int status;
     static TBUFF *last;
