@@ -1,7 +1,7 @@
 /*
  * Common utility functions for vile syntax/highlighter programs
  *
- * $Header: /usr/build/vile/vile/filters/RCS/filters.c,v 1.88 2003/05/24 00:49:25 tom Exp $
+ * $Header: /usr/build/vile/vile/filters/RCS/filters.c,v 1.98 2005/03/13 18:00:39 tom Exp $
  *
  */
 
@@ -22,7 +22,6 @@
 #endif
 
 #define VERBOSE(level,params)	if (FltOptions('v') >= level) mlforce params
-#define NONNULL(s)		((s) != 0) ? (s) : "<null>"
 
 #define HASH_LENGTH 256
 
@@ -44,7 +43,8 @@ struct _classes {
 };
 
 char *default_attr;
-int abbr_ch = '*';
+int zero_or_more = '*';		/* zero or more of the following */
+int zero_or_all = '?';		/* zero or all of the following */
 int meta_ch = '.';
 int eqls_ch = ':';
 int vile_keywords;
@@ -73,6 +73,12 @@ static unsigned len_keyword_file = 0;
  * Private functions                                                          *
  ******************************************************************************/
 
+static void
+CannotAllocate(const char *where)
+{
+    VERBOSE(1, ("%s: cannot allocate", where));
+}
+
 static const char *
 AttrsOnce(KEYWORD * entry)
 {
@@ -83,7 +89,13 @@ AttrsOnce(KEYWORD * entry)
 static void
 ExecAbbrev(char *param)
 {
-    abbr_ch = *param;
+    zero_or_more = *param;
+}
+
+static void
+ExecBrief(char *param)
+{
+    zero_or_all = *param;
 }
 
 static void
@@ -104,6 +116,7 @@ ExecDefault(char *param)
     if (is_class(param)) {
 	free(default_attr);
 	default_attr = strmalloc(param);
+	VERBOSE(1, ("set default_attr '%s' %p\n", default_attr, default_attr));
     } else {
 	*s = save;
 	VERBOSE(1, ("not a class:%s", param));
@@ -148,6 +161,7 @@ static void
 ExecTable(char *param)
 {
     flt_make_symtab(param);
+    flt_read_keywords(MY_NAME);	/* provide default values for this table */
 }
 
 static KEYWORD *
@@ -205,6 +219,10 @@ OpenKeywords(char *classname)
 
     need = sizeof(suffix) + strlen(classname) + 2;
     str_keyword_file = do_alloc(str_keyword_file, need, &len_keyword_file);
+    if (str_keyword_file == 0) {
+	CannotAllocate("OpenKeywords");
+	return 0;
+    }
     sprintf(str_keyword_file, "%s%s", classname, suffix);
 
     if (strchr(str_keyword_file, PATHSEP) != 0) {
@@ -219,6 +237,10 @@ OpenKeywords(char *classname)
 	+ 20;
 
     str_keyword_name = do_alloc(str_keyword_name, need, &len_keyword_name);
+    if (str_keyword_name == 0) {
+	CannotAllocate("OpenKeywords");
+	return 0;
+    }
 
     FIND_IT((str_keyword_name, "%s%c%s%s", PATHDOT, PATHSEP, DOT_TO_HIDE_IT, str_keyword_file));
     FIND_IT((str_keyword_name, "%s%c%s%s", path, PATHSEP, DOT_TO_HIDE_IT, str_keyword_file));
@@ -226,7 +248,7 @@ OpenKeywords(char *classname)
 
     FIND_IT((str_keyword_name, "%s%c%s%s", path, PATHSEP, leaf, str_keyword_file));
 
-    path = getenv("VILE_STARTUP_PATH");
+    path = vile_getenv("VILE_STARTUP_PATH");
 #ifdef VILE_STARTUP_PATH
     if (path == 0)
 	path = VILE_STARTUP_PATH;
@@ -236,6 +258,10 @@ OpenKeywords(char *classname)
 
 	need = strlen(path) + strlen(str_keyword_file) + 2;
 	str_keyword_name = do_alloc(str_keyword_name, need, &len_keyword_name);
+	if (str_keyword_name == 0) {
+	    CannotAllocate("OpenKeywords");
+	    return 0;
+	}
 	while (path[n] != 0) {
 	    for (m = n; path[m] != 0 && path[m] != PATHCHR; m++)
 		/*LOOP */ ;
@@ -259,6 +285,7 @@ ParseDirective(char *line)
     } table[] = {
 	/* *INDENT-OFF* */
 	{ "abbrev",  ExecAbbrev   },
+	{ "brief",   ExecBrief    },
 	{ "class",   ExecClass    },
 	{ "default", ExecDefault  },
 	{ "equals",  ExecEquals   },
@@ -328,6 +355,7 @@ class_attr(char *name)
 	VERBOSE(hash_id->kw_used, ("class_attr(%s) = %s",
 				   name, AttrsOnce(hash_id)));
 	name = result = hash_id->kw_attr;
+	VERBOSE(1, ("-> %p\n", result));
     }
     return result;
 }
@@ -355,8 +383,12 @@ void
 flt_bfr_append(char *text, int length)
 {
     flt_bfr_text = do_alloc(flt_bfr_text, flt_bfr_used + length, &flt_bfr_size);
-    strncpy(flt_bfr_text + flt_bfr_used, text, length);
-    flt_bfr_used += length;
+    if (flt_bfr_text != 0) {
+	strncpy(flt_bfr_text + flt_bfr_used, text, length);
+	flt_bfr_used += length;
+    } else {
+	CannotAllocate("flt_bfr_append");
+    }
 }
 
 void
@@ -380,6 +412,7 @@ void
 flt_bfr_error(void)
 {
     if (flt_bfr_used) {
+	flt_error("unterminated buffer");
 	flt_bfr_attr = class_attr(NAME_ERROR);
 	flt_bfr_finish();
     }
@@ -456,7 +489,8 @@ flt_initialize(void)
 	free(default_attr);
     default_attr = strmalloc(NAME_KEYWORD);
 
-    abbr_ch = '*';
+    zero_or_more = '*';
+    zero_or_all = '?';
     meta_ch = '.';
     eqls_ch = ':';
     FltOptions('v') = 0;
@@ -468,9 +502,23 @@ void
 flt_make_symtab(char *classname)
 {
     if (!set_symbol_table(classname)) {
-	CLASS *p = typecallocn(CLASS, 1);
+	CLASS *p;
+
+	if ((p = typecallocn(CLASS, 1)) == 0) {
+	    CannotAllocate("flt_make_symtab");
+	    return;
+	}
+
 	p->name = strmalloc(classname);
 	p->data = typecallocn(KEYWORD *, HASH_LENGTH);
+	if (p->name == 0 || p->data == 0) {
+	    if (p->name != 0)
+		free(p->name);
+	    free(p);
+	    CannotAllocate("flt_make_symtab");
+	    return;
+	}
+
 	p->next = classes;
 	classes = p;
 	hashtable = p->data;
@@ -494,7 +542,6 @@ flt_make_symtab(char *classname)
 	insert_keyword(NAME_NUMBER, ATTR_NUMBER, 1);
 	insert_keyword(NAME_PREPROC, ATTR_PREPROC, 1);
 	insert_keyword(NAME_TYPES, ATTR_TYPES, 1);
-
     }
 }
 
@@ -567,33 +614,19 @@ hash_function(const char *id)
     return ((CharOf(id[0]) ^ (CharOf(id[1]) << 3) ^ (CharOf(id[2]) >> 1)) & 0xff);
 }
 
-void
-insert_keyword(const char *ident, const char *attribute, int classflag)
+static KEYWORD *
+alloc_keyword(const char *ident, const char *attribute, int classflag)
 {
     KEYWORD *first;
     KEYWORD *nxt;
     int Index;
-    char *mark;
-
-    if ((mark = strchr(ident, abbr_ch)) != 0
-	&& (mark != ident)) {
-	char *temp = strmalloc(ident);
-
-	mark = temp + (mark - ident);
-	while (*mark == abbr_ch) {
-	    *mark = 0;
-	    insert_keyword(temp, attribute, classflag);
-	    if ((mark[0] = mark[1]) != 0) {
-		*(++mark) = '*';
-	    }
-	}
-	free(temp);
-	return;
-    }
 
     if ((nxt = FindIdentifier(ident)) != 0) {
 	Free(nxt->kw_attr);
-	nxt->kw_attr = strmalloc(attribute);
+	if ((nxt->kw_attr = strmalloc(attribute)) == NULL) {
+	    free(nxt);
+	    nxt = 0;
+	}
     } else {
 	nxt = first = NULL;
 	Index = hash_function(ident);
@@ -605,15 +638,71 @@ insert_keyword(const char *ident, const char *attribute, int classflag)
 	    nxt->kw_flag = classflag;
 	    nxt->kw_used = 2;
 	    nxt->next = first;
-	    hashtable[Index] = nxt;
-	} else {
-	    VERBOSE(1, ("insert_keyword: cannot allocate"));
-	    return;
+	    if (nxt->kw_name != 0
+		&& nxt->kw_attr != 0) {
+		hashtable[Index] = nxt;
+	    } else {
+		if (nxt->kw_name != 0)
+		    free(nxt->kw_name);
+		free(nxt);
+		nxt = 0;
+	    }
 	}
     }
-    VERBOSE(3, ("...\tname \"%s\"\tattr \"%s\"",
-		nxt->kw_name,
-		NONNULL(nxt->kw_attr)));
+    return nxt;
+}
+
+void
+insert_keyword(const char *ident, const char *attribute, int classflag)
+{
+    KEYWORD *nxt;
+    char *mark;
+    char *temp;
+
+    VERBOSE(2, ("insert_keyword(%s, %s, %d)\n",
+		ident,
+		attribute,
+		classflag));
+
+    if ((mark = strchr(ident, zero_or_more)) != 0
+	&& (mark != ident)) {
+	if ((temp = strmalloc(ident)) != 0) {
+
+	    mark = temp + (mark - ident);
+	    while (*mark == zero_or_more) {
+		*mark = 0;
+		insert_keyword(temp, attribute, classflag);
+		if ((mark[0] = mark[1]) != 0) {
+		    *(++mark) = zero_or_more;
+		}
+	    }
+	    free(temp);
+	} else {
+	    CannotAllocate("insert_keyword");
+	}
+    } else if ((mark = strchr(ident, zero_or_all)) != 0
+	       && (mark != ident)) {
+	if ((temp = strmalloc(ident)) != 0) {
+
+	    mark = temp + (mark - ident);
+	    if (*mark == zero_or_all) {
+		*mark = 0;
+		insert_keyword(temp, attribute, classflag);
+		while ((mark[0] = mark[1]) != 0)
+		    ++mark;
+		insert_keyword(temp, attribute, classflag);
+	    }
+	    free(temp);
+	} else {
+	    CannotAllocate("insert_keyword");
+	}
+    } else if ((nxt = alloc_keyword(ident, attribute, classflag)) == 0) {
+	CannotAllocate("insert_keyword");
+    } else {
+	VERBOSE(3, ("...\tname \"%s\"\tattr \"%s\"",
+		    nxt->kw_name,
+		    NONNULL(nxt->kw_attr)));
+    }
 }
 
 KEYWORD *
@@ -649,6 +738,7 @@ keyword_attr(char *name)
 	while ((hash_id = is_class(result)) != 0)
 	    result = hash_id->kw_attr;
     }
+    VERBOSE(1, ("keyword_attr(%s) = %p %s\n", name, result, NONNULL(result)));
     return result;
 }
 
@@ -662,7 +752,7 @@ lowercase_of(char *text)
     name = do_alloc(name, strlen(text), &used);
     for (n = 0; text[n] != 0; n++) {
 	if (isalpha(CharOf(text[n])) && isupper(CharOf(text[n])))
-	    name[n] = tolower(text[n]);
+	    name[n] = tolower(CharOf(text[n]));
 	else
 	    name[n] = text[n];
     }
@@ -718,7 +808,15 @@ parse_keyword(char *name, int classflag)
 	    VERBOSE(2, ("using attr \"%s\"", args));
 	}
 	if ((hash_id = FindIdentifier(args)) != 0) {
-	    insert_keyword(name, hash_id->kw_attr, classflag);
+	    /*
+	     * Insert the classname rather than the hash_id->kw_attr value,
+	     * since insert_keyword makes a copy of the string we pass to it.
+	     * Retrieving the attribute from the copy will give the unique
+	     * attribute string belonging to the class, so it is possible at
+	     * runtime to compare pointers from keyword_attr() to determine if
+	     * two tokens have the same class.  sql-filt.l uses this feature.
+	     */
+	    insert_keyword(name, args, classflag);
 	}
     }
 }
@@ -729,8 +827,10 @@ readline(FILE *fp, char **ptr, unsigned *len)
     char *buf = *ptr;
     unsigned used = 0;
 
-    if (buf == 0)
-	buf = (char *) malloc(*len = BUFSIZ);
+    if (buf == 0) {
+	*len = BUFSIZ;
+	buf = typeallocn(char, *len);
+    }
     while (!feof(fp)) {
 	int ch = fgetc(fp);
 	if (ch == EOF || feof(fp) || ferror(fp)) {
@@ -738,7 +838,9 @@ readline(FILE *fp, char **ptr, unsigned *len)
 	}
 	if (used + 2 >= *len) {
 	    *len = 3 * (*len) / 2;
-	    buf = (char *) realloc(buf, *len);
+	    buf = typereallocn(char, buf, *len);
+	    if (buf == 0)
+		return 0;
 	}
 	buf[used++] = ch;
 	if (ch == '\n')
