@@ -1,5 +1,5 @@
 /*
- * $Header: /usr/build/vile/vile/filters/RCS/pl-filt.c,v 1.70 2003/05/24 15:19:03 tom Exp $
+ * $Header: /usr/build/vile/vile/filters/RCS/pl-filt.c,v 1.82 2005/07/11 22:35:34 tom Exp $
  *
  * Filter to add vile "attribution" sequences to perl scripts.  This is a
  * translation into C of an earlier version written for LEX/FLEX.
@@ -7,6 +7,7 @@
 
 #include <filters.h>
 
+#define DEBUG
 #ifdef DEBUG
 DefineOptFilter("perl", "d");
 #else
@@ -38,6 +39,11 @@ typedef enum {
     ,eIGNORED
 } States;
 
+typedef struct {
+    int may_have_pattern;
+    int has_no_pattern;
+} AfterKey;
+
 static char *Action_attr;
 static char *Comment_attr;
 static char *Error_attr;
@@ -48,6 +54,8 @@ static char *String_attr;
 static char *Preproc_attr;
 static char *Number_attr;
 
+static AfterKey nullKey;
+
 /*
  * The in-memory copy of the input file.
  */
@@ -121,7 +129,7 @@ is_ESCAPED(char *s)
 {
     int found = 0;
     if (*s == BACKSLASH) {
-	if (ATLEAST(s, 2) && strchr("'\"`\\", s[1]))
+	if (ATLEAST(s, 2) && strchr("'`\\", s[1]))
 	    found = 2;
 	else
 	    found = 1;
@@ -228,29 +236,33 @@ is_QIDENT(char *s)
     return s - base;
 }
 
+/*
+ * If double-quoted, look for variables after a "$", "%" or "@".
+ * If not double-quoted, look also for syntax such as $foo'bar
+ */
 static int
-is_NORMALVARS(char *s)
+is_NORMALVARS(char *s, int dquoted)
 {
     char *base = s;
     int ch;
-    int quoted = 0;
+    int squoted = 0;
     int part1 = 0;
     int part2 = 0;
 
     while (MORE(s)) {
 	ch = CharOf(*s);
 	if (s == base) {
-	    if (strchr("$%@", ch) == 0) {
+	    if (strchr(dquoted ? "$" : "$%@", ch) == 0) {
 		break;
 	    }
-	} else if (quoted) {
+	} else if (squoted && !dquoted) {
 	    if (isalnum(ch))
 		part2 = 1;
 	    else
 		break;
 	} else {
-	    if (ch == SQUOTE) {
-		quoted = 1;
+	    if (ch == SQUOTE && !dquoted) {
+		squoted = 1;
 	    } else if (isalnum(ch) || ch == '_') {
 		part1 = 1;
 	    } else {
@@ -259,7 +271,7 @@ is_NORMALVARS(char *s)
 	}
 	s++;
     }
-    return (part1 && (quoted == part2)) ? (s - base) : 0;
+    return (part1 && (dquoted || (squoted == part2))) ? (s - base) : 0;
 }
 
 static int
@@ -392,12 +404,40 @@ is_NUMBER(char *s, int *err)
     return value ? (s - base) : 0;
 }
 
+static char *
+put_NUMBER(char *s, int ok, int *err)
+{
+    if (*err) {
+	flt_error("illegal number");
+	flt_puts(s, ok, Error_attr);
+    } else {
+	flt_puts(s, ok, Number_attr);
+    }
+    s += ok;
+    return s;
+}
+
 static int
-is_IDENT(char *s)
+is_ELLIPSIS(char *s)
+{
+    int dots;
+    int ok = 0;
+
+    for (dots = 0; dots < 3; ++dots) {
+	if (!MORE(s) || (*s++ != '.')) {
+	    break;
+	}
+	ok = dots + 1;
+    }
+    return (ok > 1) ? ok : 0;
+}
+
+static int
+is_IDENT(char *s, int quoted)
 {
     int found;
 
-    if ((found = is_NORMALVARS(s)) == 0)
+    if ((found = is_NORMALVARS(s, quoted)) == 0)
 	found = is_OTHERVARS(s);
     return found;
 }
@@ -655,13 +695,13 @@ is_QUOTE(char *s, int *delims)
 
     if (s != base) {
 	int test = after_blanks(s);
-	DPRINTF(("is_Quote(%.*s:%c)", s - base, base, test));
+	DPRINTF(("is_Quote(%.*s:%c)", (int) (s - base), base, test));
 	if (test == '#' && isspace(CharOf(*s)))
 	    test = 0;
 	if ((test == 0) || (strchr(QUOTE_DELIMS, test) == 0))
 	    s = base;
-	DPRINTF(("is_QUOTE(%.*s)",
-		 (s != base) ? (s - base) : 1,
+	DPRINTF(("is_QUOTE(%.*s)\n",
+		 (s != base) ? (int) (s - base) : 1,
 		 (s != base) ? base : ""));
     }
     return (s - base);
@@ -881,6 +921,28 @@ is_String(char *s, int *err)
     return found;
 }
 
+/*
+ * FIXME: this only handles subscripts that consist of a single identifier,
+ * with no expressions.
+ */
+static int
+is_Subscript(char *s, int len, char *delimp)
+{
+    if (delimp != 0) {
+	int delim = *delimp;
+	int delim2 = delim;
+	char *next;
+
+	if (delim != 0 && (next = strchr(LOOKUP_TERM, delim)) != 0) {
+	    delim2 = next[5];
+
+	    if (s[len] == delim2)
+		return 1;
+	}
+    }
+    return 0;
+}
+
 static char *
 put_newline(char *s)
 {
@@ -910,7 +972,7 @@ put_embedded(char *s, int len, char *attr)
 
     for (j = k = 0; j < len; j++) {
 	if ((j == 0 || (s[j - 1] != BACKSLASH))
-	    && (id = is_IDENT(s + j)) != 0) {
+	    && (id = is_IDENT(s + j, 1)) != 0) {
 	    if (var_embedded(s + j)) {
 		if (j != k)
 		    flt_puts(s + k, j - k, attr);
@@ -949,36 +1011,75 @@ put_remainder(char *s, char *attr, int quoted)
  * Check for special cases of keywords after which we may expect a pattern
  * that does not have to be inside parentheses.
  */
-static int
-check_keyword(char *s, int ok)
+static void
+check_keyword(char *s, int ok, AfterKey * state)
 {
+    state->may_have_pattern = 0;
+    state->has_no_pattern = 0;
+
     switch (ok) {
     case 2:
-	return !strncmp(s, "if", ok);
+	state->may_have_pattern = (!strncmp(s, "if", ok)
+				   || !strncmp(s, "eq", ok)
+				   || !strncmp(s, "ge", ok)
+				   || !strncmp(s, "gt", ok)
+				   || !strncmp(s, "le", ok)
+				   || !strncmp(s, "lt", ok)
+				   || !strncmp(s, "ne", ok)
+				   || !strncmp(s, "or", ok));
+	break;
+    case 3:
+	state->has_no_pattern = !strncmp(s, "sub", ok);
+	state->may_have_pattern = (!strncmp(s, "and", ok)
+				   || !strncmp(s, "cmp", ok)
+				   || !strncmp(s, "not", ok)
+				   || !strncmp(s, "xor", ok));
+	break;
     case 4:
-	return !strncmp(s, "grep", ok);
+	state->may_have_pattern = !strncmp(s, "grep", ok);
+	break;
     case 5:
-	return !strncmp(s, "split", ok)
-	    || !strncmp(s, "until", ok)
-	    || !strncmp(s, "while", ok);
+	state->may_have_pattern = (!strncmp(s, "split", ok)
+				   || !strncmp(s, "until", ok)
+				   || !strncmp(s, "while", ok));
+	break;
     case 6:
-	return !strncmp(s, "unless", ok);
+	state->may_have_pattern = !strncmp(s, "unless", ok);
+	break;
     }
-    return 0;
 }
 
+/*
+ * Identifier may be a keyword, or a user identifier.
+ */
 static char *
-put_IDENT(char *s, int ok, int *had_op, int *if_wrd)
+put_IDENT(char *s, int ok, AfterKey * if_wrd)
 {
     char *attr = 0;
     char save = s[ok];
 
-    *had_op = 0;
     s[ok] = '\0';
     attr = keyword_attr(s);
     s[ok] = save;
     flt_puts(s, ok, (attr != 0 && *attr != '\0') ? attr : Ident2_attr);
-    *if_wrd = check_keyword(s, ok);
+    check_keyword(s, ok, if_wrd);
+    return s + ok;
+}
+
+/*
+ * Identifier must be a user identifier.
+ */
+static char *
+put_NOKEYWORD(char *s, int ok, AfterKey * if_wrd)
+{
+    char *attr = 0;
+    char save = s[ok];
+
+    s[ok] = '\0';
+    attr = keyword_attr(s);
+    s[ok] = save;
+    flt_puts(s, ok, "");
+    check_keyword(s, ok, if_wrd);
     return s + ok;
 }
 
@@ -1055,29 +1156,36 @@ init_filter(int before GCC_UNUSED)
 {
 }
 
+#define opRightArrow() (had_op != 0 && old_op == had_op - 1 && *old_op == '-' && *had_op == '>')
+#define opBeforePattern() (had_op != 0 && strchr("{(|&=~!", *had_op) != 0)
+
+#define saveOp(p) { DPRINTF(("\nsaveOp @%d\n", __LINE__)); old_op = had_op; had_op = p; }
+#define clearOp() { DPRINTF(("\nclearOp @%d\n", __LINE__)); old_op = had_op = 0; }
+
 static void
 do_filter(FILE *input GCC_UNUSED)
 {
     static unsigned used;
     static char *line;
 
-    unsigned actual = 0;
-    size_t request = 0;
+    AfterKey if_old;
+    AfterKey if_wrd;
     States state = eCODE;
-    char *s;
     char *marker = 0;
-    unsigned mark_len = 0;
+    char *s;
+    int err;
+    char *had_op = 0;
+    char *old_op = 0;
+    int ignore;
     int in_line = -1;
     int in_stmt = 0;
     int ok;
-    int err;
-    int save;
-    int quoted = 0;
     int parens = 0;
-    int had_op = 0;
-    int if_wrd = 0;
-    int if_old = 0;
-    int ignore;
+    int quoted = 0;
+    int save;
+    size_t request = 0;
+    unsigned actual = 0;
+    unsigned mark_len = 0;
 
     Action_attr = class_attr(NAME_ACTION);
     Comment_attr = class_attr(NAME_COMMENT);
@@ -1113,7 +1221,8 @@ do_filter(FILE *input GCC_UNUSED)
 	the_last = the_file + the_size;
 
 	s = the_file;
-	if_old = if_wrd = 0;
+	if_old = nullKey;
+	if_wrd = nullKey;
 	while (MORE(s)) {
 	    if (*s == '\n' || s == the_file) {
 		in_line = -1;
@@ -1121,14 +1230,20 @@ do_filter(FILE *input GCC_UNUSED)
 		in_line++;
 	    }
 	    if_old = if_wrd;
-	    if_wrd = 0;
-	    DPRINTF(("(%s(%c) line:%d.%d(%d) if:%d op:%d)\n",
-		     stateName(state), *s, in_line, in_stmt, parens, if_wrd, had_op));
+	    if_wrd = nullKey;
+	    DPRINTF(("(%s(%c) line:%d.%d(%d) if:%d.%d op:%c%c)%s\n",
+		     stateName(state), *s, in_line, in_stmt, parens,
+		     if_wrd.may_have_pattern,
+		     if_wrd.has_no_pattern,
+		     old_op ? *old_op : ' ',
+		     had_op ? *had_op : ' ',
+		     opRightArrow()? " arrow" : ""));
 	    switch (state) {
 	    case eBACKTIC:
 		if ((ok = end_BACKTIC(s)) != 0) {
 		    s = put_embedded(s, ok, "");
 		} else {
+		    flt_error("missing backtic");
 		    s = put_remainder(s, Error_attr, 0);
 		}
 		if (*s == BQUOTE) {
@@ -1161,6 +1276,7 @@ do_filter(FILE *input GCC_UNUSED)
 		} else if ((ok = begin_PATTERN(s)) != 0) {
 		    flt_puts(s, ok, "");
 		    s += ok;
+		    DPRINTF(("\nePATTERN:%d\n", __LINE__));
 		    state = ePATTERN;
 		} else if (in_line < 0
 			   && (ok = begin_POD(s))) {
@@ -1181,78 +1297,105 @@ do_filter(FILE *input GCC_UNUSED)
 		    flt_puts(s, ok, "");
 		    s += ok;
 		    if_wrd = if_old;
-		} else if ((if_old || had_op || (in_stmt == 1)) && isPattern(*s)) {
+		} else if ((if_old.may_have_pattern
+			    || opBeforePattern()
+			    || (in_stmt == 1))
+			   && isPattern(*s)) {
+		    DPRINTF(("\nePATTERN:%d\n", __LINE__));
 		    state = ePATTERN;
 		} else if (*s == L_CURLY) {
-		    had_op = 1;
+		    saveOp(s);
 		    flt_putc(*s++);
 		} else if (*s == L_PAREN) {
 		    parens++;
-		    had_op = 1;
+		    saveOp(s);
 		    flt_putc(*s++);
 		    if (isPattern(*s)) {
+			DPRINTF(("\nePATTERN:%d\n", __LINE__));
 			state = ePATTERN;
 		    }
 		} else if (*s == R_PAREN) {
 		    if (--parens < 0)
 			parens = 0;
 		    flt_putc(*s++);
-		    had_op = 0;
+		    clearOp();
+		} else if ((ok = is_ELLIPSIS(s)) != 0) {
+		    flt_puts(s, ok, "");
+		    s += ok;
 		} else if ((ok = is_NUMBER(s, &err)) != 0) {
-		    had_op = 0;
-		    flt_puts(s, ok, err ? Error_attr : Number_attr);
-		    s += ok;
+		    clearOp();
+		    s = put_NUMBER(s, ok, &err);
 		} else if ((ok = is_KEYWORD(s)) != 0) {
-		    if ((s != the_file) && (s[-1] == '*')) {
-			/* typeglob */
-			s = put_IDENT(s, ok, &had_op, &if_wrd);
-			break;
-		    }
-		    if (is_QUOTE(s, &ignore)) {
+		    if ((s != the_file)
+			&& (s[-1] == '*')) {	/* typeglob */
+			s = put_IDENT(s, ok, &if_wrd);
+			clearOp();
+		    } else if ((s != the_file)
+			       && (if_old.has_no_pattern)) {
+			s = put_NOKEYWORD(s, ok, &if_wrd);
+			clearOp();
+		    } else if (!opRightArrow()
+			       && !is_Subscript(s, ok, had_op)
+			       && is_QUOTE(s, &ignore)) {
+			DPRINTF(("\nePATTERN:%d\n", __LINE__));
 			state = ePATTERN;
-			break;
-		    }
-		    if (is_FORMAT(s, ok)) {
-			quoted = 0;
-			state = eHERE;
-			mark_len = 0;
-			marker = strcpy(do_alloc(0, 2, &mark_len), ".");
-		    }
-		    save = s[ok];
-		    s[ok] = 0;
-		    if (!strcmp(s, "__END__") || !strcmp(s, "__DATA__"))
-			state = eIGNORED;
-		    had_op = 0;
-		    flt_puts(s, ok, keyword_attr(s));
-		    if_wrd = check_keyword(s, ok);
-		    s[ok] = save;
-		    s += ok;
-		} else if ((ok = is_Option(s)) != 0) {
-		    had_op = 0;
-		    flt_puts(s, ok, Keyword_attr);
-		    if_wrd = check_keyword(s, ok);
-		    s += ok;
-		} else if ((ok = is_IDENT(s)) != 0) {
-		    s = put_IDENT(s, ok, &had_op, &if_wrd);
-		} else if ((ok = is_String(s, &err)) != 0) {
-		    had_op = 0;
-		    if (*s == DQUOTE) {
-			s = put_embedded(s, ok, err ? Error_attr : String_attr);
 		    } else {
-			flt_puts(s, ok, err ? Error_attr : String_attr);
+			if (is_FORMAT(s, ok)) {
+			    quoted = 0;
+			    state = eHERE;
+			    mark_len = 0;
+			    marker = strcpy(do_alloc(0, 2, &mark_len), ".");
+			}
+			save = s[ok];
+			s[ok] = 0;
+			if (!strcmp(s, "__END__") || !strcmp(s, "__DATA__"))
+			    state = eIGNORED;
+			clearOp();
+			flt_puts(s, ok, keyword_attr(s));
+			check_keyword(s, ok, &if_wrd);
+			s[ok] = save;
+			s += ok;
+		    }
+		} else if ((ok = is_Option(s)) != 0) {
+		    clearOp();
+		    flt_puts(s, ok, Keyword_attr);
+		    check_keyword(s, ok, &if_wrd);
+		    s += ok;
+		} else if ((ok = is_IDENT(s, 0)) != 0) {
+		    s = put_IDENT(s, ok, &if_wrd);
+		    clearOp();
+		} else if ((ok = is_String(s, &err)) != 0) {
+		    clearOp();
+		    if (*s == DQUOTE) {
+			if (err) {
+			    flt_error("unexpected quote");
+			    s = put_embedded(s, ok, Error_attr);
+			} else {
+			    s = put_embedded(s, ok, String_attr);
+			}
+		    } else {
+			if (err) {
+			    flt_error("missing quote");
+			    flt_puts(s, ok, Error_attr);
+			} else {
+			    flt_puts(s, ok, String_attr);
+			}
 			s += ok;
 		    }
 		} else {
 		    if (parens) {
-			if (strchr("|&=~!", *s) != 0)
-			    had_op = 1;
-			else if (!isspace(CharOf(*s)))
-			    had_op = 0;
+			if (strchr("|&=~!->", *s) != 0) {
+			    saveOp(s);
+			} else if (!isspace(CharOf(*s))) {
+			    clearOp();
+			}
 		    } else {
-			if (*s == ';')
+			if (*s == ';') {
 			    in_stmt = 0;
-			if (strchr("|&=~!", *s) != 0)
-			    had_op = 1;
+			}
+			if (strchr("|&=~!->", *s) != 0) {
+			    saveOp(s);
+			}
 		    }
 		    flt_putc(*s++);
 		}
@@ -1275,15 +1418,17 @@ do_filter(FILE *input GCC_UNUSED)
 		s = skip_BLANKS(s);
 		if (MORE(s)) {
 		    if ((ok = is_NAME(s)) != 0 && !is_QUOTE(s, &ignore)) {
-			s = put_IDENT(s, ok, &had_op, &if_wrd);
-		    } else if ((ok = is_IDENT(s)) != 0 && !is_QUOTE(s, &ignore)) {
-			s = put_IDENT(s, ok, &had_op, &if_wrd);
+			s = put_IDENT(s, ok, &if_wrd);
+			clearOp();
+		    } else if ((ok = is_IDENT(s, 0)) != 0 && !is_QUOTE(s, &ignore)) {
+			s = put_IDENT(s, ok, &if_wrd);
+			clearOp();
 		    } else if ((ok = add_to_PATTERN(s)) != 0) {
 			s = write_PATTERN(s, ok);
 		    } else {
 			flt_putc(*s++);
 		    }
-		    had_op = 0;
+		    clearOp();
 		    state = eCODE;
 		}
 		break;
