@@ -1,5 +1,5 @@
 /*
- * $Header: /usr/build/vile/vile/filters/RCS/rubyfilt.c,v 1.11 2003/05/24 00:49:25 tom Exp $
+ * $Header: /usr/build/vile/vile/filters/RCS/rubyfilt.c,v 1.38 2005/05/14 12:38:54 tom Exp $
  *
  * Filter to add vile "attribution" sequences to ruby scripts.  This is a
  * translation into C of an earlier version written for LEX/FLEX.
@@ -18,6 +18,10 @@ DefineFilter("ruby");
 #endif
 
 #define isIdent(c)   (isalnum(CharOf(c)) || c == '_')
+#define isIdent0(c)  (isalpha(CharOf(c)) || c == '_')
+
+#define MORE(s)	     ((s) < the_last)
+#define ATLEAST(s,n) (the_last - (s) > (n))
 
 #ifdef DEBUG
 #define DPRINTF(params) if(FltOptions('d'))printf params
@@ -25,12 +29,32 @@ DefineFilter("ruby");
 #define DPRINTF(params)		/*nothing */
 #endif
 
+#define Parsed(var,tok) var = tok; \
+	DPRINTF(("...%s: %d\n", tokenType(var), ok))
+
 typedef enum {
-    eCODE
+    eALIAS
+    ,eCLASS
+    ,eCODE
+    ,eDEF
     ,eHERE
     ,ePOD
 } States;
 
+typedef enum {
+    tNULL
+    ,tBLANK
+    ,tCHAR
+    ,tCOMMENT
+    ,tHERE
+    ,tKEYWORD
+    ,tNUMBER
+    ,tOPERATOR
+    ,tREGEXP
+    ,tSTRING
+    ,tVARIABLE
+} TType;
+
 static char *Comment_attr;
 static char *Error_attr;
 static char *Ident_attr;
@@ -39,6 +63,8 @@ static char *Keyword_attr;
 static char *String_attr;
 static char *Number_attr;
 
+static int var_embedded(char *);
+
 /*
  * The in-memory copy of the input file.
  */
@@ -50,19 +76,70 @@ static size_t the_size;
 static char *
 stateName(States state)
 {
-    char *result;
+    char *result = "?";
+
     switch (state) {
+    case eALIAS:
+	result = "ALIAS";
+	break;
+    case eCLASS:
+	result = "CLASS";
+	break;
     case eCODE:
 	result = "CODE";
 	break;
+    case eDEF:
+	result = "DEF";
+	break;
     case eHERE:
 	result = "HERE";
 	break;
     case ePOD:
 	result = "POD";
 	break;
-    default:
-	result = "?";
+    }
+    return result;
+}
+
+static char *
+tokenType(TType type)
+{
+    char *result = "?";
+
+    switch (type) {
+    case tNULL:
+	result = "No Token";
+	break;
+    case tBLANK:
+	result = "BLANK";
+	break;
+    case tCHAR:
+	result = "CHAR";
+	break;
+    case tCOMMENT:
+	result = "COMMENT";
+	break;
+    case tHERE:
+	result = "HERE";
+	break;
+    case tKEYWORD:
+	result = "KEYWORD";
+	break;
+    case tNUMBER:
+	result = "NUMBER";
+	break;
+    case tOPERATOR:
+	result = "OPERATOR";
+	break;
+    case tREGEXP:
+	result = "REGEXP";
+	break;
+    case tSTRING:
+	result = "STRING";
+	break;
+    case tVARIABLE:
+	result = "VARIABLE";
+	break;
     }
     return result;
 }
@@ -75,7 +152,7 @@ static int
 is_BLANK(char *s)
 {
     int found = 0;
-    while ((s != the_last) && isBlank(*s)) {
+    while (MORE(s) && isBlank(*s)) {
 	found++;
 	s++;
     }
@@ -87,18 +164,19 @@ is_ESCAPED(char *s)
 {
     int found = 0;
     if (*s == BACKSLASH) {
-	found = 1;
+	found = 2;
     }
     return found;
 }
 
 static int
-is_STRINGS(char *s, int *err, int left_delim, int right_delim)
+is_STRINGS(char *s, int *err, int left_delim, int right_delim, int single)
 {
     char *base = s;
     int found = 0;
     int escape = 0;
     int level = 0;
+    int len;
 
     *err = 0;
     if (*s == left_delim) {	/* should always be true */
@@ -107,7 +185,7 @@ is_STRINGS(char *s, int *err, int left_delim, int right_delim)
 	    ++level;
 
 	for (;;) {
-	    if (s >= the_last) {
+	    if (!MORE(s)) {
 		*err = 1;	/* unterminated string */
 		break;
 	    }
@@ -115,6 +193,8 @@ is_STRINGS(char *s, int *err, int left_delim, int right_delim)
 		escape = 1;
 	    } else if (escape) {
 		escape = 0;
+	    } else if (!single && (len = var_embedded(s)) != 0) {
+		s += len - 1;
 	    } else {
 		if (left_delim != right_delim) {
 		    if (*s == left_delim) {
@@ -148,16 +228,22 @@ is_QIDENT(char *s)
 {
     char *base = s;
     int ch;
+    int leading = 1;
 
-    while (s != the_last) {
+    while (MORE(s)) {
 	ch = CharOf(*s);
-	if (isIdent(ch)
-	    || ch == BQUOTE
-	    || ch == SQUOTE
-	    || ch == DQUOTE)
+	if (leading && isIdent0(ch)) {
+	    leading = 0;
 	    s++;
-	else
+	} else if (!leading && isIdent(ch)) {
+	    s++;
+	} else if (ch == BQUOTE
+		   || ch == SQUOTE
+		   || ch == DQUOTE) {
+	    s++;
+	} else {
 	    break;
+	}
     }
     return s - base;
 }
@@ -171,8 +257,8 @@ is_KEYWORD(char *s)
 {
     int found = 0;
 
-    if (!isdigit(CharOf(s[0]))) {
-	while ((s + found < the_last) && isIdent(s[found])) {
+    if (isIdent0(CharOf(s[0]))) {
+	while (ATLEAST(s, found) && isIdent(s[found])) {
 	    ++found;
 	}
     }
@@ -188,9 +274,13 @@ is_GLOBAL(char *s)
 {
     int found = 0;
 
-    if (*s == '$' && (++s < the_last)) {
+    if (*s == '$' && MORE(++s)) {
 	if (*s != '\0' && strchr("-_./,\"\\#%=~|$?&`'+*[];!@<>():", *s)) {
 	    found = 1;
+	} else if (isdigit(CharOf(*s))) {
+	    while (ATLEAST(s, found)
+		   && isdigit(CharOf(s[found])))
+		++found;
 	} else {
 	    found = is_KEYWORD(s);
 	}
@@ -211,7 +301,7 @@ is_INSTANCE(char *s)
     int found = 0;
 
     if (*s == '@') {
-	while ((*s++ == '@') && (s < the_last)) {
+	while ((*s++ == '@') && MORE(s)) {
 	    ;
 	}
 	if ((found = is_KEYWORD(s)) != 0)
@@ -245,13 +335,13 @@ is_CHAR(char *s, int *err)
 {
     int found = 0;
 
-    if (*s == '?' && (s + 5 < the_last)) {
+    if (*s == '?' && ATLEAST(s, 5)) {
 	if ((*++s == BACKSLASH)
 	    && ((*++s == 'M') || (*s == 'C'))
 	    && (s[1] == '-')) {
 	    *err = 0;
 	    found = 5;
-	    if (*s == 'M' && (s + 5 < the_last)
+	    if (*s == 'M' && ATLEAST(s, 5)
 		&& s[2] == BACKSLASH
 		&& s[3] == 'C'
 		&& s[4] == '-') {
@@ -270,6 +360,10 @@ is_CHAR(char *s, int *err)
  *	HEXADECIMAL	0x[0-9a-fA-F_]+
  *	REAL		[-+]?([0-9_]*\.[0-9][0-9_]*)([eE][+-]?[0-9_]+)?
  *	NUMBER		{SIGN}?({DECIMAL}|{OCTAL}|{HEXADECIMAL})[L]?|{REAL}
+ *
+ * But note that ruby allows a function immediately after a number separated
+ * by a ".", e.g.,
+ *	1.upto(n)
  */
 static int
 is_NUMBER(char *s, int *err)
@@ -281,7 +375,7 @@ is_NUMBER(char *s, int *err)
     int dot = 0;
 
     *err = 0;
-    while (s != the_last) {
+    while (MORE(s)) {
 	int ch = CharOf(*s);
 
 	if ((s == base) && (ch == '+' || ch == '-')) {
@@ -297,8 +391,8 @@ is_NUMBER(char *s, int *err)
 		}
 	    }
 	} else if (ch == '.') {
-	    if (the_last - s > 1
-		&& s[1] == '.') {
+	    if (ATLEAST(s, 1)
+		&& (s[1] == '.' || isalpha(CharOf(s[1])))) {
 		break;
 	    }
 	    if (radix == 8)
@@ -313,7 +407,7 @@ is_NUMBER(char *s, int *err)
 	    /* EMPTY */ ;
 	} else if (radix == 0) {
 	    if (ch == '0') {
-		if ((s + 1) == the_last) {
+		if (!ATLEAST(s, 1)) {
 		    radix = 10;
 		} else if (s[1] == 'x') {
 		    radix = 16;
@@ -339,7 +433,7 @@ is_NUMBER(char *s, int *err)
 		value = 1;
 	    } else {
 		if (value) {
-		    while (s != the_last) {
+		    while (MORE(s)) {
 			ch = CharOf(*s);
 			if (isalnum(ch)) {
 			    *err = 1;
@@ -369,9 +463,9 @@ is_COMMENT(char *s)
     char *t = is_BLANK(s) + s;
 
     if (*t++ == '#') {
-	while (t != the_last) {
+	while (MORE(t)) {
 	    if (*t == '\n') {
-		if ((t + 1) == the_last
+		if (!ATLEAST(t, 1)
 		    || t[1] != '#')
 		    break;
 	    }
@@ -395,7 +489,7 @@ end_marker(char *s, char *marker, int only)
 {
     int len = strlen(marker);
 
-    return (the_last - s > len
+    return (ATLEAST(s, len)
 	    && !strncmp(s, marker, len)
 	    && isspace(CharOf(s[len]))
 	    && (!only || (s[len] == '\n')));
@@ -430,7 +524,7 @@ begin_HERE(char *s, int *quoted, int *strip_here)
     int ok;
     int strip = 0;
 
-    if (the_last - s > 3
+    if (ATLEAST(s, 3)
 	&& s[0] == '<'
 	&& s[1] == '<'
 	&& !isBlank(s[2])) {
@@ -466,7 +560,7 @@ skip_BLANKS(char *s)
 {
     char *base = s;
 
-    while (s != the_last) {
+    while (MORE(s)) {
 	if (!isspace(CharOf(*s))) {
 	    break;
 	}
@@ -478,6 +572,86 @@ skip_BLANKS(char *s)
     return s;
 }
 
+/*
+ * Return the number of characters if an operator is found, otherwise zero.
+ */
+static int
+is_OPERATOR(char *s)
+{
+    int found = 0;
+#define OPS(s) { s, sizeof(s) - 1 }
+    static const struct {
+	const char *ops;
+	int len;
+    } table[] = {
+	/* 3 */
+	OPS("&&="),
+	    OPS("**="),
+	    OPS("..."),
+	    OPS("<<="),
+	    OPS("<=>"),
+	    OPS("==="),
+	    OPS(">>="),
+	    OPS("[]="),
+	    OPS("||="),
+	/* 2 */
+	    OPS("!="),
+	    OPS("!~"),
+	    OPS("%="),
+	    OPS("&&"),
+	    OPS("&="),
+	    OPS("**"),
+	    OPS("*="),
+	    OPS("+="),
+	    OPS("-="),
+	    OPS(".."),
+	    OPS("/="),
+	    OPS("::"),
+	    OPS("<<"),
+	    OPS("<="),
+	    OPS("=="),
+	    OPS("=>"),
+	    OPS("=~"),
+	    OPS(">="),
+	    OPS(">>"),
+	    OPS("[]"),
+	    OPS("^="),
+	    OPS("|="),
+	    OPS("||"),
+	/* 1 */
+	    OPS("!"),
+	    OPS("&"),
+	    OPS("("),
+	    OPS(")"),
+	    OPS("*"),
+	    OPS("+"),
+	    OPS(","),
+	    OPS("-"),
+	    OPS(";"),
+	    OPS("="),
+	    OPS("["),
+	    OPS("]"),
+	    OPS("^"),
+	    OPS("{"),
+	    OPS("|"),
+	    OPS("}"),
+	    OPS("~"),
+    };
+
+    if (ispunct(*s)) {
+	unsigned n;
+	for (n = 0; n < TABLESIZE(table); ++n) {
+	    if (ATLEAST(s, table[n].len)
+		&& table[n].ops[0] == *s
+		&& !memcmp(s, table[n].ops, table[n].len)) {
+		found = table[n].len;
+		break;
+	    }
+	}
+    }
+    return found;
+}
+
 /*
  * Do the real work for is_Regexp().  Documentation is vague, but the parser
  * appears to use the same quoting type for each nesting level.
@@ -489,8 +663,9 @@ is_REGEXP(char *s, int left_delim, int right_delim)
     int found = 0;
     int len;
     int level = 0;
+    int range = 0;
 
-    while (s < the_last) {
+    while (MORE(s)) {
 	if (left_delim != right_delim) {
 	    if (*s == left_delim) {
 		++level;
@@ -502,17 +677,25 @@ is_REGEXP(char *s, int left_delim, int right_delim)
 		/* otherwise, fallthru to the other right_delim check */
 	    }
 	}
-	if (s != base && *s == right_delim) {
+	if ((len = is_ESCAPED(s)) != 0) {
+	    s += len;
+	} else if (*s == R_BLOCK && range) {
+	    range = 0;
 	    ++s;
-	    while (s < the_last && isalpha(CharOf(*s))) {
+	} else if (*s == L_BLOCK && !range) {
+	    range = 1;
+	    ++s;
+	} else if ((len = var_embedded(s)) != 0) {
+	    s += len;
+	} else if (range) {
+	    ++s;
+	} else if (s != base && *s == right_delim) {
+	    ++s;
+	    while (MORE(s) && isalpha(CharOf(*s))) {
 		++s;
 	    }
 	    found = s - base;
 	    break;
-	} else if ((len = is_VARIABLE(s)) != 0) {
-	    s += len;
-	} else if ((len = is_ESCAPED(s)) != 0) {
-	    s += len;
 	} else {
 	    ++s;
 	}
@@ -556,7 +739,7 @@ is_Regexp(char *s, int *delim)
     if (*s == '/') {
 	*delim = balanced_delimiter(s);
 	found = is_REGEXP(s, *s, *delim);
-    } else if (s + 4 < the_last
+    } else if (ATLEAST(s, 4)
 	       && s[0] == '%'
 	       && s[1] == 'r'
 	       && !isalnum(CharOf(s[2]))) {
@@ -571,35 +754,55 @@ is_Regexp(char *s, int *delim)
  * Parse the various types of quoted strings.
  */
 static int
-is_String(char *s, int *err)
+is_String(char *s, int *delim, int *err)
 {
     int found = 0;
 
-    if (s + 2 < the_last) {
+    *delim = 0;
+    if (ATLEAST(s, 2)) {
 	switch (*s) {
 	case '%':
-	    if (s + 4 < the_last) {
+	    if (ATLEAST(s, 4)) {
 		switch (s[1]) {
 		case 'q':
+		case 'w':
+		    found = is_STRINGS(s + 2,
+				       err,
+				       s[2],
+				       balanced_delimiter(s + 2),
+				       1);
+		    if (found != 0) {
+			found += 2;
+			*delim = SQUOTE;
+		    }
+		    break;
+		case 'x':
 		case 'Q':
 		    found = is_STRINGS(s + 2,
 				       err,
 				       s[2],
-				       balanced_delimiter(s + 2));
+				       balanced_delimiter(s + 2),
+				       0);
 		    if (found != 0) {
 			found += 2;
+			*delim = DQUOTE;
 		    }
 		    break;
 		}
 	    }
 	    break;
-	case BQUOTE:
 	case SQUOTE:
+	    if ((found = is_STRINGS(s, err, *s, *s, 1)) != 0)
+		*delim = SQUOTE;
+	    break;
+	case BQUOTE:
 	case DQUOTE:
-	    found = is_STRINGS(s, err, *s, *s);
+	    if ((found = is_STRINGS(s, err, *s, *s, 0)) != 0)
+		*delim = DQUOTE;
 	    break;
 	case BACKSLASH:
 	    found = is_ESCAPED(s);
+	    *delim = SQUOTE;
 	    break;
 	}
     }
@@ -611,7 +814,7 @@ line_size(char *s)
 {
     char *base = s;
 
-    while (s != the_last) {
+    while (MORE(s)) {
 	if (*s == '\n')
 	    break;
 	s++;
@@ -622,7 +825,7 @@ line_size(char *s)
 static char *
 put_newline(char *s)
 {
-    if (s != the_last)
+    if (MORE(s))
 	flt_putc(*s++);
     return s;
 }
@@ -635,14 +838,15 @@ static int
 var_embedded(char *s)
 {
     int found = 0;
+    int delim;
 
-    if (*s == '#' && (++s < the_last)) {
+    if (*s == '#' && MORE(++s)) {
 	if (*s == L_CURLY) {
-	    for (found = 1; s + found < the_last; ++found) {
+	    for (found = 1; ATLEAST(s, found); ++found) {
 		/* FIXME: this check for %q{xxx} is too naive... */
-		if (s[found] == '%' && (strchr("qQ", s[found + 1]) != 0)) {
+		if (s[found] == '%' && (strchr("wqQ", s[found + 1]) != 0)) {
 		    int ignore;
-		    found += is_String(s + found, &ignore);
+		    found += is_String(s + found, &delim, &ignore);
 		} else if (s[found] == R_CURLY) {
 		    ++found;
 		    break;
@@ -655,6 +859,11 @@ var_embedded(char *s)
     return found ? found + 1 : 0;
 }
 
+/*
+ * FIXME: var_embedded() can recognize a multi-line embedded variable, but
+ * this function is called for each line.  Probably should redo the output
+ * so this is called at the end of the here-document, etc.
+ */
 static char *
 put_embedded(char *s, int len, char *attr)
 {
@@ -663,7 +872,8 @@ put_embedded(char *s, int len, char *attr)
 
     for (j = k = 0; j < len; j++) {
 	if ((j == 0 || (s[j - 1] != BACKSLASH))
-	    && (id = var_embedded(s + j)) != 0) {
+	    && (id = var_embedded(s + j)) != 0
+	    && (id + j) < len) {
 	    if (j != k)
 		flt_puts(s + k, j - k, attr);
 	    flt_puts(s + j, id, Ident2_attr);
@@ -718,6 +928,15 @@ put_KEYWORD(char *s, int ok, int *had_op)
     return s + ok;
 }
 
+static char *
+put_OPERATOR(char *s, int ok, int *had_op)
+{
+    flt_puts(s, ok, "");
+    if (strchr("(|&=~!,;", *s) != 0)
+	*had_op = 1;
+    return s + ok;
+}
+
 static char *
 put_VARIABLE(char *s, int ok)
 {
@@ -735,29 +954,40 @@ static char *
 put_REGEXP(char *s, int length, int delim)
 {
     char *base = s;
+    char *first = s;
     char *last;
     int len;
+    int range = 0;
 
     if (*s == '%') {
 	flt_puts(s, 2, Keyword_attr);
 	s += 2;
+	first = s;
     }
     flt_bfr_begin(String_attr);
     while (s < base + length) {
-	if (s != base && *s == delim) {
-	    flt_bfr_append(s, 1);
-	    last = ++s;
-	    while (s < the_last && isalpha(CharOf(*s))) {
-		++s;
-	    }
-	    flt_bfr_embed(s, s - last, Keyword_attr);
-	    break;
+	if ((len = is_ESCAPED(s)) != 0) {
+	    flt_bfr_append(s, len);
+	    s += len;
+	} else if (*s == R_BLOCK && range) {
+	    range = 0;
+	    flt_bfr_append(s++, 1);
+	} else if (*s == L_BLOCK && !range) {
+	    range = 1;
+	    flt_bfr_append(s++, 1);
 	} else if ((len = var_embedded(s)) != 0) {
 	    flt_bfr_embed(s, len, Ident2_attr);
 	    s += len;
-	} else if ((len = is_ESCAPED(s)) != 0) {
-	    flt_bfr_append(s, len);
-	    s += len;
+	} else if (range) {
+	    flt_bfr_append(s++, 1);
+	} else if (s != first && *s == delim) {
+	    flt_bfr_append(s, 1);
+	    last = ++s;
+	    while (MORE(s) && isalpha(CharOf(*s))) {
+		++s;
+	    }
+	    flt_bfr_embed(last, s - last, Keyword_attr);
+	    break;
 	} else {
 	    flt_bfr_append(s, 1);
 	    ++s;
@@ -784,6 +1014,8 @@ do_filter(FILE *input GCC_UNUSED)
     unsigned actual = 0;
     size_t request = 0;
     States state = eCODE;
+    TType this_tok = tNULL;
+    TType last_tok = tNULL;
     char *s;
     char *marker = 0;
     int in_line = -1;
@@ -792,7 +1024,7 @@ do_filter(FILE *input GCC_UNUSED)
     int delim;
     int strip_here = 0;		/* strip leading blanks in here-document */
     int quote_here = 0;		/* tokens in here-document are quoted */
-    int had_op = 0;
+    int had_op = 1;		/* true to allow regex */
 
     Comment_attr = class_attr(NAME_COMMENT);
     Error_attr = class_attr(NAME_ERROR);
@@ -826,17 +1058,79 @@ do_filter(FILE *input GCC_UNUSED)
 	the_last = the_file + the_size;
 
 	s = the_file;
-	while (s != the_last) {
+	while (MORE(s)) {
 	    if (*s == '\n') {
 		in_line = -1;
+		if (state == eCODE)
+		    had_op = 1;
 	    } else {
 		in_line++;
 	    }
 	    DPRINTF(("(%s(%.*s) line:%d op:%d)\n",
 		     stateName(state), is_KEYWORD(s) + 1, s, in_line, had_op));
 	    switch (state) {
+		/*
+		 * alias method-name method-name
+		 * alias global-variable-name global-variable-name
+		 */
+	    case eALIAS:
+	    case eDEF:
+		if ((ok = is_COMMENT(s)) != 0) {
+		    Parsed(this_tok, tCOMMENT);
+		    s = put_COMMENT(s, ok);
+		} else if ((ok = is_BLANK(s)) != 0) {
+		    Parsed(this_tok, tBLANK);
+		    flt_puts(s, ok, "");
+		    s += ok;
+		} else if ((ok = is_KEYWORD(s)) != 0) {
+		    Parsed(this_tok, tKEYWORD);
+		    s = put_KEYWORD(s, ok, &had_op);
+		    state = eCODE;
+		} else if ((ok = is_OPERATOR(s)) != 0) {
+		    Parsed(this_tok, tOPERATOR);
+		    s = put_OPERATOR(s, ok, &had_op);
+		    state = eCODE;
+		} else {
+		    flt_putc(*s++);
+		    state = eCODE;
+		}
+		had_op = 1;	/* kludge - in case a "def" precedes regex */
+		break;
+		/*
+		 * Class definitions use '<' specially, like a reverse arrow:
+		 *      class ClassName < SuperClass
+		 *      class << Object
+		 *
+		 * The first case is not a syntax problem, but the
+		 * singleton-class definition conflicts with here-documents.
+		 */
+	    case eCLASS:
+		if ((ok = is_COMMENT(s)) != 0) {
+		    Parsed(this_tok, tCOMMENT);
+		    s = put_COMMENT(s, ok);
+		} else if ((ok = is_BLANK(s)) != 0) {
+		    Parsed(this_tok, tBLANK);
+		    flt_puts(s, ok, "");
+		    s += ok;
+		} else if ((ok = is_KEYWORD(s)) != 0) {
+		    Parsed(this_tok, tKEYWORD);
+		    s = put_KEYWORD(s, ok, &had_op);
+		    state = eCODE;
+		} else if ((ok = is_OPERATOR(s)) != 0) {
+		    Parsed(this_tok, tOPERATOR);
+		    s = put_OPERATOR(s, ok, &had_op);
+		    state = eCODE;
+		} else {
+		    flt_putc(*s++);
+		    state = eCODE;
+		}
+		had_op = 1;	/* kludge - in case a "class" precedes regex */
+		break;
 	    case eCODE:
-		if ((marker = begin_HERE(s, &quote_here, &strip_here)) != 0) {
+		if ((last_tok == tKEYWORD || last_tok == tOPERATOR)
+		    && (marker = begin_HERE(s, &quote_here, &strip_here)) != 0) {
+		    ok = 1;	/* FIXME: should allow other tokens on line */
+		    Parsed(this_tok, tHERE);
 		    state = eHERE;
 		    s = put_remainder(s, String_attr, quote_here);
 		} else if ((in_line < 0 && begin_POD(s + 1))
@@ -846,43 +1140,81 @@ do_filter(FILE *input GCC_UNUSED)
 		    flt_putc(*s++);	/* write the newline */
 		    s = put_remainder(s, Comment_attr, 1);
 		} else if ((ok = is_COMMENT(s)) != 0) {
-		    DPRINTF(("...COMMENT: %d\n", ok));
+		    Parsed(this_tok, tCOMMENT);
 		    s = put_COMMENT(s, ok);
 		} else if ((ok = is_BLANK(s)) != 0) {
-		    DPRINTF(("...BLANK: %d\n", ok));
+		    Parsed(this_tok, tBLANK);
 		    flt_puts(s, ok, "");
 		    s += ok;
 		} else if (had_op && (ok = is_Regexp(s, &delim)) != 0) {
-		    DPRINTF(("...REGEXP: %d\n", ok));
+		    Parsed(this_tok, tREGEXP);
 		    s = put_REGEXP(s, ok, delim);
 		} else if ((ok = is_CHAR(s, &err)) != 0) {
-		    DPRINTF(("...CHAR: %d\n", ok));
+		    Parsed(this_tok, tCHAR);
 		    had_op = 0;
-		    flt_puts(s, ok, err ? Error_attr : Number_attr);
+		    if (err) {
+			flt_error("not a number");
+			flt_puts(s, ok, Error_attr);
+		    } else {
+			flt_puts(s, ok, Number_attr);
+		    }
 		    s += ok;
 		} else if ((ok = is_NUMBER(s, &err)) != 0) {
-		    DPRINTF(("...NUMBER: %d\n", ok));
+		    Parsed(this_tok, tNUMBER);
 		    had_op = 0;
-		    flt_puts(s, ok, err ? Error_attr : Number_attr);
+		    if (err) {
+			flt_error("not a number");
+			flt_puts(s, ok, Error_attr);
+		    } else {
+			flt_puts(s, ok, Number_attr);
+		    }
 		    s += ok;
 		} else if ((ok = is_KEYWORD(s)) != 0) {
-		    DPRINTF(("...KEYWORD: %d\n", ok));
+		    Parsed(this_tok, tKEYWORD);
+		    if (ok == 5 && !strncmp(s, "alias", ok))
+			state = eALIAS;
+		    else if (ok == 5 && !strncmp(s, "class", ok))
+			state = eCLASS;
+		    else if (ok == 3 && !strncmp(s, "def", ok))
+			state = eDEF;
 		    s = put_KEYWORD(s, ok, &had_op);
 		} else if ((ok = is_VARIABLE(s)) != 0) {
-		    DPRINTF(("...VARIABLE: %d\n", ok));
+		    Parsed(this_tok, tVARIABLE);
 		    s = put_VARIABLE(s, ok);
-		} else if ((ok = is_String(s, &err)) != 0) {
 		    had_op = 0;
-		    if (*s == DQUOTE) {
-			s = put_embedded(s, ok, err ? Error_attr : String_attr);
+		} else if (ATLEAST(s, (ok = 2)) && !strncmp(s, "?\"", ok)) {
+		    Parsed(this_tok, tVARIABLE);
+		    s = put_VARIABLE(s, ok);	/* csv.rb uses it, undocumented */
+		    had_op = 0;
+		} else if ((ok = is_String(s, &delim, &err)) != 0) {
+		    Parsed(this_tok, tSTRING);
+		    had_op = 0;
+		    if (*s == '%') {
+			flt_puts(s, 2, Keyword_attr);
+			s += 2;
+			ok -= 2;
+		    }
+		    if (delim == DQUOTE) {
+			if (err) {
+			    flt_error("unexpected quote");
+			    s = put_embedded(s, ok, Error_attr);
+			} else {
+			    s = put_embedded(s, ok, String_attr);
+			}
 		    } else {
-			flt_puts(s, ok, err ? Error_attr : String_attr);
+			if (err) {
+			    flt_error("unterminated string");
+			    flt_puts(s, ok, Error_attr);
+			} else {
+			    flt_puts(s, ok, String_attr);
+			}
 			s += ok;
 		    }
+		} else if ((ok = is_OPERATOR(s)) != 0) {
+		    Parsed(this_tok, tOPERATOR);
+		    s = put_OPERATOR(s, ok, &had_op);
 		} else {
-		    if (strchr("()|&=~!,;", *s) != 0)
-			had_op = 1;
-		    else if (!isspace(CharOf(*s)))
+		    if (!isspace(CharOf(*s)))
 			had_op = 0;
 		    flt_putc(*s++);
 		}
@@ -901,6 +1233,29 @@ do_filter(FILE *input GCC_UNUSED)
 		s = put_remainder(s, Comment_attr, 1);
 		break;
 	    }
+
+	    switch (this_tok) {
+	    case tNULL:
+	    case tBLANK:
+	    case tCOMMENT:
+		continue;
+	    case tCHAR:
+	    case tHERE:
+	    case tNUMBER:
+	    case tREGEXP:
+	    case tSTRING:
+		last_tok = this_tok;
+		break;
+	    case tKEYWORD:
+		last_tok = this_tok;
+		break;
+	    case tOPERATOR:
+		last_tok = this_tok;
+		break;
+	    case tVARIABLE:
+		last_tok = this_tok;
+		break;
+	    }
 	}
 	free(the_file);
     }
