@@ -3,9 +3,9 @@
  * written for vile by paul fox.
  * rewritten to use regular expressions by T.Dickey
  *
- * Copyright (c) 1990-2003 by Paul Fox and Thomas Dickey
+ * Copyright (c) 1990-2005 by Paul Fox and Thomas Dickey
  *
- * $Header: /usr/build/vile/vile/RCS/finderr.c,v 1.118 2003/07/27 17:07:07 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/finderr.c,v 1.130 2005/05/27 22:24:25 tom Exp $
  *
  */
 
@@ -28,6 +28,8 @@ typedef struct {
     char *exp_text;
     regexp *exp_comp;
     int words[W_LAST];
+    int lBase;
+    int cBase;
 } ERR_PATTERN;
 
 static LINE *getdot(BUFFER *bp);
@@ -36,6 +38,9 @@ static void putdotback(BUFFER *bp, LINE *dotp);
 static char febuff[NBUFN];	/* name of buffer to find errors in */
 static int newfebuff = TRUE;	/* is the name new since last time? */
 
+/*
+ * These variables are set as a side-effect of decode_exp().
+ */
 static TBUFF *fe_verb;
 static TBUFF *fe_file;
 static TBUFF *fe_text;
@@ -44,6 +49,17 @@ static int lL_base;
 static long fe_colm;
 static long fe_line;
 
+/*
+ * Beginning of line.
+ */
+#define P_BOL	"^"
+
+/*
+ * Text from running "ant", which puts the program name before each message,
+ * in square brackets.
+ */
+#define P_ANT	P_BOL "\\(\\s\\+\\[[^]]\\+\\]\\s\\+\\)\\?"
+
 /*
  * This is the list of predefined regular expressions for the error
  * finder.  The user can substitute a new list at runtime by loading
@@ -78,9 +94,9 @@ static long fe_line;
 static const
 char *const predefined[] =
 {
-    "^\"%[^\" \t]\", line %L:%T",	/* various C compilers */
-    "^%F:\\s*%L:%C:\\s*%T",	/* antic */
-    "^%F:\\s*%L:\\s*%T",	/* "grep -n" */
+    P_ANT "\"%[^\" \t]\", line %L:%T",	/* various C compilers */
+    P_ANT "%F:\\s*%L:%C:\\s*%T",	/* antic */
+    P_ANT "%F:\\s*%L:\\s*%T",	/* "grep -n" */
 
 #if SYS_VMS
     "[ \t]*At line number %L in %[^;].*",	/* crude support for DEC C
@@ -130,8 +146,10 @@ char *const predefined[] =
 #endif
 
     /* Borland C++ */
+#if CC_NEWDOSCC
     "^Error\\( [^ ]\\+\\)\\? %F %L: %T",
     "^Warning\\( [^ ]\\+\\)\\? %F %L: %T",
+#endif
 
     "^%B:%L:%T",		/* "pp" in scratch buf */
     "^[^:]\\+: %V directory `%[^']'",	/* GNU make */
@@ -214,7 +232,10 @@ marks_in(const char *expr)
 		 else dst = lsprintf(dst, "%s", S)
 #define APP_C    if (pass != 1) *dst++ = *src
 
-static void
+/*
+ * Convert the pattern; returns false if malloc's fail.
+ */
+static int
 convert_pattern(ERR_PATTERN * errp, LINE *lp)
 {
     static const char before[] = "\\(";
@@ -229,12 +250,13 @@ convert_pattern(ERR_PATTERN * errp, LINE *lp)
     int word;
     int mark;
     int range;
+    int status = TRUE;
     size_t want = llength(lp);
     char *first = lp->l_text;
     char *last = first + want;
 
     (void) memset(errp, 0, sizeof(*errp));
-    TPRINTF(("error-pattern %*S\n", want, first));
+    TPRINTF(("error-pattern %.*s\n", (int) want, first));
 
     /* In the first pass, find the number of fields we'll substitute.
      * Then allocate a new string that's a genuine regular expression
@@ -255,11 +277,13 @@ convert_pattern(ERR_PATTERN * errp, LINE *lp)
 		    mark = W_VERB;
 		    break;
 		case 'F':
-		    APP_S(before);
-		    APP_T(tb_values(filename_expr));
-		    APP_S(after);
-		    errp->words[W_FILE] = ++word;
-		    word += marks_in(tb_values(filename_expr));
+		    if (tb_values(filename_expr) != 0) {
+			APP_S(before);
+			APP_T(tb_values(filename_expr));
+			APP_S(after);
+			errp->words[W_FILE] = ++word;
+			word += marks_in(tb_values(filename_expr));
+		    }
 		    break;
 		case 'B':
 		    APP_S("\\(\\[[^:]\\+]\\)");
@@ -271,13 +295,13 @@ convert_pattern(ERR_PATTERN * errp, LINE *lp)
 		    break;
 		case 'c':
 		case 'C':
-		    cC_base = ((*src) == 'C');
+		    errp->cBase = ((*src) == 'C');
 		    APP_S(number);
 		    errp->words[W_COLM] = ++word;
 		    break;
 		case 'l':
 		case 'L':
-		    lL_base = ((*src) == 'L');
+		    errp->lBase = ((*src) == 'L');
 		    APP_S(number);
 		    errp->words[W_LINE] = ++word;
 		    break;
@@ -333,8 +357,10 @@ convert_pattern(ERR_PATTERN * errp, LINE *lp)
 	    beginDisplay();
 	    dst = temp = typeallocn(char, want + 1);
 	    endofDisplay();
-	    if (dst == 0)
+	    if (dst == 0) {
+		status = no_memory("convert_pattern");
 		break;
+	    }
 	} else {
 	    *dst = EOS;
 	}
@@ -344,13 +370,18 @@ convert_pattern(ERR_PATTERN * errp, LINE *lp)
 	TRACE(("COMPILE %s\n", temp));
 	for (word = 0; word < W_LAST; word++)
 	    TRACE(("word[%d] = %d (%s)\n", word, errp->words[word],
-		   get_token_name(word)));
+		   get_token_name((ErrTokens) word)));
 #endif
 	TPRINTF(("-> %s\n", temp));
 	exp = regcomp(temp, strlen(temp), TRUE);
+	/* FIXME:  this might be null if the pattern was incorrect, or if we
+	 * ran out of memory.  We only want the latter condition.
+	 */
     }
     errp->exp_text = temp;
     errp->exp_comp = exp;
+
+    return status;
 }
 
 /*
@@ -393,6 +424,7 @@ load_patterns(void)
     BUFFER *bp;
     LINE *lp;
     size_t n;
+    int status = TRUE;
 
     /* find the error-expressions buffer */
     if ((bp = find_b_name(ERRORS_BufName)) == 0) {
@@ -420,15 +452,23 @@ load_patterns(void)
 	beginDisplay();
 	exp_count = bp->b_linecount;
 	exp_table = typeallocn(ERR_PATTERN, exp_count);
-	for (n = 0; n < W_LAST; n++)
-	    exp_table->words[n] = -1;
 	endofDisplay();
 
-	n = 0;
-	for_each_line(lp, bp)
-	    convert_pattern(&exp_table[n++], lp);
+	if (exp_table != 0) {
+	    for (n = 0; n < W_LAST; n++)
+		exp_table->words[n] = -1;
+	    n = 0;
+	    for_each_line(lp, bp) {
+		if (!convert_pattern(&exp_table[n++], lp)) {
+		    status = FALSE;
+		    break;
+		}
+	    }
+	} else {
+	    status = no_memory("load_patterns");
+	}
     }
-    return TRUE;
+    return status;
 }
 
 /*
@@ -480,6 +520,8 @@ decode_exp(ERR_PATTERN * exp)
 	    *(lookup[j].number) = 0;
 	}
     }
+    cC_base = exp->cBase;
+    lL_base = exp->lBase;
     fe_colm = cC_base;
 
     /*
@@ -546,6 +588,58 @@ decode_exp(ERR_PATTERN * exp)
     return failed;
 }
 
+#define set_tabstop_val(bp,value) set_b_val(bp, VAL_TAB, value)
+
+static void
+goto_column(void)
+{
+    int saved_tabstop = tabstop_val(curbp);
+
+    if (error_tabstop > 0)
+	set_tabstop_val(curbp, error_tabstop);
+
+    gocol(fe_colm ? fe_colm - cC_base : 0);
+
+    if (error_tabstop > 0)
+	set_tabstop_val(curbp, saved_tabstop);
+}
+
+#define DIRLEVELS 20
+static int dir_level = 0;
+static char *dir_stack[DIRLEVELS];
+
+static void
+freeDirs(void)
+{
+    beginDisplay();
+    while (dir_level) {
+	free(dir_stack[dir_level]);
+	--dir_level;
+    }
+    endofDisplay();
+}
+
+static void
+updateDirs(const char *errverb, const char *errfile)
+{
+    if (!strcmp("Entering", errverb)) {
+	if (dir_level < DIRLEVELS) {
+	    beginDisplay();
+	    ++dir_level;
+	    dir_stack[dir_level] = strmalloc(errfile);
+	    endofDisplay();
+	}
+    } else if (!strcmp("Leaving", errverb)) {
+	if (dir_level > 0) {
+	    beginDisplay();
+	    if (dir_stack[dir_level] != 0)
+		free(dir_stack[dir_level]);
+	    --dir_level;
+	    endofDisplay();
+	}
+    }
+}
+
 /* Edits the file and goes to the line pointed at by the next compiler error in
  * the "[output]" window.  It unfortunately doesn't mark the lines for you, so
  * adding lines to the file throws off the later numbering.  Solutions to this
@@ -575,10 +669,6 @@ finderr(int f GCC_UNUSED, int n GCC_UNUSED)
 
     static LINE *odotp = 0;
 
-#define DIRLEVELS 20
-    static int l = 0;
-    static char *dirs[DIRLEVELS];
-
     if (!comp_err_exps(FALSE, 1))
 	return (FALSE);
 
@@ -592,8 +682,7 @@ finderr(int f GCC_UNUSED, int n GCC_UNUSED)
 	oerrline = -1;
 	oerrfile = tb_init(&oerrfile, EOS);
 	oerrtext = tb_init(&oerrtext, EOS);
-	while (l)
-	    free(dirs[l--]);
+	freeDirs();
 	odotp = 0;
 	endofDisplay();
     }
@@ -603,10 +692,7 @@ finderr(int f GCC_UNUSED, int n GCC_UNUSED)
 
 	LINE *tdotp;
 
-	beginDisplay();
-	while (l)
-	    free(dirs[l--]);
-	endofDisplay();
+	freeDirs();
 
 	tdotp = lforw(buf_head(sbp));
 
@@ -635,16 +721,7 @@ finderr(int f GCC_UNUSED, int n GCC_UNUSED)
 		    errfile = tb_values(fe_file);
 
 		    if (errverb != 0 && errfile != 0) {
-			beginDisplay();
-			if (!strcmp("Entering", errverb)) {
-			    if (l < DIRLEVELS) {
-				dirs[++l] = strmalloc(errfile);
-			    }
-			} else if (!strcmp("Leaving", errverb)) {
-			    if (l > 0)
-				free(dirs[l--]);
-			}
-			endofDisplay();
+			updateDirs(errverb, errfile);
 		    }
 		} else if (interrupted()) {
 		    kbd_alarm();
@@ -685,18 +762,8 @@ finderr(int f GCC_UNUSED, int n GCC_UNUSED)
 			&& errtext != 0
 			&& strcmp(tb_values(oerrtext), errtext))
 			break;
-		} else if (errverb != 0
-			   && errfile != 0) {
-		    beginDisplay();
-		    if (!strcmp("Entering", errverb)) {
-			if (l < DIRLEVELS) {
-			    dirs[++l] = strmalloc(errfile);
-			}
-		    } else if (!strcmp("Leaving", errverb)) {
-			if (l > 0)
-			    free(dirs[l--]);
-		    }
-		    endofDisplay();
+		} else if (errverb != 0 && errfile != 0) {
+		    updateDirs(errverb, errfile);
 		}
 	    }
 	}
@@ -706,11 +773,8 @@ finderr(int f GCC_UNUSED, int n GCC_UNUSED)
 	} else if (lforw(dotp) == buf_head(sbp)) {
 	    mlwarn("[No more errors in %s buffer]", febuff);
 	    /* start over at the top of file */
-	    beginDisplay();
 	    putdotback(sbp, lforw(buf_head(sbp)));
-	    while (l)
-		free(dirs[l--]);
-	    endofDisplay();
+	    freeDirs();
 	    return FALSE;
 	}
 	dotp = lforw(dotp);
@@ -723,7 +787,7 @@ finderr(int f GCC_UNUSED, int n GCC_UNUSED)
 
     odotp = dotp;
 
-    (void) lengthen_path(pathcat(ferrfile, dirs[l], errfile));
+    (void) lengthen_path(pathcat(ferrfile, dir_stack[dir_level], errfile));
 
     if (strcmp(ferrfile, curbp->b_fname)) {
 	/* if we must change windows */
@@ -746,7 +810,7 @@ finderr(int f GCC_UNUSED, int n GCC_UNUSED)
 	mlforce("%s", errtext);
 	len = strlen(errtext);
     } else {
-	mlforce("Error: %*S", dotp->l_used, dotp->l_text);
+	mlforce("Error: %.*s", dotp->l_used, dotp->l_text);
 	errtext = dotp->l_text;
 	len = dotp->l_used;
     }
@@ -756,11 +820,14 @@ finderr(int f GCC_UNUSED, int n GCC_UNUSED)
     }
     /* it's an absolute move */
     curwp->w_lastdot = DOT;
-    if ((fe_line + lL_base) >= curbp->b_lines_on_disk)
+    if (fe_line <= lL_base) {
+	status = gotobob(f, n);
+    } else if ((fe_line - lL_base) >= curbp->b_lines_on_disk) {
 	status = gotoeob(f, n);
-    else
+    } else {
 	status = gotoline(TRUE, -(curbp->b_lines_on_disk - fe_line + lL_base));
-    gocol(fe_colm ? fe_colm - cC_base : 0);
+    }
+    goto_column();
 
     oerrline = fe_line;
     (void) tb_scopy(&oerrfile, errfile);
@@ -847,10 +914,12 @@ make_err_regex_list(int dum1 GCC_UNUSED, void *ptr GCC_UNUSED)
     BUFFER *bp;
     LINE *lp;
 
-    bprintf("--- Error Meta-Expressions and Resulting Regular Expressions %*P",
-	    term.cols - 1, '-');
+    bprintf("--- Error Meta-Expressions and Resulting Regular Expressions ");
+    bpadc('-', term.cols - DOT.o);
+
     if (exp_table == 0)
 	load_patterns();
+
     if (exp_table != 0
 	&& (bp = find_b_name(ERRORS_BufName)) != 0) {
 	size_t j = 0;
@@ -859,13 +928,13 @@ make_err_regex_list(int dum1 GCC_UNUSED, void *ptr GCC_UNUSED)
 	    int k, first = TRUE;
 	    if (j >= exp_count)
 		break;
-	    bprintf("\n%7d ", j);
-	    bprintf("%*S\n", llength(lp), lp->l_text);
-	    bprintf("%*S%s", ERR_PREFIX, " ", exp_table[j].exp_text);
+	    bprintf("\n%7d ", (int) j);
+	    bprintf("%.*s\n", llength(lp), lp->l_text);
+	    bprintf("%.*s%s", ERR_PREFIX, " ", exp_table[j].exp_text);
 	    for (k = 0; k < W_LAST; k++) {
 		if (exp_table[j].words[k] != 0) {
 		    if (first) {
-			bprintf("\n%*S", ERR_PREFIX, " ");
+			bprintf("\n%.*s", ERR_PREFIX, " ");
 			first = FALSE;
 		    } else {
 			bprintf(", ");
