@@ -13,7 +13,7 @@
  *
  *	modify (ifdef-style) 'expand_leaf()' to allow ellipsis.
  *
- * $Header: /usr/build/vile/vile/RCS/glob.c,v 1.78 2002/11/05 20:39:42 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/glob.c,v 1.87 2005/06/03 00:03:29 tom Exp $
  *
  */
 
@@ -94,19 +94,28 @@ static size_t myLen = 0;	/* length and index of the expanded list */
 static char **myVec = 0;	/* the expanded list */
 
 /*--------------------------------------------------------------------------*/
-int
-string_has_wildcards(const char *item)
+static void
+strip_escapes(char *buffer)
+{
+    int j, k, ch;
+
+    for (j = k = 0; (ch = buffer[j]) != EOS; ++j) {
+	if (ch == BACKSLASH)
+	    ++j;
+	if ((buffer[k++] = buffer[j]) == EOS)
+	    break;
+    }
+    buffer[k] = EOS;
+}
+
+/*
+ * Use this function to decide if we should perform wildcard expansion.
+ */
+static int
+string_has_globs(const char *item)
 {
-#if OPT_VMS_PATH || SYS_WINNT	/* either host can support ~/whatever */
-    if (*item == '~')
-	return TRUE;
-#endif
 #if OPT_VMS_PATH || SYS_UNIX || OPT_MSDOS_PATH
     while (*item != EOS) {
-#if UNIX_GLOBBING
-	if (iswild(*item))
-	    return TRUE;
-#endif
 	if (*item == GLOB_SINGLE || *item == GLOB_MULTI)
 	    return TRUE;
 #if OPT_GLOB_ELLIPSIS || SYS_VMS
@@ -118,15 +127,44 @@ string_has_wildcards(const char *item)
 	if (*item == GLOB_RANGE[0])
 	    return TRUE;
 #endif
+#endif
+	item++;
+    }
+#endif
+    return FALSE;
+}
+
+/*
+ * Use this function to decide if we should do wildcard/tilde/variable
+ * expansion.
+ */
+int
+string_has_wildcards(const char *item)
+{
+    const char *base = item;
+
+#if OPT_VMS_PATH || SYS_WINNT	/* either host can support ~/whatever */
+    if (*item == CH_TILDE)
+	return TRUE;
+#endif
+#if OPT_VMS_PATH || SYS_UNIX || OPT_MSDOS_PATH
+    while (*item != EOS) {
+#if UNIX_GLOBBING
+	if (iswild(*item))
+	    return TRUE;
+#endif
+#if !OPT_VMS_PATH
 #if OPT_GLOB_ENVIRON
-	if (*item == '$' && (isname(item[1]) || isdelim(item[1])))
+	if (*item == '$'
+	    && (item == base || !isEscaped(item))
+	    && (isname(item[1]) || isdelim(item[1])))
 	    return TRUE;
 #endif
 #endif
 	item++;
     }
 #endif
-    return FALSE;
+    return string_has_globs(base);
 }
 
 /*
@@ -170,7 +208,7 @@ cs_char(int ch)
     return isUpper(ch) ? toLower(ch) : ch;
 }
 #else
-#define cs_char(ch) (ch)
+#define cs_char(ch) CharOf(ch)
 #endif
 
 static int
@@ -227,15 +265,14 @@ glob_match_leaf(char *leaf, char *pattern)
 		    break;
 		}
 		if (*pattern == '-' && pattern != first) {
-		    int lo = pattern[-1];
-		    int hi = pattern[1];
+		    int lo = CharOf(pattern[-1]);
+		    int hi = CharOf(pattern[1]);
 		    if (hi == GLOB_RANGE[1])
-			hi = '~';
+			hi = '~';	/* last-printable ASCII */
 		    if (((cs_char(lo) <= cs_char(*leaf))
 			 && (cs_char(*leaf) <= cs_char(hi))) != negate)
 			found = TRUE;
-		    if (pattern[1] != GLOB_RANGE[1])
-			pattern++;
+		    pattern++;
 		} else if ((cs_char(*pattern++) == cs_char(*leaf)) != negate)
 		    found = TRUE;
 	    }
@@ -287,7 +324,7 @@ wild_leaf(char *pattern)
 	for (k = j + 1; (c = pattern[k]) != EOS; k++) {
 	    if (is_slashc(c)) {
 		pattern[k] = EOS;
-		ok = string_has_wildcards(pattern + j);
+		ok = string_has_globs(pattern + j);
 		pattern[k] = c;
 		if (ok)
 		    return pattern + j;
@@ -302,7 +339,7 @@ wild_leaf(char *pattern)
 	else
 	    j++;		/* leaf is empty */
     }
-    return string_has_wildcards(pattern + j) ? pattern + j : 0;
+    return string_has_globs(pattern + j) ? pattern + j : 0;
 }
 
 #if !SYS_OS2
@@ -397,7 +434,7 @@ expand_leaf(char *path,		/* built-up pathname, top-level */
 	    if (!glob_match_leaf(leaf, wild))
 		continue;
 	    if (next != 0) {	/* there are more leaves */
-		if (!string_has_wildcards(next)) {
+		if (!string_has_globs(next)) {
 		    s = skip_string(leaf);
 		    *s++ = SLASHC;
 		    (void) strcpy(s, next);
@@ -507,7 +544,7 @@ expand_leaf(char *path,		/* built-up pathname, top-level */
 		continue;
 
 	    if (next != 0) {	/* there are more leaves */
-		if (!string_has_wildcards(next)) {
+		if (!string_has_globs(next)) {
 		    s = skip_string(leaf);
 		    *s++ = SLASHC;
 		    (void) strcpy(s, next);
@@ -650,6 +687,10 @@ glob_from_pipe(const char *pattern)
  *	$NAME
  *	$(NAME)
  *	${NAME}
+ *
+ * but ignores
+ *	NAME$
+ *	\$NAME
  */
 static void
 expand_environ(char *pattern)
@@ -660,15 +701,19 @@ expand_environ(char *pattern)
     char save[NFILEN];
 
     for (j = 0; pattern[j] != EOS; j++) {
-	if (pattern[j] == '$') {
+	if (j != 0 && isEscaped(pattern + j))
+	    continue;
 
-	    k = j + 1;
+	k = j + 1;
+	if (pattern[j] == '$' && pattern[k] != EOS) {
 	    if (pattern[k] == '(')
 		delim = ')';
 	    else if (pattern[k] == '{')
 		delim = '}';
-	    else
+	    else if (isAlnum(pattern[k]))
 		delim = EOS;
+	    else
+		continue;
 
 	    if (delim != EOS)
 		k++;
@@ -804,13 +849,15 @@ expand_pattern(char *item)
 #endif
 #endif
 	expand_environ(pattern);
-	if (string_has_wildcards(pattern)) {
+	if (string_has_globs(pattern)) {
+	    strip_escapes(pattern);
 	    if ((result = expand_leaf(builtup, pattern)) != FALSE
 		&& (myLen - first > 1)) {
 		qsort((char *) &myVec[first], myLen - first,
 		      sizeof(*myVec), compar);
 	    }
 	} else {
+	    strip_escapes(pattern);
 	    result = record_a_match(pattern);
 	}
     }
@@ -970,10 +1017,19 @@ expand_wild_args(int *argcp, char ***argvp)
     }
     if (comma || option) {
 	char **newvec = typeallocn(char *, comma + *argcp + 1);
+
+	if (newvec == 0) {
+	    no_memory("expand_wild_args");
+	    return;
+	}
+
 	for (j = k = 0; j < *argcp; j++) {
 	    char *the_arg = strmalloc((*argvp)[j]);
 	    char *item, *tmp;
 
+	    if (the_arg == 0)
+		break;
+
 	    /*
 	     * strip off supported VMS options, else don't muck
 	     * with pseudo-unix path delimiters so that the
@@ -1016,7 +1072,16 @@ expand_wild_args(int *argcp, char ***argvp)
     }
 #endif
     if (glob_needed(*argvp)) {
-	char **newargs = glob_expand(*argvp);
+	char **newargs;
+	int n;
+
+	for (n = 0; n < *argcp; ++n) {
+	    char *test = add_backslashes2((*argvp)[n], "\\$");
+	    if (test != (*argvp)[n]) {
+		(*argvp)[n] = strmalloc(test);
+	    }
+	}
+	newargs = glob_expand(*argvp);
 	if (newargs != 0) {
 	    *argvp = newargs;
 	    *argcp = glob_length(newargs);
