@@ -55,7 +55,7 @@
  *	not (yet) correspond to :-commands.  Before implementing, probably will
  *	have to make TESTC a settable mode.
  *
- * $Header: /usr/build/vile/vile/RCS/history.c,v 1.76 2003/07/27 16:55:39 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/history.c,v 1.82 2005/01/20 01:58:52 tom Exp $
  *
  */
 
@@ -137,6 +137,7 @@ static int
 willExtend(const char *src, int srclen)
 {
     if ((tb_length(MyText) == 0)
+	&& src != 0
 	&& (srclen > 0)) {
 	return (src[0] == SHPIPE_LEFT[0]) || isRepeatable(src[0]);
     }
@@ -151,7 +152,7 @@ willExtend(const char *src, int srclen)
 static int
 sameLine(LINE *lp, char *src, int srclen)
 {
-    if (srclen <= 0)
+    if (lp == 0 || !lisreal(lp) || src == 0 || srclen <= 0)
 	return 0;
     else {
 	int dstlen = llength(lp);
@@ -228,11 +229,11 @@ parseArg(HST * parm, LINE *lp)
  * parsing.
  */
 static int
-needQuotes(TBUFF * src)
+needQuotes(TBUFF *src)
 {
 #if HST_QUOTES
     if (tb_length(MyText)
-	|| must_quote_token(tb_values(src), tb_length(src)))
+	&& must_quote_token(tb_values(src), tb_length(src)))
 	return TRUE;
 #endif
     return FALSE;
@@ -251,7 +252,7 @@ stripQuotes(char *src, int len, int eolchar, int *actual)
 
     if (len > 0) {
 	TRACE(("stripQuotes(%.*s)\n", len, src));
-	if ((dst = (char *) malloc((size_t) len + 1)) != 0) {
+	if ((dst = typeallocn(char, (size_t) len + 1)) != 0) {
 	    int j, k;
 	    int quoted = FALSE;
 	    int escaped = FALSE;
@@ -297,10 +298,10 @@ stripQuotes(char *src, int len, int eolchar, int *actual)
 #endif
 
 static void
-glueBufferToResult(TBUFF ** dst, TBUFF * src)
+glueBufferToResult(TBUFF **dst, TBUFF *src)
 {
     int shell_cmd = ((tb_length(*dst) != 0 && isShellOrPipe(tb_values(*dst)))
-		     || (tb_length(*dst) != 0 && isShellOrPipe(tb_values(src))));
+		     || (tb_length(src) != 0 && isShellOrPipe(tb_values(src))));
 
     if (willGlue())
 	(void) tb_append(dst, MyGlue);
@@ -343,7 +344,7 @@ hst_glue(int c)
 }
 
 void
-hst_append(TBUFF * cmd, int glue)
+hst_append(TBUFF *cmd, int glue, int can_extend)
 {
     static int skip = 1;	/* e.g., after "!" */
 
@@ -361,7 +362,7 @@ hst_append(TBUFF * cmd, int glue)
 	   tb_visible(cmd),
 	   glue));
     TRACE(("...MyText        :%d:%s\n", tb_length(MyText), tb_visible(MyText)));
-    if (willExtend(tb_args(cmd))
+    if (can_extend && willExtend(tb_args(cmd))
 	&& tb_length(cmd) > (size_t) skip) {
 	kbd_pushback(cmd, skip);
     }
@@ -372,17 +373,17 @@ hst_append(TBUFF * cmd, int glue)
 }
 
 void
-hst_append_s(char *cmd, int glue)
+hst_append_s(char *cmd, int glue, int can_extend)
 {
     TBUFF *p = tb_string(cmd);
-    hst_append(p, glue);
+    hst_append(p, glue, can_extend);
     tb_free(&p);
 }
 
 void
 hst_remove(const char *cmd)
 {
-    if (MyLevel == 1 && *cmd != EOS) {
+    if (MyLevel == 1 && cmd != 0 && *cmd != EOS) {
 	TBUFF *temp = 0;
 	unsigned len = tb_length(tb_scopy(&temp, cmd)) - 1;
 
@@ -498,12 +499,6 @@ hst_find(HST * parm, BUFFER *bp, LINE *lp, int direction)
 		continue;
 	}
 
-	/* avoid picking up lines with range-spec, since this is too
-	 * cumbersome to splice in 'namedcmd()'.
-	 */
-	if (islinespecchar(lp->l_text[0]))
-	    continue;
-
 	/* '/' and '?' are not (yet) :-commands.  Don't display them
 	 * in the command-name scrolling.
 	 */
@@ -588,7 +583,7 @@ display_LINE(HST * parm, LINE *lp)
  * Update the display using a TBUFF as source
  */
 static void
-display_TBUFF(HST * parm, TBUFF * tp)
+display_TBUFF(HST * parm, TBUFF *tp)
 {
     hst_display(parm, tb_args(tp));
 }
@@ -635,13 +630,12 @@ hst_scroll(LINE *lp1, HST * parm)
  * escape-character is entered.
  */
 int
-edithistory(
-	       TBUFF ** buffer,
-	       unsigned *position,
-	       int *given,
-	       UINT options,
-	       int (*endfunc) (EOL_ARGS),
-	       int eolchar)
+edithistory(TBUFF **buffer,
+	    unsigned *position,
+	    int *given,
+	    UINT options,
+	    int (*endfunc) (EOL_ARGS),
+	    int eolchar)
 {
     HST param;
     BUFFER *bp;
