@@ -7,7 +7,7 @@
  * Most code probably by Dan Lawrence or Dave Conroy for MicroEMACS
  * Extensions for vile by Paul Fox
  *
- * $Header: /usr/build/vile/vile/RCS/insert.c,v 1.141 2003/06/18 21:38:27 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/insert.c,v 1.143 2005/07/13 23:42:48 tom Exp $
  *
  */
 
@@ -650,7 +650,7 @@ ins(void)
 }
 
 static int
-isallspace(LINEPTR ln, int lb, int ub)
+isallspace(LINE *ln, int lb, int ub)
 {
     while (lb <= ub) {
 	if (!isSpace(lgetc(ln, ub)))
@@ -1171,7 +1171,8 @@ inspound(void)
 	if (autoindented > llength(DOT.l))
 	    autoindented = llength(DOT.l);
 	DOT.o = w_left_margin(curwp);
-	(void) ldelete((B_COUNT) autoindented, FALSE);
+	if (autoindented > 0)
+	    (void) ldelete((B_COUNT) autoindented, FALSE);
     }
     autoindented = -1;
 
