@@ -17,12 +17,12 @@
  * distributable status.  This version of vile is distributed under the
  * terms of the GNU Public License (see COPYING).
  *
- * Copyright (c) 1992-2003 by Paul Fox and Thomas Dickey
+ * Copyright (c) 1992-2005 by Paul Fox and Thomas Dickey
  *
  */
 
 /*
- * $Header: /usr/build/vile/vile/RCS/main.c,v 1.510 2003/07/27 15:18:54 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/main.c,v 1.540 2005/07/11 22:25:32 tom Exp $
  */
 
 #define realdef			/* Make global definitions not external */
@@ -33,12 +33,15 @@
 #include	"nefunc.h"
 #include	"nefsms.h"
 
-#include	<assert.h>
 #include	<ctype.h>
 
 #if OPT_LOCALE
 #include	<locale.h>
 
+#ifdef HAVE_LANGINFO_CODESET
+#include	<langinfo.h>
+#endif
+
 #ifdef HAVE_WCTYPE
 #include	<wctype.h>
 #define sys_iscntrl(n)  iswcntrl(n)
@@ -152,19 +155,30 @@ MainProgram(int argc, char *argv[])
 	char *env = "";
 
 	/*
-	 * Force 8-bit locale for display drivers where we only support 8-bits
+	 * Force 8-bit locale for display drivers where we only support 8-bits.
+	 * This is a special case, assumes that the name of the 8-bit locale
+	 * can be found by stripping the "UTF-8" string, and also that both
+	 * locales are installed.
 	 */
-#if DISP_TERMCAP
+#define nonnull(s) ((s) ? (s) : "<null>")
+#if DISP_TERMCAP || DISP_X11
 	if (((env = getenv("LC_ALL")) != 0 && *env != 0) ||
 	    ((env = getenv("LC_CTYPE")) != 0 && *env != 0) ||
 	    ((env = getenv("LANG")) != 0 && *env != 0)) {
-	    char *utf = strstr(env, ".UTF-8");
+	    char *utf;
+	    char *tmp;
 
-	    if (utf != 0) {
-		char *tmp = strmalloc(env);
+	    if (((utf = strstr(env, ".UTF-8")) != 0
+		 || (utf = strstr(env, ".utf-8")) != 0
+		 || (utf = strstr(env, ".UTF8")) != 0
+		 || (utf = strstr(env, ".utf8")) != 0)
+		&& (tmp = strmalloc(env)) != 0) {
 		tmp[utf - env] = EOS;
-		env = tmp;
 		utf8_locale = TRUE;
+#if DISP_TERMCAP && OPT_ICONV_FUNCS
+		tcap_setup_locale(env, tmp);
+#endif
+		env = tmp;
 	    } else {
 		env = "";
 	    }
@@ -172,15 +186,45 @@ MainProgram(int argc, char *argv[])
 	    env = "";
 	}
 #endif
-	setlocale(LC_ALL, env);	/* set locale according to environment vars */
+	/* set locale according to environment vars */
+	vl_locale = setlocale(LC_ALL, env);
+
+#ifdef HAVE_LANGINFO_CODESET
+	/*
+	 * Check the encoding and try to decide if it looks like an 8-bit
+	 * encoding.  If it does not, fallback to POSIX locale and its
+	 * corresponding encoding.
+	 */
+	vl_encoding = nl_langinfo(CODESET);
+	if (vl_locale == 0
+	    || vl_encoding == 0
+	    || (strstr(vl_encoding, "ASCII") == 0
+		&& strstr(vl_encoding, "ANSI") == 0
+		&& strncmp(vl_encoding, "ISO-8859", 8) != 0
+		&& strncmp(vl_encoding, "ISO 8859", 8) != 0
+		&& strncmp(vl_encoding, "ISO_8859", 8) != 0
+		&& strncmp(vl_encoding, "ISO8859", 8) != 0
+		&& strncmp(vl_encoding, "8859", 4) != 0)) {
+	    vl_locale = setlocale(LC_ALL, "C");
+	    vl_encoding = nl_langinfo(CODESET);
+	}
+#endif
+	if (vl_locale == 0)
+	    vl_locale = "built-in";
+	if (vl_encoding == 0)
+	    vl_encoding = "8bit";	/* meaningless, but we need a value */
+
+	/* make our own copy of the strings */
+	vl_locale = strmalloc(vl_locale);
+	vl_encoding = strmalloc(vl_encoding);
     }
 #endif /* OPT_LOCALE */
 
 #if OPT_NAMEBST
     build_namebst(nametbl, 0, nametblsize - 1);
 #endif
-    global_val_init();		/* global buffer values */
     charinit();			/* character types -- we need these early  */
+    global_val_init();		/* global buffer values */
     winit(FALSE);		/* command-line */
 #if !SYS_UNIX
     expand_wild_args(&argc, &argv);
@@ -188,7 +232,7 @@ MainProgram(int argc, char *argv[])
     prog_arg = argv[0];		/* this contains our only clue to exec-path */
 #if SYS_MSDOS || SYS_OS2 || SYS_OS2_EMX || SYS_WINNT
     if (strchr(pathleaf(prog_arg), '.') == 0) {
-	char *t = malloc(strlen(prog_arg) + 5);
+	char *t = typeallocn(char, strlen(prog_arg) + 5);
 	lsprintf(t, "%s.exe", prog_arg);
 	prog_arg = t;
     }
@@ -530,7 +574,7 @@ MainProgram(int argc, char *argv[])
 
     /* this comes out to 70 on an 80 (or greater) column display */
     {
-	register int fill;
+	int fill;
 	fill = (7 * term.cols) / 8;	/* must be done after vtinit() */
 	if (fill > 70)
 	    fill = 70;
@@ -571,7 +615,7 @@ MainProgram(int argc, char *argv[])
     } else {
 
 	/* else vileinit is the contents of their VILEINIT variable */
-	vileinit = getenv("VILEINIT");
+	vileinit = vile_getenv("VILEINIT");
 	if (vileinit != NULL) {	/* set... */
 	    BUFFER *vbp, *obp;
 	    UINT oflags = 0;
@@ -714,7 +758,7 @@ MainProgram(int argc, char *argv[])
 
 #if OPT_POPUP_MSGS
     if (global_g_val(GMDPOPUP_MSGS) && (startstat != TRUE)) {
-	bp = bfind(MESSAGES_BufName, BFSCRTCH);
+	bp = bfind(MESSAGES_BufName, BFINVS);
 	bsizes(bp);
 	TRACE(("Checking size of popup messages: %d\n", bp->b_linecount));
 	if (bp->b_linecount > 1) {
@@ -741,7 +785,7 @@ MainProgram(int argc, char *argv[])
      * initialized.  Give it one last chance.
      */
     if ((startstat != TRUE) && tb_length(mlsave))
-	mlforce("%*S", tb_length(mlsave), tb_values(mlsave));
+	mlforce("%.*s", (int) tb_length(mlsave), tb_values(mlsave));
 
     /* process commands */
     main_loop();
@@ -940,15 +984,17 @@ tidy_exit(int code)
 
 #ifndef strmalloc
 char *
-strmalloc(const char *s)
+strmalloc(const char *src)
 {
-    char *ns;
-    beginDisplay();
-    ns = castalloc(char, strlen(s) + 1);
-    if (ns != 0)
-	(void) strcpy(ns, s);
-    endofDisplay();
-    return ns;
+    char *dst = 0;
+    if (src != 0) {
+	beginDisplay();
+	dst = castalloc(char, strlen(src) + 1);
+	if (dst != 0)
+	    (void) strcpy(dst, src);
+	endofDisplay();
+    }
+    return dst;
 }
 #endif
 
@@ -1014,6 +1060,8 @@ no_memory(const char *s)
 #define DFT_CSUFFIX "\\.\\(\\([CchisS]\\)\\|CC\\|cc\\|cpp\\|cxx\\|hxx\\|scm\\)$"
 #endif
 
+#define DFT_STRIPSUFFIX "\\(\\.orig\\|~\\)$"
+
 #define DFT_FENCE_CHARS "{}()[]"
 #define DFT_CINDENT_CHARS ":#" DFT_FENCE_CHARS
 
@@ -1087,6 +1135,9 @@ init_mode_value(struct VAL *d, MODECLASS v_class, int v_which)
 #ifdef GMDPOPUP_MSGS
 	    setINT(GMDPOPUP_MSGS, -TRUE);	/* popup-msgs */
 #endif
+#ifdef GVAL_READER_POLICY
+	    setINT(GVAL_READER_POLICY, RP_BOTH);
+#endif
 #ifdef GMDRESOLVE_LINKS
 	    setINT(GMDRESOLVE_LINKS, FALSE);	/* set noresolve-links by default in case we've got NFS problems */
 #endif
@@ -1129,6 +1180,9 @@ init_mode_value(struct VAL *d, MODECLASS v_class, int v_which)
 #ifdef GVAL_FINDCFG
 	    setTXT(GVAL_FINDCFG, "");
 #endif
+#ifdef GVAL_FOR_BUFFERS
+	    setINT(GVAL_FOR_BUFFERS, FB_MIXED);
+#endif
 #ifdef GVAL_GLOB
 	    setTXT(GVAL_GLOB, "!echo %s");
 #endif
@@ -1145,6 +1199,12 @@ init_mode_value(struct VAL *d, MODECLASS v_class, int v_which)
 	    setTXT(GVAL_REDIRECT_KEYS,
 		   "F5::S,F10::S,F11::S,F7::F,F5:C:,F9::Y");
 #endif
+#ifdef GVAL_RECENT_FILES
+	    setINT(GVAL_RECENT_FILES, 0);
+#endif
+#ifdef GVAL_RECENT_FOLDERS
+	    setINT(GVAL_RECENT_FOLDERS, 0);
+#endif
 #ifdef GVAL_SCROLLPAUSE
 	    setINT(GVAL_SCROLLPAUSE, 0);
 #endif
@@ -1210,6 +1270,9 @@ init_mode_value(struct VAL *d, MODECLASS v_class, int v_which)
 #ifdef MDCRYPT
 	    setINT(MDCRYPT, FALSE);	/* crypt */
 #endif
+#ifdef MDFILTERMSGS
+	    setINT(MDFILTERMSGS, FALSE);	/* syntax-filter messages */
+#endif
 #ifdef MDHILITE
 	    setINT(MDHILITE, TRUE);	/* syntax-highlighting */
 #endif
@@ -1225,12 +1288,18 @@ init_mode_value(struct VAL *d, MODECLASS v_class, int v_which)
 #ifdef VAL_AUTOCOLOR
 	    setINT(VAL_AUTOCOLOR, 0);	/* auto syntax coloring timeout */
 #endif
-#ifdef VAL_HILITEMATCH
-	    setINT(VAL_HILITEMATCH, 0);		/* no hilite */
+#ifdef VAL_C_SWIDTH
+	    setINT(VAL_C_SWIDTH, 8);	/* C file shiftwidth */
+#endif
+#ifdef VAL_C_TAB
+	    setINT(VAL_C_TAB, 8);	/* C file tab stop */
 #endif
 #ifdef VAL_FENCE_LIMIT
 	    setINT(VAL_FENCE_LIMIT, 10);	/* fences iteration timeout */
 #endif
+#ifdef VAL_HILITEMATCH
+	    setINT(VAL_HILITEMATCH, 0);		/* no hilite */
+#endif
 #ifdef VAL_LOCKER
 	    setTXT(VAL_LOCKER, "");	/* Name locker */
 #endif
@@ -1240,11 +1309,8 @@ init_mode_value(struct VAL *d, MODECLASS v_class, int v_which)
 #ifdef VAL_RECORD_LENGTH
 	    setINT(VAL_RECORD_LENGTH, 0);
 #endif
-#ifdef VAL_C_SWIDTH
-	    setINT(VAL_C_SWIDTH, 8);	/* C file shiftwidth */
-#endif
-#ifdef VAL_C_TAB
-	    setINT(VAL_C_TAB, 8);	/* C file tab stop */
+#ifdef VAL_STRIPSUFFIX
+	    setPAT(VAL_STRIPSUFFIX, DFT_STRIPSUFFIX);
 #endif
 	default:
 	    setIntValue(d, 0);
@@ -1297,7 +1363,7 @@ init_mode_value(struct VAL *d, MODECLASS v_class, int v_which)
 
 #define DFT_HELP_FILE "vile.hlp"
 
-#define DFT_MENU_FILE "vilemenu.rc"
+#define DFT_MENU_FILE ".vilemenu"
 
 #ifdef VILE_LIBDIR_PATH
 #define DFT_LIBDIR_PATH VILE_LIBDIR_PATH
@@ -1312,7 +1378,7 @@ init_mode_value(struct VAL *d, MODECLASS v_class, int v_which)
 #endif
 
 #define DFT_MLFORMAT \
-"%-%i%- %b %m:: :%f:is : :%=%F: : :%l:(:,:%c::) :%p::% :%S%-%-%|"
+"%-%i%- %b %m:: :%f:is : :%=%F: : :%l:(:,:%c::) :%p::% :%C:char ::%S%-%-%|"
 
 #define DFT_POSFORMAT \
 "Line %{$curline}d of %{$blines}d,\
@@ -1324,7 +1390,7 @@ static char *
 default_help_file(void)
 {
     char *result;
-    if ((result = getenv("VILE_HELP_FILE")) == 0)
+    if ((result = vile_getenv("VILE_HELP_FILE")) == 0)
 	result = DFT_HELP_FILE;
     return result;
 }
@@ -1333,7 +1399,7 @@ static char *
 default_libdir_path(void)
 {
     char *result;
-    if ((result = getenv("VILE_LIBDIR_PATH")) == 0)
+    if ((result = vile_getenv("VILE_LIBDIR_PATH")) == 0)
 	result = DFT_LIBDIR_PATH;
     return result;
 }
@@ -1344,13 +1410,13 @@ default_menu_file(void)
 {
     static char default_menu[] = DFT_MENU_FILE;
     char temp[NSTRING];
-    char *menurc = getenv("VILE_MENU");
+    char *menurc = vile_getenv("VILE_MENU");
 
-    if (menurc == NULL || *menurc == EOS) {
+    if (isEmpty(menurc)) {
 	sprintf(temp, "%.*s_MENU", (int) (sizeof(temp) - 6), prognam);
 	mkupper(temp);
 	menurc = getenv(temp);
-	if (menurc == NULL || *menurc == EOS) {
+	if (isEmpty(menurc)) {
 	    menurc = default_menu;
 	}
     }
@@ -1363,7 +1429,7 @@ static char *
 default_startup_file(void)
 {
     char *result;
-    if ((result = getenv("VILE_STARTUP_FILE")) == 0)
+    if ((result = vile_getenv("VILE_STARTUP_FILE")) == 0)
 	result = DFT_STARTUP_FILE;
     return result;
 }
@@ -1374,7 +1440,7 @@ default_startup_path(void)
     char *result = 0;
     char *s;
 
-    if ((s = getenv("VILE_STARTUP_PATH")) != 0) {
+    if ((s = vile_getenv("VILE_STARTUP_PATH")) != 0) {
 	append_to_path_list(&result, s);
     } else {
 #if SYS_MSDOS || SYS_OS2 || SYS_WINNT
@@ -1544,7 +1610,7 @@ global_val_init(void)
     set_submode_val("c", MDCINDENT, TRUE);	/* C-style indent */
     set_submode_txt("c", VAL_CINDENT_CHARS, DFT_CINDENT_CHARS);		/* C-style indent flags */
 
-    set_majormode_rexp("c", MVAL_SUFFIXES, DFT_CSUFFIX);
+    set_majormode_rexp("c", MVAL_MODE_SUFFIXES, DFT_CSUFFIX);
 #endif
 
 #if OPT_EVAL
@@ -1773,9 +1839,9 @@ siguninit(void)
 static void
 do_num_proc(int *cp, int *fp, int *np)
 {
-    register int c, f, n;
-    register int mflag;
-    register int oldn;
+    int c, f, n;
+    int mflag;
+    int oldn;
 
     c = *cp;
 
@@ -1821,9 +1887,9 @@ do_num_proc(int *cp, int *fp, int *np)
 static void
 do_rept_arg_proc(int *cp, int *fp, int *np)
 {
-    register int c, f, n;
-    register int mflag;
-    register int oldn;
+    int c, f, n;
+    int mflag;
+    int oldn;
     c = *cp;
 
     if (c != reptc)
@@ -1942,7 +2008,7 @@ zzquit(int f, int n)
 int
 quickexit(int f, int n)
 {
-    register int status;
+    int status;
     if ((status = writeall(f, n, FALSE, TRUE, FALSE, FALSE)) == TRUE)
 	status = quithard(f, n);	/* conditionally quit */
     return status;
@@ -2039,6 +2105,10 @@ quit(int f, int n GCC_UNUSED)
 	FreeAndNull(startup_path);
 	FreeAndNull(gregexp);
 	tb_free(&tb_matched_pat);
+#if OPT_LOCALE
+	FreeAndNull(vl_locale);
+	FreeAndNull(vl_encoding);
+#endif /* OPT_LOCALE */
 #if OPT_MLFORMAT
 	FreeAndNull(modeline_format);
 #endif
@@ -2047,9 +2117,9 @@ quit(int f, int n GCC_UNUSED)
 #endif
 	FreeAndNull(helpfile);
 	FreeAndNull(startup_file);
-
 #if SYS_UNIX
-	if (strcmp(exec_pathname, "."))
+	if (exec_pathname != 0
+	    && strcmp(exec_pathname, "."))
 	    FreeAndNull(exec_pathname);
 #endif
 #if OPT_FILTER
@@ -2211,7 +2281,7 @@ reptc_func(int f GCC_UNUSED, int n GCC_UNUSED)
 void
 charinit(void)
 {
-    register int c;
+    int c;
 
     TRACE((T_CALLED "charinit() lo=%d, hi=%d\n",
 	   global_g_val(GVAL_PRINT_LOW),
@@ -2220,6 +2290,10 @@ charinit(void)
     /* If we're using the locale functions, set our flags based on its
      * tables.  Note that just because you have 'setlocale()' doesn't mean
      * that the tables are present or correct.  But this is a start.
+     *
+     * NOTE:  Solaris8 and some versions of M$ incorrectly classify tab as a
+     * printable character (ANSI C says control characters are not printable).
+     * Ignore that (the former fixes it in Solaris9).
      */
 #if OPT_LOCALE
     for (c = 0; c < N_chars; c++) {
@@ -2230,7 +2304,7 @@ charinit(void)
 	    vlCTYPE(c) |= vl_digit;
 	if (sys_islower(c))
 	    vlCTYPE(c) |= vl_lower;
-	if (sys_isprint(c))
+	if (sys_isprint(c) && c != '\t')
 	    vlCTYPE(c) |= vl_print;
 	if (sys_ispunct(c))
 	    vlCTYPE(c) |= vl_punct;
@@ -2242,8 +2316,8 @@ charinit(void)
 	if (sys_isxdigit(c))
 	    vlCTYPE(c) |= vl_xdigit;
 #endif
-	vl_uppercase[c] = toupper(c);
-	vl_lowercase[c] = tolower(c);
+	vl_uppercase[c] = (char) toupper(c);
+	vl_lowercase[c] = (char) tolower(c);
     }
 #else /* ! OPT_LOCALE */
     (void) memset((char *) vl_chartypes_, 0, sizeof(vl_chartypes_));
@@ -2519,13 +2593,12 @@ track_realloc(char *p, unsigned size)
 
 /* track mallocs */
 char *
-track_malloc(
-		unsigned size)
+track_malloc(unsigned size)
 {
     char *p;
 
     size += sizeof(HEAPSIZE);
-    if ((p = malloc(size)) != 0) {
+    if ((p = typeallocn(char, size)) != 0) {
 	(void) memset(p, 0, size);	/* so we can use for calloc */
 	p = new_ramsize(p, size);
 	display_heap_usage();
@@ -2623,9 +2696,10 @@ showmemory(int f, int n)
 char *
 strncpy0(char *t, const char *f, size_t l)
 {
-    (void) strncpy(t, f, l);
-    if (l)
+    if (t != 0 && f != 0 && l != 0) {
+	(void) strncpy(t, f, l);
 	t[l - 1] = EOS;
+    }
     return t;
 }
 
@@ -2637,12 +2711,59 @@ strncpy0(char *t, const char *f, size_t l)
 char *
 vl_strncpy(char *dest, const char *src, size_t destlen)
 {
-    size_t srclen = strlen(src) + 1;
+    size_t srclen = (src != 0) ? (strlen(src) + 1) : 0;
     if (srclen > destlen)
 	srclen = destlen;
     return strncpy0(dest, src, srclen);
 }
 
+/*
+ * Wrapper for environment variables that tell vile where to find its
+ * components.
+ */
+char *
+vile_getenv(const char *name)
+{
+    char *result = getenv(name);
+#if defined(_WIN32)
+    if (result == 0) {
+	static HKEY rootkeys[] =
+	{HKEY_CURRENT_USER, HKEY_LOCAL_MACHINE};
+
+	int j;
+	HKEY hkey;
+	DWORD dwSzBuffer;
+	char buffer[256];
+
+	for (j = 0; j < TABLESIZE(rootkeys); ++j) {
+	    if (RegOpenKeyEx(rootkeys[j],
+			     VILE_SUBKEY "\\Environment",
+			     0,
+			     KEY_READ,
+			     &hkey) == ERROR_SUCCESS) {
+		dwSzBuffer = sizeof(buffer);
+		if (RegQueryValueEx(hkey,
+				    name,
+				    NULL,
+				    NULL,
+				    (LPBYTE) buffer,
+				    &dwSzBuffer) == ERROR_SUCCESS
+		    && dwSzBuffer != 0) {
+
+		    buffer[dwSzBuffer - 1] = 0;
+		    result = strmalloc(buffer);
+		    (void) RegCloseKey(hkey);
+		    break;
+		}
+
+		(void) RegCloseKey(hkey);
+	    }
+	}
+    }
+#endif
+    return result;
+}
+
 #if defined(SA_RESTART)
 /* several systems (SCO, SunOS) have sigaction without SA_RESTART */
 /*
@@ -2741,7 +2862,8 @@ make_startup_file(char *name)
     FILE *fp;
     char temp[NFILEN];
 
-    if (strcmp(pathcat(temp, home_dir(), startup_file), startup_file)
+    if (startup_file != 0
+	&& strcmp(pathcat(temp, home_dir(), startup_file), startup_file)
 	&& ((fp = fopen(temp, "w")) != 0)) {
 	fprintf(fp, "; generated by %s -I\n", prog_arg);
 	fprintf(fp, "source %s\n", name);
@@ -2801,6 +2923,45 @@ valid_window(WINDOW *test)
 }
 #endif
 
+#ifndef valid_line_bp
+int
+valid_line_bp(LINE *test, BUFFER *bp)
+{
+    beginDisplay();
+    if (test != NULL) {
+	int valid = FALSE;
+	LINE *lp;
+
+	if (test == buf_head(bp)) {
+	    valid = TRUE;
+	} else {
+	    for_each_line(lp, bp) {
+		if (lp == test) {
+		    valid = TRUE;
+		    break;
+		}
+	    }
+	}
+	if (!valid)
+	    imdying(SIGTERM);
+    }
+    endofDisplay();
+
+    return (test != NULL);
+}
+#endif
+
+#ifndef valid_line_wp
+int
+valid_line_wp(LINE *test, WINDOW *wp)
+{
+    return (test != NULL)
+	&& valid_window(wp)
+	&& valid_buffer(wp->w_bufp)
+	&& valid_line_bp(test, wp->w_bufp);
+}
+#endif
+
 #ifdef VILE_ERROR_ABORT
 #undef ExitProgram		/* in case it is oleauto_exit() */
 
@@ -2809,7 +2970,7 @@ ExitProgram(int code)
 {
     char *env;
     if (code != GOODEXIT
-	&& (env = getenv("VILE_ERROR_ABORT")) != 0
+	&& (env = vile_getenv("VILE_ERROR_ABORT")) != 0
 	&& *env != '\0') {
 	siginit(FALSE);
 	abort();
