@@ -5,7 +5,7 @@
  *	included in main.c
  *
  *	Copyright (c) 1990 by Paul Fox
- *	Copyright (c) 1995-2003 by Paul Fox and Thomas Dickey
+ *	Copyright (c) 1995-2004 by Paul Fox and Thomas Dickey
  *
  *	See the file "cmdtbl" for input data formats, and "estruct.h" for
  *	the output structures.
@@ -15,7 +15,7 @@
  * by Tom Dickey, 1993.    -pgf
  *
  *
- * $Header: /usr/build/vile/vile/RCS/mktbls.c,v 1.130 2003/05/27 00:54:48 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/mktbls.c,v 1.132 2004/12/03 00:01:22 tom Exp $
  *
  */
 
@@ -911,7 +911,7 @@ dump_all_modes(void)
 	"#endif /* realdef */",
 	"",
 	"#ifdef realdef",
-	"EXTERN_CONST char *const all_modes[] = {",
+	"DECL_EXTERN_CONST(char *const all_modes[]) = {",
     };
     static const char *const bottom[] =
     {
@@ -1135,7 +1135,7 @@ dump_mmodes(void)
 	"} M_VALUES;",
 	"",
 	"#ifdef realdef",
-	"EXTERN_CONST struct VALNAMES m_valnames[MAX_M_VALUES+1] = {",
+	"DECL_EXTERN_CONST(struct VALNAMES m_valnames[MAX_M_VALUES+1]) = {",
     };
     static const char *const bottom[] =
     {
@@ -1186,7 +1186,7 @@ dump_all_submodes(void)
 	"",
 	"#if OPT_MAJORMODE",
 	"#ifdef realdef",
-	"EXTERN_CONST char *const all_submodes[] = {",
+	"DECL_EXTERN_CONST(char *const all_submodes[]) = {",
     };
     static const char *const bottom[] =
     {
@@ -1281,7 +1281,7 @@ dump_qmodes(void)
 	"} Q_VALUES;",
 	"",
 	"#ifdef realdef",
-	"EXTERN_CONST struct VALNAMES q_valnames[MAX_Q_VALUES+1] = {",
+	"DECL_EXTERN_CONST(struct VALNAMES q_valnames[MAX_Q_VALUES+1]) = {",
     };
     static const char *const bottom[] =
     {
@@ -1330,7 +1330,7 @@ dump_bmodes(void)
 	"} B_VALUES;",
 	"",
 	"#ifdef realdef",
-	"EXTERN_CONST struct VALNAMES b_valnames[] = {",
+	"DECL_EXTERN_CONST(struct VALNAMES b_valnames[]) = {",
     };
     static const char *const bottom[] =
     {
@@ -1394,7 +1394,7 @@ init_statevars(void)
 	"/*\tlist of recognized state variables\t*/",
 	"",
 	"#ifdef realdef",
-	"EXTERN_CONST char *const statevars[] = {"
+	"DECL_EXTERN_CONST(char *const statevars[]) = {"
     };
     static int done;
 
@@ -1441,7 +1441,7 @@ dump_statevars(void)
 	"",
 	"",
 	"#ifdef realdef",
-	"StateFunc *statevar_func[] = {",
+	"DECL_EXTERN(StateFunc *statevar_func[]) = {",
 	""
     };
     static const char *const tail[] =
@@ -1515,7 +1515,7 @@ init_fsms(void)
     Fprintf(nefsms, "#ifndef realdef\n");
     Fprintf(nefsms, "extern const FSM_CHOICES fsm_%s_choices[];\n", name);
     Fprintf(nefsms, "#else\n");
-    Fprintf(nefsms, "EXTERN_CONST FSM_CHOICES fsm_%s_choices[] = %c\n",
+    Fprintf(nefsms, "DECL_EXTERN_CONST(FSM_CHOICES fsm_%s_choices[]) = %c\n",
 	    name, L_CURL);
 }
 
@@ -1623,7 +1623,10 @@ save_funcs(
     Sprintf(temp, "extern int %s ( int f, int n );", func);
     InsertOnEnd(&all_funcs, temp);
 
-    s = append(strcpy(temp, "\tEXTERN_CONST CMDFUNC f_"), func);
+    s = strcpy(temp, "\t");
+    s = append(s, "DECL_EXTERN_CONST(CMDFUNC f_");
+    s = append(s, func);
+    s = append(s, ")");
     (void) PadTo(32, temp);
     s = append(s, "= {\n\tINIT_UNION(");
     s = append(s, func);
@@ -1681,7 +1684,7 @@ dump_gmodes(void)
 	"} G_VALUES;",
 	"",
 	"#ifdef realdef",
-	"EXTERN_CONST struct VALNAMES g_valnames[] = {",
+	"DECL_EXTERN_CONST(struct VALNAMES g_valnames[]) = {",
     };
     static const char *const bottom[] =
     {
@@ -1745,7 +1748,7 @@ init_ufuncs(void)
 	"",
 	"typedef struct UFUNC {",
 	"\tconst char *f_name;\t/* name of function */",
-	"\tUFuncCode f_code;",
+	"\tunsigned f_code;",
 	"} UFUNC;",
 	"",
 	"#define NARGMASK	0x000f",
@@ -1757,7 +1760,7 @@ init_ufuncs(void)
 	"#define SRET		0x0400",
 	"",
 	"#ifdef realdef",
-	"EXTERN_CONST UFUNC vl_ufuncs[] = {",
+	"DECL_EXTERN_CONST(UFUNC vl_ufuncs[]) = {",
     };
     static int done;
     LIST *p;
@@ -1802,7 +1805,7 @@ dump_ufuncs(void)
 	    init_ufuncs();
 	Sprintf(temp, "\t{\"%s\",", p->Name);
 	(void) PadTo(15, temp);
-	Sprintf(temp + strlen(temp), "(UFuncCode)(%s)},", p->Data);
+	Sprintf(temp + strlen(temp), "(unsigned)(%s)},", p->Data);
 	if (p->Note[0]) {
 	    (void) PadTo(32, temp);
 	    Sprintf(temp + strlen(temp), "/* %s */", p->Note);
@@ -1844,7 +1847,7 @@ dump_wmodes(void)
 	"} W_VALUES;",
 	"",
 	"#ifdef realdef",
-	"EXTERN_CONST struct VALNAMES w_valnames[] = {",
+	"DECL_EXTERN_CONST(struct VALNAMES w_valnames[]) = {",
     };
     static const char *bottom[] =
     {
