@@ -1,7 +1,7 @@
 /*	npopen:  like popen, but grabs stderr, too
  *		written by John Hutchinson, heavily modified by Paul Fox
  *
- * $Header: /usr/build/vile/vile/RCS/npopen.c,v 1.89 2002/11/02 16:41:25 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/npopen.c,v 1.92 2005/01/23 00:59:40 tom Exp $
  *
  */
 
@@ -69,7 +69,7 @@ npopen(char *cmd, const char *type)
 
 #if SYS_UNIX
 int
-inout_popen(FILE ** fr, FILE ** fw, char *cmd)
+inout_popen(FILE **fr, FILE **fw, char *cmd)
 {
     int rp[2];
     int wp[2];
@@ -145,7 +145,7 @@ inout_popen(FILE ** fr, FILE ** fw, char *cmd)
 }
 
 void
-npclose(FILE * fp)
+npclose(FILE *fp)
 {
     int child;
     (void) fflush(fp);
@@ -185,7 +185,8 @@ exec_sh_c(char *cmd)
     for (i = 3; i < NOFILE; i++)
 	(void) close(i);
 
-    if ((sh = user_SHELL()) == NULL || *sh == EOS) {
+    sh = user_SHELL();
+    if (isEmpty(sh)) {
 	sh = bin_sh;
 	shname = pathleaf(sh);
     } else {
@@ -242,7 +243,7 @@ system_SHELL(char *cmd)
 {
     int cpid;
 
-    TRACE(("system_SHELL(%s)\n", cmd != 0 ? cmd : ""));
+    TRACE(("system_SHELL(%s)\n", NONNULL(cmd)));
 
     cpid = softfork();
     if (cpid < 0) {
@@ -348,7 +349,7 @@ deleteTemp(void)
 }
 
 static void
-closePipe(FILE *** pp)
+closePipe(FILE ***pp)
 {
     if (*pp != 0) {
 	if (**pp != 0) {
@@ -443,7 +444,7 @@ npopen(char *cmd, const char *type)
  * on SYS_MSDOS, we don't need both at the same instant.
  */
 int
-inout_popen(FILE ** fr, FILE ** fw, char *cmd)
+inout_popen(FILE **fr, FILE **fw, char *cmd)
 {
     char *type = (fw != 0) ? "w" : "r";
     static FILE *pp[2] =
@@ -503,7 +504,7 @@ npflush(void)
 }
 
 void
-npclose(FILE * fp)
+npclose(FILE *fp)
 {
 #if SYS_MSDOS
     if (myWrtr != 0 && myPipe == 0)
