@@ -1,12 +1,13 @@
 /*
  * Uses the Win32 console API.
  *
- * $Header: /usr/build/vile/vile/RCS/ntconio.c,v 1.76 2002/10/07 23:33:57 cmorgan Exp $
+ * $Header: /usr/build/vile/vile/RCS/ntconio.c,v 1.82 2005/01/20 11:58:44 tom Exp $
  *
  */
 
 #include "estruct.h"
 #include "edef.h"
+#include "chgdfunc.h"
 
 #include <process.h>
 
@@ -132,7 +133,8 @@ nticursor(int cmode)
 static void
 nttitle(const char *title)
 {				/* set the current window title */
-    SetConsoleTitle(title);
+    if (title != 0)
+	SetConsoleTitle(title);
 }
 #endif
 
@@ -297,25 +299,30 @@ ntscroll(int from, int to, int n)
 #endif
 }
 
-#if	OPT_FLASH
+#if OPT_FLASH
 static void
 flash_display(void)
 {
     DWORD length = term.cols * term.rows;
     DWORD got;
-    WORD *buf1 = malloc(sizeof(*buf1) * length);
-    WORD *buf2 = malloc(sizeof(*buf2) * length);
+    WORD *buf1;
+    WORD *buf2;
+
     static COORD origin;
-    ReadConsoleOutputAttribute(hConsoleOutput, buf1, length, origin, &got);
-    ReadConsoleOutputAttribute(hConsoleOutput, buf2, length, origin, &got);
-    for (got = 0; got < length; got++) {
-	buf2[got] ^= (FOREGROUND_INTENSITY | BACKGROUND_INTENSITY);
+
+    if ((buf1 = typeallocn(WORD, length)) != 0
+	&& (buf2 = typeallocn(WORD, length)) != 0) {
+	ReadConsoleOutputAttribute(hConsoleOutput, buf1, length, origin, &got);
+	ReadConsoleOutputAttribute(hConsoleOutput, buf2, length, origin, &got);
+	for (got = 0; got < length; got++) {
+	    buf2[got] ^= (FOREGROUND_INTENSITY | BACKGROUND_INTENSITY);
+	}
+	WriteConsoleOutputAttribute(hConsoleOutput, buf2, length, origin, &got);
+	Sleep(200);
+	WriteConsoleOutputAttribute(hConsoleOutput, buf1, length, origin, &got);
+	free(buf1);
+	free(buf2);
     }
-    WriteConsoleOutputAttribute(hConsoleOutput, buf2, length, origin, &got);
-    Sleep(200);
-    WriteConsoleOutputAttribute(hConsoleOutput, buf1, length, origin, &got);
-    free(buf1);
-    free(buf2);
 }
 #endif
 
@@ -579,6 +586,27 @@ ntkclose(void)
 #endif
 }
 
+#define isModified(state) (state & \
+		(LEFT_CTRL_PRESSED \
+		 | RIGHT_CTRL_PRESSED \
+		 | LEFT_ALT_PRESSED \
+		 | RIGHT_ALT_PRESSED \
+		 | SHIFT_PRESSED))
+
+static int
+modified_key(int key, DWORD state)
+{
+    key |= mod_KEY;
+    if (state & (LEFT_CTRL_PRESSED | RIGHT_CTRL_PRESSED))
+	key |= mod_CTRL;
+    if (state & (LEFT_ALT_PRESSED | RIGHT_ALT_PRESSED))
+	key |= mod_ALT;
+    if (state & SHIFT_PRESSED)
+	key |= mod_SHIFT;
+
+    return key;
+}
+
 static struct {
     int windows;
     int vile;
@@ -648,9 +676,16 @@ decode_key_event(INPUT_RECORD * irp)
 	   irp->Event.KeyEvent.wVirtualScanCode,
 	   irp->Event.KeyEvent.dwControlKeyState));
 
-    if ((key = (UCHAR) irp->Event.KeyEvent.uChar.AsciiChar) != 0)
+    if ((key = (UCHAR) irp->Event.KeyEvent.uChar.AsciiChar) != 0) {
+	if (isCntrl(key)) {
+	    DWORD cstate = state & ~(LEFT_CTRL_PRESSED | RIGHT_CTRL_PRESSED);
+	    if (isModified(cstate))
+		key = modified_key(key, cstate);
+	}
 	return key;
+    }
 
+    key = irp->Event.KeyEvent.wVirtualKeyCode;
     if ((state & (RIGHT_CTRL_PRESSED | LEFT_CTRL_PRESSED)) != 0
 	&& (state & (RIGHT_CTRL_PRESSED
 		     | LEFT_CTRL_PRESSED
@@ -658,9 +693,9 @@ decode_key_event(INPUT_RECORD * irp)
 	/*
 	 * Control-shift-6 is control/^, control/~ or control/`.
 	 */
-	if (irp->Event.KeyEvent.wVirtualKeyCode == '6'
-	    || irp->Event.KeyEvent.wVirtualKeyCode == '^'
-	    || irp->Event.KeyEvent.wVirtualKeyCode == '\036') {
+	if (key == '6'
+	    || key == '^'
+	    || key == '\036') {
 	    TRACE(("...decode_key_event ^^\n"));
 	    return '\036';
 	}
@@ -676,17 +711,8 @@ decode_key_event(INPUT_RECORD * irp)
 	     * pageup/pagedown and arrow key bindings that would be lost if we
 	     * used the Win32-only definition.
 	     */
-	    if (state &
-		(LEFT_CTRL_PRESSED | RIGHT_CTRL_PRESSED
-		 | LEFT_ALT_PRESSED | RIGHT_ALT_PRESSED
-		 | SHIFT_PRESSED)) {
-		key = mod_KEY | keyxlate[i].vile;
-		if (state & (LEFT_CTRL_PRESSED | RIGHT_CTRL_PRESSED))
-		    key |= mod_CTRL;
-		if (state & (LEFT_ALT_PRESSED | RIGHT_ALT_PRESSED))
-		    key |= mod_ALT;
-		if (state & SHIFT_PRESSED)
-		    key |= mod_SHIFT;
+	    if (isModified(state)) {
+		key = modified_key(keyxlate[i].vile, state);
 		if (keyxlate[i].vile == '2') {
 		    if ((key & mod_CTRL) && ((key & mod_ALT) == 0)) {
 			/* either ^2 or ^@, => nul char */
@@ -1143,6 +1169,18 @@ handle_mouse_event(MOUSE_EVENT_RECORD mer)
 		    );
 	    }
 	    break;
+
+#ifdef MOUSE_WHEELED
+	case MOUSE_WHEELED:
+	    /*
+	     * Trial and error experimentation shows that dwButtonState
+	     * has its high bit set when the wheel moves back and not
+	     * set otherwise.
+	     */
+	    mvupwind(TRUE, ((long) mer.dwButtonState < 0) ? -3 : 3);
+	    update(TRUE);
+	    return;
+#endif /* MOUSE_WHEELED */
 	}
 
 	for_ever {
@@ -1381,7 +1419,7 @@ parse_icursor_string(char *str, int *revert_cursor)
  * The valid range of <insmode_height> and <cmdmode_height> is 1-100.
  */
 int
-chgd_icursor(VALARGS * args, int glob_vals, int testing)
+chgd_icursor(BUFFER *bp, VALARGS * args, int glob_vals, int testing)
 {
     if (!testing) {
 	int revert_cursor;
