@@ -4,14 +4,16 @@
  *	Copyright (c) 1990, 1995-1999 by Paul Fox, except for delins(), which is
  *	Copyright (c) 1986 by University of Toronto, as noted below.
  *
- * $Header: /usr/build/vile/vile/RCS/oneliner.c,v 1.102 2003/03/18 02:20:27 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/oneliner.c,v 1.108 2005/07/13 23:43:54 tom Exp $
  */
 
 #include	"estruct.h"
 #include	"edef.h"
 
-#define PLIST	0x01
-#define PNUMS	0x02
+#define PNONE	0x00
+#define PLIST	0x01		/* shows line in list-mode */
+#define PNUMS	0x02		/* shows line in number-mode */
+#define PGREP	0x04		/* like "grep -n": buffer name, line number */
 
 static int delins(regexp * exp, char *sourc, int lensrc);
 static int substline(regexp * exp, int nth_occur, int printit, int globally, int *confirmp);
@@ -25,11 +27,11 @@ static int lines_changed, total_changes;
 static int
 pregion(UINT flag)
 {
-    register WINDOW *wp;
-    register BUFFER *bp;
-    register int status;
+    WINDOW *wp;
+    BUFFER *bp;
+    int status;
     REGION region;
-    register LINEPTR linep;
+    LINE *linep;
     BUFFER *oldbp = curbp;
 
     if ((status = get_fl_region(&region)) != TRUE) {
@@ -62,15 +64,21 @@ pregion(UINT flag)
     do {
 	if (!addline(bp, linep->l_text, llength(linep)))
 	    break;		/* out of memory */
-	if (flag & PNUMS) {
+	if (flag & (PNUMS | PGREP)) {
 	    BUFFER *savebp = curbp;
 	    WINDOW *savewp = curwp;
 	    curbp = bp;
 	    curwp = bp2any_wp(bp);
-	    DOT.l = lback(buf_head(bp));
-	    DOT.o = 0;
-	    bprintf("%s:%d:", oldbp->b_bname, line_no(oldbp, linep));
-	    DOT.o = 0;
+	    if (flag & PGREP) {
+		DOT.l = lback(buf_head(bp));
+		DOT.o = 0;
+		bprintf("%s:%d:", oldbp->b_bname, line_no(oldbp, linep));
+		DOT.o = 0;
+	    } else {
+		make_local_w_val(curwp, WMDNUMBER);
+		set_w_val(curwp, WMDNUMBER, TRUE);
+		flag &= ~PNUMS;
+	    }
 	    curbp = savebp;
 	    curwp = savewp;
 	}
@@ -80,6 +88,7 @@ pregion(UINT flag)
     set_bname(bp, P_LINES_BufName);
     set_rdonly(bp, non_filename(), MDVIEW);
     set_b_val(bp, VAL_TAB, tabstop_val(curbp));
+    b_clr_counted(bp);
 
     for_each_visible_window(wp) {
 	if (wp->w_bufp == bp) {
@@ -98,15 +107,21 @@ llineregion(void)
 }
 
 int
-pplineregion(void)
+nlineregion(void)
 {
     return pregion(PNUMS);
 }
 
+int
+pplineregion(void)
+{
+    return pregion(PGREP);
+}
+
 int
 plineregion(void)
 {
-    return pregion(0);
+    return pregion(PNONE);
 }
 
 static regexp *substexp;
@@ -159,7 +174,7 @@ substreg1(int needpats, int use_opts, int is_globalsub)
     int c, status;
     static int printit, globally, nth_occur, confirm;
     REGION region;
-    LINEPTR oline;
+    LINE *oline;
     int getopts = FALSE;
     TBUFF *newpattern;
 
@@ -178,14 +193,16 @@ substreg1(int needpats, int use_opts, int is_globalsub)
 
 	if (gregexp) {
 	    FreeIfNeeded(substexp);
-	    substexp = castalloc(regexp, (size_t) (gregexp->size));
+	    if ((substexp = castalloc(regexp, (size_t) (gregexp->size))) == 0) {
+		no_memory("substreg1");
+		return FALSE;
+	    }
 	    (void) memcpy((char *) substexp, (char *) gregexp,
 			  (size_t) gregexp->size);
 	}
 
 	newpattern = 0;
-	if (tb_length(replacepat) != 0)
-	    tb_copy(&newpattern, replacepat);
+	tb_init(&newpattern, EOS);
 	status = readpattern("replacement string: ",
 			     &newpattern, (regexp **) 0, c, FALSE);
 	if (status == ABORT) {
@@ -277,7 +294,7 @@ substreg1(int needpats, int use_opts, int is_globalsub)
 static void
 showpat(regexp * rp, int on)
 {
-    register LINEPTR lp;
+    LINE *lp;
     int row;
 
     for (lp = curwp->w_line.l, row = curwp->w_toprow;
@@ -296,8 +313,8 @@ substline(regexp * exp, int nth_occur, int printit, int globally, int *confirmp)
 {
     int foundit;
     int again = 0;
-    register int s;
-    register int which_occur = 0;
+    int s;
+    int which_occur = 0;
     int matched_at_eol = FALSE;
     int yes, c, skipped;
 
@@ -398,7 +415,7 @@ substline(regexp * exp, int nth_occur, int printit, int globally, int *confirmp)
 	}
     } while (globally && sameline(scanboundpos, DOT));
     if (foundit && printit) {
-	register WINDOW *wp = curwp;
+	WINDOW *wp = curwp;
 	(void) setmark();
 	s = plineregion();
 	if (s != TRUE)
@@ -472,10 +489,10 @@ delins(regexp * exp, char *sourc, int lensrc)
 	len_delins = dlength + 1;
     }
 
-    (void) memcpy(buf_delins, exp->startp[0], (size_t) dlength);
+    (void) memcpy(buf_delins, exp->startp[0], dlength);
     buf_delins[dlength] = EOS;
 
-    if (ldelete((long) dlength, FALSE) != TRUE) {
+    if (ldelete(dlength, FALSE) != TRUE) {
 	mlforce("[Error while deleting]");
 	return FALSE;
     }
