@@ -2,7 +2,7 @@
  *		The routines in this file handle the conversion of pathname
  *		strings.
  *
- * $Header: /usr/build/vile/vile/RCS/path.c,v 1.135 2003/05/25 20:21:27 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/path.c,v 1.150 2005/07/09 15:01:56 Peter.Prymmer Exp $
  *
  *
  */
@@ -176,7 +176,7 @@ is_vms_pathname(const char *path, int option)
     const char *base = path;
     int this = 0, next = -1;
 
-    if (*path == EOS)		/* this can happen with null buffer-name */
+    if (isEmpty(path))		/* this can happen with null buffer-name */
 	return FALSE;
 
     while (ispath(*path)) {
@@ -223,7 +223,7 @@ is_vms_pathname(const char *path, int option)
 	    break;
 	case '!':
 	case '/':
-	case '~':
+	case CH_TILDE:
 	    return FALSE;	/* a DEC-shell name */
 	default:
 	    if (!ispath(*path))
@@ -261,7 +261,7 @@ is_vms_pathname(const char *path, int option)
 char *
 vms_pathleaf(char *path)
 {
-    register char *s;
+    char *s;
     for (s = skip_string(path);
 	 s > path && !strchr(":]", s[-1]);
 	 s--) ;
@@ -280,7 +280,7 @@ vms_pathleaf(char *path)
 char *
 unix_pathleaf(char *path)
 {
-    register char *s = last_slash(path);
+    char *s = last_slash(path);
     if (s == 0) {
 #if OPT_MSDOS_PATH
 	if ((s = is_msdos_drive(path)) == 0)
@@ -313,10 +313,15 @@ pathcat(char *dst, const char *path, char *leaf)
     char *s;
     size_t have;
 
+    if (dst == 0)
+	return 0;
+
+    if (leaf == 0)
+	leaf = "";
+
     leaf = vl_strncpy(save_leaf, leaf, NFILEN);		/* leaf may be in dst */
 
-    if (path == 0
-	|| *path == EOS) {
+    if (isEmpty(path)) {
 	(void) strcpy(dst, leaf);
     } else {
 
@@ -368,7 +373,7 @@ pathcat(char *dst, const char *path, char *leaf)
 char *
 last_slash(char *fn)
 {
-    register char *s;
+    char *s;
 
     if (*fn != EOS)
 	for (s = skip_string(fn); s > fn; s--)
@@ -398,7 +403,7 @@ static UPATH *user_paths;
 static char *
 save_user(const char *name, const char *path)
 {
-    register UPATH *q;
+    UPATH *q;
 
     if (name != NULL
 	&& path != NULL
@@ -412,7 +417,7 @@ save_user(const char *name, const char *path)
 	} else {
 	    FreeIfNeeded(q->name);
 	    FreeIfNeeded(q->path);
-	    free((char *) q);
+	    free(q);
 	}
     }
     return NULL;
@@ -422,9 +427,9 @@ static char *
 find_user(const char *name)
 {
 #if SYS_UNIX
-    register struct passwd *p;
+    struct passwd *p;
 #endif
-    register UPATH *q;
+    UPATH *q;
 
     if (name != NULL) {
 	for (q = user_paths; q != NULL; q = q->next) {
@@ -509,7 +514,7 @@ home_dir(void)
 char *
 home_path(char *path)
 {
-    if (*path == '~') {
+    if (*path == CH_TILDE) {
 	char temp[NFILEN];
 	char *s;
 	char *d;
@@ -798,8 +803,8 @@ resolve_directory(char *path_name, char **file_namep)
 
     while ((thisdev != rootdev)
 	   || (thisino != rootino)) {
-	register DIR *dp;
-	register DIRENT *de;
+	DIR *dp;
+	DIRENT *de;
 	dev_t dotdev;
 	ino_t dotino;
 	char mount_point;
@@ -1129,7 +1134,7 @@ canonpath(char *ss)
 	    s = ss;
 #endif
 
-#if SYS_UNIX
+#if SYS_UNIX && !OPT_VMS_PATH
 	(void) home_path(s);
 #endif
 
@@ -1312,7 +1317,7 @@ shorten_path(char *path, int keep_cwd)
 # endif
 #endif
 
-    if (!path || *path == EOS)
+    if (isEmpty(path))
 	return path;
 
     if (isInternalName(path))
@@ -1438,7 +1443,7 @@ shorten_path(char *path, int keep_cwd)
 static int
 mixed_case(const char *path)
 {
-    register int c;
+    int c;
     int had_upper = FALSE;
     int had_lower = FALSE;
     while ((c = *path++) != EOS) {
@@ -1460,10 +1465,10 @@ lengthen_path(char *path)
 #if SYS_VMS
     struct FAB my_fab;
     struct NAM my_nam;
-    char my_esa[NAM$C_MAXRSS];	/* expanded: sys$parse */
-    char my_rsa[NAM$C_MAXRSS];	/* result: sys$search */
+    char my_esa[NAM$C_MAXRSS + 1];	/* expanded: sys$parse */
+    char my_rsa[NAM$C_MAXRSS + 1];	/* result: sys$search */
 #endif
-    register int len;
+    int len;
     const char *cwd;
     char *f;
     char temp[NFILEN];
@@ -1491,7 +1496,7 @@ lengthen_path(char *path)
 #endif
 	    return path;
     }
-#if SYS_UNIX
+#if SYS_UNIX && !OPT_VMS_PATH
     (void) home_path(f);
 #endif
 
@@ -1499,11 +1504,11 @@ lengthen_path(char *path)
     /*
      * If the file exists, we can ask VMS to tell the full pathname.
      */
-    if ((*path != EOS) && maybe_pathname(path)) {
+    if ((*path != EOS) && maybe_pathname(path) && is_vms_pathname(path, -TRUE)) {
 	int fd;
 	long status;
 	char temp[NFILEN], leaf[NFILEN];
-	register char *s;
+	char *s;
 
 	if (!strchr(path, '*') && !strchr(path, '?')) {
 	    if ((fd = open(SL_TO_BSL(path), O_RDONLY, 0)) >= 0) {
@@ -1549,9 +1554,18 @@ lengthen_path(char *path)
     }
 #else
 # if OPT_VMS_PATH
-    /* this is only for testing! */
-    if (fakevms_filename(path))
-	return path;
+    if ((*path != EOS) && maybe_pathname(path)) {
+	/*
+	 * getname() returns a full pathname, like realpath().
+	 */
+	if (realpath(path, temp)) {
+	    strcpy(path, temp);
+	}
+	/* this is only for testing! */
+	if (fakevms_filename(path)) {
+	    return path;
+	}
+    }
 # endif
 #endif
 
@@ -1641,7 +1655,7 @@ is_absolute_pathname(char *path)
 
 #if SYS_UNIX || OPT_MSDOS_PATH || SYS_VMS
 #if SYS_UNIX
-    if (path[0] == '~')
+    if (path[0] == CH_TILDE)
 	return TRUE;
 #endif
     if (is_slashc(path[0]))
@@ -1769,11 +1783,11 @@ int
 is_directory(char *path)
 {
 #if OPT_VMS_PATH
-    register char *s;
+    char *s;
 #endif
     struct stat sb;
 
-    if (path == NULL || *path == EOS)
+    if (isEmpty(path))
 	return FALSE;
 
 #if OPT_VMS_PATH
@@ -1856,7 +1870,7 @@ const char *
 parse_pathlist(const char *list, char *result)
 {
     if (list != NULL && *list != EOS) {
-	register int len = 0;
+	int len = 0;
 
 	while (*list && (*list != vl_pathchr)) {
 	    if (len < NFILEN - 1)
@@ -2002,8 +2016,13 @@ sl_to_bsl(const char *p)
     if (p != 0) {
 	static TBUFF *cnv;
 	size_t len = strlen(p);
-	char *t = tb_values(tb_alloc(&cnv, (len | 127) + 1));
-	char *s = slconv(p, t, '/', '\\');
+	char *s, *t;
+
+	if (tb_alloc(&cnv, (len | 127) + 1) == 0)
+	    return (char *) p;	/* not quite giving up, but close */
+
+	t = tb_values(cnv);
+	s = slconv(p, t, '/', '\\');
 	if ((s = is_msdos_drive(s)) == 0)
 	    s = t;
 	/* Trim trailing slash if it's not the first */
@@ -2037,18 +2056,21 @@ bsl_to_sl_inplace(char *p)
 char *
 version_of(char *fname)
 {
-    char *s = strchr(fname, ';');
-    if (s == 0) {
-	fname = pathleaf(fname);
-	if ((s = strchr(fname, '.')) == 0
-	    || (*++s == EOS)
-	    || (s = strchr(fname, '.')) == 0)
-	    s = skip_string(fname);
-	if (strcmp(s, "*")) {	/* either "*" or a number */
-	    char *d;
-	    (void) strtol(s, &d, 10);
-	    if (*d != EOS)
-		s = skip_string(s);
+    char *s;
+    if (fname != 0) {
+	s = strchr(fname, ';');
+	if (s == 0) {
+	    fname = pathleaf(fname);
+	    if ((s = strchr(fname, '.')) == 0
+		|| (*++s == EOS)
+		|| (s = strchr(fname, '.')) == 0)
+		s = skip_string(fname);
+	    if (strcmp(s, "*")) {	/* either "*" or a number */
+		char *d;
+		(void) strtol(s, &d, 10);
+		if (*d != EOS)
+		    s = skip_string(s);
+	    }
 	}
     }
     return s;
@@ -2235,17 +2257,18 @@ append_libdir_to_path(void)
     char buf[NFILEN];
 
     if (libdir_path != 0
-	&& (tmp = getenv("PATH")) != 0) {
-	env = strmalloc(tmp);
+	&& (tmp = getenv("PATH")) != 0
+	&& (env = strmalloc(tmp)) != 0) {
 	cp = libdir_path;
 	while ((cp = parse_pathlist(cp, buf)) != 0) {
 	    append_to_path_list(&env, buf);
 	}
 	if (strcmp(tmp, env)) {
-	    tmp = (char *) malloc(6 + strlen(env));
-	    lsprintf(tmp, "PATH=%s", env);
-	    putenv(tmp);
-	    TRACE(("putenv %s\n", tmp));
+	    if ((tmp = typeallocn(char, 6 + strlen(env))) != 0) {
+		lsprintf(tmp, "PATH=%s", env);
+		putenv(tmp);
+		TRACE(("putenv %s\n", tmp));
+	    }
 	} else {
 	    free(env);
 	}
@@ -2264,11 +2287,11 @@ path_leaks(void)
 {
 #if SYS_UNIX
     while (user_paths != NULL) {
-	register UPATH *paths = user_paths;
+	UPATH *paths = user_paths;
 	user_paths = paths->next;
 	free(paths->name);
 	free(paths->path);
-	free((char *) paths);
+	free(paths);
     }
 #endif
 }
