@@ -2,7 +2,7 @@
  * This file contains the command processing functions for a number of random
  * commands. There is no functional grouping here, for sure.
  *
- * $Header: /usr/build/vile/vile/RCS/random.c,v 1.272 2003/05/04 22:49:20 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/random.c,v 1.287 2005/07/14 00:09:57 tom Exp $
  *
  */
 
@@ -24,6 +24,10 @@
 #endif
 #endif
 
+#if SYS_VMS
+#include	<cvtdef.h>	/* for float constants */
+#endif
+
 #include	"dirstuff.h"
 
 #if CC_TURBO
@@ -105,8 +109,8 @@ liststuff(const char *name,
 	  int iarg,
 	  void *vargp)
 {
-    register BUFFER *bp;
-    register int s;
+    BUFFER *bp;
+    int s;
     WINDOW *wp;
     int alreadypopped;
     BUFFER *ocurbp = curbp;
@@ -170,12 +174,12 @@ showcpos(int f GCC_UNUSED, int n GCC_UNUSED)
 #if OPT_POSFORMAT
     static TBUFF *result;
 #else
-    register LINE *lp;		/* current line */
-    register B_COUNT numchars = 0;	/* # of chars in file */
-    register L_NUM numlines = 0;	/* # of lines in file */
-    register B_COUNT predchars = 0;	/* # chars preceding point */
-    register L_NUM predlines = 0;	/* # lines preceding point */
-    register int curchar = '\n';	/* character under cursor */
+    LINE *lp;			/* current line */
+    B_COUNT numchars = 0;	/* # of chars in file */
+    L_NUM numlines = 0;		/* # of lines in file */
+    B_COUNT predchars = 0;	/* # chars preceding point */
+    L_NUM predlines = 0;	/* # lines preceding point */
+    int curchar = '\n';		/* character under cursor */
     long ratio;
     C_NUM col;
     C_NUM savepos;		/* temp save for current offset */
@@ -231,17 +235,18 @@ showcpos(int f GCC_UNUSED, int n GCC_UNUSED)
     ecol = mk_to_vcol(DOT, FALSE, curbp, 0, 0);
     DOT.o = savepos;
 
-    ratio = 0;			/* Ratio before dot. */
-    if (numchars != 0)
-	ratio = (100L * predchars) / numchars;
+    ratio = PERCENT(predchars, numchars);
 
     /* summarize and report the info */
-    if (numlines == 0 && numchars == 0)
+    if (numlines == 0 && numchars == 0) {
 	mlforce("File is empty");
-    else
+    } else {
 	mlforce("Line %d of %d, Col %d of %d, Char %ld of %ld (%ld%%) char is 0x%x or 0%o",
-		predlines + 1, numlines, col + 1, ecol,
-		predchars + 1, numchars, ratio, curchar, curchar);
+		predlines + 1, numlines,
+		col + 1, ecol,
+		predchars + 1, numchars,
+		ratio, curchar, curchar);
+    }
 #endif
     return TRUE;
 }
@@ -298,7 +303,7 @@ vl_line_count(BUFFER *the_buffer)
     L_NUM numlines = 0;		/* # of lines in file */
     if (the_buffer != 0) {
 #if SMALLER
-	register LINE *lp;	/* current line */
+	LINE *lp;		/* current line */
 
 	for_each_line(lp, the_buffer)
 	    ++ numlines;
@@ -312,11 +317,11 @@ vl_line_count(BUFFER *the_buffer)
 
 /* return the number of the given line */
 L_NUM
-line_no(BUFFER *the_buffer, LINEPTR the_line)
+line_no(BUFFER *the_buffer, LINE *the_line)
 {
     L_NUM line_num = 0;
 
-    if (the_line != null_ptr) {
+    if (the_line != 0) {
 #if SMALLER
 	LINE *lp;
 
@@ -342,7 +347,7 @@ line_no(BUFFER *the_buffer, LINEPTR the_line)
 B_COUNT
 char_no(BUFFER *the_buffer, MARK the_mark)
 {
-    register LINE *lp;
+    LINE *lp;
     B_COUNT curchar = 0;
     int rslen = len_record_sep(curbp);
 
@@ -379,8 +384,8 @@ int
 gotochr(int f, int n)
 {
     LINE *lp;
-    int len;
-    int len_rs = len_record_sep(curbp);
+    B_COUNT len;
+    B_COUNT len_rs = len_record_sep(curbp);
 
     if (!f) {
 	DOT.l = lback(buf_head(curbp));
@@ -389,12 +394,13 @@ gotochr(int f, int n)
 	B_COUNT goal = n;
 	for_each_line(lp, curbp) {
 	    len = line_length(lp);
-	    if ((goal -= len) <= 0) {
+	    if (goal <= len) {
 		DOT.l = lp;
 		if ((DOT.o = goal + len - 1) >= llength(lp))
 		    DOT.o = llength(lp) - 1;
 		break;
 	    }
+	    goal -= len;
 	}
 	if (goal > 0)
 	    return FALSE;
@@ -414,8 +420,8 @@ gotochr(int f, int n)
 int
 getcol(MARK mark, int actual)
 {
-    register C_NUM c, i;
-    register C_NUM col = 0;
+    C_NUM c, i;
+    C_NUM col = 0;
 
     if (mark.l != 0
 	&& llength(mark.l) > 0) {
@@ -489,7 +495,7 @@ getoff(C_NUM goal, C_NUM * rcolp)
 int
 gocol(int n)
 {
-    register int offs;		/* column number the cursor is on */
+    int offs;			/* column number the cursor is on */
 
     if (DOT.l != 0) {
 	offs = getoff(n, (C_NUM *) 0);
@@ -547,12 +553,13 @@ twiddle(int f GCC_UNUSED, int n GCC_UNUSED)
 int
 forceblank(int f, int n)
 {
-    register LINE *lp1;
-    register LINE *lp2;
+    LINE *lp1;
+    LINE *lp2;
     B_COUNT nld;
     B_COUNT n_arg;
     C_NUM nchar;
     int s = TRUE;
+    B_COUNT len_rs = len_record_sep(curbp);
 
     if (!f || n < 0)
 	n = 0;
@@ -572,8 +579,8 @@ forceblank(int f, int n)
     while ((lp2 = lforw(lp2)) != buf_head(curbp) &&
 	   firstchar(lp2) < 0) {
 	++nld;
-	if (nld > n_arg)
-	    nchar += llength(lp2) + 1;
+	if (nld > n_arg && lisreal(lp2))
+	    nchar += line_length(lp2);
     }
 
     DOT.l = lforw(lp1);
@@ -584,8 +591,10 @@ forceblank(int f, int n)
     } else if (n_arg < nld) {	/* delete (nld - n_arg) lines */
 	DOT.l = lp2;
 	DOT.o = 0;
-	backchar(TRUE, nchar);
-	s = ldelete((B_COUNT) nchar, FALSE);
+	if (nchar) {
+	    backchar(TRUE, nchar);
+	    s = ldelete((B_COUNT) nchar, FALSE);
+	}
     } else {			/* insert (n_arg - nld) lines */
 	n_arg = n_arg - nld;
 	while (s && n_arg--)
@@ -848,16 +857,23 @@ catnap(int milli, int watchinput)
 
 #if SYS_VMS
 	float seconds = milli / 1000.;
+	unsigned flags = 0;
+#ifdef USE_IEEE_FLOAT
+	unsigned type = CVT$K_IEEE_S;
+#else
+	unsigned type = CVT$K_VAX_F;
+#endif
 	if (watchinput) {
 	    float tenth = .1;
 	    while (seconds > 0.1) {
-		lib$wait(&tenth);
+		lib$wait(&tenth, &flags, &type);
 		if (term.typahead())
 		    return TRUE;
 		seconds -= tenth;
 	    }
 	}
-	lib$wait(&seconds);
+	if (seconds > 0.0)
+	    lib$wait(&seconds, &flags, &type);
 #define have_slept 1
 #endif
 
@@ -902,7 +918,7 @@ current_directory(int force)
 #ifdef HAVE_GETCWD
 #ifdef HAVE_REALPATH
     /*
-     * For slow (read:  simulated) filesystems, a getcwd() can be very slow. 
+     * For slow (read:  simulated) filesystems, a getcwd() can be very slow.
      * We always make a call to this function during startup, but wish to avoid
      * the overhead of getcwd() in that case.  However, it is not unusual to
      * invoke vile from other applications, which do not set $PWD.  In that
@@ -960,6 +976,11 @@ current_directory(int force)
     update_dos_drv_dir(cwd);
 #endif
 
+#if OPT_VMS_PATH && !SYS_VMS
+    strcat(cwd, "/");
+    (void) unix2vms_path(cwd, cwd);
+#endif
+
     TRACE(("current_directory(%s)\n", cwd));
     return cwd;
 }
@@ -1083,7 +1104,7 @@ update_dos_drv_dir(char *cwd)
 int
 vl_chdir(int f GCC_UNUSED, int n GCC_UNUSED)
 {
-    register int status;
+    int status;
     static TBUFF *last;
     char cdirname[NFILEN];
 
@@ -1144,6 +1165,29 @@ cd_and_pwd(char *path)
 
 	/* if dirstack on screen, update it */
 	(void) display_dirstack(DIRS_OPT);
+
+#if SYS_WINNT && DISP_NTWIN
+	/* if $recent-folders > 0, save directory path in registry */
+	{
+	    char cwd[FILENAME_MAX + 1];
+
+	    /*
+	     * must ask OS for cwd because "path" may be something like, say:
+	     *
+	     *   ../some_dir
+	     *
+	     * which is not usable later when "." changes.
+	     */
+	    getcwd(cwd, sizeof(cwd));
+
+	    /*
+	     * Store path in Unix format, since that's the external
+	     * presentation format used by the editor.
+	     */
+	    bsl_to_sl_inplace(cwd);
+	    store_recent_file_or_folder(cwd, FALSE);
+	}
+#endif
 	return TRUE;
     }
     return FALSE;
@@ -1372,8 +1416,7 @@ ch_fname(BUFFER *bp, const char *fname)
 	    }
 
 	    if (!bp->b_fname) {
-		bp->b_fname = strmalloc(np);
-		if (!bp->b_fname) {
+		if ((bp->b_fname = strmalloc(np)) == 0) {
 		    bp->b_fname = out_of_mem;
 		    bp->b_fnlen = strlen(bp->b_fname);
 		    no_memory("ch_fname");
@@ -1402,6 +1445,27 @@ ch_fname(BUFFER *bp, const char *fname)
 }
 
 #if OPT_HOOKS
+static char *
+name_of_hook(HOOK * hook)
+{
+    char *result = "unknown";
+    if (hook == &cdhook)
+	result = "$cd-hook";
+    else if (hook == &readhook)
+	result = "$read-hook";
+    else if (hook == &writehook)
+	result = "$write-hook";
+    else if (hook == &bufhook)
+	result = "$buf-hook";
+    else if (hook == &exithook)
+	result = "$exit-hook";
+    else if (hook == &autocolorhook)
+	result = "autocolor-hook";
+    else if (hook == &majormodehook)
+	result = "$majormode-hook";
+    return result;
+}
+
 int
 run_a_hook(HOOK * hook)
 {
@@ -1414,6 +1478,8 @@ run_a_hook(HOOK * hook)
 	save_pre_op_dot = pre_op_dot;
 
 	DisableHook(hook);
+	TPRINTF(("running %s HOOK with %s, current %s\n",
+		 name_of_hook(hook), hook->proc, curbp->b_bname));
 	status = docmd(hook->proc, TRUE, FALSE, 1);
 	EnableHook(hook);
 
@@ -1923,19 +1989,20 @@ vl_dirs_add(int f GCC_UNUSED, int n GCC_UNUSED)
 long
 vl_atol(char *str, int base, int *failed)
 {
+    long result = 0;
+
     *failed = FALSE;
 
+    if (str != 0) {
 #if defined(EDOM) && defined(ERANGE)
-    {
 	char *prem;
-	long rslt = 0;
 
 	str = skip_blanks(str);
 	if (!*str)
 	    *failed = TRUE;
 	else {
 	    set_errno(0);
-	    rslt = strtol(str, &prem, base);
+	    result = strtol(str, &prem, base);
 	    if (errno == EDOM || errno == ERANGE)
 		*failed = TRUE;
 	    else {
@@ -1943,15 +2010,15 @@ vl_atol(char *str, int base, int *failed)
 		    *failed = TRUE;	/* trailing garbage */
 	    }
 	}
-	return (rslt);
-    }
 #else
-    /*
-     * too bad, so sad.  if this routine must be implemented using K&R,
-     * it should be doable with sscanf().
-     */
-    return atoi(str);
+	/*
+	 * too bad, so sad.  if this routine must be implemented using K&R,
+	 * it should be doable with sscanf().
+	 */
+	result = atoi(str);
 #endif
+    }
+    return result;
 }
 
 /*
