@@ -3,7 +3,7 @@
  * and mark.  Some functions are commands.  Some functions are just for
  * internal use.
  *
- * $Header: /usr/build/vile/vile/RCS/region.c,v 1.126 2003/06/17 23:39:48 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/region.c,v 1.132 2005/07/14 00:36:03 tom Exp $
  *
  */
 
@@ -328,7 +328,8 @@ shift_right_line(void *flagp GCC_UNUSED, int l GCC_UNUSED, int r GCC_UNUSED)
     if (is_empty_line(DOT)
 	|| (is_c_mode(curbp)
 	    && llength(DOT.l) > 0
-	    && char_at(DOT) == '#')) {
+	    && char_at(DOT) == '#'
+	    && b_val(curbp, MDCINDENT))) {
 	return TRUE;
     }
     s = shiftwid_val(curbp);
@@ -528,7 +529,8 @@ entabline(void *flagp GCC_UNUSED, int l GCC_UNUSED, int r GCC_UNUSED)
 		    while (ooff < llength(lp))
 			(void) lreplc(lp, noff++, lgetc(lp, ooff++));
 		    DOT.o = noff;
-		    (void) ldelete(ooff - noff, FALSE);
+		    if (ooff > noff)
+			(void) ldelete(ooff - noff, FALSE);
 		}
 		(void) gocol(savecol);
 		return TRUE;
@@ -571,7 +573,8 @@ trimline(void *flag GCC_UNUSED, int l GCC_UNUSED, int r GCC_UNUSED)
     int off;
     LINE *lp;
     int odoto, s;
-    int delcnt, was_at_eol;
+    B_COUNT delcnt;
+    int was_at_eol;
 
     lp = DOT.l;
 
@@ -582,15 +585,16 @@ trimline(void *flag GCC_UNUSED, int l GCC_UNUSED, int r GCC_UNUSED)
        that's okay, since the math still works. */
     off = lastchar(lp);
 
-    delcnt = llength(lp) - (off + 1);
-    if (!delcnt)
+    if (llength(lp) <= (off + 1))
 	return TRUE;
 
+    delcnt = llength(lp) - (off + 1);
+
     odoto = DOT.o;
     was_at_eol = (odoto == llength(lp));
 
     DOT.o = off + 1;
-    s = ldelete((B_COUNT) delcnt, FALSE);
+    s = ldelete(delcnt, FALSE);
 
     if (odoto > off) {		/* do we need to back up? */
 	odoto = llength(lp);
@@ -653,7 +657,8 @@ blankline(void *flagp, int l, int r)
 
     if (llength(lp) <= r) {
 	/* then the rect doesn't extend to the end of line */
-	ldelete((B_COUNT) (llength(lp) - l), FALSE);
+	if (llength(lp) > l)
+	    ldelete((B_COUNT) (llength(lp) - l), FALSE);
 
 	/* so there's nothing beyond the rect, so insert at
 	   most r-l chars of the string, or nothing */
@@ -664,10 +669,12 @@ blankline(void *flagp, int l, int r)
 	} else {
 	    len = 0;
 	}
-    } else {
+    } else if (r > l) {
 	/* the line goes on, so delete and reinsert exactly */
-	ldelete((B_COUNT) (r - l), FALSE);
 	len = r - l;
+	ldelete((B_COUNT) len, FALSE);
+    } else {
+	len = 0;
     }
 
     s = lstrinsert(string, len);
@@ -1031,7 +1038,7 @@ typedef struct {
 } ENCODEREG;
 
 static void
-encode_one_attribute(TBUFF ** result, int count, char *hypercmd, VIDEO_ATTR attr)
+encode_one_attribute(TBUFF **result, int count, char *hypercmd, VIDEO_ATTR attr)
 {
     char temp[80];
 
@@ -1139,6 +1146,10 @@ encode_attributes(LINE *lp, BUFFER *bp, REGION * top_region)
 		    my_list = typeallocn(AREGION, my_size = 10);
 		else if (my_used + 1 > my_size)
 		    my_list = typereallocn(AREGION, my_list, my_size *= 2);
+		if (my_list == 0) {
+		    no_memory("encode_attributes");
+		    return 0;
+		}
 		my_list[my_used++] = ar_temp;
 	    }
 	}
@@ -1282,7 +1293,7 @@ operattrencode(int f, int n)
 int
 oper_wrenc(int f, int n)
 {
-    register int s;
+    int s;
     char fname[NFILEN];
 
     if (ukb != 0) {
