@@ -18,7 +18,7 @@
  * transferring the selection are not dealt with in this file.  Procedures
  * for dealing with the representation are maintained in this file.
  *
- * $Header: /usr/build/vile/vile/RCS/select.c,v 1.147 2003/06/30 00:31:33 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/select.c,v 1.158 2005/07/14 00:07:09 tom Exp $
  *
  */
 
@@ -46,7 +46,7 @@ static void detach_attrib(BUFFER *bp, AREGION * arp);
 static int attribute_cntl_a_sequences(void);
 static void free_line_attribs(BUFFER *bp);
 static int add_line_attrib(BUFFER *bp, REGION * rp, REGIONSHAPE rs,
-			   VIDEO_ATTR vattr, TBUFF * hypercmd);
+			   VIDEO_ATTR vattr, TBUFF *hypercmd);
 static void purge_line_attribs(BUFFER *bp, REGION * rp, REGIONSHAPE rs,
 			       int owner);
 
@@ -674,8 +674,8 @@ static void
 sweepmsg(const char *msg)
 {
     char temp[NLINE];
-    (void) kcod2pstr(fnc2kcod(&f_multimotion), temp);
-    mlforce("[%s (end with %*S)]", msg, *temp, temp + 1);
+    (void) kcod2pstr(fnc2kcod(&f_multimotion), temp, sizeof(temp));
+    mlforce("[%s (end with %.*s)]", msg, CharOf(*temp), temp + 1);
 }
 
 static int
@@ -719,7 +719,7 @@ on_double_click(void)
     sel_release();
     if (!is_at_end_of_line(DOT)
 	&& !isSpace(char_at(DOT))) {
-	while (DOT.o >= 0) {
+	while (DOT.o > 0) {
 	    DOT.o--;
 	    if (isSpace(char_at(DOT))) {
 		DOT.o++;
@@ -1475,10 +1475,10 @@ attribute_cntl_a_seqs_in_region(REGION * rp, REGIONSHAPE shape)
 /*
  * Setup for iteration over region to attribute, ensure that DOT < MK.
  */
-LINEPTR
+LINE *
 setup_region(void)
 {
-    LINEPTR pastline;		/* pointer to line just past EOP */
+    LINE *pastline;		/* pointer to line just past EOP */
 
     if (!sameline(MK, DOT)) {
 	REGION region;
@@ -1648,7 +1648,7 @@ set_mark_after(int count, int rslen)
 static int
 attribute_cntl_a_sequences(void)
 {
-    LINEPTR pastline;
+    LINE *pastline;
     C_NUM offset;		/* offset in cur line of place to attribute */
     int count;
 
@@ -1668,14 +1668,16 @@ attribute_cntl_a_sequences(void)
 	    if (char_at(DOT) == CONTROL_A) {
 		offset = decode_attribute(DOT.l->l_text, llength(DOT.l),
 					  DOT.o, &count);
+		if (offset > DOT.o) {
 #if EFFICIENCY_HACK
-		new_attribs = curbp->b_attribs;
-		curbp->b_attribs = orig_attribs;
-		ldelete((B_COUNT) (offset - DOT.o), FALSE);
-		curbp->b_attribs = new_attribs;
+		    new_attribs = curbp->b_attribs;
+		    curbp->b_attribs = orig_attribs;
+		    ldelete((B_COUNT) (offset - DOT.o), FALSE);
+		    curbp->b_attribs = new_attribs;
 #else
-		ldelete((B_COUNT) (offset - DOT.o), FALSE);
+		    ldelete((B_COUNT) (offset - DOT.o), FALSE);
 #endif
+		}
 		set_mark_after(count, len_record_sep(curbp));
 		if (apply_attribute())
 		    (void) attributeregion();
@@ -1707,7 +1709,7 @@ discard_syntax_highlighting(void)
 static int
 attribute_from_filter(void)
 {
-    LINEPTR pastline;
+    LINE *pastline;
     int skip;
     int nbytes;
     int done;
@@ -1803,7 +1805,7 @@ attribute_directly(void)
 
 	    if (filtername == 0
 		&& curbp->majr != 0)
-		filtername = curbp->majr->name;
+		filtername = curbp->majr->shortname;
 
 	    if (filtername != 0
 		&& flt_start(filtername)) {
@@ -1817,6 +1819,12 @@ attribute_directly(void)
 	}
 	attach_attrib(selbufp, &selregion);
 	attach_attrib(startbufp, &startregion);
+#if OPT_HILITEMATCH
+	if (curbp->b_highlight & HILITE_ON) {
+	    curbp->b_highlight |= HILITE_DIRTY;
+	    attrib_matches();
+	}
+#endif
     }
 #endif
     return code;
@@ -1894,65 +1902,80 @@ find_line_attr_idx(VIDEO_ATTR vattr)
 }
 
 /* Attempt to shift a portion of a line either left or right for
-   inserts or deletes.  The idea is to preserve the line attributes
-   as much as possible until autocolor gets around to recoloring the
-   line */
+ * inserts or deletes.  The idea is to preserve the line attributes
+ * as much as possible until autocolor gets around to recoloring the
+ * line.
+ *
+ * Returns false if we ran out of memory.
+ */
 /* ARGSUSED */
-void
-lattr_shift(BUFFER *bp GCC_UNUSED, LINEPTR lp, int doto, int shift)
+int
+lattr_shift(BUFFER *bp GCC_UNUSED, LINE *lp, int doto, int shift)
 {
+    int status = TRUE;
     UCHAR *lap;
-    if (!lp->l_attrs)
-	return;
-    lap = lp->l_attrs;
-    if (shift > 0) {
-	int f, t, len;
-	len = strlen((char *) lap);
-	t = len - 1;
-	if (t <= 0)
-	    return;
-	for (f = t; f >= doto && f > t - shift; f--)
-	    if (lap[f] != 1) {
-		int newlen;
 
-		beginDisplay();
-		newlen = len + shift - (t - f);
-		lap = castrealloc(UCHAR, lap, newlen + 1);
-		lp->l_attrs = lap;
-		lap[newlen] = 0;
-		t = newlen - 1;
-		f = t - shift;
-		endofDisplay();
-		break;
+    if (lp->l_attrs != 0) {
+	lap = lp->l_attrs;
+	if (shift > 0) {
+	    int f, t, len;
+	    len = strlen((char *) lap);
+	    t = len - 1;
+	    if (t > 0) {
+		for (f = t; f >= doto && f > t - shift; f--) {
+		    if (lap[f] != 1) {
+			int newlen;
+
+			beginDisplay();
+			newlen = len + shift - (t - f);
+			if ((lap = castrealloc(UCHAR, lap, newlen + 1)) != 0) {
+			    lp->l_attrs = lap;
+			    lap[newlen] = 0;
+			    t = newlen - 1;
+			    f = t - shift;
+			}
+			endofDisplay();
+			break;
+		    }
+		}
+		if (lap != 0) {
+		    while (f > doto) {
+			lap[t--] = lap[f--];
+		    }
+		} else {
+		    no_memory("lattr_shift");
+		    status = FALSE;
+		}
+	    }
+	} else if (shift < 0) {
+	    int f, t;
+	    int saw_attr = 0;
+	    shift = -shift;
+	    /* Move t to doto, but don't run off end */
+	    for (t = 0; t < doto && lap[t]; t++)
+		saw_attr |= (lap[t] != 1);
+	    if (lap[t] != 0) {
+		/* Position f, but don't run off end */
+		for (f = t; f < doto + shift && lap[f]; f++) ;
+		/* Shift via copying, but observe what it is we shift */
+		while (lap[f]) {
+		    saw_attr |= (lap[f] != 1);
+		    lap[t++] = lap[f++];
+		}
+		/* Try to get rid of the line attributes entirely */
+		if (!saw_attr) {
+		    beginDisplay();
+		    FreeAndNull(lp->l_attrs);
+		    endofDisplay();
+		} else {
+		    /* Normal out the stuff at the end. */
+		    while (t < f)
+			lap[t++] = 1;
+		}
 	    }
-	while (f > doto) {
-	    lap[t--] = lap[f--];
 	}
-    } else if (shift < 0) {
-	int f, t;
-	int saw_attr = 0;
-	shift = -shift;
-	/* Move t to doto, but don't run off end */
-	for (t = 0; t < doto && lap[t]; t++)
-	    saw_attr |= (lap[t] != 1);
-	if (lap[t] == 0)
-	    return;
-	/* Position f, but don't run off end */
-	for (f = t; f < doto + shift && lap[f]; f++) ;
-	/* Shift via copying, but observe what it is we shift */
-	while (lap[f]) {
-	    saw_attr |= (lap[f] != 1);
-	    lap[t++] = lap[f++];
-	}
-	/* Try to get rid of the line attributes entirely */
-	if (!saw_attr) {
-	    FreeAndNull(lp->l_attrs);
-	    return;
-	}
-	/* Normal out the stuff at the end. */
-	while (t < f)
-	    lap[t++] = 1;
     }
+    return status;
 }
 
 #endif /* OPT_LINE_ATTRS */
@@ -1979,10 +2002,10 @@ free_line_attribs(BUFFER *bp)
 
 static int
 add_line_attrib(BUFFER *bp, REGION * rp, REGIONSHAPE rs, VIDEO_ATTR vattr,
-		TBUFF * hypercmdp)
+		TBUFF *hypercmdp)
 {
 #if OPT_LINE_ATTRS
-    LINEPTR lp;
+    LINE *lp;
     WINDOW *wp;
     int vidx;
     int i;
@@ -2026,12 +2049,15 @@ add_line_attrib(BUFFER *bp, REGION * rp, REGIONSHAPE rs, VIDEO_ATTR vattr,
 	}
     } else {
 	/* Must allocate and initialize memory for the line attributes */
-	lp->l_attrs = castalloc(UCHAR, llength(lp) + 1);
-	lp->l_attrs[llength(lp)] = 0;
-	for (i = llength(lp) - 1; i >= 0; i--)
-	    lp->l_attrs[i] = 1;
-	if (last > llength(lp))
-	    last = llength(lp);
+	beginDisplay();
+	if ((lp->l_attrs = castalloc(UCHAR, llength(lp) + 1)) != 0) {
+	    lp->l_attrs[llength(lp)] = 0;
+	    for (i = llength(lp) - 1; i >= 0; i--)
+		lp->l_attrs[i] = 1;
+	    if (last > llength(lp))
+		last = llength(lp);
+	}
+	endofDisplay();
     }
     endofDisplay();
 
@@ -2055,11 +2081,11 @@ static void
 purge_line_attribs(BUFFER *bp, REGION * rp, REGIONSHAPE rs, int owner)
 {
 #if OPT_LINE_ATTRS
-    LINEPTR ls = rp->r_orig.l;
-    LINEPTR le = rp->r_end.l;
+    LINE *ls = rp->r_orig.l;
+    LINE *le = rp->r_end.l;
     int os = rp->r_orig.o;
     int oe = rp->r_end.o;
-    LINEPTR lp;
+    LINE *lp;
     int i;
     int do_update = 0;
 
