@@ -3,7 +3,7 @@
  *	for getting and setting the values of the vile state variables,
  *	plus helper utility functions.
  *
- * $Header: /usr/build/vile/vile/RCS/statevar.c,v 1.75 2003/07/27 17:11:36 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/statevar.c,v 1.94 2005/03/13 18:20:19 tom Exp $
  */
 
 #include	"estruct.h"
@@ -40,8 +40,8 @@ static const char *
 DftEnv(const char *name, const char *dft)
 {
 #if OPT_EVAL && OPT_SHELL
-    name = vile_getenv(name);
-    return (*name == EOS) ? dft : name;
+    name = safe_getenv(name);
+    return isEmpty(name) ? dft : name;
 #else
     return dft;
 #endif
@@ -50,16 +50,20 @@ DftEnv(const char *name, const char *dft)
 static void
 SetEnv(char **namep, const char *value)
 {
+    char *newvalue;
+
     beginDisplay();
+    if ((newvalue = strmalloc(value)) != 0) {
 #if OPT_EVAL && OPT_SHELL
-    FreeIfNeeded(*namep);
+	FreeIfNeeded(*namep);
 #endif
-    *namep = strmalloc(value);
+	*namep = newvalue;
+    }
     endofDisplay();
 }
 
 static int
-any_ro_BOOL(TBUFF ** rp, const char *vp, int value)
+any_ro_BOOL(TBUFF **rp, const char *vp, int value)
 {
     if (rp) {
 	render_boolean(rp, value);
@@ -72,7 +76,7 @@ any_ro_BOOL(TBUFF ** rp, const char *vp, int value)
 }
 
 static int
-any_rw_BOOL(TBUFF ** rp, const char *vp, int *value)
+any_rw_BOOL(TBUFF **rp, const char *vp, int *value)
 {
     if (rp) {
 	render_boolean(rp, *value);
@@ -86,7 +90,7 @@ any_rw_BOOL(TBUFF ** rp, const char *vp, int *value)
 }
 
 static int
-any_ro_INT(TBUFF ** rp, const char *vp, int value)
+any_ro_INT(TBUFF **rp, const char *vp, int value)
 {
     if (rp) {
 	render_int(rp, value);
@@ -99,7 +103,21 @@ any_ro_INT(TBUFF ** rp, const char *vp, int value)
 }
 
 static int
-any_ro_STR(TBUFF ** rp, const char *vp, const char *value)
+any_rw_INT(TBUFF **rp, const char *vp, int *value)
+{
+    if (rp) {
+	render_int(rp, *value);
+	return TRUE;
+    } else if (vp) {
+	*value = scan_int(vp);
+	return TRUE;
+    } else {
+	return FALSE;
+    }
+}
+
+static int
+any_ro_STR(TBUFF **rp, const char *vp, const char *value)
 {
     if (rp) {
 	if (value != 0) {
@@ -113,7 +131,7 @@ any_ro_STR(TBUFF ** rp, const char *vp, const char *value)
 }
 
 static int
-any_rw_STR(TBUFF ** rp, const char *vp, TBUFF ** value)
+any_rw_STR(TBUFF **rp, const char *vp, TBUFF **value)
 {
     if (rp) {
 	if (value != 0 && *value != 0) {
@@ -128,7 +146,7 @@ any_rw_STR(TBUFF ** rp, const char *vp, TBUFF ** value)
 }
 
 static int
-any_rw_EXPR(TBUFF ** rp, const char *vp, TBUFF ** value)
+any_rw_EXPR(TBUFF **rp, const char *vp, TBUFF **value)
 {
     if (rp) {
 	if (value != 0) {
@@ -149,7 +167,7 @@ any_rw_EXPR(TBUFF ** rp, const char *vp, TBUFF ** value)
 }
 
 static int
-any_rw_TXT(TBUFF ** rp, const char *vp, char **value)
+any_rw_TXT(TBUFF **rp, const char *vp, char **value)
 {
     if (rp) {
 	tb_scopy(rp, *value);
@@ -163,12 +181,12 @@ any_rw_TXT(TBUFF ** rp, const char *vp, char **value)
 }
 
 static int
-any_REPL(TBUFF ** rp, const char *vp, CHARTYPE type)
+any_REPL(TBUFF **rp, const char *vp, CHARTYPE type)
 {
     if (rp) {
 	lgrabtext(rp, type);
 	return TRUE;
-    } else if (vp && curbp) {
+    } else if (vp && valid_buffer(curbp)) {
 	return lrepltext(type, vp, strlen(vp));
     } else {
 	return FALSE;
@@ -177,7 +195,7 @@ any_REPL(TBUFF ** rp, const char *vp, CHARTYPE type)
 
 #if OPT_HOOKS
 static int
-any_HOOK(TBUFF ** rp, const char *vp, HOOK * hook)
+any_HOOK(TBUFF **rp, const char *vp, HOOK * hook)
 {
     if (rp) {
 	tb_scopy(rp, hook->proc);
@@ -192,11 +210,15 @@ any_HOOK(TBUFF ** rp, const char *vp, HOOK * hook)
 #endif
 
 const char *
-vile_getenv(const char *s)
+safe_getenv(const char *s)
 {
 #if OPT_EVAL && OPT_SHELL
-    register char *v = getenv(s);
-    return v ? v : "";
+    char *v = getenv(s);
+#if defined(_WIN32)
+    if (v == 0)
+	v = vile_getenv(s);
+#endif
+    return NONNULL(v);
 #else
     return "";
 #endif
@@ -278,8 +300,8 @@ get_directory(void)
 #endif
 
 /* access the default kill buffer */
-static char *
-getkill(TBUFF ** rp)
+static void
+getkill(TBUFF **rp)
 {
     tb_init(rp, EOS);
     if (kbs[0].kbufh != 0) {
@@ -287,8 +309,6 @@ getkill(TBUFF ** rp)
 	tb_bappend(rp, (char *) (kbs[n].kbufh->d_chunk), kbs[n].kused);
     }
     tb_append(rp, EOS);
-
-    return tb_values(*rp);
 }
 
 #if OPT_PATHLOOKUP
@@ -301,7 +321,7 @@ get_cdpath(void)
 }
 
 int
-var_CDPATH(TBUFF ** rp, const char *vp)
+var_CDPATH(TBUFF **rp, const char *vp)
 {
     if (rp) {
 	tb_scopy(rp, get_cdpath());
@@ -345,13 +365,13 @@ cfgopts(void)
 # endif
 #endif
 #if DISP_X11
-# ifdef OL_WIDGETS
+# if OL_WIDGETS
 	"openlook",
 # endif
-# ifdef MOTIF_WIDGETS
+# if MOTIF_WIDGETS
 	"motif",
 # endif
-# ifdef ATHENA_WIDGETS
+# if ATHENA_WIDGETS
 	"athena",
 #  ifdef HAVE_LIB_XAW
 	"xaw",
@@ -385,7 +405,7 @@ cfgopts(void)
 
 #if OPT_EVAL
 int
-var_ABUFNAME(TBUFF ** rp, const char *vp)
+var_ABUFNAME(TBUFF **rp, const char *vp)
 {
     BUFFER *bp;
     return any_ro_STR(rp, vp, ((bp = find_alt()) != 0) ? bp->b_bname : "");
@@ -393,24 +413,29 @@ var_ABUFNAME(TBUFF ** rp, const char *vp)
 
 #if OPT_HOOKS
 int
-var_AUTOCOLORHOOK(TBUFF ** rp, const char *vp)
+var_AUTOCOLORHOOK(TBUFF **rp, const char *vp)
 {
     return any_HOOK(rp, vp, &autocolorhook);
 }
 #endif
 
 int
-var_BCHARS(TBUFF ** rp, const char *vp)
+var_BCHARS(TBUFF **rp, const char *vp)
 {
-    return curbp ? any_ro_INT(rp, vp, curbp->b_bytecount) : FALSE;
+    bsizes(curbp);
+    return (valid_buffer(curbp)
+	    ? any_ro_INT(rp, vp, curbp->b_bytecount)
+	    : FALSE);
 }
 
 int
-var_BFLAGS(TBUFF ** rp, const char *vp)
+var_BFLAGS(TBUFF **rp, const char *vp)
 {
+    char temp[80];
+
     if (rp) {
-	tb_init(rp, EOS);
-	buffer_flags(tb_values(*rp), curbp);
+	buffer_flags(temp, curbp);
+	tb_scopy(rp, temp);
 	return TRUE;
     } else if (vp) {
 	return ABORT;		/* read-only */
@@ -420,9 +445,12 @@ var_BFLAGS(TBUFF ** rp, const char *vp)
 }
 
 int
-var_BLINES(TBUFF ** rp, const char *vp)
+var_BLINES(TBUFF **rp, const char *vp)
 {
-    return curbp ? any_ro_INT(rp, vp, curbp->b_linecount) : FALSE;
+    bsizes(curbp);
+    return (valid_buffer(curbp)
+	    ? any_ro_INT(rp, vp, curbp->b_linecount)
+	    : FALSE);
 }
 
 #if DISP_NTWIN
@@ -442,7 +470,7 @@ parse_rgb(const char **src, int dft)
 }
 
 int
-var_BRIGHTNESS(TBUFF ** rp, const char *vp)
+var_BRIGHTNESS(TBUFF **rp, const char *vp)
 {
     char temp[80];
 
@@ -463,20 +491,20 @@ var_BRIGHTNESS(TBUFF ** rp, const char *vp)
 
 #if OPT_HOOKS
 int
-var_BUFHOOK(TBUFF ** rp, const char *vp)
+var_BUFHOOK(TBUFF **rp, const char *vp)
 {
     return any_HOOK(rp, vp, &bufhook);
 }
 #endif
 
 int
-var_CBUFNAME(TBUFF ** rp, const char *vp)
+var_CBUFNAME(TBUFF **rp, const char *vp)
 {
     if (rp) {
 	tb_scopy(rp, curbp->b_bname);
 	return TRUE;
     } else if (vp) {
-	if (curbp) {
+	if (valid_buffer(curbp)) {
 	    set_bname(curbp, vp);
 	    curwp->w_flag |= WFMODE;
 	}
@@ -487,33 +515,37 @@ var_CBUFNAME(TBUFF ** rp, const char *vp)
 }
 
 int
-var_BWINDOWS(TBUFF ** rp, const char *vp)
+var_BWINDOWS(TBUFF **rp, const char *vp)
 {
-    return curbp ? any_ro_INT(rp, vp, curbp->b_nwnd) : FALSE;
+    return (valid_buffer(curbp)
+	    ? any_ro_INT(rp, vp, curbp->b_nwnd)
+	    : FALSE);
 }
 
 #if OPT_HOOKS
 int
-var_CDHOOK(TBUFF ** rp, const char *vp)
+var_CDHOOK(TBUFF **rp, const char *vp)
 {
     return any_HOOK(rp, vp, &cdhook);
 }
 #endif
 
 int
-var_CFGOPTS(TBUFF ** rp, const char *vp)
+var_CFGOPTS(TBUFF **rp, const char *vp)
 {
     return any_ro_STR(rp, vp, cfgopts());
 }
 
 int
-var_CFNAME(TBUFF ** rp, const char *vp)
+var_CFNAME(TBUFF **rp, const char *vp)
 {
     if (rp) {
-	tb_scopy(rp, (curbp && curbp->b_fname) ? curbp->b_fname : "");
+	tb_scopy(rp, ((valid_buffer(curbp) && curbp->b_fname)
+		      ? curbp->b_fname
+		      : ""));
 	return TRUE;
     } else if (vp) {
-	if (curbp) {
+	if (valid_buffer(curbp)) {
 	    ch_fname(curbp, vp);
 	    curwp->w_flag |= WFMODE;
 	}
@@ -524,10 +556,10 @@ var_CFNAME(TBUFF ** rp, const char *vp)
 }
 
 int
-var_CHAR(TBUFF ** rp, const char *vp)
+var_CHAR(TBUFF **rp, const char *vp)
 {
     if (rp) {
-	if (curbp && !is_empty_buf(curbp)) {
+	if (valid_buffer(curbp) && !is_empty_buf(curbp)) {
 	    if (is_at_end_of_line(DOT))
 		render_int(rp, '\n');
 	    else
@@ -537,9 +569,7 @@ var_CHAR(TBUFF ** rp, const char *vp)
 	}
 	return TRUE;
     } else if (vp) {
-	if (curbp == NULL || b_val(curbp, MDVIEW)) {
-	    return rdonly();
-	} else {
+	if (valid_buffer(curbp) && !b_val(curbp, MDVIEW)) {
 	    int s, c;
 	    mayneedundo();
 	    (void) ldelete(1L, FALSE);	/* delete 1 char */
@@ -550,6 +580,8 @@ var_CHAR(TBUFF ** rp, const char *vp)
 		s = linsert(1, c);
 	    (void) backchar(FALSE, 1);
 	    return s;
+	} else {
+	    return rdonly();
 	}
     } else {
 	return FALSE;
@@ -558,7 +590,7 @@ var_CHAR(TBUFF ** rp, const char *vp)
 
 #if OPT_ENCRYPT
 int
-var_CRYPTKEY(TBUFF ** rp, const char *vp)
+var_CRYPTKEY(TBUFF **rp, const char *vp)
 {
     if (rp) {
 	tb_scopy(rp, WRITE_ONLY);
@@ -566,8 +598,10 @@ var_CRYPTKEY(TBUFF ** rp, const char *vp)
     } else if (vp) {
 	beginDisplay();
 	FreeIfNeeded(cryptkey);
-	cryptkey = (char *) malloc(NKEYLEN);
+	cryptkey = typeallocn(char, NKEYLEN);
 	endofDisplay();
+	if (cryptkey == 0)
+	    return no_memory("var_CRYPTKEY");
 	vl_make_encrypt_key(cryptkey, vp);
 	return TRUE;
     } else {
@@ -578,12 +612,12 @@ var_CRYPTKEY(TBUFF ** rp, const char *vp)
 
 #if !SMALLER
 int
-var_CURCHAR(TBUFF ** rp, const char *vp)
+var_CURCHAR(TBUFF **rp, const char *vp)
 {
     if (rp) {
 	render_int(rp, vl_getcchar() + 1);
 	return TRUE;
-    } else if (vp && curbp) {
+    } else if (vp && valid_buffer(curbp)) {
 	return gotochr(TRUE, strtol(vp, 0, 0));
     } else {
 	return FALSE;
@@ -592,12 +626,12 @@ var_CURCHAR(TBUFF ** rp, const char *vp)
 #endif
 
 int
-var_CURCOL(TBUFF ** rp, const char *vp)
+var_CURCOL(TBUFF **rp, const char *vp)
 {
     if (rp) {
 	render_int(rp, getccol(FALSE) + 1);
 	return TRUE;
-    } else if (vp && curbp) {
+    } else if (vp && valid_buffer(curbp)) {
 	return gotocol(TRUE, strtol(vp, 0, 0));
     } else {
 	return FALSE;
@@ -605,12 +639,12 @@ var_CURCOL(TBUFF ** rp, const char *vp)
 }
 
 int
-var_CURLINE(TBUFF ** rp, const char *vp)
+var_CURLINE(TBUFF **rp, const char *vp)
 {
     if (rp) {
 	render_int(rp, getcline());
 	return TRUE;
-    } else if (vp && curbp) {
+    } else if (vp && valid_buffer(curbp)) {
 	return gotoline(TRUE, strtol(vp, 0, 0));
     } else {
 	return FALSE;
@@ -619,7 +653,7 @@ var_CURLINE(TBUFF ** rp, const char *vp)
 
 #if OPT_SHELL
 int
-var_CWD(TBUFF ** rp, const char *vp)
+var_CWD(TBUFF **rp, const char *vp)
 {
     if (rp) {
 	tb_scopy(rp, current_directory(FALSE));
@@ -633,12 +667,12 @@ var_CWD(TBUFF ** rp, const char *vp)
 #endif
 
 int
-var_CWLINE(TBUFF ** rp, const char *vp)
+var_CWLINE(TBUFF **rp, const char *vp)
 {
     if (rp) {
 	render_int(rp, getlinerow());
 	return TRUE;
-    } else if (vp && curbp) {
+    } else if (vp && valid_buffer(curbp)) {
 	return forwline(TRUE, strtol(vp, 0, 0) - getlinerow());
     } else {
 	return FALSE;
@@ -646,14 +680,14 @@ var_CWLINE(TBUFF ** rp, const char *vp)
 }
 
 int
-var_DEBUG(TBUFF ** rp, const char *vp)
+var_DEBUG(TBUFF **rp, const char *vp)
 {
     return any_rw_BOOL(rp, vp, &tracemacros);
 }
 
 #if OPT_SHELL
 int
-var_DIRECTORY(TBUFF ** rp, const char *vp)
+var_DIRECTORY(TBUFF **rp, const char *vp)
 {
     if (rp) {
 	tb_scopy(rp, get_directory());
@@ -668,31 +702,47 @@ var_DIRECTORY(TBUFF ** rp, const char *vp)
 #endif
 
 int
-var_CMD_COUNT(TBUFF ** rp, const char *vp)
+var_CMD_COUNT(TBUFF **rp, const char *vp)
 {
     return any_ro_INT(rp, vp, cmd_count);
 }
 
 int
-var_DISCMD(TBUFF ** rp, const char *vp)
+var_DISCMD(TBUFF **rp, const char *vp)
 {
     return any_rw_BOOL(rp, vp, &vl_msgs);
 }
 
 int
-var_DISINP(TBUFF ** rp, const char *vp)
+var_DISINP(TBUFF **rp, const char *vp)
 {
     return any_rw_BOOL(rp, vp, &vl_echo);
 }
 
+#if OPT_TRACE
 int
-var_EOC(TBUFF ** rp, const char *vp)
+var_GOAL_COLUMN(TBUFF **rp, const char *vp)
+{
+    return any_rw_INT(rp, vp, &curgoal);
+}
+#endif
+
+#if OPT_LOCALE
+int
+var_ENCODING(TBUFF **rp, const char *vp)
+{
+    return any_ro_STR(rp, vp, vl_encoding);
+}
+#endif
+
+int
+var_EOC(TBUFF **rp, const char *vp)
 {
     return any_ro_BOOL(rp, vp, ev_end_of_cmd ? 1 : 0);
 }
 
 int
-var_FILENAME_IC(TBUFF ** rp, const char *vp)
+var_FILENAME_IC(TBUFF **rp, const char *vp)
 {
 #if OPT_CASELESS || SYS_VMS
     static int myvalue = TRUE;
@@ -704,7 +754,7 @@ var_FILENAME_IC(TBUFF ** rp, const char *vp)
 
 #if OPT_FINDERR
 int
-var_FILENAME_EXPR(TBUFF ** rp, const char *vp)
+var_FILENAME_EXPR(TBUFF **rp, const char *vp)
 {
     int code = any_rw_EXPR(rp, vp, &filename_expr);
     if (rp != 0 && code == TRUE) {
@@ -715,13 +765,13 @@ var_FILENAME_EXPR(TBUFF ** rp, const char *vp)
 }
 
 int
-var_ERROR_EXPR(TBUFF ** rp, const char *vp)
+var_ERROR_EXPR(TBUFF **rp, const char *vp)
 {
     return any_rw_EXPR(rp, vp, &error_expr);
 }
 
 int
-var_ERROR_MATCH(TBUFF ** rp, const char *vp)
+var_ERROR_MATCH(TBUFF **rp, const char *vp)
 {
     if (rp) {
 	tb_scopy(rp, tb_values(error_match));
@@ -734,7 +784,7 @@ var_ERROR_MATCH(TBUFF ** rp, const char *vp)
 }
 
 int
-var_ERROR_BUFFER(TBUFF ** rp, const char *vp)
+var_ERROR_BUFFER(TBUFF **rp, const char *vp)
 {
     if (rp) {
 	tb_scopy(rp, get_febuff());
@@ -746,10 +796,16 @@ var_ERROR_BUFFER(TBUFF ** rp, const char *vp)
 	return FALSE;
     }
 }
+
+int
+var_ERROR_TABSTOP(TBUFF **rp, const char *vp)
+{
+    return any_rw_INT(rp, vp, &error_tabstop);
+}
 #endif
 
 int
-var_EXEC_PATH(TBUFF ** rp, const char *vp)
+var_EXEC_PATH(TBUFF **rp, const char *vp)
 {
     return any_ro_STR(rp, vp, exec_pathname);
 }
@@ -761,14 +817,14 @@ var_EXEC_PATH(TBUFF ** rp, const char *vp)
 #endif
 
 int
-var_EXEC_SUFFIX(TBUFF ** rp, const char *vp)
+var_EXEC_SUFFIX(TBUFF **rp, const char *vp)
 {
     return any_ro_STR(rp, vp, the_EXEC_SUFFIX);
 }
 
 #if OPT_HOOKS
 int
-var_EXITHOOK(TBUFF ** rp, const char *vp)
+var_EXITHOOK(TBUFF **rp, const char *vp)
 {
     return any_HOOK(rp, vp, &exithook);
 }
@@ -776,7 +832,7 @@ var_EXITHOOK(TBUFF ** rp, const char *vp)
 
 #if SYS_WINNT
 int
-var_FAVORITES(TBUFF ** rp, const char *vp)
+var_FAVORITES(TBUFF **rp, const char *vp)
 {
     return any_ro_STR(rp, vp, get_favorites());
 }
@@ -784,7 +840,7 @@ var_FAVORITES(TBUFF ** rp, const char *vp)
 
 #if OPT_FINDPATH
 int
-var_FINDPATH(TBUFF ** rp, const char *vp)
+var_FINDPATH(TBUFF **rp, const char *vp)
 {
     if (rp) {
 	tb_scopy(rp, get_findpath());
@@ -799,17 +855,17 @@ var_FINDPATH(TBUFF ** rp, const char *vp)
 
 /* debug aid -- this may go away */
 int
-var_FINDCMD(TBUFF ** rp, const char *vp)
+var_FINDCMD(TBUFF **rp, const char *vp)
 {
     char *cmd = last_findcmd();
 
-    return (any_ro_STR(rp, vp, (cmd) ? cmd : ""));
+    return (any_ro_STR(rp, vp, NONNULL(cmd)));
 }
 #endif
 
 #if DISP_X11||DISP_NTWIN
 int
-var_FONT(TBUFF ** rp, const char *vp)
+var_FONT(TBUFF **rp, const char *vp)
 {
 #if SYS_WINNT && DISP_NTWIN
     if (rp) {
@@ -835,20 +891,20 @@ var_FONT(TBUFF ** rp, const char *vp)
 #endif
 
 int
-var_FWD_SEARCH(TBUFF ** rp, const char *vp)
+var_FWD_SEARCH(TBUFF **rp, const char *vp)
 {
     return any_rw_BOOL(rp, vp, &last_srch_direc);
 }
 
 int
-var_HELPFILE(TBUFF ** rp, const char *vp)
+var_HELPFILE(TBUFF **rp, const char *vp)
 {
     return any_rw_TXT(rp, vp, &helpfile);
 }
 
 #if DISP_X11
 int
-var_ICONNAM(TBUFF ** rp, const char *vp)
+var_ICONNAM(TBUFF **rp, const char *vp)
 {
     if (rp) {
 	tb_scopy(rp, x_get_icon_name());
@@ -863,13 +919,13 @@ var_ICONNAM(TBUFF ** rp, const char *vp)
 #endif
 
 int
-var_IDENTIF(TBUFF ** rp, const char *vp)
+var_IDENTIF(TBUFF **rp, const char *vp)
 {
     return any_REPL(rp, vp, vl_ident);
 }
 
 int
-var_KBDMACRO(TBUFF ** rp, const char *vp)
+var_KBDMACRO(TBUFF **rp, const char *vp)
 {
     if (rp) {
 	get_kbd_macro(rp);
@@ -882,7 +938,7 @@ var_KBDMACRO(TBUFF ** rp, const char *vp)
 }
 
 int
-var_KILL(TBUFF ** rp, const char *vp)
+var_KILL(TBUFF **rp, const char *vp)
 {
     if (rp) {
 	getkill(rp);
@@ -895,21 +951,13 @@ var_KILL(TBUFF ** rp, const char *vp)
 }
 
 int
-var_LASTKEY(TBUFF ** rp, const char *vp)
+var_LASTKEY(TBUFF **rp, const char *vp)
 {
-    if (rp) {
-	render_int(rp, lastkey);
-	return TRUE;
-    } else if (vp) {
-	lastkey = scan_int(vp);
-	return TRUE;
-    } else {
-	return FALSE;
-    }
+    return any_rw_INT(rp, vp, &lastkey);
 }
 
 int
-var_LCOLS(TBUFF ** rp, const char *vp)
+var_LCOLS(TBUFF **rp, const char *vp)
 {
     if (rp) {
 	int col;
@@ -934,41 +982,49 @@ var_LCOLS(TBUFF ** rp, const char *vp)
 }
 
 int
-var_LIBDIR_PATH(TBUFF ** rp, const char *vp)
+var_LIBDIR_PATH(TBUFF **rp, const char *vp)
 {
     return any_rw_TXT(rp, vp, &libdir_path);
 }
 
 int
-var_LINE(TBUFF ** rp, const char *vp)
+var_LINE(TBUFF **rp, const char *vp)
 {
     return any_REPL(rp, vp, (CHARTYPE) 0);
 }
 
 int
-var_LLENGTH(TBUFF ** rp, const char *vp)
+var_LLENGTH(TBUFF **rp, const char *vp)
 {
-    return curbp ? any_ro_INT(rp, vp, llength(DOT.l)) : FALSE;
+    return valid_buffer(curbp) ? any_ro_INT(rp, vp, llength(DOT.l)) : FALSE;
 }
 
+#if OPT_LOCALE
+int
+var_LOCALE(TBUFF **rp, const char *vp)
+{
+    return any_ro_STR(rp, vp, vl_locale);
+}
+#endif
+
 #if OPT_MAJORMODE
 int
-var_MAJORMODE(TBUFF ** rp, const char *vp)
+var_MAJORMODE(TBUFF **rp, const char *vp)
 {
-    return any_ro_STR(rp, vp, ((curbp != 0 && curbp->majr != 0)
-			       ? curbp->majr->name
+    return any_ro_STR(rp, vp, ((valid_buffer(curbp) && curbp->majr != 0)
+			       ? curbp->majr->shortname
 			       : ""));
 }
 
 int
-var_MAJORMODEHOOK(TBUFF ** rp, const char *vp)
+var_MAJORMODEHOOK(TBUFF **rp, const char *vp)
 {
     return any_HOOK(rp, vp, &majormodehook);
 }
 #endif
 
 int
-var_MATCH(TBUFF ** rp, const char *vp)
+var_MATCH(TBUFF **rp, const char *vp)
 {
     return any_ro_STR(rp, vp,
 		      (tb_length(tb_matched_pat)
@@ -978,21 +1034,21 @@ var_MATCH(TBUFF ** rp, const char *vp)
 
 #if OPT_MENUS
 int
-var_MENU_FILE(TBUFF ** rp, const char *vp)
+var_MENU_FILE(TBUFF **rp, const char *vp)
 {
     return any_rw_TXT(rp, vp, &menu_file);
 }
 #endif
 
 int
-var_MODE(TBUFF ** rp, const char *vp)
+var_MODE(TBUFF **rp, const char *vp)
 {
     return any_ro_STR(rp, vp, current_modename());
 }
 
 #if OPT_MLFORMAT
 int
-var_MLFORMAT(TBUFF ** rp, const char *vp)
+var_MLFORMAT(TBUFF **rp, const char *vp)
 {
     if (rp) {
 	if (modeline_format == 0)
@@ -1011,14 +1067,14 @@ var_MLFORMAT(TBUFF ** rp, const char *vp)
 #endif
 
 int
-var_MODIFIED(TBUFF ** rp, const char *vp)
+var_MODIFIED(TBUFF **rp, const char *vp)
 {
-    return any_ro_BOOL(rp, vp, curbp && b_is_changed(curbp));
+    return any_ro_BOOL(rp, vp, valid_buffer(curbp) && b_is_changed(curbp));
 }
 
 #if OPT_COLOR
 int
-var_NCOLORS(TBUFF ** rp, const char *vp)
+var_NCOLORS(TBUFF **rp, const char *vp)
 {
     if (rp) {
 	render_int(rp, ncolors);
@@ -1030,14 +1086,14 @@ var_NCOLORS(TBUFF ** rp, const char *vp)
 }
 #else
 int
-var_NCOLORS(TBUFF ** rp GCC_UNUSED, const char *vp GCC_UNUSED)
+var_NCOLORS(TBUFF **rp GCC_UNUSED, const char *vp GCC_UNUSED)
 {
     return FALSE;
 }
 #endif
 
 int
-var_NTILDES(TBUFF ** rp, const char *vp)
+var_NTILDES(TBUFF **rp, const char *vp)
 {
     if (rp) {
 	render_int(rp, ntildes);
@@ -1054,20 +1110,20 @@ var_NTILDES(TBUFF ** rp, const char *vp)
 
 #if OPT_SHELL
 int
-var_OCWD(TBUFF ** rp, const char *vp)
+var_OCWD(TBUFF **rp, const char *vp)
 {
     return any_ro_STR(rp, vp, previous_directory());
 }
 #endif
 
 int
-var_OS(TBUFF ** rp, const char *vp)
+var_OS(TBUFF **rp, const char *vp)
 {
     return any_ro_STR(rp, vp, opersys);
 }
 
 int
-var_PAGELEN(TBUFF ** rp, const char *vp)
+var_PAGELEN(TBUFF **rp, const char *vp)
 {
     if (rp) {
 	render_int(rp, term.rows);
@@ -1086,7 +1142,7 @@ var_PAGELEN(TBUFF ** rp, const char *vp)
 
 /* separator for lists of pathnames */
 int
-var_PATHCHR(TBUFF ** rp, const char *vp)
+var_PATHCHR(TBUFF **rp, const char *vp)
 {
     if (rp) {
 	tb_append(rp, vl_pathchr);
@@ -1103,7 +1159,7 @@ var_PATHCHR(TBUFF ** rp, const char *vp)
 
 /* separator for levels of pathnames */
 int
-var_PATHSEP(TBUFF ** rp, const char *vp)
+var_PATHSEP(TBUFF **rp, const char *vp)
 {
     if (rp) {
 	tb_append(rp, vl_pathsep);
@@ -1120,7 +1176,7 @@ var_PATHSEP(TBUFF ** rp, const char *vp)
 
 #if OPT_POSFORMAT
 int
-var_POSFORMAT(TBUFF ** rp, const char *vp)
+var_POSFORMAT(TBUFF **rp, const char *vp)
 {
     if (rp) {
 	if (position_format == 0)
@@ -1138,7 +1194,7 @@ var_POSFORMAT(TBUFF ** rp, const char *vp)
 #endif
 
 int
-var_CURWIDTH(TBUFF ** rp, const char *vp)
+var_CURWIDTH(TBUFF **rp, const char *vp)
 {
     if (rp) {
 	render_int(rp, term.cols);
@@ -1156,7 +1212,7 @@ var_CURWIDTH(TBUFF ** rp, const char *vp)
 }
 
 int
-var_PALETTE(TBUFF ** rp, const char *vp)
+var_PALETTE(TBUFF **rp, const char *vp)
 {
     if (rp) {
 	tb_scopy(rp, (tb_length(tb_curpalette)
@@ -1171,25 +1227,25 @@ var_PALETTE(TBUFF ** rp, const char *vp)
 }
 
 int
-var_PATCHLEVEL(TBUFF ** rp, const char *vp)
+var_PATCHLEVEL(TBUFF **rp, const char *vp)
 {
-    return any_ro_STR(rp, vp, PATCHLEVEL);
+    return any_ro_STR(rp, vp, VILE_PATCHLEVEL);
 }
 
 int
-var_PATHNAME(TBUFF ** rp, const char *vp)
+var_PATHNAME(TBUFF **rp, const char *vp)
 {
     return any_REPL(rp, vp, vl_pathn);
 }
 
 int
-var_PENDING(TBUFF ** rp, const char *vp)
+var_PENDING(TBUFF **rp, const char *vp)
 {
     return any_ro_BOOL(rp, vp, keystroke_avail());
 }
 
 int
-var_PROCESSID(TBUFF ** rp, const char *vp)
+var_PROCESSID(TBUFF **rp, const char *vp)
 {
     if (rp) {
 #if SYS_UNIX
@@ -1212,26 +1268,26 @@ var_PROCESSID(TBUFF ** rp, const char *vp)
 }
 
 int
-var_PROGNAME(TBUFF ** rp, const char *vp)
+var_PROGNAME(TBUFF **rp, const char *vp)
 {
     return any_ro_STR(rp, vp, prognam);
 }
 
 int
-var_PROMPT(TBUFF ** rp, const char *vp)
+var_PROMPT(TBUFF **rp, const char *vp)
 {
     return any_rw_STR(rp, vp, &prompt_string);
 }
 
 int
-var_QIDENTIF(TBUFF ** rp, const char *vp)
+var_QIDENTIF(TBUFF **rp, const char *vp)
 {
     return any_REPL(rp, vp, vl_qident);
 }
 
 #if OPT_HOOKS
 int
-var_RDHOOK(TBUFF ** rp, const char *vp)
+var_RDHOOK(TBUFF **rp, const char *vp)
 {
     return any_HOOK(rp, vp, &readhook);
 }
@@ -1241,7 +1297,7 @@ var_RDHOOK(TBUFF ** rp, const char *vp)
  * Note that replacepat is stored without a trailing null.
  */
 int
-var_REPLACE(TBUFF ** rp, const char *vp)
+var_REPLACE(TBUFF **rp, const char *vp)
 {
     if (rp) {
 	tb_copy(rp, replacepat);
@@ -1259,12 +1315,12 @@ var_REPLACE(TBUFF ** rp, const char *vp)
  * Note that searchpat is stored without a trailing null.
  */
 int
-var_SEARCH(TBUFF ** rp, const char *vp)
+var_SEARCH(TBUFF **rp, const char *vp)
 {
     if (rp) {
 	tb_copy(rp, searchpat);
 	return TRUE;
-    } else if (vp && curbp) {
+    } else if (vp && valid_buffer(curbp)) {
 	(void) tb_init(&searchpat, EOS);
 	(void) tb_sappend(&searchpat, vp);
 	beginDisplay();
@@ -1280,7 +1336,7 @@ var_SEARCH(TBUFF ** rp, const char *vp)
 }
 
 int
-var_SEED(TBUFF ** rp, const char *vp)
+var_SEED(TBUFF **rp, const char *vp)
 {
     if (rp) {
 	render_int(rp, seed);
@@ -1296,7 +1352,7 @@ var_SEED(TBUFF ** rp, const char *vp)
 
 #if OPT_SHELL
 int
-var_SHELL(TBUFF ** rp, const char *vp)
+var_SHELL(TBUFF **rp, const char *vp)
 {
     if (rp) {
 	tb_scopy(rp, get_shell());
@@ -1311,7 +1367,7 @@ var_SHELL(TBUFF ** rp, const char *vp)
 #endif
 
 int
-var_SRES(TBUFF ** rp, const char *vp)
+var_SRES(TBUFF **rp, const char *vp)
 {
     if (rp) {
 	tb_scopy(rp, screen_desc);
@@ -1324,13 +1380,13 @@ var_SRES(TBUFF ** rp, const char *vp)
 }
 
 int
-var_STARTUP_FILE(TBUFF ** rp, const char *vp)
+var_STARTUP_FILE(TBUFF **rp, const char *vp)
 {
     return any_rw_TXT(rp, vp, &startup_file);
 }
 
 int
-var_STARTUP_PATH(TBUFF ** rp, const char *vp)
+var_STARTUP_PATH(TBUFF **rp, const char *vp)
 {
     return any_rw_TXT(rp, vp, &startup_path);
 }
@@ -1344,66 +1400,61 @@ setcmdstatus(int s)
 }
 
 int
-var_STATUS(TBUFF ** rp, const char *vp)
+var_STATUS(TBUFF **rp, const char *vp)
 {
     return any_rw_BOOL(rp, vp, &lastcmdstatus);
 }
 
 #if OPT_TITLE
 int
-var_TITLE(TBUFF ** rp, const char *vp)
+var_TITLE(TBUFF **rp, const char *vp)
 {
     if (rp) {
-#if SYS_WINNT
-	tb_scopy(rp, w32_wdw_title());
-	return TRUE;
-#endif
 #if DISP_X11
 	tb_scopy(rp, x_get_window_name());
 	return TRUE;
+#elif SYS_WINNT
+	tb_scopy(rp, w32_wdw_title());
+	return TRUE;
 #endif
     } else if (vp) {
+#if DISP_X11 || SYS_WINNT
 	tb_scopy(&request_title, vp);
 	return TRUE;
+#else
+	return ABORT;		/* read-only */
+#endif
     }
     return FALSE;
 }
 
 int
-var_TITLEFORMAT(TBUFF ** rp, const char *vp)
+var_TITLEFORMAT(TBUFF **rp, const char *vp)
 {
     return any_rw_STR(rp, vp, &title_format);
 }
 #endif /* OPT_TITLE */
 
 int
-var_TPAUSE(TBUFF ** rp, const char *vp)
+var_TPAUSE(TBUFF **rp, const char *vp)
 {
-    if (rp) {
-	render_int(rp, term.pausecount);
-	return TRUE;
-    } else if (vp) {
-	term.pausecount = strtol(vp, 0, 0);
-	return TRUE;
-    } else {
-	return FALSE;
-    }
+    return any_rw_INT(rp, vp, &term.pausecount);
 }
 
 int
-var_VERSION(TBUFF ** rp, const char *vp)
+var_VERSION(TBUFF **rp, const char *vp)
 {
     return any_ro_STR(rp, vp, version);
 }
 
 int
-var_WITHPREFIX(TBUFF ** rp, const char *vp)
+var_WITHPREFIX(TBUFF **rp, const char *vp)
 {
     return any_ro_STR(rp, vp, tb_values(with_prefix));
 }
 
 int
-var_WLINES(TBUFF ** rp, const char *vp)
+var_WLINES(TBUFF **rp, const char *vp)
 {
     if (rp) {
 	render_int(rp, curwp->w_ntrows);
@@ -1416,14 +1467,14 @@ var_WLINES(TBUFF ** rp, const char *vp)
 }
 
 int
-var_WORD(TBUFF ** rp, const char *vp)
+var_WORD(TBUFF **rp, const char *vp)
 {
     return any_REPL(rp, vp, vl_nonspace);
 }
 
 #if OPT_HOOKS
 int
-var_WRHOOK(TBUFF ** rp, const char *vp)
+var_WRHOOK(TBUFF **rp, const char *vp)
 {
     return any_HOOK(rp, vp, &writehook);
 }
@@ -1431,7 +1482,7 @@ var_WRHOOK(TBUFF ** rp, const char *vp)
 
 #if DISP_X11 && OPT_SHELL
 int
-var_XDISPLAY(TBUFF ** rp, const char *vp)
+var_XDISPLAY(TBUFF **rp, const char *vp)
 {
     if (rp) {
 	tb_scopy(rp, get_xdisplay());
@@ -1445,7 +1496,7 @@ var_XDISPLAY(TBUFF ** rp, const char *vp)
 }
 
 int
-var_XSHELL(TBUFF ** rp, const char *vp)
+var_XSHELL(TBUFF **rp, const char *vp)
 {
     if (rp) {
 	tb_scopy(rp, get_xshell());
@@ -1459,7 +1510,7 @@ var_XSHELL(TBUFF ** rp, const char *vp)
 }
 
 int
-var_XSHELLFLAGS(TBUFF ** rp, const char *vp)
+var_XSHELLFLAGS(TBUFF **rp, const char *vp)
 {
     if (rp) {
 	tb_scopy(rp, get_xshellflags());
@@ -1479,7 +1530,7 @@ void
 ev_leaks(void)
 {
 #if OPT_EVAL
-    register UVAR *p;
+    UVAR *p;
     while ((p = temp_vars) != 0)
 	rmv_tempvar(p->u_name);
 
