@@ -3,7 +3,7 @@
  * characters, and write characters in a barely buffered fashion on the display.
  * All operating systems.
  *
- * $Header: /usr/build/vile/vile/RCS/termio.c,v 1.195 2003/05/25 20:21:52 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/termio.c,v 1.200 2005/07/11 22:44:41 tom Exp $
  *
  */
 
@@ -128,6 +128,10 @@ error "No termios or sgtty"
 #define USE_POLL 0
 #endif
 
+#ifndef USE_SELECT
+#define USE_SELECT 0
+#endif
+
 #ifndef USE_SGTTY
 #define USE_SGTTY 0
 #endif
@@ -142,14 +146,14 @@ error "No termios or sgtty"
 
 #ifndef O_NDELAY
 #undef USE_FCNTL
-#define USE_FCNTL 1
+#define USE_FCNTL 0
 #endif
 
 #if USE_FCNTL
 
 #include	<fcntl.h>
-char kbd_char;			/* one char of typeahead, gotten during poll    */
-int kbd_char_good;		/* is a char in kbd_char?                       */
+static char kbd_char;		/* one char of typeahead, gotten during poll    */
+static int kbd_char_good;	/* is a char in kbd_char?                       */
 
 /*
  * putting the input tty in polling mode lets us check for
@@ -214,9 +218,13 @@ set_kbd_polling(int yes)
 # endif
 #endif
 
-struct termios otermios, ntermios;
+static int saved_tty = FALSE;
+static struct termios otermios;
 
-char tobuf[TBUFSIZ];		/* terminal output buffer */
+#if !DISP_X11
+static struct termios ntermios;
+static char tobuf[TBUFSIZ];	/* terminal output buffer */
+#endif
 
 void
 vl_save_tty(void)
@@ -240,13 +248,16 @@ vl_save_tty(void)
 #else
     wkillc = tocntrl('W');
 #endif
+    saved_tty = TRUE;
 }
 
 void
 vl_restore_tty(void)
 {
-    tcdrain(1);
-    tcsetattr(0, TCSADRAIN, &otermios);
+    if (saved_tty) {
+	tcdrain(1);
+	tcsetattr(0, TCSADRAIN, &otermios);
+    }
 }
 
 void
@@ -379,6 +390,7 @@ ttunclean(void)
 
 /* original terminal characteristics and characteristics to use inside */
 struct termio otermio, ntermio;
+static int saved_tty = FALSE;
 
 #ifdef HAVE_SETBUFFER		/* setbuffer() isn't on most termio systems */
 char tobuf[TBUFSIZ];		/* terminal output buffer */
@@ -405,12 +417,14 @@ vl_save_tty(void)
 #else /* no SIGTSTP */
     suspc = tocntrl('Z');
 #endif
+    saved_tty = TRUE;
 }
 
 void
 vl_restore_tty(void)
 {
-    ioctl(0, TCSETAF, (char *) &otermio);
+    if (saved_tty)
+	ioctl(0, TCSETAF, (char *) &otermio);
 }
 
 void
@@ -514,11 +528,16 @@ char tobuf[TBUFSIZ];		/* terminal output buffer */
 
 #undef	CTRL
 #include	<sgtty.h>	/* for stty/gtty functions */
+
+static int saved_tty = FALSE;
+
 struct sgttyb ostate;		/* saved tty state */
 struct sgttyb nstate;		/* values for editor mode */
 struct sgttyb rnstate;		/* values for raw editor mode */
+
 int olstate;			/* Saved local mode values */
 int nlstate;			/* new local mode values */
+
 struct ltchars oltchars;	/* Saved terminal special character set */
 struct ltchars nltchars =
 {-1, -1, -1, -1, -1, -1};	/* a lot of nothing */
@@ -544,17 +563,20 @@ vl_save_tty(void)
 #ifdef	TIOCLGET
     ioctl(0, TIOCLGET, (char *) &olstate);
 #endif
+    saved_tty = TRUE;
 }
 
 void
 vl_restore_tty(void)
 {
-    ioctl(0, TIOCSETN, (char *) &ostate);
-    ioctl(0, TIOCSETC, (char *) &otchars);
-    ioctl(0, TIOCSLTC, (char *) &oltchars);
+    if (saved_tty) {
+	ioctl(0, TIOCSETN, (char *) &ostate);
+	ioctl(0, TIOCSETC, (char *) &otchars);
+	ioctl(0, TIOCSLTC, (char *) &oltchars);
 #ifdef	TIOCLSET
-    ioctl(0, TIOCLSET, (char *) &olstate);
+	ioctl(0, TIOCLSET, (char *) &olstate);
 #endif
+    }
 }
 
 void
@@ -1044,8 +1066,14 @@ ttopen(void)
 	tidy_exit(status);
     newmode[0] = oldmode[0];
     newmode[1] = oldmode[1];
-    newmode[1] &= ~(TT$M_TTSYNC | TT$M_HOSTSYNC);
-    newmode[2] = oldmode[2] | TT2$M_PASTHRU;
+    newmode[1] |= TT$M_NOBRDCST;	/* turn on no-broadcast */
+    newmode[1] &= ~TT$M_TTSYNC;
+    newmode[1] &= ~TT$M_ESCAPE;	/* turn off escape-processing */
+    newmode[1] &= ~TT$M_HOSTSYNC;
+    newmode[1] &= ~TT$M_NOTYPEAHD;	/* turn off no-typeahead */
+    newmode[2] = oldmode[2];
+    newmode[2] |= TT2$M_PASTHRU;	/* turn on pass-through */
+    newmode[2] |= TT2$M_ALTYPEAHD;	/* turn on big typeahead buffer */
     status = sys$qiow(EFN, iochan, IO$_SETMODE, &iosb, 0, 0,
 		      newmode, sizeof(newmode), 0, 0, 0, 0);
     if (status != SS$_NORMAL
@@ -1162,11 +1190,15 @@ read_vms_tty(int length)
     QIO_SB iosb;
     int term[2] =
     {0, 0};
+    unsigned mask = (IO$_READVBLK
+		     | IO$M_NOECHO
+		     | IO$M_NOFILTR
+		     | IO$M_TRMNOECHO);
 
     status = sys$qiow(EFN, iochan,
-		      (length == 1)
-		      ? IO$_READLBLK | IO$M_NOECHO | IO$M_NOFILTR
-		      : IO$_READLBLK | IO$M_NOECHO | IO$M_NOFILTR | IO$M_TIMED,
+		      ((length == 1)
+		       ? mask
+		       : mask | IO$M_TIMED),
 		      &iosb, 0, 0, ibuf, length, 0, term, 0, 0);
 
     if (status != SS$_NORMAL)
