@@ -17,14 +17,13 @@
  *   "FAILED" may not be used to test an OLE return code.  Use SUCCEEDED
  *   instead.
  *
- * $Header: /usr/build/vile/vile/RCS/w32ole.cpp,v 1.20 2002/10/20 12:56:42 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/w32ole.cpp,v 1.22 2005/01/26 23:20:01 tom Exp $
  */
 
 #include "w32vile.h"
 
 #include <ole2.h>
 #include <comutil.h>
-#include <assert.h>
 #include <ctype.h>
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -66,8 +65,8 @@ oleauto_init(OLEAUTO_OPTIONS *opts)
     LPCLASSFACTORY pcf = NULL;
 
     olebuf_len = ansibuf_len = 512;
-    ansibuf    = (char *) malloc(ansibuf_len);
-    olebuf     = (OLECHAR *) malloc(olebuf_len * sizeof(OLECHAR));
+    ansibuf    = typeallocn(char, ansibuf_len);
+    olebuf     = typeallocn(OLECHAR, olebuf_len);
     if (! (olebuf && ansibuf))
         return (FALSE);
 
@@ -192,7 +191,7 @@ ConvertToAnsi(OLECHAR *szW)
         if (ansibuf)
             free(ansibuf);
         ansibuf_len = ansibuf_len * 2 + len;
-        if ((ansibuf = (char *) malloc(ansibuf_len)) == NULL)
+        if ((ansibuf = typeallocn(char, ansibuf_len)) == NULL)
             return (ansibuf);  /* We're gonna' die */
     }
     wcstombs(ansibuf, szW, len);
@@ -210,7 +209,7 @@ ConvertToUnicode(char *szA)
         if (olebuf)
             free(olebuf);
         olebuf_len = olebuf_len * 2 + len;
-        if ((olebuf = (OLECHAR *) malloc(olebuf_len * sizeof(OLECHAR))) == NULL)
+        if ((olebuf = typeallocn(OLECHAR, olebuf_len)) == NULL)
             return (olebuf);  /* We're gonna' die */
     }
     mbstowcs(olebuf, szA, len);
@@ -788,8 +787,8 @@ syncfile(int f, int n)
          */
 
         olebuf_len = ansibuf_len = 512;
-        ansibuf    = (char *) malloc(ansibuf_len);
-        olebuf     = (OLECHAR *) malloc(olebuf_len * sizeof(OLECHAR));
+        ansibuf    = typeallocn(char, ansibuf_len);
+        olebuf     = typeallocn(OLECHAR, olebuf_len);
         if (! (olebuf && ansibuf))
         {
             no_memory("syncfile()");
