@@ -3,7 +3,7 @@
  * paragraph at a time.  There are all sorts of word mode commands.  If I
  * do any sentence mode commands, they are likely to be put in this file.
  *
- * $Header: /usr/build/vile/vile/RCS/word.c,v 1.74 2002/04/30 23:24:11 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/word.c,v 1.79 2005/07/14 00:02:44 tom Exp $
  *
  */
 
@@ -21,8 +21,8 @@
 int
 wrapword(int f, int n)
 {
-    register int cnt = 0;	/* size of word wrapped to next line */
-    register int c;		/* character temporary */
+    int cnt = 0;		/* size of word wrapped to next line */
+    int c;			/* character temporary */
     B_COUNT to_delete;
     C_NUM to_append = 0;
 
@@ -39,7 +39,7 @@ wrapword(int f, int n)
      * (n!=0), try to split the line at the last embedded whitespace.
      */
     if (f) {
-	register LINE *lp = DOT.l;
+	LINE *lp = DOT.l;
 	to_delete = 0L;
 	if (DOT.o >= 0 && !n && !isSpace(lgetc(lp, DOT.o))) {
 	    for (c = DOT.o; c >= 0; c--) {
@@ -273,11 +273,14 @@ backword(int f, int n)
     return (forwchar(TRUE, 1));
 }
 
-int
-joinregion(void)
+/*
+ * If not 'exact', will adjust whitespace at the join-position.
+ */
+static int
+join_region(int exact)
 {
-    register int status;
-    register int doto, c;
+    int status;
+    int doto, c;
     LINE *end;
     REGION region;
     int done = FALSE;
@@ -302,14 +305,18 @@ joinregion(void)
 	    (void) setmark();
 	    if (forwline(FALSE, 1) != TRUE)
 		break;
-	    (void) firstnonwhite(FALSE, 1);
+	    if (exact) {
+		(void) gotobol(FALSE, 1);
+	    } else {
+		(void) firstnonwhite(FALSE, 1);
+	    }
 
 	    done = ((DOT.l == end) || (lforw(DOT.l) == end));
 	    if (killregionmaybesave(FALSE) != TRUE)
 		break;
 
 	    doto = DOT.o;
-	    if (doto == 0) {
+	    if (doto == 0 || exact) {
 		/*EMPTY */ ;
 	    }
 	    /* join at column 0 to empty line */
@@ -329,6 +336,18 @@ joinregion(void)
     return status;
 }
 
+int
+joinregion(void)
+{
+    return join_region(FALSE);
+}
+
+int
+joinregion_x(void)
+{
+    return join_region(TRUE);
+}
+
 int
 joinlines(int f, int n)
 {
@@ -385,19 +404,19 @@ comment_prefix(void)
 		&& memcmp("/*", tb_values(*wp), 2) == 0)
 
 static int
-do_formatting(TBUFF ** wp, TBUFF ** cp)
+do_formatting(TBUFF **wp, TBUFF **cp)
 {
-    register int c;		/* current char during scan     */
-    register int clength;	/* position on line during fill */
-    register int plength;	/* prefix to ignore during fill */
-    register size_t i;		/* index during word copy       */
-    register int newlen;	/* tentative new line length    */
-    register int finished;	/* Are we at the End-Of-Paragraph? */
-    register int firstflag;	/* first word? (needs no space) */
-    register int is_comment;	/* doing a comment block?       */
-    register int at_nl = TRUE;	/* just saw a newline?          */
-    register LINEPTR pastline;	/* pointer to line just past EOP */
-    register int sentence;	/* was the last char a period?  */
+    int c;			/* current char during scan     */
+    int clength;		/* position on line during fill */
+    int plength;		/* prefix to ignore during fill */
+    size_t i;			/* index during word copy       */
+    int newlen;			/* tentative new line length    */
+    int finished;		/* Are we at the End-Of-Paragraph? */
+    int firstflag;		/* first word? (needs no space) */
+    int is_comment;		/* doing a comment block?       */
+    int at_nl = TRUE;		/* just saw a newline?          */
+    LINE *pastline;		/* pointer to line just past EOP */
+    int sentence;		/* was the last char a period?  */
     int secondindent;
     int s;
     int fillcolumn = getfillcol(curbp);
@@ -501,7 +520,7 @@ do_formatting(TBUFF ** wp, TBUFF ** cp)
 			}
 			if (DOT.l != pastline
 			    && !dot_at_section_break()) {
-			    int spcs = DOT.o;
+			    B_COUNT spcs = DOT.o;
 			    DOT.o = 0;
 			    s = ldelete(spcs, FALSE);
 			    if (s != TRUE)
@@ -632,12 +651,12 @@ formatregion(void)
 int
 wordcount(int f GCC_UNUSED, int n GCC_UNUSED)
 {
-    register LINE *lp;		/* current line to scan */
-    register int offset;	/* current char to scan */
+    LINE *lp;			/* current line to scan */
+    int offset;			/* current char to scan */
     long size;			/* size of region left to count */
-    register int ch;		/* current character to scan */
-    register int wordflag;	/* are we in a word now? */
-    register int lastflag;	/* were we just in a word? */
+    int ch;			/* current character to scan */
+    int wordflag;		/* are we in a word now? */
+    int lastflag;		/* were we just in a word? */
     long nwhite;		/* total # of blanks */
     long nwords;		/* total # of words */
     long nchars;		/* total number of chars */
@@ -646,7 +665,7 @@ wordcount(int f GCC_UNUSED, int n GCC_UNUSED)
     int status;			/* status return code */
     REGION region;		/* region to look at */
     const char *ending = get_record_sep(curbp);
-    int len_rs = strlen(ending);
+    int len_rs = len_record_sep(curbp);
 
     /* make sure we have a region to count */
     if ((status = getregion(&region)) != TRUE) {
@@ -700,4 +719,4 @@ wordcount(int f GCC_UNUSED, int n GCC_UNUSED)
 
     return (TRUE);
 }
-#endif
+#endif /* OPT_WORDCOUNT */
